//Copyright (c) 2006, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

package jep;

import java.awt.*;
import java.net.*;
import java.util.*;
import java.applet.*;
import java.io.*;
import sun.plugin.AppletViewer;
import sun.applet.AppletClassLoader;
import sun.awt.AppContext;
import sun.awt.PeerEvent;
import sun.awt.SunToolkit;
import javax.swing.UIManager;
import apple.awt.CColorPaint;
import apple.awt.MyEventQueue;
import apple.laf.FixAquaImageFactory;
import apple.laf.CColorPaintUIResource;

public class AppletFramePanel extends AppletViewer {
	private static boolean showDebugInfo = false;
	protected HashMap attributes = null;
	protected AppletFrame frame = null;
	Hashtable fixedDefaultsTable = new Hashtable();
	Object[] fixedDefaultsArray = null;
	// This value is used to determine how (or whether) instances of this
	// class are serialized.  In the original it's set to something that's
	// presumably only meaningful to Apple.  If we don't set it at all (i.e.
	// if we don't override the value it has in its superclass), Java will
	// decide how to serialize this class.  If we set it to zero we can stop
	// any instance of this class from ever being serialized.  See the docs
	// for ObjectStreamClass.getSerialVersionUID().
	//private static final long serialVersionUID = 0L;
	protected int appletResizeCount = 0;

	// As of JEP 0.9.4 we do more cleanup as an applet is destroyed.  See
	// AppletFrame.destroyApplet() for more information.
	public void cleanup() {
		frame = null;
		removeAppletStatusListener(null);
		synchronized(atts) {
			atts.clear();
		}
		synchronized(attributes) {
			attributes.clear();
		}
		synchronized(fixedDefaultsTable) {
			fixedDefaultsTable.clear();
		}
	}

	protected boolean destroying() {
		if ((frame == null) || frame.destroying) {
			return true;
		} else {
			return false;
		}
	}

	// In the original, the "attributes" class is Hashtable.  Hashtable has
	// built-in synchronization, while HashMap doesn't.  But we're already
	// doing all the synchonization we need in the getParameter() and
	// setParameter() methods below, and Java's AppletViewer class uses
	// HashMap.  (Sun's AppletViewer class has getParameter() and
	// setParameter() methods, but Sun strangely gave them no access
	// modifiers.  So we, like Apple, are forced to write our own and make
	// them "public".)
	public AppletFramePanel(URL docbase, HashMap attributes, AppletFrame frame) {
		if (showDebugInfo) {
			System.err.println("Calling AppletFramePanel(): " +
				"docbase (" + ((docbase == null) ? "null" : docbase.toString()) + ") " +
				"attributes (" + ((attributes == null) ? "null" : attributes.toString()) + ")");
		}
		setDocumentBase(docbase.toString());
		this.attributes = attributes;
		this.frame = frame;
		// In Java 1.6.0 the basic CColorPaint constructor takes five
		// parameters.  On OS X Tiger (and earlier), in other Java versions
		// this constructor takes four parameters.  The same is true of the
		// WWDC Leopard Developer Preview (build 9A421).  But with build 9A283
		// CColorPaint's constructor also takes five parameters in Java 1.5.0.
		// Until Leopard settles down, we'll need to try first one constructor,
		// then (if that fails) the other.
		try {
			Class[] parmTypes = new Class[4];
			Object[] parms =
				{new Long(AppletHandlerFactory.getNSColor(255, 255, 255)),
				new Integer(255), new Integer(255), new Integer(255)};
			parmTypes[0] = long.class;
			parmTypes[1] = int.class;
			parmTypes[2] = int.class;
			parmTypes[3] = int.class;
			setBackground((Color) CColorPaint.class.
				getConstructor(parmTypes).newInstance(parms));
		} catch (Throwable e) {
			try {
				Class[] parmTypes = new Class[5];
				Object[] parms =
					{new Long(AppletHandlerFactory.getNSColor(255, 255, 255)),
					new Integer(255), new Integer(255), new Integer(255),
					new Integer(255)};
				parmTypes[0] = long.class;
				parmTypes[1] = int.class;
				parmTypes[2] = int.class;
				parmTypes[3] = int.class;
				parmTypes[4] = int.class;
				setBackground((Color) CColorPaint.class.
					getConstructor(parmTypes).newInstance(parms));
			} catch (Throwable f) {
			}
		}
		// This seems not to be initialized correctly in our superclass, or in
		// Apple's original code.
		status = 0;
	}

	// An applet's start() method is now sometimes called earlier than it used
	// to be (see startApplet in AppletView.m).  For some reason this means
	// that a lot of the native Cocoa code called from this class
	// (AppletFramePanel) sometimes runs without any autorelease pool in
	// place.  The following code (addImpl(), paintForegrnd(), run() and
	// additions to validate() below) tries to fix this.  (I've only observed
	// _NSAutoreleaseNoPool() error messages on a Mac Intel box, but I assume
	// that this problem can (in principle) happen on any Mac running OS X
	// 10.4 (at least).)
	protected Object addImplLock = new Object();
	protected ThreadLocal addImplPool = new ThreadLocal() {
		protected Object initialValue() {
			return new Long(0);
		}
	};
	protected long getAddImplPool() {
		return ((Long) (addImplPool.get())).longValue();
	}
	protected void setAddImplPool(long pool) {
		addImplPool.set(new Long(pool));
	}

	protected void addImpl(Component comp, Object constraints, int index) {
		synchronized(addImplLock) {
			if (getAddImplPool() == 0)
				setAddImplPool(AppletHandlerFactory.allocAutoreleasePool());
		}
		super.addImpl(comp, constraints, index);
	}

	public void paintForegrnd(Graphics g) {
		long pool = AppletHandlerFactory.allocAutoreleasePool();
		super.paintForegrnd(g);
		AppletHandlerFactory.releaseAutoreleasePool(pool);
	}

	public void run() {
		long pool = AppletHandlerFactory.allocAutoreleasePool();
		AppletFrame.setCurrentAppletFrame(frame);
		super.run();
		AppletFrame.setCurrentAppletFrame(null);
		AppletHandlerFactory.releaseAutoreleasePool(pool);
	}

	// In our superclass this method creates the applet and starts its thread
	// and event spool, but at this point the applet's init() method hasn't
	// yet been called (i.e. none of the applet's own code has yet started
	// running).
    protected Applet createApplet(final AppletClassLoader loader)
		throws ClassNotFoundException, IllegalAccessException, IOException,
		InstantiationException, InterruptedException 
	{
		Applet newApplet = super.createApplet(loader);
		if (newApplet == null)
			return null;
		frame.changeFrameAppContext(SunToolkit.targetToAppContext(newApplet));
		frame.setSimpleLayout();
		if (destroying())
			return newApplet;
		// We no longer need these fixes in Java 1.4.2 Update 1.
		if (!AppletHandlerFactory.isJava142Update1()) {
			synchronized(fixedDefaultsTable) {
				// Yield after we post the FixStuff event to make it happen as
				// soon as possible (i.e. before the applet starts running its
				// own code), but make the part of FixStuff that changes LAF
				// color defaults synchronize on the same object we've
				// synchronized on here -- so that fixLafColorDefaults()
				// finishes before the FixStuff's own color-fixing code (which
				// needs to run in the applet's context) begins.
				SunToolkit.postEvent(SunToolkit.targetToAppContext(newApplet),
					new PeerEvent(newApplet, new FixStuff(),
					PeerEvent.PRIORITY_EVENT));
				Thread.sleep(100);
				if (destroying())
					return newApplet;
				if (frame.isCarbonApp && AppletHandlerFactory.isJava142()) {
					fixLafColorDefaults();
				}
			}
		// But we do need a different fix.
		} else {
			SunToolkit.postEvent(SunToolkit.targetToAppContext(newApplet),
				new PeerEvent(newApplet, new FixOtherStuff(),
				PeerEvent.PRIORITY_EVENT));
			Thread.sleep(100);
			if (destroying())
				return newApplet;
		}
		// Log that we're creating an applet.  This may help people who can't
		// figure out whether or not they're using applets!
		String appletInfo = null;
		// Some applets have broken getAppletInfo() implementations
		try {
			appletInfo = newApplet.getAppletInfo();
		} catch (Exception e) {
		}
		String name = newApplet.getClass().getName();
		String info = new Date() + " JEP creating applet " + name +
			" (" + getCodeBase().toString() + ")" +
			((appletInfo != null) ? ("\n" + appletInfo) : "");
		tellAppletCreated(frame.getCocoaParentViewLong(), newApplet, info, name);
		System.out.println(info);
		return newApplet;
	}

	protected native void tellAppletCreated(long cocoaParentView, Applet applet, String info, String name);

	// With Java 1.4.2 in a Carbon browser, the Java equivalents of Apple's
	// system colors often don't work properly in Swing applets.  This
	// method "fixes" the colors that AquaImageFactory's methods provide,
	// then "resets" the default "Aqua" Swing look and feel settings that
	// depend on them (at the middle level -- see Sun's UIManager.java for
	// what this means).  Afterwards we also "fix" all the other colors among
	// the LAF defaults (those that don't belong to the CColorPaintUIResource
	// class) -- this time at the third or "user" level.  (This task begins
	// here but gets finished in FixStuff's run() method below.)  Doing
	// it this way ensures that our "fixes" don't disappear when the applet
	// changes the default LAF to something else and then back to "Aqua".
	// (This fix isn't needed any longer in Java 1.4.2 Update 1.)
	protected void fixLafColorDefaults() {
		if (destroying())
			return;
		FixAquaImageFactory.fix();
		try {
			UIManager.setLookAndFeel(UIManager.getLookAndFeel());
		} catch (Throwable e) {
		}
		// This fetches the "middle level" settings (as opposed to the
		// "system" and "user" level ones) -- see Sun's UIManager.java.
		Enumeration keys = UIManager.getLookAndFeelDefaults().keys();
		// Now we pick out the non-CColorPaintUIResource colors, change them,
		// and store them for use in FixStuff's run() method below.
		while (keys.hasMoreElements()) {
			Object key = keys.nextElement();
			Object value = UIManager.get(key);
			if ((value instanceof Color) && 
				!(value instanceof CColorPaintUIResource))
			{
				CColorPaint color = new CColorPaint(
					AppletHandlerFactory.getNSColor(((Color)value).getRed(),
						((Color)value).getGreen(), ((Color)value).getBlue()),
					((Color)value).getRed(), ((Color)value).getGreen(),
						((Color)value).getBlue());
				fixedDefaultsTable.put(key, color);
			}
		}
		Object[] entries = fixedDefaultsTable.entrySet().toArray();
		fixedDefaultsArray = new Object[entries.length * 2];
		for (int i = 1; i <= entries.length; ++i) {
			fixedDefaultsArray[2*(i-1)] = ((Map.Entry)entries[i-1]).getKey();
			fixedDefaultsArray[2*(i-1)+1] = ((Map.Entry)entries[i-1]).getValue();
		}
		if (showDebugInfo) {
			System.err.println("AppletFramePanel fixLafColorDefaults(): " +
				fixedDefaultsTable.toString());
		}
	}

	public void validate() {
		if (destroying() || AppletHandlerFactory.isMainThread())
			return;
		super.validate();
		synchronized(addImplLock) {
			long pool = getAddImplPool();
			if (pool != 0) {
				AppletHandlerFactory.releaseAutoreleasePool(pool);
				setAddImplPool(0);
			}
		}
	}

	public void invalidate() {
		if (destroying() || AppletHandlerFactory.isMainThread())
			return;
		super.invalidate();
	}

	public String getParameter(String name) {
		if (attributes == null)
			return null;
		name = name.toLowerCase(java.util.Locale.ENGLISH);
		synchronized(attributes) {
			String value = (String) attributes.get(name);
			if (value != null)
				value = trimWhiteSpaces(value);
			return value;
		}
	}

	public void setParameter(String name, Object value) {
		if (attributes == null)
			return;
		name = name.toLowerCase(java.util.Locale.ENGLISH);
		synchronized(attributes) {
			attributes.put(name, trimWhiteSpaces(value.toString()));
		}
	}

	// Make Applet.resize() (which calls AppletFramePanel.appletResize())
	// trigger the JEP's support for directly (or "natively") resizing an
	// applet.  See AppletFrame.resizeApplet() for more info.  The handler
	// for APPLET_RESIZE events (in sun.plugin.AppletViewer) is broken --
	// it resets the applet to its current size.  So we need to do its job
	// ourselves here.
	public void appletResize(int width, int height) {
		if (destroying())
			return;
		Applet applet = getApplet();
		if ((width == 0) || (height == 0) || (applet == null))
			return;
		if (!AppletHandlerFactory.isJava142Update1()) {
			appletResizeDirectly(width, height);
			return;
		}
		AppContext appletContext = SunToolkit.targetToAppContext(applet);
		boolean isAppletContext = AppContext.getAppContext().equals(appletContext);
		if (isAppletContext) {
			appletResizeDirectly(width, height);
		} else {
			if (++appletResizeCount == 1)
				AppletHandlerFactory.suspendAppletUpdates(frame.getCocoaParentViewLong());
			SunToolkit.postEvent(appletContext,
				new PeerEvent(applet, new AppletResize(width, height),
				PeerEvent.PRIORITY_EVENT));
		}
	}

	protected void appletResizeDirectly(int width, int height) {
		if (destroying())
			return;
		if (!frame.clipLock.lock(1000)) {
			System.err.println("AppletFramePanel.appletResize(): lock timed out");
			Throwable barf = new Throwable();
			barf.fillInStackTrace();
			barf.printStackTrace();
			return;
		}
		if ((width == getWidth()) && (height == getHeight())) {
			frame.clipLock.unlock();
			return;
		}
		setParameter("width", Integer.toString(width));
		setParameter("height", Integer.toString(height));
		LayoutManager layout = frame.getLayout();
		if (layout instanceof SimpleLayout)
			((SimpleLayout) layout).resetSize(width, height);
		layout = getLayout();
		if (layout instanceof SimpleLayout)
			((SimpleLayout) layout).resetSize(width, height);
		frame.setBounds(0, 0, width, height);
		setBounds(0, 0, width, height);
		Applet applet = frame.getApplet();
		if (applet != null)
			applet.setBounds(0, 0, width, height);
		AppletHandlerFactory.resetHybridView(frame.getCocoaParentViewLong());
		frame.clipLock.unlock();
	}

	private String trimWhiteSpaces(String string) {
		if (string == null)
			return null;

		StringBuffer buffer = new StringBuffer();

		for (int i = 0; i < string.length(); ++i) {
			char c = string.charAt(i);
			if (c == '\n' || c == '\f' || c == '\r' || c == '\t') {
				continue;
			} else {
				buffer.append(c);
			}
		}

		return buffer.toString().trim();
	}

	public String toString() {
		return super.toString() + " (jep)";
	}

	private class FixStuff implements Runnable {

		// Neither of these fixes is needed any longer in Java 1.4.2 Update 1.
		public void run() {
			if (destroying())
				return;
			// Fix the event queue on each applet's thread.  (We may already
			// have done on a different thread if more than one applet shares
			// the same AppContext -- as they will if they share the same
			// origin.)
			AppContext appContext = AppContext.getAppContext();
			EventQueue currentQueue = (EventQueue)
				appContext.get(AppContext.EVENT_QUEUE_KEY);
			if (!(currentQueue instanceof MyEventQueue)) {
				MyEventQueue newQueue = new MyEventQueue();
				currentQueue.push(newQueue);
				appContext.put(AppContext.EVENT_QUEUE_KEY, newQueue);
			}
			// With Java 1.4.2 in a Carbon browser, some of the color settings
			// for Swing's default "look and feel" (apple.laf.AquaLookAndFeel)
			// don't work properly.  This block and the fixLafColorDefaults()
			// method above (plus FixAquaImageFactory.fix()) are a workaround
			// for this bug.  All the colors I have direct access to (in
			// AquaImageFactory and AquaLookAndFeel) I'm able to fix.  But
			// there are some others I don't know how to access, which still
			// don't work right:  Examples are the bottom parts of the
			// InternalFrame objects in the SwingSet2 demo.
			synchronized(AppletFramePanel.this.fixedDefaultsTable) {
				if (destroying())
					return;
				if (AppletFramePanel.this.fixedDefaultsArray != null) {
					UIManager.getDefaults().
						putDefaults(AppletFramePanel.this.fixedDefaultsArray);
				}
			}
		}

	}

	private class FixOtherStuff implements Runnable {

		public void run() {
			if (destroying())
				return;
			AppContext appContext = AppContext.getAppContext();
			EventQueue currentQueue = (EventQueue)
				appContext.get(AppContext.EVENT_QUEUE_KEY);
			if (!(currentQueue instanceof MySimpleEventQueue)) {
				MySimpleEventQueue newQueue = new MySimpleEventQueue();
				currentQueue.push(newQueue);
				appContext.put(AppContext.EVENT_QUEUE_KEY, newQueue);
			}
		}

	}

	private class AppletResize implements Runnable {
		private final int width, height;

		AppletResize(int newWidth, int newHeight) {
			width = newWidth;
			height = newHeight;
		}

		public void run() {
			if (--appletResizeCount == 0) {
				AppletFramePanel.this.appletResizeDirectly(width, height);
				AppletHandlerFactory.resumeAppletUpdates(
					AppletFramePanel.this.frame.getCocoaParentViewLong());
			} else if (appletResizeCount < 0) {
				appletResizeCount = 0;
			}
			if (showDebugInfo) {
				System.err.println("AppletFramePanel AppletResize run() called, width is " +
					String.valueOf(width) + " height is " + String.valueOf(height));
			}
		}

	}

	static {
		if ("true".equalsIgnoreCase(System.getProperty("jep.debuginfo", "false"))) {
			showDebugInfo = true;
		}
	}

}

