//Copyright (c) 2006, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// Garbage collection is completely broken on Apple's current implementations
// of Java SE 6 if you make an applet's Frame object also its "handler" (what
// I used to call AppletHolder objects).  I'm not sure why this is (though
// it's clearly a bug).  But the bug seems to be triggered by creating the
// Frame object (and its associated Panel and peer objects) directly from JNI
// code (as was necessary).  Now I use AppletHandler objects to do the
// "handling" and AppletFrame objects to do the "framing", and on Java SE 6 I
// only create Frame objects (and other associated objects) on a Java secondary
// thread (the peer objects are created by AppletFramePanel.appletInit()).  It
// turns out this also works for the other Java versions on OS X Tiger and
// Leopard.  But on Panther and Jaguar you can't run AppletFramePanel.
// appletInit() on a (secondary) Java thread (you get wierdness like JVM
// popups not working with Camino).  So on these OSs I've continued to create
// AppletFrame and AppletFramePanel objects (together with their AppletHandler
// counterparts) directly in JNI code (AppletView doInitialization2) on the
// main thread, and to call AppletFramePanel.appletInit() on non-Java secondary
// threads (in AppletView doInitialization3).

// I've renamed some classes (and files) to make tne new organization clearer:
// AppletHolderFactory has been renamed to AppletHandlerFactory, AppletHolder
// to AppletFrame, AppletHolderPanel to AppletFramePanel, and
// AppletHolderContext to AppletFrameContext.

package jep;

import java.applet.*;
import java.awt.*;
import java.util.*;
import java.net.*;
import java.lang.reflect.*;
import sun.plugin.AppletStatusListener;

public class AppletHandler implements AppletStatusListener {
	protected AppletFrame frame = null;
	protected long cocoaParentView = 0;
	protected int pluginInstance = 0; // MRJPluginInstance *
	protected URL docbase = null;
	protected HashMap attributes = null;
	protected boolean isCarbonApp = false;
	protected boolean createdInvisible = false;
	protected InitApplet initer = new InitApplet();

	public AppletHandler(long cocoaParentView, int pluginInstance, URL docbase,
		HashMap attributes, boolean isCarbonApp, boolean createdInvisible)
		throws MalformedURLException
	{
		this.cocoaParentView = cocoaParentView;
		this.pluginInstance = pluginInstance;
		this.docbase = new URL(docbase.toExternalForm());
		this.attributes = new HashMap(attributes);
		this.isCarbonApp = isCarbonApp;
		this.createdInvisible = createdInvisible;
		if (AppletHandlerFactory.getOSXVersion() < 0x1040) {
			try {
				Class[] parmTypes = new Class[6];
				Object[] parms = {new Long(cocoaParentView),
					new Integer(pluginInstance), this.docbase,
					this.attributes, new Boolean(isCarbonApp),
					new Boolean(createdInvisible)};
				parmTypes[0] = long.class;
				parmTypes[1] = int.class;
				parmTypes[2] = URL.class;
				parmTypes[3] = HashMap.class;
				parmTypes[4] = boolean.class;
				parmTypes[5] = boolean.class;
				frame = (AppletFrame) AppletHandlerFactory.AppletFrameClass.
					getConstructor(parmTypes).newInstance(parms);
				frame.getPanel().addAppletStatusListener(AppletHandler.this);
			} catch (NoSuchMethodException e) {
				System.err.println("AppletHandler(): exception: " + e.getMessage());
				e.printStackTrace();
			} catch (InstantiationException e) {
				System.err.println("AppletHandler(): exception: " + e.getMessage());
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				System.err.println("AppletHandler(): exception: " + e.getMessage());
				e.printStackTrace();
			} catch (InvocationTargetException e) {
				System.err.println("AppletHandler(): exception: " + e.getMessage());
				e.printStackTrace();
			}
		}
	}

	public void showAppletInfo() {
		frame.showAppletInfo();
	}

	public Applet getApplet() {
		try {
			return frame.getPanel().getApplet();
		} catch (NullPointerException e) {
			return null;
		}
	}

	public int getLoadingStatus() {
		try {
			return frame.getPanel().getLoadingStatus();
		} catch (NullPointerException e) {
			return 0;
		}
	}

	public void validateClips() {
		frame.validateClips();
	}

	public void zeroClips() {
		frame.zeroClips();
	}

	public void validateApplet() {
		frame.validateApplet();
	}

	public void invalidateAndValidate() {
		frame.invalidateAndValidate();
	}

	public void setAppletVisible(boolean visible, boolean flagOnly, boolean wait) {
		frame.setAppletVisible(visible, flagOnly, wait);
	}

	public boolean isAppletVisible() {
		if (frame == null)
			return false;
		return frame.isAppletVisible();
	}

	public void repaint() {
		frame.repaint();
	}

	public boolean isValid() {
		return frame.isValid();
	}

	public Rectangle getAppletBounds() {
		return frame.getAppletBounds();
	}

	public void setAppletBounds(int x, int y, int width, int height) {
		frame.setAppletBounds(x, y, width, height);
	}

	public void setAppletBounds(Rectangle bounds) {
		frame.setAppletBounds(bounds);
	}

	public void resizeApplet(int width, int height) {
		frame.resizeApplet(width, height);
	}

	public void setSimpleLayout() {
		frame.setSimpleLayout();
	}

	public void setIgnoreRepaint(boolean flag) {
		frame.setIgnoreRepaint(flag);
	}

	public void setUpdatesOff(boolean b) {
		frame.setUpdatesOff(b);
	}

	public void setAsyncUpdatesOff(boolean b) {
		frame.setAsyncUpdatesOff(b);
	}

	public void imposeClip(int clipX, int clipY, int clipWidth, int clipHeight) {
		frame.imposeClip(clipX, clipY, clipWidth, clipHeight);
	}

	public void removeClip() {
		frame.removeClip();
	}

	public void startApplet() {
		frame.startApplet();
	}

	public void stopApplet() {
		frame.stopApplet();
	}

	public void appletStartApplet() {
		frame.appletStartApplet();
	}

	public void appletStopApplet() {
		frame.appletStopApplet();
	}

	public void destroyApplet() {
		frame.destroyApplet();
		initer.notifyError();
		cocoaParentView = 0L;
		frame = null;
		docbase = null;
		attributes = null;
	}

	public void activateFrame() {
		frame.activateFrame();
	}

	// Starts the process of loading and initializing the applet.  This must
	// be called on a secondary non-Java thread (from JNI code) in Panther
	// and Jaguar.
	public void initApplet() {
		if (AppletHandlerFactory.getOSXVersion() < 0x1040) {
			frame.getPanel().appletInit();
		} else {
			initer.start();
		}
	}

	public void statusChanged(int status) {
		try {
			updateStatusFromJava(cocoaParentView, status);
		} catch (NullPointerException e) {
		}
		if ((status == AppletFramePanel.APPLET_ERROR) ||
			(status == AppletFramePanel.APPLET_DESTROY))
		{
			initer.notifyError();
		} else if (status == AppletFramePanel.APPLET_INIT) {
			initer.notifyInitDone();
		}
	}

	private native static void updateStatusFromJava(long cocoaParentView, int status);

	protected class InitApplet extends Thread {
		private boolean initDone = false;
		private boolean error = false;

		public synchronized void notifyInitDone() {
			if (!error)
				initDone = true;
			notifyAll();
		}

		public synchronized void notifyError() {
			error = true;
			initDone = false;
			notifyAll();
		}

		public synchronized void waitTilDone() {
			if (initDone || error)
				return;
			// Wait for up to 60 seconds for an applet to load
			try {
				wait(60000);
			} catch (InterruptedException e) {
			}
			return;
		}

		public void run() {
			try {
				Class[] parmTypes = new Class[6];
				Object[] parms = {new Long(cocoaParentView),
					new Integer(pluginInstance), docbase, attributes,
					new Boolean(isCarbonApp), new Boolean(createdInvisible)};
				parmTypes[0] = long.class;
				parmTypes[1] = int.class;
				parmTypes[2] = URL.class;
				parmTypes[3] = HashMap.class;
				parmTypes[4] = boolean.class;
				parmTypes[5] = boolean.class;
				frame = (AppletFrame) AppletHandlerFactory.AppletFrameClass.
					getConstructor(parmTypes).newInstance(parms);
				frame.getPanel().addAppletStatusListener(AppletHandler.this);
				activateFrame();
				frame.getPanel().appletInit();
			} catch (Throwable e) {
				System.err.println("AppletHandler.InitApplet().run(): exception: " + e.getMessage());
				e.printStackTrace();
				reportAppletError(cocoaParentView);
				return;
			}
			waitTilDone();
			if (initDone) {
				maybeDoAppletStart(cocoaParentView);
			} else {
				reportAppletError(cocoaParentView);
			}
		}

	}

	private native static void maybeDoAppletStart(long cocoaParentView);
	private native static void reportAppletError(long cocoaParentView);

}

