//Copyright (c) 2006, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

package jep;

import java.lang.reflect.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.Toolkit;
import java.awt.EventQueue;
import java.security.Security;
import sun.plugin.AppletViewer;
import sun.awt.AppContext;
import apple.awt.MyEventQueue;

public class AppletHandlerFactory {
	protected static boolean showDebugInfo = false;
	protected static boolean sPropsInited = false;
	static Thread mainThread = null;
	static AppContext mainAppContext = null;
	static boolean appleNativesLoaded = false;
	static String javaVersion = null;
	static String javaRuntimeVersion = null;
	public static Class AppletFrameClass = null;
	public static Class MyCAppletFrameClass = null;
	public static Method myCAppletFrameCreateNew = null;
	public static Class MyCPanelClass = null;
	public static Method myCPanelCreateNew = null;
	public static Class MyCCanvasClass = null;
	public static Method myCCanvasCreateNew = null;
	public static Class MyCLabelClass = null;
	public static Method myCLabelCreateNew = null;
	public static Class MyCScrollPaneClass = null;
	public static Method myCScrollPaneCreateNew = null;
	public static Class GraphicsClipperClass = null;
	public static Method graphicsClipperCreateNew = null;
	public static Class PaintEventClass = null;

	private static void setMainThread() {
		mainThread = Thread.currentThread();
		mainAppContext = AppContext.getAppContext();
	}

	public static void resetMainEventQueue() {
		EventQueue mainQueue = (EventQueue)
			mainAppContext.get(AppContext.EVENT_QUEUE_KEY);
		if (!AppletHandlerFactory.isJava142Update1()) {
			if (!(mainQueue instanceof MyEventQueue)) {
				MyEventQueue newQueue = new MyEventQueue();
				mainQueue.push(newQueue);
				mainAppContext.put(AppContext.EVENT_QUEUE_KEY, newQueue);
			}
		} else {
			if (!(mainQueue instanceof MySimpleEventQueue)) {
				MySimpleEventQueue newQueue = new MySimpleEventQueue();
				mainQueue.push(newQueue);
				mainAppContext.put(AppContext.EVENT_QUEUE_KEY, newQueue);
			}
		}
	}

	public static boolean isMainThread() {
		if (mainThread == null)
			return false;
		return mainThread.equals(Thread.currentThread());
	}

	// If we return an AppletHandler object from a function called from JNI (as
	// this function did before JEP 0.9.5), the Java garbage collector never
	// frees it.  But if we pass our newly created AppletHandler object as a
	// parameter to a JNI native function (tellHolderCreated()), the garbage
	// collector does what it's supposed to do.
	private static boolean createAppletHandler(long cocoaParentView,
		int pluginInstance, String docbase, HashMap attributes,
		boolean isCarbonApp, boolean createdInvisible)
	{
		initProperties();
		AppletHandler handler = null;
		boolean retval = true;
		if (showDebugInfo) {
			System.err.println("Calling AppletHandlerFactory createAppletHandler(): " +
				"cocoaParentView (0x" + Long.toHexString(cocoaParentView) + ") " +
				"pluginInstance (0x" + Integer.toHexString(pluginInstance) + ") " +
				"docbase (" + ((docbase == null) ? "null" : docbase) + ") " +
				"attributes (" + ((attributes == null) ? "null" : attributes.toString()) + ")");
		}
		try {
			handler = new AppletHandler(cocoaParentView, pluginInstance,
				new URL(docbase), attributes, isCarbonApp, createdInvisible);
		} catch (MalformedURLException e) {
			System.err.println("AppletHandlerFactory: bad document URL: " + docbase);
			retval = false;
		} catch (Throwable e) {
			System.err.println("AppletHandlerFactory: exception: " + e.getMessage());
			e.printStackTrace();
			retval = false;
		}
		if (showDebugInfo) {
			System.err.println("AppletHandlerFactory createAppletHandler() returned " +
				((handler == null) ? "null" : handler.toString()));
		}
		if (retval)
			tellHandlerCreated(cocoaParentView, handler);
		return retval;
	}

	public static boolean isJava142() {
		if (javaVersion.compareTo("1.4.2") >= 0)
			return true;
		return false;
	}

	public static boolean isJava142Update1() {
		if (javaVersion.compareTo("1.4.2_04") >= 0)
			return true;
		return false;
	}

	public static boolean isJava142Update2() {
		if ((javaVersion.compareTo("1.4.2_05") >= 0) &&
			(javaRuntimeVersion.compareTo("1.4.2_05-141.3") >= 0))
			return true;
		return false;
	}

	public static boolean isJava142Tiger() {
		if ((getOSXVersion() >= 0x1040) &&
			(javaVersion.compareTo("1.4.2_07") >= 0))
			return true;
		return false;
	}

	public static boolean isJava150() {
		if (javaVersion.compareTo("1.5.0") >= 0)
			return true;
		return false;
	}

	public static boolean isJava160() {
		if (javaVersion.compareTo("1.6.0") >= 0)
			return true;
		return false;
	}

	public static String getJavaVersionString() {
		return javaVersion;
	}

	public static boolean getPropertyBoolean(String key, boolean dflt) {
		String property = System.getProperty(key);
		if (property == null)
			return dflt;
		boolean retval = dflt;
		if ("true".equalsIgnoreCase(property) ||
			"yes".equalsIgnoreCase(property) ||
			"on".equalsIgnoreCase(property))
			retval = true;
		if ("false".equalsIgnoreCase(property) ||
			"no".equalsIgnoreCase(property) ||
			"off".equalsIgnoreCase(property))
			retval = false;
		return retval;
	}

	public native static long getNSColor(int red, int green, int blue);

	public native static long getNSSystemColor(String colorName, int[] rgbValues);

	public native static int getCurrentPThreadID();

	public native static int getOSXVersion();

	public native static void disableScreenUpdates();

	public native static void enableScreenUpdates();

	public native static long allocAutoreleasePool();

	public native static void releaseAutoreleasePool(long pool);

	public native static boolean isAppletViewDisplayable(long cocoaParentView);

	public native static void resetHybridView(long cocoaParentView);

	public native static void suspendAppletUpdates(long cocoaParentView);

	public native static void resumeAppletUpdates(long cocoaParentView);

	public native static void cocoaDrawApplet(long cocoaParentView);

	public native static void deleteAppletJNIGlobalRefs(long cocoaParentView);

	private native static void tellHandlerCreated(long cocoaParentView, AppletHandler handler);

	public native static void processMainThreadEvents(long millis, long jobID);

	public native static void processQueuedMainThreadEvents();

	public native static long beforeProcessing(Object key);

	public native static void stopProcessing(long jobID);

	public native static void callRegisterNatives();

	protected static void initProperties() {
		if (sPropsInited)
			return;
		sPropsInited = true;
		if (showDebugInfo)
			System.err.println("Calling AppletHandlerFactory initProperties()");
		// Sun's JVMs all try to ensure that there's just one instance of the
		// system "Toolkit" (a subclass of java.awt.Toolkit that performs low-
		// level, platform-specific drawing tasks).  The JVM also lets you use
		// the "awt.toolkit" system property to specify which subclass of
		// java.awt.Toolkit to use (by default Apple uses apple.awt.CToolkit).
		// For various reasons we need to override Apple's default (see
		// MyCToolkit.java for more information).  This property must of
		// course be set very early -- doing it here seems to be early enough.
		System.setProperty("awt.toolkit", "apple.awt.MyCToolkit");
		// We need this for some graphics that are drawn asynchronously
		// (notably by http://java.sun.com/j2se/jre_check/index.jsp).
		// But (as of JEP 0.9.5+e) this applet has now disappeared from the
		// net, and setting EnableLazyDrawing to "false" seems to cause
		// problems with at least two other applets:
		// https://bugzilla.mozilla.org/show_bug.cgi?id=335804
		// https://bugzilla.mozilla.org/show_bug.cgi?id=302843
		//if (isJava142Update1() && !isJava142Tiger())
		//	System.setProperty("apple.awt.graphics.EnableLazyDrawing", "false");
		// As of JEP 0.9.5+f, we use JavaPluginCocoa.bundle's registerNatives()
		// call to register its "native" mathods.  We need to call it on a
		// secondary thread, after we set the "awt.toolkit" property, but
		// before we load the AWT libraries (i.e. before we have the JVM load
		// them, since libawt.jnilib's image has already been loaded).
		if (isJava142Update1())
			callRegisterNatives();
		// Apple made incompatible changes to the EmbeddableAppletFrame
		// interface (used by AppletFrame) and to the CCanvas class between
		// Java 1.4.1 and Java 1.4.2 -- EmbeddableAppletFrame's single method
		// (getCocoaParentView()) returns an int in 1.4.1 but a long in 1.4.2,
		// and CCanvas's constructor's innerView parameter is an int in 1.4.1
		// but a long in 1.4.2 (this and other, parallel, changes were
		// presumably made to add support for 64-bit processors).  So we need
		// to write two versions of each of these classes and to add a level
		// of indirection to our code that uses them, so we can use a version
		// check to choose between them at run-time.
		try {
			// As of JEP 0.9.4 we load the "java.awt.Frame" class (which
			// triggers loading the AWT libraries) here, instead of doing it
			// in AppletView loadAWTLibraries.
			Class FrameClass = Class.forName("java.awt.Frame");
			Class PanelClass = Class.forName("java.awt.Panel");
			Class CanvasClass = Class.forName("java.awt.Canvas");
			Class LabelClass = Class.forName("java.awt.Label");
			Class ScrollPaneClass = Class.forName("java.awt.ScrollPane");
			if (isJava142Tiger()) {
				AppletFrameClass = Class.forName("jep.AppletFrame142Tiger");
				MyCCanvasClass = Class.forName("apple.awt.MyCCanvas142");
				MyCAppletFrameClass = Class.forName("apple.awt.MyCAppletFrame142Tiger");
				MyCPanelClass = Class.forName("apple.awt.MyCPanel142");
				MyCLabelClass = Class.forName("apple.awt.MyCLabel142");
				MyCScrollPaneClass = Class.forName("apple.awt.MyCScrollPane142");
				GraphicsClipperClass = Class.forName("apple.awt.GraphicsClipper142");
			} else if (isJava142()) {
				AppletFrameClass = Class.forName("jep.AppletFrame142");
				MyCCanvasClass = Class.forName("apple.awt.MyCCanvas142");
				MyCAppletFrameClass = Class.forName("apple.awt.MyCAppletFrame142");
				MyCPanelClass = Class.forName("apple.awt.MyCPanel142");
				MyCLabelClass = Class.forName("apple.awt.MyCLabel142");
				MyCScrollPaneClass = Class.forName("apple.awt.MyCScrollPane142");
				GraphicsClipperClass = Class.forName("apple.awt.GraphicsClipper142");
			} else {
				AppletFrameClass = Class.forName("jep.AppletFrame141");
				MyCCanvasClass = Class.forName("apple.awt.MyCCanvas141");
				MyCAppletFrameClass = Class.forName("apple.awt.MyCAppletFrame141");
				MyCPanelClass = Class.forName("apple.awt.MyCPanel141");
				MyCLabelClass = Class.forName("apple.awt.MyCLabel");
				MyCScrollPaneClass = Class.forName("apple.awt.MyCScrollPane141");
				GraphicsClipperClass = Class.forName("apple.awt.GraphicsClipper");
			}
			if (!isJava142Update1()) {
				PaintEventClass = Class.forName("apple.awt.CocoaEvent");
			}
			// Calling getMethod on our own classes' methods sometimes results
			// in access violations (when it's called on an applet thread).
			// So now (as of JEP 0.9.6/0.9.5+d) we pre-load here the version-
			// specific methods we need to call later.  This partly resolves
			// http://sourceforge.net/forum/forum.php?thread_id=1443262&forum_id=371125.
			Class[] myCAppletFrameCreateNewParmTypes = 
				{FrameClass, AppletFrame.class};
			myCAppletFrameCreateNew = MyCAppletFrameClass.getMethod(
				"createNew", myCAppletFrameCreateNewParmTypes);
			Class[] myCPanelCreateNewParmTypes = 
				{PanelClass, AppletFrame.class};
			myCPanelCreateNew = MyCPanelClass.getMethod(
				"createNew", myCPanelCreateNewParmTypes);
			Class[] myCCanvasCreateNewParmTypes = 
				{CanvasClass, long.class, AppletFrame.class};
			myCCanvasCreateNew = MyCCanvasClass.getMethod(
				"createNew", myCCanvasCreateNewParmTypes);
			Class[] myCLabelCreateNewParmTypes = 
				{LabelClass, AppletFrame.class};
			myCLabelCreateNew = MyCLabelClass.getMethod(
				"createNew", myCLabelCreateNewParmTypes);
			Class[] myCScrollPaneCreateNewParmTypes = 
				{ScrollPaneClass, AppletFrame.class};
			myCScrollPaneCreateNew = MyCScrollPaneClass.getMethod(
				"createNew", myCScrollPaneCreateNewParmTypes);
			Class ComponentModelClass = Class.forName("apple.awt.ComponentModel");
			Class SunGraphics2DClass = Class.forName("sun.java2d.SunGraphics2D");
			Class[] graphicsClipperCreateNewParmTypes =
				{SunGraphics2DClass, AppletFrame.class, ComponentModelClass};
			graphicsClipperCreateNew = GraphicsClipperClass.getMethod(
				"createNew", graphicsClipperCreateNewParmTypes);
		} catch (Throwable e) {
			System.err.println("AppletHandlerFactory: initProperties(): exception: " +
				e.getMessage());
			e.printStackTrace();
		}
		// Make sure our replacements for the ServiceProvider or ServiceManager
		// classes get loaded before Apple's equivalents.
		try {
			if (isJava150()) {
				Class.forName("com.sun.deploy.services.ServiceManager");
			} else {
				Class.forName("sun.plugin.services.ServiceProvider");
			}
		} catch (Throwable e) {
			System.err.println("AppletHandlerFactory: initProperties(): exception: " +
				e.getMessage());
			e.printStackTrace();
		}
		// giveOurselvesAllPermissions() only effects code called from the
		// main thread (as opposed to the applet threads) -- which is probably
		// just as well for security.  But unless we do it here, some of our
		// classes that have "static" blocks might get loaded for the first
		// time on one of the applet threads, and that code might not have
		// sufficient permissions.
		try {
			Class.forName("apple.awt.MyEventQueue");
			Class.forName("java.awt.MyEventQueue2");
			Class.forName("java.awt.SimpleLayout");
			// Preloading these two classes partially clears up access errors
			// that happen when a Swing object is created in a JavaScript-to-
			// Java LiveConnect call.  See
			// http://sourceforge.net/forum/forum.php?thread_id=1443262&forum_id=371125.
			Class.forName("sun.awt.datatransfer.DataTransferer");
			Class.forName("sun.awt.im.InputContext");
		} catch (Throwable e) {
			System.err.println("AppletHandlerFactory: initProperties(): exception: " +
				e.getMessage());
			e.printStackTrace();
		}
		if (isJava142Tiger())
			LiveConnect.initialize();
		giveOurselvesAllPermissions();
		restrictPackageAccess();
		loadAppleNatives();
		if (isJava150()) {
			System.setProperty("browser", "sun.plugin");
			StringBuffer version = new StringBuffer(javaVersion);
			int length = version.length();
			for (int i = 1; i <= length; ++i) {
				if (version.charAt(i - 1) == '_') {
					version.delete(i - 1, length);
					break;
				}
			}
			System.setProperty("javaplugin.version", version.toString());
			for (int i = 1; i <= version.length(); ++i) {
				if (version.charAt(i - 1) == '.')
					version.deleteCharAt(i - 1);
			}
			System.setProperty("javaplugin.nodotversion", version.toString());
		} else {
			try {
				Class.forName("sun.plugin.AppletViewer").
					getMethod("setupVersionProperties", (Class[]) null).
					invoke(null, (Object[]) null);
			} catch (Throwable e) {
				System.err.println("AppletHandlerFactory: initProperties(): exception: " +
					e.getMessage());
				e.printStackTrace();
			}
		}
		MyJavaRunTime.initJavaRunTime();
	}

	// This needs to be called either on the main thread or before a
	// SecurityManager is created (in AppletViewer.initProperties(), called
	// from MyJavaRunTime.initJavaRunTime()).  The contents of our policy file
	// gets added to other, existing policy files -- usually the system policy
	// file and (if one exists) the user's policy file.  Since our policy file
	// is added last, it doesn't override anything in the other policy files.
	protected static void giveOurselvesAllPermissions() {
		int n = 1;
		while (Security.getProperty("policy.url." + n) != null)
			++n;
		String ourPolicyProperty = "policy.url." + n;
		String ourPluginHome = System.getProperty("jep.pluginhome");
		Security.setProperty(ourPolicyProperty,
			"file:" + ourPluginHome + "/Contents/MacOS/JavaEmbeddingPlugin.policy");
		if (showDebugInfo) {
			System.err.println("AppletHandlerFactory: giveOurselvesAllPermissions(): ourPolicyProperty is " +
				ourPolicyProperty);
			System.err.println("AppletHandlerFactory: giveOurselvesAllPermissions(): ourPluginHome is " +
				ourPluginHome);
		}
	}

	// We should restrict access to the "jep" package, to avoid possible
	// security holes.
	protected static void restrictPackageAccess() {
		String packageAccess = Security.getProperty("package.access");
		packageAccess += ",jep.";
		Security.setProperty("package.access", packageAccess);
		if (showDebugInfo) {
			System.err.println("AppletHandlerFactory restrictPackageAccess(): " +
				"\"package.access\" now " + packageAccess);
		}
	}

	// If we're using a Java version earlier than Java 1.4.2 Update 1, we need
	// to load libAppletsInCocoa.jnilib, which contains native methods used by
	// other parts of Apple's JVM.  (If we're using Java 1.4.2 Update 1 or
	// later, the native methods have been moved to JavaPluginCocoa.bundle,
	// which we load using loadJavaPluginCocoa in AppletView.m.)
	protected static void loadAppleNatives() {
		if (isJava142Update1()) {
			appleNativesLoaded = true;
			return;
		}
		StringBuffer userPath = null;
		StringBuffer systemPath = null;
		if (isJava142()) {
			userPath = new StringBuffer().append(System.getProperty("user.home")).
				append("/Library/Internet Plug-Ins/JavaPluginCocoa.bundle/Contents/MacOS/libAppletsInCocoa.jnilib");
			systemPath = new StringBuffer().
				append("/Library/Internet Plug-Ins/JavaPluginCocoa.bundle/Contents/MacOS/libAppletsInCocoa.jnilib");
			try {
				System.load(userPath.toString());
				appleNativesLoaded = true;
				if (showDebugInfo) {
					System.err.println("AppletHandlerFactory loadAppleNatives: Loaded " +
						userPath.toString());
				}
			} catch (Throwable e) {
				try {
					System.load(systemPath.toString());
					appleNativesLoaded = true;
					if (showDebugInfo) {
						System.err.println("AppletHandlerFactory loadAppleNatives: Loaded " +
							systemPath.toString());
					}
				} catch (Throwable f) {
					System.err.println("AppletHandlerFactory loadAppleNatives: Couldn't load libAppletsInCocoa.jnilib");
				}
			}
			return;
		}
		try {
			systemPath = new StringBuffer().
				append("/System/Library/Frameworks/JavaVM.framework/Libraries/libAppletsInCocoa.jnilib");
			System.load(systemPath.toString());
			appleNativesLoaded = true;
			if (showDebugInfo)
				System.err.println("AppletHandlerFactory loadAppleNatives: Loaded " + systemPath.toString());
		} catch (Throwable e) {
			// This isn't a fatal error in Java 1.4.1.
			if (showDebugInfo)
				System.err.println("AppletHandlerFactory loadAppleNatives: Couldn't load libAppletsInCocoa.jnilib");
		}
	}

	public String toString() {
		return getClass().getName() + " (jep)";
	}

	static {
		javaVersion = System.getProperty("java.version");
		javaRuntimeVersion = System.getProperty("java.runtime.version");
		if ("true".equalsIgnoreCase(System.getProperty("jep.debuginfo", "false"))) {
			showDebugInfo = true;
		}
	}

}

