/*
 * Copyright (c) 2004, Steven Michaud, All Rights Reserved
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#import "Controller.h"

#ifdef __cplusplus
extern "C" {
#endif

extern NSLock *customEventSpoolLock;
extern NSMutableArray *customEventSpool;
extern NSLock *cEventSpoolLock;
extern NSMutableArray *cEventSpool;

@class JEPController;

@interface AppletView : NSView
{
@public
	NSString *appletDocbase;
	NSDictionary *appletAttrs;
	JEPController *myController;
	NSString *myAppletName;
	jobject myAppletHandler;
	jobject myApplet;
	jobject myJSObject;
	JavaVM *javaVM;
	int appletState;
	BOOL doingInitOnMainThread;
	BOOL isInvisible;
	BOOL visibilitySetExplicitly;
	BOOL isDestroying;
	BOOL awaitingAppletWindowCleanup;
	NSRect currentBounds;
	NSRect initialBounds;
	NSRecursiveLock *myLock;
	BOOL returnedCreatedInfo;
	BOOL returnedInitializedInfo;
	BOOL returnedStartedInfo;
}

+ (void)clipWithTranslation:(NSView *)drawRectView controller:(JEPController *)controller;
+ (BOOL)isJava142:(JNIEnv *)javaEnv;
+ (BOOL)isJava142Update1:(JNIEnv *)javaEnv;
+ (BOOL)isJava142Update2:(JNIEnv *)javaEnv;
+ (BOOL)isJava142Tiger:(JNIEnv *)javaEnv;
+ (BOOL)isJava150:(JNIEnv *)javaEnv;
+ (BOOL)isJava160:(JNIEnv *)javaEnv;
+ (NSString *)getJavaVersionString:(JNIEnv *)javaEnv;
+ (OSStatus)maybeCreateJavaVM:(JavaVMOption *)jvmExtraOpts
	extraOptsNum:(int)jvmExtraOptsNum vmPtr:(JavaVM **)vmPtr
	envPtr:(void **)envPtr initAWT:(BOOL)initAWT;
+ (void)registerJEPFactoryNatives:(JNIEnv *)javaEnv;
+ (void)registerJEPNatives:(JNIEnv *)javaEnv;
+ (void)setMainJavaThread:(JNIEnv *)javaEnv;
+ (void)loadAWTLibraries;
+ (NSMutableArray *)getRuntimeParams;
+ (NSMutableDictionary *)getRuntimeParamsAsDictionary;
+ (BOOL)getRuntimeParamBoolean:(NSString *)name;
+ (NSString *)getRuntimeParamString:(NSString *)name;
+ (BOOL)getJavaEnv:(JNIEnv **)envPtr;
+ (BOOL)getJavaEnv:(JNIEnv **)envPtr vmPtr:(JavaVM **)vmPtr;
+ (BOOL)getJavaEnvIfAttached:(JNIEnv **)envPtr;
+ (NSString *)toString:(jobject)javaThing javaEnv:(JNIEnv *)javaEnv;
+ (NSString *)getJavaThread:(JNIEnv *)javaEnv;
+ (OSStatus)printMessage:(NSString *)message javaEnv:(JNIEnv *)javaEnv;
+ (OSStatus)printStackTrace:(JNIEnv *)javaEnv;
+ (void)postFakeEvent;
+ (void)processCustomEventSpool:(BOOL)justOneEvent;
+ (void)performOnMainThread:(SEL)aSelector onTarget:(id)target
	withObject:(id)anObject waitUntilDone:(BOOL)wait;
+ (void)performOnMainThread:(SEL)aSelector onTarget:(id)target
	withObject:(id)anObject waitUntilDone:(BOOL)wait inModes:(NSArray *)modes;
+ (void)processCEventSpool:(BOOL)justOneEvent;
+ (void)performCEventOnMainThread:(void (*) (void *))runnable
	userInfo:(void *)userInfo waitUntilDone:(BOOL)wait;
+ (void)postJEPSendingEventsStartedNotification;
+ (void)postJEPSendingEventsEndedNotification;
- (BOOL)destroying;
- (BOOL)shouldProcessAllEvents:(JNIEnv *)javaEnv;
- (AppletView *)initWithParams:(JEPController *)controller
	docbase:(NSURL *)docbase attributes:(NSDictionary *)attributes;
- (void)deleteJNIGlobalRefs:(JNIEnv *)javaEnv;
- (BOOL)getJavaEnv:(JNIEnv **)envPtr;
- (BOOL)checkJNIException:(JNIEnv *)env;
- (void)handleJNIException:(jobject)throwable javaEnv:(JNIEnv *)env;
- (void)repaint;
- (void)validate;
- (void)invalidate;
- (BOOL)isValid;
- (void)validateClips;
- (void)zeroClips;
- (NSRect)translateCurrentClip:(NSView *)viewToClip;
- (void)activateFrame;
- (void)initApplet;
- (void)startApplet;
- (void)appletStartApplet;
- (void)stopApplet;
- (void)appletStopApplet;
- (void)destroyApplet;
- (void)resizeApplet:(int)newWidth newHeight:(int)newHeight;
- (void)printComponentInfo;
- (void)setSimpleLayout;
- (void)setIgnoreRepaint:(BOOL)flag;
- (void)setUpdatesOff:(BOOL)flag;
- (void)setAsyncUpdatesOff:(BOOL)flag;
- (void)setAppletVisible:(BOOL)visible flagOnly:(BOOL)flagOnly wait:(BOOL)wait;
- (BOOL)isAppletVisible;
- (jobject)getAppletFromJava;
- (int)getStatusFromJava;
- (NSRect)getAppletBounds;
- (void)setClip:(NSRect)clip;
- (void)imposeClip:(NSRect)clip;
- (void)unsetClip;
- (BOOL)isCurrentClipEmpty;
- (BOOL)areAppletBoundsEmpty;
- (BOOL)isDisplayable;
- (void)doMatchingLeftMouseUp:(EventRef)unmatchedLeftMouseDown;
/* The following are all "private" methods (they're made private by not
 * being listed in this header file).
 */
//- (void)doInitialization1:(JEPController *)controller;
//- (BOOL)doInitialization2;
//+ (void)doInitialization3:(AppletView *)aView;
//- (void)returnCreatedInfo;
//- (void)returnInitializedInfo;
//- (void)returnStartedInfo;
//- (void)maybeDoAppletStart;
//- (void)activateFrame;
//- (void)initApplet;

@end

@interface CustomEvent : NSObject
{
	id target;
	SEL selector;
	id object;
	BOOL wait;
	NSArray *modes;
	BOOL done;
}

- (id)initWithParams:(id)aTarget aSelector:(SEL)aSelector
	anObject:(id)anObject waitUntilDone:(BOOL)waitUntilDone
	modes:(NSArray *)modesArray;
- (id)getTarget;
- (SEL)getSelector;
- (id)getObject;
- (BOOL)getWaitUntilDone;
- (NSArray *)getModes;
- (BOOL)getIsDone;
- (void)dispatch;

@end

@interface CEvent : NSObject
{
	void (*runnable) (void *);
	void *info;
	BOOL wait;
	BOOL done;
}

- (id)initWithParams:(void (*) (void *))aRunnable userInfo:(void *)userInfo
	waitUntilDone:(BOOL)waitUntilDone;
- (void (*) (void *))getRunnable;
- (void *)getInfo;
- (BOOL)getWaitUntilDone;
- (BOOL)getIsDone;
- (void)dispatch;

@end

#ifdef __cplusplus
}
#endif
