/*
 * Copyright (c) 2006, Steven Michaud, All Rights Reserved
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* The contents of this file are needed by NSEvent _initWithCGSEvent:eventRef:,
 * which is in turn needed (along with GetEventPlatformEventRecord()) to
 * replace Apple's buggy mouse and keyboard event handlers for its Cocoa-
 * Carbon interface.  All the information here was extracted from the debug
 * versions of Apple's HIToolbox framework on Jaguar, Panther and Tiger.
 */
/* It also contains a bunch of other undocumented structures and functions
 * that I've deciphered from HIToolbox_debug, which have a number of uses.
 */

#ifdef __cplusplus
extern "C" {
#endif

/* These are available on Jaguar and Panther, but not on Tiger or Leopard.
 * (The functions are present in Tiger's and Leopard's HIToolbox framework
 * ... but they aren't "external" symbols!)
 */
extern void _DebugPrintEventQueue(EventQueueRef inEventQueue);
// Among other things, "prints" (to stdout) all of inEvent's parameters.
extern void _DebugPrintEvent(EventRef inEvent);

/* Local copy, which will either use a system call (if it's available) or
 * do the work directly.
 */
extern void DebugPrintEvent(EventRef inEvent);

typedef enum _CGSEventType {
	kCGSFirstEventType = 0,
	kCGSEventNull = 0,
	kCGSEventLeftMouseDown = 1,
	kCGSEventLeftMouseUp = 2,
	kCGSEventRightMouseDown = 3,
	kCGSEventRightMouseUp = 4,
	kCGSEventMouseMoved = 5,
	kCGSEventLeftMouseDragged = 6,
	kCGSEventRightMouseDragged = 7,
	kCGSEventMouseEntered = 8,
	kCGSEventMouseExited = 9,
	kCGSEventKeyDown = 10,
	kCGSEventKeyUp = 11,
	kCGSEventFlagsChanged = 12,
	kCGSEventKitDefined = 13,
	kCGSEventSystemDefined = 14,
	kCGSEventApplicationDefined = 15,
	kCGSEventTimer = 16,
	kCGSEventCursorUpdate = 17,
	kCGSEventJournaling = 18,
	kCGSEventSuspend = 19,
	kCGSEventResume = 20,
	kCGSEventNotification = 21,
	kCGSEventScrollWheel = 22,
	kCGSEventTabletPointer = 23,
	kCGSEventTabletProximity = 24,
	kCGSEventOtherMouseDown = 25,
	kCGSEventOtherMouseUp = 26,
	kCGSEventOtherMouseDragged = 27,
	kCGSLastEventType = 27,
} CGSEventType;

typedef struct _CGSEventRecord {
	unsigned char placeholder[256];
} CGSEventRecord;

struct _CGSTabletPointData {
    int x;
    int y;
    int z;
    unsigned short buttons;
    unsigned short pressure;
    struct {
        short x;
        short y;
    } tilt;
    unsigned short rotation;
    short tangentialPressure;
    unsigned short deviceID;
    short vendor1;
    short vendor2;
    short vendor3;
};

struct _CGSTabletProximityData {
    unsigned short vendorID;
    unsigned short tabletID;
    unsigned short pointerID;
    unsigned short deviceID;
    unsigned short systemTabletID;
    unsigned short vendorPointerType;
    unsigned int vendorSerialNumber;
    unsigned long long uniqueID;
    unsigned int capabilityMask;
    unsigned char pointerType;
    unsigned char enterProximity;
    short reserved1;
};

struct __CGEventSourceData {
    int source;
    unsigned int sourceUID;
    unsigned int sourceGID;
    int flags;
    unsigned long long userData;
    unsigned int sourceState;
    unsigned short localEventSuppressionInterval;
    unsigned char suppressionIntervalFlags;
    unsigned char remoteMouseDragFlags;
    unsigned long long serviceID;
};

struct _CGEventProcess {
    int pid;
    unsigned int psnHi;
    unsigned int psnLo;
    void *targetID;
    int flags;
};

typedef struct _CGSEventRecord_Leopard {
    unsigned short major;
    unsigned short minor;
    unsigned int length;
    unsigned int type;
    struct CGPoint location;
    struct CGPoint windowLocation;
    unsigned long long time;
    unsigned int flags;
    unsigned int window;
    unsigned int connection;
    struct __CGEventSourceData eventSource;
    struct _CGEventProcess target;
    union {
        struct {
            unsigned char subx;
            unsigned char suby;
            short eventNum;
            int click;
            unsigned char pressure;
            char buttonNumber;
            unsigned char subtype;
            unsigned char _reserved2;
            short deltaX;
            short deltaY;
            union {
                struct _CGSTabletPointData point;
                struct _CGSTabletProximityData proximity;
            } tablet;
            unsigned int instantMouser:1;
            unsigned int _bitPadding:31;
            unsigned int windowUnderMousePointer;
            unsigned int windowUnderMousePointerThatCanHandleThisEvent;
            int _padding[1];
        } mouse;
        struct {
            unsigned char subx;
            unsigned char suby;
            short eventNum;
            int click;
            unsigned char pressure;
            char buttonNumber;
            unsigned char subType;
            unsigned char _reserved2;
            short deltaX;
            short deltaY;
            union {
                struct _CGSTabletPointData point;
                struct _CGSTabletProximityData proximity;
            } tablet;
            unsigned int instantMouser:1;
            unsigned int _bitPadding:31;
            unsigned int windowUnderMousePointer;
            unsigned int windowUnderMousePointerThatCanHandleThisEvent;
            int _padding[1];
        } move;
        struct {
            unsigned short uniCharCount;
            short repeat;
            unsigned short charSet;
            unsigned short charCode;
            unsigned short keyCode;
            short keyData;
            short specialKey;
            unsigned char _padding;
            unsigned char uniCharFlags;
            unsigned int keyboardType;
            unsigned int uniKeyState;
            unsigned short uniChar[20];
        } key;
        struct {
            short reserved;
            short eventNum;
            int trackingNum;
            int userData;
            unsigned int triggerType;
            long long unknown1;
            long long unknown2;
            int _padding[8];
        } tracking;
        struct {
            unsigned short notifyCode;
            unsigned short flags;
            unsigned int targetHiPSN;
            unsigned int targetLoPSN;
            int status;
            int misc1;
            int misc2;
            int misc3;
            int misc4;
            int pid;
            int attributes;
            int flavor;
            int _padding[5];
        } process;
        struct {
            short deltaAxis1;
            short deltaAxis2;
            short deltaAxis3;
            short reserved1;
            int fixedDeltaAxis1;
            int fixedDeltaAxis2;
            int fixedDeltaAxis3;
            int pointDeltaAxis1;
            int pointDeltaAxis2;
            int pointDeltaAxis3;
            int _padding[4];
            unsigned int instantMouser:1;
            unsigned int unknown:1;
            unsigned int _bitPadding:30;
            unsigned int windowUnderMousePointer;
            unsigned int windowUnderMousePointerThatCanHandleThisEvent;
            int _padding2[1];
        } scrollWheel;
        struct _CGSTabletPointData tablet;
        struct _CGSTabletProximityData proximity;
        struct {
            short reserved;
            short subtype;
            union {
                float F[15];
                int L[15];
                short S[30];
                char C[60];
            } misc;
        } compound;
        struct {
            short unknown1;
            short unknown2;
            unsigned int unknown3;
            unsigned long long unknown4;
            int unknown5[12];
        } unknown;
    } data;
    unsigned int unknown1[6];
    unsigned short unknown2;
    unsigned short unknown3;
    unsigned short *unknown4;
} CGSEventRecord_Leopard;

typedef struct _CGSEventRecord_Tiger {
    unsigned short major;
    unsigned short minor;
    unsigned int length;
    unsigned int type;
    struct CGPoint location;
    struct CGPoint windowLocation;
    unsigned long long time;
    unsigned int flags;
    void *window;
    void *connection;
    struct __CGEventSourceData eventSource;
    struct _CGEventProcess target;
    union {
        struct {
            unsigned char subx;
            unsigned char suby;
            short eventNum;
            int click;
            unsigned char pressure;
            char buttonNumber;
            unsigned char subtype;
            unsigned char _reserved2;
            short deltaX;
            short deltaY;
            union {
                struct _CGSTabletPointData point;
                struct _CGSTabletProximityData proximity;
            } tablet;
            unsigned int instantMouser:1;
            unsigned int _bitPadding:31;
            void *windowUnderMousePointer;
            void *windowUnderMousePointerThatCanHandleThisEvent;
            int _padding[1];
        } mouse;
        struct {
            unsigned char subx;
            unsigned char suby;
            short eventNum;
            int click;
            unsigned char pressure;
            char buttonNumber;
            unsigned char subType;
            unsigned char _reserved2;
            short deltaX;
            short deltaY;
            union {
                struct _CGSTabletPointData point;
                struct _CGSTabletProximityData proximity;
            } tablet;
            unsigned int instantMouser:1;
            unsigned int _bitPadding:31;
            void *windowUnderMousePointer;
            void *windowUnderMousePointerThatCanHandleThisEvent;
            int _padding[1];
        } move;
        struct {
            unsigned short uniCharCount;
            short repeat;
            unsigned short charSet;
            unsigned short charCode;
            unsigned short keyCode;
            short keyData;
            short specialKey;
            unsigned char _padding;
            unsigned char uniCharFlags;
            unsigned int keyboardType;
            unsigned int uniKeyState;
            unsigned short uniChar[20];
        } key;
        struct {
            short reserved;
            short eventNum;
            int trackingNum;
            int userData;
            unsigned int triggerType;
            int _padding[12];
        } tracking;
        struct {
            unsigned short notifyCode;
            unsigned short flags;
            unsigned int targetHiPSN;
            unsigned int targetLoPSN;
            int status;
            int misc1;
            int misc2;
            int misc3;
            int misc4;
            int pid;
            int attributes;
            int flavor;
            int _padding[5];
        } process;
        struct {
            short deltaAxis1;
            short deltaAxis2;
            short deltaAxis3;
            short reserved1;
            int fixedDeltaAxis1;
            int fixedDeltaAxis2;
            int fixedDeltaAxis3;
            int pointDeltaAxis1;
            int pointDeltaAxis2;
            int pointDeltaAxis3;
            int _padding[4];
            unsigned int instantMouser:1;
            unsigned int _bitPadding:31;
            void *windowUnderMousePointer;
            void *windowUnderMousePointerThatCanHandleThisEvent;
            int _padding2[1];
        } scrollWheel;
        struct _CGSTabletPointData tablet;
        struct _CGSTabletProximityData proximity;
        struct {
            short reserved;
            short subtype;
            union {
                float F[15];
                int L[15];
                short S[30];
                char C[60];
            } misc;
        } compound;
    } data;
} CGSEventRecord_Tiger;

typedef struct _CGSEventRecord_Panther {
    unsigned short major;
    unsigned short minor;
    unsigned int length;
    unsigned int type;
    struct CGPoint location;
    struct CGPoint windowLocation;
    unsigned long long time;
    unsigned int flags;
    void *window;
    void *connection;
    union {
        struct {
            unsigned char subx;
            unsigned char suby;
            short eventNum;
            int click;
            unsigned char pressure;
            char buttonNumber;
            unsigned char subType;
            unsigned char _reserved2;
            short deltaX;
            short deltaY;
            union {
                struct _CGSTabletPointData point;
                struct _CGSTabletProximityData proximity;
            } tablet;
            unsigned int instantMouser:1;
            unsigned int _bitPadding:31;
            void *windowUnderMousePointer;
            void *windowUnderMousePointerThatCanHandleThisEvent;
            int _padding[1];
        } mouse;
        struct {
            unsigned char subx;
            unsigned char suby;
            short eventNum;
            int click;
            unsigned char pressure;
            char buttonNumber;
            unsigned char subType;
            unsigned char _reserved2;
            short deltaX;
            short deltaY;
            union {
                struct _CGSTabletPointData point;
                struct _CGSTabletProximityData proximity;
            } tablet;
            unsigned int _instantMouser:1;
            unsigned int _bitPadding:31;
            void *windowUnderMousePointer;
            void *windowUnderMousePointerThatCanHandleThisEvent;
            int _padding[1];
        } move;
        struct {
            short reserved;
            short repeat;
            unsigned short charSet;
            unsigned short charCode;
            unsigned short keyCode;
            short keyData;
            short specialKey;
            short _pad;
            unsigned int keyboardType;
            int _padding[11];
        } key;
        struct {
            short reserved;
            short eventNum;
            int trackingNum;
            int userData;
            unsigned int triggerType;
            int _padding[12];
        } tracking;
        struct {
            unsigned short notifyCode;
            unsigned short flags;
            unsigned int targetHiPSN;
            unsigned int targetLoPSN;
            int status;
            int misc1;
            int misc2;
            int misc3;
            int misc4;
            int pid;
            int attributes;
            int flavor;
            int _padding[5];
        } process;
        struct {
            short deltaAxis1;
            short deltaAxis2;
            short deltaAxis3;
            short reserved1;
            int reserved2;
            int _padding[9];
            unsigned int instantMouser:1;
            unsigned int _bitPadding:31;
            void *windowUnderMousePointer;
            void *windowUnderMousePointerThatCanHandleThisEvent;
            int _padding2[1];
        } scrollWheel;
        struct {
            int x;
            int y;
            int z;
            unsigned short buttons;
            unsigned short pressure;
            struct {
                short x;
                short y;
            } tilt;
            unsigned short rotation;
            short tangentialPressure;
            unsigned short deviceID;
            short vendor1;
            short vendor2;
            short vendor3;
            int _padding[8];
        } tablet;
        struct {
            unsigned short vendorID;
            unsigned short tabletID;
            unsigned short pointerID;
            unsigned short deviceID;
            unsigned short systemTabletID;
            unsigned short vendorPointerType;
            unsigned int pointerSerialNumber;
            unsigned long long uniqueID;
            unsigned int capabilityMask;
            unsigned char pointerType;
            unsigned char enterProximity;
            short reserved1;
            int _padding[8];
        } proximity;
        struct {
            short reserved;
            short subtype;
            union {
                float F[15];
                int L[15];
                short S[30];
                char C[60];
            } misc;
        } compound;
    } data;
} CGSEventRecord_Panther;

typedef struct _CGSEventRecord_Jaguar {
    unsigned short major;
    unsigned short minor;
    unsigned int length;
    unsigned int type;
    struct CGPoint location;
    struct CGPoint windowLocation;
    unsigned long long time;
    unsigned int flags;
    void *window;
    void *connection;
    union {
        struct {
            unsigned char subx;
            unsigned char suby;
            short eventNum;
            int click;
            unsigned char pressure;
            char buttonNumber;
            unsigned char subType;
            unsigned char _reserved2;
            short deltaX;
            short deltaY;
            union {
                struct _CGSTabletPointData point;
                struct _CGSTabletProximityData proximity;
            } tablet;
            unsigned int instantMouser:1;
            unsigned int _bitPadding:31;
            int _padding[3];
        } mouse;
        struct {
            unsigned char subx;
            unsigned char suby;
            short eventNum;
            int click;
            unsigned char pressure;
            char buttonNumber;
            unsigned char subType;
            unsigned char _reserved2;
            short deltaX;
            short deltaY;
            union {
                struct _CGSTabletPointData point;
                struct _CGSTabletProximityData proximity;
            } tablet;
            unsigned int instantMouser:1;
            unsigned int _bitPadding:31;
            int _padding[3];
        } move;
        struct {
            short reserved;
            short repeat;
            unsigned short charSet;
            unsigned short charCode;
            unsigned short keyCode;
            short keyData;
            short specialKey;
            short _pad;
            unsigned int keyboardType;
            int _padding[11];
        } key;
        struct {
            short reserved;
            short eventNum;
            int trackingNum;
            int userData;
            int _padding[13];
        } tracking;
        struct {
            unsigned short notifyCode;
            unsigned short flags;
            unsigned int targetHiPSN;
            unsigned int targetLoPSN;
            int status;
            int misc1;
            int misc2;
            int misc3;
            int misc4;
            int pid;
            int _padding[7];
        } process;
        struct {
            short deltaAxis1;
            short deltaAxis2;
            short deltaAxis3;
            short reserved1;
            int reserved2;
            int _padding[9];
            unsigned int instantMouser:1;
            unsigned int _bitPadding:31;
            int _padding2[3];
        } scrollWheel;
        struct {
            int x;
            int y;
            int z;
            unsigned short buttons;
            unsigned short pressure;
            struct {
                short x;
                short y;
            } tilt;
            unsigned short rotation;
            short tangentialPressure;
            unsigned short deviceID;
            short vendor1;
            short vendor2;
            short vendor3;
            int _padding[8];
        } tablet;
        struct {
            unsigned short vendorID;
            unsigned short tabletID;
            unsigned short pointerID;
            unsigned short deviceID;
            unsigned short systemTabletID;
            unsigned short vendorPointerType;
            unsigned int pointerSerialNumber;
            unsigned long long uniqueID;
            unsigned int capabilityMask;
            unsigned char pointerType;
            unsigned char enterProximity;
            short reserved1;
            int _padding[8];
        } proximity;
        struct {
            short reserved;
            short subtype;
            union {
                float F[15];
                int L[15];
                short S[30];
                char C[60];
            } misc;
        } compound;
    } data;
} CGSEventRecord_Jaguar;

NSEvent *CreateNSEvent(CGSEventRecord *eventRec, EventRef anEvent);

/* A kEventParamPostOptions parameter is added to an event that was sent using
 * SendToEventTargetWithOptions() with non-standard options.
 */
enum {
	kEventParamPostOptions = 'popt',	// typeEventTargetOptions
	typeEventTargetOptions = 'etop'
};

/* As best I can tell, the OS always sends an event of class kEventClassTarget
 * and kind kEventTargetGetNext to a (prospective) target before the OS sends
 * a "real" event to that target.  The kEventTargetGetNext event usually
 * (always?) has a single parameter kEventParamEventRef (of typeEventRef) --
 * the "real" event in question.  If you add a kEventParamEventTarget
 * parameter, the OS will try to send the event to that target instead (you'll
 * override the OS's default behavior).  So if you register for this event on
 * enough targets (I use the event dispatcher and application targets), you
 * should see all "real" events without having to register for them
 * specifically (though some events do get sent to both the event dispatcher
 * and application targets).  (Actually there are some exceptions -- for
 * example you don't see kEventWindowUpdate events.)
 */
enum {
	kEventClassTarget = 'etrg',
	kEventTargetGetNext = 1,
	kEventTargetInstallStandard = 2,
	kEventTargetRemoveStandard = 3,
	kEventTargetHandlerInstalled = 4,	// No parameters
	kEventTargetHandlerRemoved = 5,		// ? only used on Tiger, no parameters
	kEventParamEventTarget = 'etrg',	// typeEventTargetRef
};

// What's used to store the data for each event parameter on Panther and Tiger.
typedef struct _EventParam {
	UInt32 name;
	UInt32 type;
	UInt32 size;
	UInt32 flags;
	unsigned char data[16];
} EventParam;

// What's used to store the data for each event parameter on Jaguar.
typedef struct _AVLNode *AVLNodePtr;
typedef struct _AVLNode {
    AVLNodePtr subtree[2];
    short int bal;
    OSType signature;
    OSType key;
    OSType dataType;
    UInt32 dataSize;
    UInt8 flags;
    char data[4];
} AVLNode;

// Used for EventHandler.fIndex (Tiger only).
typedef struct _TEventTypeIndex {
	EventTypeSpec *fIndex;
	UInt32 fCount;
} TEventTypeIndex;

/* The "private" contents of an EventHandlerRef.  You can recursively follow
 * the "next" field to see all the handlers installed on a given target before
 * your handler was installed on it.  You can also invoke any of these handlers
 * yourself (using the appropriate handler's "proc" and "userData" fields, and
 * creating a usable EventHandlerCall structure).
 */

typedef struct _EventHandler_Jaguar {
	EventTargetRef owner;
	EventHandlerRef next;
	EventHandlerUPP proc;
	void *userData;
	unsigned char kind;
    CFStringRef debugName;
} EventHandler_Jaguar;

typedef struct _EventHandler_Panther {
	EventTargetRef owner;
	EventHandlerRef next;
	EventHandlerUPP proc;
	void *userData;
	unsigned char kind;
    CFStringRef debugName;
} EventHandler_Panther;

typedef struct _EventHandler_Tiger {
	EventTargetRef owner;
	EventHandlerRef next;
	EventHandlerUPP proc;
	void *userData;
	unsigned char kind;
	UInt16 dispatchRefCount;
	Boolean removedDuringDispatch;
	CFStringRef debugName;
	TEventTypeIndex fIndex;
	UInt32 fExternalRef; // _Not_ an EventHandlerRef (against HIToolbox_debug).
} EventHandler_Tiger;

/* Structures used by the "index" field in the private contents of an
 * EventHandlerRef.  This "index" appears to be arranged by eventClass
 * and eventKind.
 */

typedef struct _FPListItem_Jaguar {
	UInt32 eventClass;
	UInt16 eventKindBegin;
	UInt16 eventKindEnd;
	EventHandlerRef handler;
} FPListItem_Jaguar;

typedef struct _FPListItem_Panther {
	UInt32 eventClass;
	UInt32 eventKind;
	EventHandlerRef handler;
	UInt32 whatever;
} FPListItem_Panther;

typedef struct _FastPointerList {
	short unsigned int currCount;
	short unsigned int maxCount;
	void *list;
} FPList, HandlerKindIndex, *HandlerIndex, *HandlerStackPool;

/* The "private" contents of an EventTargetRef.  The "installedHandlers" field
 * is just the most recently installed handler.  To get all of them, recurse
 * through the "next" field.
 */

typedef struct _EventTarget_Jaguar {
	void *targetData;
	UInt16 refCount;
	EventTargetRef (*getNextTargetHook) (void);
	void (*installStdHandlerHook) (void);
	void (*removeStdHandlerHook) (void);
	EventHandlerUPP defaultHandler;
	void *defaultHandlerData;
	CFStringRef debugName;
	Boolean traceEvents;
	EventHandlerRef installedHandlers;
    HandlerIndex index;
} EventTarget_Jaguar;

typedef struct _EventTarget_Panther {
	UInt16 refCount;
	void (*handlerInstalledHook) (void);
	void (*handlerRemovedHook) (void);
    EventHandlerUPP defaultHandler;
    void *defaultHandlerData;
    CFStringRef debugName;
    Boolean traceEvents;
    EventHandlerRef installedHandlers;
    HandlerIndex index;
} EventTarget_Panther;

/* The "index" field is unused (presumably because sUseHandlerIndex (a non-
 * external Boolean) is set to "false"), and sometimes contains garbage.  It's
 * functionality is superceded (on Tiger) by EventHandler.fIndex.
 */
typedef struct _EventTarget_Tiger {
	UInt16 refCount;
	Boolean traceEvents;
	UInt8 dispatchFlags;
	UInt32 modificationCount : 32;
	void (*handlerInstalledHook) (void);
	void (*handlerRemovedHook) (void);
	EventTargetRef (*nextTargetHook) (void);
	void *hookRefcon;
	EventHandlerUPP defaultHandler;
	void *defaultHandlerData;
	CFStringRef debugName;
    EventHandlerRef installedHandlers;
    HandlerIndex index; // Sometimes contains garbage.  Not used.
    HandlerStackPool stackPool; // Sometimes contains garbage.  Not used?
} EventTarget_Tiger;

// The "private" contents of an EventHandlerCallRef.
typedef struct _EventHandlerCall {
	EventTargetRef currTarget;
	EventHandlerRef currHandler;
	Boolean calledThru;
	UInt32 targetData;
} EventHandlerCall;

/* EventPriv.next and EventPriv.prev may either point to real events or to
 * one of the following parts of the EventQueue structure -- "header",
 * "priorityTails.highPriorityTail", or "priorityTails.standardPriorityTail".
 * The latter have the same size as an EventPriv structure, and use a few
 * of its fields (notably "next", "prev", "priority" and sometimes "classID").
 * But basically they seem to be placeholders.
 */

// The "private" contents of an EventRef on Jaguar.
typedef struct _EventPriv_Jaguar {
	EventQueueRef queue;
	EventRef next;
	EventRef prev;
	UInt32 classID;
	UInt32 kind;
	UInt32 flags;
	double time;
	UInt32 priority;
	UInt32 owners;
	UInt32 attributes;
	UInt32 busyCount;
	Boolean hasPlatEvent;
	CGSEventRecord_Jaguar platEvent;
	AVLNodePtr parameters;
	CFAllocatorRef allocator;
	UInt32 lock; // pthread_mutex_t * ?
} EventPriv_Jaguar;

// The "private" contents of an EventRef on Panther.
typedef struct _EventPriv_Panther {
	EventQueueRef queue;
	EventRef next;
	EventRef prev;
	UInt32 classID;
	UInt32 kind;
	UInt32 flags;
	double time;
	UInt32 priority;
	UInt32 owners;
	UInt32 attributes;
	UInt32 busyCount;
	CGSEventRecord_Panther platEvent;
	UInt32 numSlots;
	UInt32 numUsed;
	EventParam *parameters;
	CFAllocatorRef allocator;
	UInt32 lock; // pthread_mutex_t * ?
} EventPriv_Panther;

// The "private" contents of an EventRef on Tiger.
typedef struct _EventPriv_Tiger {
	EventQueueRef queue;
	EventRef next;
	EventRef prev;
	UInt32 classID;
	UInt32 kind;
	UInt32 flags;
	double time;
	uint64_t machTime;
	UInt32 priority;
	UInt32 owners;
	UInt32 attributes;
	UInt32 busyCount;
	CGSEventRecord_Tiger platEvent;
	UInt32 numSlots;
	UInt32 numUsed;
	EventParam *parameters;
	CFAllocatorRef allocator;
	pthread_mutex_t lock;
} EventPriv_Tiger;

#ifdef __cplusplus
}
#endif
