//Copyright (c) 2004, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// We need to be part of the apple.laf package to have access to
// AquaImageFactory's static CColorPaintUIResource variables.
package apple.laf;

import jep.*;

// With Java 1.4.2 in a Carbon browser, the Java equivalents of Apple's
// system colors often don't work properly in Swing applets (especially if
// they belong to the CColorPaint or CColorPaintUIResource classes).  The
// first time they're displayed the colors are correct, but if you change the
// key window or key view they sometimes turn black.  I don't know the
// underlying cause -- though presumably Apple has forgotten to run one or
// more initialization routines.  But I've found a fix for all the colors I
// have access to (i.e. that I'm able to change) -- replace them with colors
// that have the same RGB values, plus an NSColor object (re)generated
// directly from those RGB values.  See fixLafColorDefaults() and FixStuff
// in AppletFramePanel.java for more information.  (This problem no longer
// exists in Java 1.4.2 Update 1.)
public class FixAquaImageFactory {
	static private boolean fixed = false;

	public FixAquaImageFactory() {
	}

	public static synchronized void fix() {
		if (fixed)
			return;
		fixed = true;

		CColorPaintUIResource color =
			AquaImageFactory.getTextSelectionBackgroundColorUIResource();
		AquaImageFactory.sTextSelectionBackgroundColor = new
			CColorPaintUIResource(
			AppletHandlerFactory.getNSColor(color.getRed(), color.getGreen(), color.getBlue()),
			color.getRed(), color.getGreen(), color.getBlue());
		color = AquaImageFactory.getTextSelectionForegroundColorUIResource();
		AquaImageFactory.sTextSelectionForegroundColor = new
			CColorPaintUIResource(
			AppletHandlerFactory.getNSColor(color.getRed(), color.getGreen(), color.getBlue()),
			color.getRed(), color.getGreen(), color.getBlue());
		color = AquaImageFactory.getWindowBackgroundColorUIResource();
		AquaImageFactory.sBackgroundColor = new
			CColorPaintUIResource(
			AppletHandlerFactory.getNSColor(color.getRed(), color.getGreen(), color.getBlue()),
			color.getRed(), color.getGreen(), color.getBlue());
		color = AquaImageFactory.getMenuSelectionForegroundColorUIResource();
		AquaImageFactory.sMenuSelectionForegroundColor = new
			CColorPaintUIResource(
			AppletHandlerFactory.getNSColor(color.getRed(), color.getGreen(), color.getBlue()),
			color.getRed(), color.getGreen(), color.getBlue());
		color = AquaImageFactory.getMenuSelectionBackgroundColorUIResource();
		AquaImageFactory.sMenuSelectionBackgroundColor = new
			CColorPaintUIResource(
			AppletHandlerFactory.getNSColor(color.getRed(), color.getGreen(), color.getBlue()),
			color.getRed(), color.getGreen(), color.getBlue());
		color = AquaImageFactory.getMenuBackgroundColorUIResource();
		AquaImageFactory.sMenuBackgroundColor = new
			CColorPaintUIResource(
			AppletHandlerFactory.getNSColor(color.getRed(), color.getGreen(), color.getBlue()),
			color.getRed(), color.getGreen(), color.getBlue());
		color = AquaImageFactory.getDisabledControlTextColorUIResource();
		AquaImageFactory.sDisabledControlTextColor = new
			CColorPaintUIResource(
			AppletHandlerFactory.getNSColor(color.getRed(), color.getGreen(), color.getBlue()),
			color.getRed(), color.getGreen(), color.getBlue());
		color = AquaImageFactory.getTabBackgroundColorUIResource();
		AquaImageFactory.sTabBackgroundColor = new
			CColorPaintUIResource(
			AppletHandlerFactory.getNSColor(color.getRed(), color.getGreen(), color.getBlue()),
			color.getRed(), color.getGreen(), color.getBlue());
	}

}

