//Copyright (c) 2004, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// This class is in the apple.awt package mostly because its friends are, too,
// and because of the slight increase in security caused by the fact that
// Apple's default java.security file protects the apple.awt package from
// ordinary access.
package apple.awt;

import jep.*;
import java.awt.*;
import java.awt.event.*;
import sun.java2d.SunGraphics2D;
import sun.java2d.SurfaceData;

// See description of MyCToolkit class.

public class MyCCanvas141 extends CCanvas {
	protected static boolean showDebugInfo = false;
	protected AppletFrame ancestor = null;

	public static MyCCanvas141 createNew(Canvas canvas, long innerView, AppletFrame ancestor) {
		MyCCanvas141 newPeer = new MyCCanvas141(canvas, innerView);
		newPeer.setAncestor(ancestor);
		return newPeer;
	}

	public MyCCanvas141(Canvas canvas, long innerView) {
		super(canvas, (int) innerView);
		if (showDebugInfo) {
			System.err.println("Called MyCCanvas141(): " +
				"Canvas (" + ((canvas == null) ? "null" : canvas.toString()) + ") " +
				"innerView (0x" + Long.toHexString(innerView) + ")");
		}
	}

	public void setAncestor(AppletFrame frame) {
		ancestor = frame;
	}

	public int initPeer(int view) {
		int retval = super.initPeer(view);
		fPeerPainterLOCKABLE = new MyPeerPaintEvent141(this, getViewPtr());
		return retval;
	}

	public void handleEvent(AWTEvent awtevent) {
		if (awtevent == null)
			return;
		if (!(awtevent instanceof PaintEvent)) {
			super.handleEvent(awtevent);
			return;
		}
		if (ancestor == null) {
			super.handleEvent(awtevent);
			return;
		}
		if (!AppletHandlerFactory.isAppletViewDisplayable(ancestor.getCocoaParentViewLong())) {
			if (showDebugInfo)
				System.err.println("MyCCanvas141 handleEvent(): Skipping paint event because applet invisible");
			return;
		}
		if (ancestor.updatesOff) {
			if (showDebugInfo)
				System.err.println("MyCCanvas141 handleEvent(): Skipping paint event because updates off");
			return;
		}
		Rectangle clipRect = ancestor.translateCurrentClip(fTarget);
		Rectangle updateRect = ((PaintEvent)awtevent).getUpdateRect();
		if (updateRect == null) {
			super.handleEvent(awtevent);
			return;
		}
		clipRect = clipRect.intersection(updateRect);
		if (showDebugInfo) {
			System.err.println("MyCCanvas141 handleEvent(): updateRect clipped from " +
				updateRect.toString() + " to " + clipRect.toString());
		}
		if (!clipRect.isEmpty()) {
			((PaintEvent)awtevent).setUpdateRect(clipRect);
			super.handleEvent(awtevent);
		}
	}

	// Standard clipping doesn't work here -- I suppose because other code
	// replaces the user clip before the graphics object we create here gets
	// used.  Resetting the "device clip" works just fine, though.
	public Graphics getGraphics() {
		SunGraphics2D g = (SunGraphics2D) super.getGraphics();
		GraphicsClipper clipper = null;
		try {
			Object[] parms = {g, ancestor, this};
			clipper = (GraphicsClipper) AppletHandlerFactory.
				graphicsClipperCreateNew.invoke(null, parms);
		} catch (Throwable e) {
			System.err.println("MyCCanvas141 getGraphics(): exception: " +
				e.getMessage());
			e.printStackTrace();
			return (Graphics) g;
		}
		clipper.validateDevClip();
		return (Graphics) g;
	}

	public String toString() {
		return getClass().getName() + " (jep)";
	}

	static {
		if ("true".equalsIgnoreCase(System.getProperty("jep.debuginfo", "false"))) {
			showDebugInfo = true;
		}
	}

}

