//Copyright (c) 2004, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// As of Apple's Java 1.4.2 Update 1 DP1, this class needs to be in the
// same package as CPanel (i.e. apple.awt).  That's because CPanel's
// constructor is no longer "public", and in fact has no access modifier
// at all -- which means that it can only be accessed from within the same
// package.  Fortunately we're allowed to insinuate this class into the
// apple.awt package -- it may help that we give ourselves "all permissions"
// in AppletHandlerFactory.initProperties(), and that our jarfile is in the
// JVM's xbootclasspath.
package apple.awt;

import jep.*;
import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;
import sun.awt.RepaintArea;
import sun.java2d.SunGraphics2D;
import sun.java2d.SurfaceData;

// See description of MyCToolkit class.

public class MyCPanel extends CPanel {
	protected static boolean showDebugInfo = false;
	protected AppletFrame ancestor = null;

	public MyCPanel(Component component) {
		super(component);
		if (showDebugInfo) {
			System.err.println("Called MyCPanel(): " +
				"Component (" + ((component == null) ? "null" : component.toString()) + ")");
		}
	}

	public void setAncestor(AppletFrame frame) {
		ancestor = frame;
	}

	public void handleEvent(AWTEvent awtevent) {
		if (awtevent == null)
			return;
		if (!(awtevent instanceof PaintEvent)) {
			super.handleEvent(awtevent);
			return;
		}
		if (ancestor == null) {
			super.handleEvent(awtevent);
			return;
		}
		if (!AppletHandlerFactory.isAppletViewDisplayable(ancestor.getCocoaParentViewLong())) {
			if (showDebugInfo)
				System.err.println("MyCPanel handleEvent(): Skipping paint event because applet invisible");
			return;
		}
		if (ancestor.updatesOff) {
			if (showDebugInfo)
				System.err.println("MyCPanel handleEvent(): Skipping paint event because updates off");
			return;
		}
		Rectangle clipRect = ancestor.translateCurrentClip(fTarget);
		Rectangle updateRect = ((PaintEvent)awtevent).getUpdateRect();
		if (updateRect == null) {
			super.handleEvent(awtevent);
			return;
		}
		clipRect = clipRect.intersection(updateRect);
		if (showDebugInfo) {
			System.err.println("MyCPanel handleEvent(): updateRect clipped from " +
				updateRect.toString() + " to " + clipRect.toString());
		}
		if (!clipRect.isEmpty()) {
			if (AppletHandlerFactory.isJava142Update1()) {
				RepaintArea temp = new RepaintArea();
				temp.add(clipRect, awtevent.getID());
				paintArea = temp;
			}
			((PaintEvent)awtevent).setUpdateRect(clipRect);
			super.handleEvent(awtevent);
		}
	}

	// Standard clipping doesn't work here -- I suppose because other code
	// replaces the user clip before the graphics object we create here gets
	// used.  Resetting the "device clip" works just fine, though.
	public Graphics getGraphics() {
		SunGraphics2D g = (SunGraphics2D) super.getGraphics();
		GraphicsClipper clipper = null;
		try {
			Object[] parms = {g, ancestor, this};
			clipper = (GraphicsClipper) AppletHandlerFactory.
				graphicsClipperCreateNew.invoke(null, parms);
		} catch (Throwable e) {
			System.err.println("MyCPanel getGraphics(): exception: " +
				e.getMessage());
			e.printStackTrace();
			return (Graphics) g;
		}
		clipper.validateDevClip();
		return (Graphics) g;
	}

	public String toString() {
		return getClass().getName() + " (jep)";
	}

	static {
		if ("true".equalsIgnoreCase(System.getProperty("jep.debuginfo", "false"))) {
			showDebugInfo = true;
		}
	}

}

