//Copyright (c) 2004, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// Unlike the MyCAppletFrame, MyCPanel and MyCCanvas classes, this one doesn't
// need to be in the apple.awt package.  But Apple's default java.security
// file protects the apple.awt package from ordinary user access, and while
// the extra security doesn't really seem to be necessary, it might help.
package apple.awt;

import jep.*;
import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.awt.peer.*;
import java.util.*;
import com.apple.eawt.CocoaComponent;
import sun.awt.SunToolkit;
import sun.awt.AppContext;
import sun.awt.PeerEvent;

// All Java-initiated (i.e. ansynchronous) drawing goes through the
// apple.awt.ContainerModel class's handleEvent() method.  (This also goes
// through its addDamagedArea() method, along with additional drawing from
// PeerPaintEvent.paintFromCocoaDrawRect(), which is probably all Cocoa-
// initiated.)  If we could override handleEvent() we'd have an ideal spot to
// clip asynchronous drawing events or (if updates were off) to discard them
// altogether.  I haven't discovered a way to override this class directly,
// but by other means we can do something almost as useful -- we can subclass
// CToolkit and substitute an instance of our subclass for the single "system"
// CToolkit (see AppletHandlerFactory.java's initProperties() method).  Then
// we can override subclasses of ContainerModel.  The CPanel class is the most
// useful of these -- it's possible that it's the only one we really need to
// subclass for asynchronous drawing.  But I've also subclassed two others
// that look like they might be important.

// As of Java 1.4.2 Update 1, the key class (the one we'd want to subclass) is
// apple.awt.ComponentModel instead of apple.awt.ContainerModel (the former is
// the latter's superclass).  Otherwise most of our strategy is the same ...
// though I've discovered that it's useful (and in Update 1 also necessary) to
// also override getGraphics().  (The reason it's necessary in Update 1 is
// that not all paint events go through handleEvent(), and as of Update 1 it
// stopped being possible to clip these events in Handlers.m's
// lockFocusIfCanDraw_AWT() method.  Fortunately all paint events _do_ seem to
// go through getGraphics().)

public class MyCToolkit extends CToolkit {
	private static boolean showDebugInfo = false;

	public MyCToolkit() {
		if (showDebugInfo)
			System.err.println("Calling MyCToolkit()");
		AppletHandlerFactory.resetMainEventQueue();
	}

	public FramePeer createFrame(Frame frame) {
		if (frame == null)
			return super.createFrame(frame);
		FramePeer peer = null;
		if (frame instanceof AppletFrame) {
			try {
				Object[] parms = {frame, frame};
				peer = (FramePeer) AppletHandlerFactory.
					myCAppletFrameCreateNew.invoke(null, parms);
			} catch (Throwable e) {
				System.err.println("MyCToolkit createFrame(): exception: " +
					e.getMessage());
				e.printStackTrace();
				return null;
			}
			SunToolkit.targetCreatedPeer(frame, peer);
		} else {
			peer = super.createFrame(frame);
			AppletFrame appletFrame = AppletFrame.getCurrentAppletFrame();
			if (appletFrame != null)
				appletFrame.addChildWindow(frame);
		}
		if (showDebugInfo) {
			System.err.println("MyCToolkit createFrame(): Frame " +
				peer.toString() + " thread " + Thread.currentThread().toString());
		}
		return peer;
	}

	public WindowPeer createWindow(Window window) {
		WindowPeer peer = super.createWindow(window);
		if (window != null) {
			AppletFrame appletFrame = AppletFrame.getCurrentAppletFrame();
			if (appletFrame != null)
				appletFrame.addChildWindow(window);
		}
		return peer;
	}

	public PanelPeer createPanel(Panel panel) {
		if (panel == null)
			return super.createPanel(panel);
		Component ancestor = panel;
		for (;;) {
			Component next = (Component) ancestor.getParent();
			if (next == null)
				break;
			ancestor = next;
		}
		PanelPeer peer = null;
		if (ancestor instanceof AppletFrame) {
			try {
				Object[] parms = {panel, ancestor};
				peer = (PanelPeer) AppletHandlerFactory.
					myCPanelCreateNew.invoke(null, parms);
			} catch (Throwable e) {
				System.err.println("MyCToolkit createPanel(): exception: " +
					e.getMessage());
				e.printStackTrace();
				return null;
			}
			SunToolkit.targetCreatedPeer(panel, peer);
		} else {
			peer = super.createPanel(panel);
		}
		if (showDebugInfo) {
			System.err.println("MyCToolkit createPanel(): Panel " +
				peer.toString() + " thread " + Thread.currentThread().toString());
		}
		return peer;
	}

	public ScrollPanePeer createScrollPane(ScrollPane scrollpane) {
		if (scrollpane == null)
			return super.createScrollPane(scrollpane);
		Component ancestor = scrollpane;
		for (;;) {
			Component next = (Component) ancestor.getParent();
			if (next == null)
				break;
			ancestor = next;
		}
		ScrollPanePeer peer = null;
		if (ancestor instanceof AppletFrame) {
			try {
				Object[] parms = {scrollpane, ancestor};
				peer = (ScrollPanePeer) AppletHandlerFactory.
					myCScrollPaneCreateNew.invoke(null, parms);
			} catch (Throwable e) {
				System.err.println("MyCToolkit createScrollPane(): exception: " +
					e.getMessage());
				e.printStackTrace();
				return null;
			}
			SunToolkit.targetCreatedPeer(scrollpane, peer);
		} else {
			peer = super.createScrollPane(scrollpane);
		}
		if (showDebugInfo) {
			System.err.println("MyCToolkit createScrollPane(): ScrollPane " +
				peer.toString() + " thread " + Thread.currentThread().toString());
		}
		return peer;
	}

	public CanvasPeer createCanvas(Canvas canvas) {
		if (canvas == null)
			return super.createCanvas(canvas);
		Component ancestor = canvas;
		for (;;) {
			Component next = (Component) ancestor.getParent();
			if (next == null)
				break;
			ancestor = next;
		}
		CanvasPeer peer = null;
		long nsView = 0;
		if (ancestor instanceof AppletFrame) {
			if (canvas instanceof CocoaComponent) {
				nsView = (long) ((CocoaComponent)canvas).createNSView();
			}
			try {
				Object[] parms = {canvas, new Long(nsView), ancestor};
				peer = (CanvasPeer) AppletHandlerFactory.
					myCCanvasCreateNew.invoke(null, parms);
			} catch (Throwable e) {
				System.err.println("MyCToolkit createCanvas(): exception: " +
					e.getMessage());
				e.printStackTrace();
				return null;
			}
			SunToolkit.targetCreatedPeer(canvas, peer);
		} else {
			peer = super.createCanvas(canvas);
		}
		if (showDebugInfo) {
			System.err.println("MyCToolkit createCanvas(): (nsView is 0x" +
				Long.toHexString(nsView) + ") Canvas " + peer.toString() +
				" thread " + Thread.currentThread().toString());
		}
		return peer;
	}

	public LabelPeer createLabel(Label label) {
		if ((label == null) || !AppletHandlerFactory.isJava142Tiger())
			return super.createLabel(label);
		Component ancestor = label;
		for (;;) {
			Component next = (Component) ancestor.getParent();
			if (next == null)
				break;
			ancestor = next;
		}
		LabelPeer peer = null;
		if (ancestor instanceof AppletFrame) {
			try {
				Object[] parms = {label, ancestor};
				peer = (LabelPeer) AppletHandlerFactory.
					myCLabelCreateNew.invoke(null, parms);
			} catch (Throwable e) {
				System.err.println("MyCToolkit createLabel(): exception: " +
					e.getMessage());
				e.printStackTrace();
				return null;
			}
			SunToolkit.targetCreatedPeer(label, peer);
		} else {
			peer = super.createLabel(label);
		}
		if (showDebugInfo) {
			System.err.println("MyCToolkit createLabel(): Label " +
				peer.toString() + " thread " + Thread.currentThread().toString());
		}
		return peer;
	}

	// Bizarrely, in Java 1.6.0 we can get hangs when we create a new
	// MOUSE_MOVED event on the main thread.  So now we make sure that this
	// event is created on the applet's event-dispatcher thread.
	public static boolean postMouseMovedEvent(Object peer, int x, int y, long when) {
		if (peer instanceof CComponent) {
			Component ancestor = ((CComponent) peer).fTarget;
			for (;;) {
				Component next = (Component) ancestor.getParent();
				if (next == null)
					break;
				ancestor = next;
			}
			if (!(ancestor instanceof AppletFrame))
				return false;
			if (((AppletFrame)ancestor).destroying)
				return true;
			Applet applet = ((AppletFrame)ancestor).getApplet();
			AppContext appletContext = null;
			if (applet != null)
				appletContext = SunToolkit.targetToAppContext(applet);
			boolean setDirectly = ((appletContext == null) ||
				AppContext.getAppContext().equals(appletContext));
			if (setDirectly) {
				((CComponent)peer).postEvent(new MouseEvent(
					((CComponent)peer).fTarget, MouseEvent.MOUSE_MOVED,
					when, 0, x, y, 0, false));
			} else {
				SunToolkit.postEvent(appletContext, new PeerEvent(applet,
					new PostMouseMovedEvent((CComponent) peer, x, y, when),
					PeerEvent.PRIORITY_EVENT));
			}
			return true;
		} else {
			return false;
		}
	}

	// See reflectScrolledClipView_AWTSV() in AppletView.m for why this is
	// needed.
	public static void setScrollPanePosition(Object peer, int x, int y) {
		if (!(peer instanceof MyCScrollPane))
			return;
		MyCScrollPane scrollPanePeer = (MyCScrollPane) peer;
		ScrollPane scrollpane = (ScrollPane) scrollPanePeer.fTarget;
		if (scrollpane == null)
			return;
		Point currentPosition = scrollpane.getScrollPosition();
		if ((currentPosition.x == x) && (currentPosition.y == y))
			return;
		Component ancestor = scrollpane;
		for (;;) {
			Component next = (Component) ancestor.getParent();
			if (next == null)
				break;
			ancestor = next;
		}
		if (!(ancestor instanceof AppletFrame))
			return;
		if (((AppletFrame)ancestor).destroying)
			return;
		Applet applet = ((AppletFrame)ancestor).getApplet();
		AppContext appletContext = null;
		if (applet != null)
			appletContext = SunToolkit.targetToAppContext(applet);
		boolean setDirectly = ((appletContext == null) ||
			AppContext.getAppContext().equals(appletContext));
		if (setDirectly) {
			scrollPanePeer.setSkipNextScrollView();
			scrollpane.setScrollPosition(x, y);
		} else {
			SunToolkit.postEvent(appletContext, new PeerEvent(applet,
				new SetScrollPanePosition((AppletFrame) ancestor, scrollPanePeer, x, y),
				PeerEvent.PRIORITY_EVENT));
		}
	}

	public static void printStackTrace() {
		System.err.print(new Date() + " ");
		Throwable barf = new Throwable();
		barf.fillInStackTrace();
		barf.printStackTrace();
	}

	public static String getJavaThread() {
		return Thread.currentThread().toString();
	}

	public String toString() {
		return getClass().getName() + " (jep)";
	}

	static {
		if ("true".equalsIgnoreCase(System.getProperty("jep.debuginfo", "false"))) {
			showDebugInfo = true;
		}
	}

	private static class PostMouseMovedEvent implements Runnable {
		private final CComponent peer;
		private final int x;
		private final int y;
		private final long when;

		PostMouseMovedEvent(CComponent peer, int x, int y, long when) {
			this.peer = peer;
			this.x = x;
			this.y = y;
			this.when = when;
		}

		public void run() {
			peer.postEvent(new MouseEvent(peer.fTarget,
				MouseEvent.MOUSE_MOVED, when, 0, x, y, 0, false));
		}

	}

	private static class SetScrollPanePosition implements Runnable {
		private final AppletFrame frame;
		private final MyCScrollPane scrollPanePeer;
		private final int x;
		private final int y;

		SetScrollPanePosition(AppletFrame frame,
			MyCScrollPane scrollPanePeer, int x, int y)
		{
			this.frame = frame;
			this.scrollPanePeer = scrollPanePeer;
			this.x = x;
			this.y = y;
		}

		public void run() {
			scrollPanePeer.setSkipNextScrollView();
			ScrollPane scrollpane = (ScrollPane) scrollPanePeer.fTarget;
			frame.validateClips();
			scrollpane.setScrollPosition(x, y);
		}

	}

}

