//Copyright (c) 2005, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// Based partly on code in apple/awt/PeerPaintEvent.class in ui.jar.
// I would have used it directly had I not needed to fix bugs in it.

package apple.awt;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class MyPeerPaintEvent141 extends PeerPaintEvent {
	protected boolean canceled = false;
	protected boolean doingDispatch = false;
	protected boolean surfaceDataStale = false;
	protected Object resizeEventLock = new Object();
	protected ComponentEvent resizeEvent = null;
	protected long nsGraphicsContext = 0L;
	protected long view = 0L;

	public MyPeerPaintEvent141(ContainerModel model, long view) {
		super(model, view);
		this.view = view;
	}

	void markSurfaceDataStale() {
		surfaceDataStale = true;
	}

	public void addResizeEvent(ComponentEvent anEvent) {
		synchronized (resizeEventLock) {
			resizeEvent = anEvent;
		}
	}

	public ComponentEvent popResizeEvent() {
		synchronized (resizeEventLock) {
			ComponentEvent event = resizeEvent;
			resizeEvent = null;
			return event;
		}
	}

	public boolean isResizePending() {
		return resizeEvent != null;
	}

	public synchronized void dispatch() {
		if (canceled)
			return;
		CocoaEvent.dispatchAlert(null);
		if (canceled)
			return;
		if (!fPeer.fTarget.isVisible()) {
			fDispatched = true;
			notifyAll();
			return;
		}
		if (nsGraphicsContext == 0L) {
			fDispatched = true;
			notifyAll();
			return;
		}
		if (canceled)
			return;
		if (view == 0L) {
			fPeer.resetViewPtr();
			view = fPeer.getViewPtr();
			if (view == 0L) {
				fDispatched = true;
				notifyAll();
				return;
			}
		}
		synchronized(fInPaintLOCK) {
			try	{
				if (canceled)
					return;
				doingDispatch = true;
				ComponentEvent resizeEvent = popResizeEvent();
				if (resizeEvent != null) {
					Container container = null;
					if (fFirstTimePaintedByCocoa) {
						fFirstTimePaintedByCocoa = false;
						container = fPeer.fTarget.getParent();
						if ((container == null) && (fPeer.fTarget instanceof Container))
							container = (Container) fPeer.fTarget;
					}
					try {
						fPeer.fTarget.dispatchEvent(resizeEvent);
						if (container != null) {
							container.invalidate();
							container.validate();
						}
					} catch (Throwable e) {
						e.printStackTrace();
					}
					if (canceled)
						return;
				}
				fPeer.setCurrentContext(nsGraphicsContext, true);
				try {
					if (surfaceDataStale) {
						surfaceDataStale = false;
						fPeer.replaceSurfaceData();
					}
					Graphics g = fPeer.getGraphics();
					fPeer.paintDamagedArea();
					fPeer.completeContext();
				} catch (Throwable e) {
					e.printStackTrace();
				} finally {
					try {
						fPeer.unsetCurrentContext(nsGraphicsContext);
					} catch (Throwable e) {
						fPeer.fCurrentNSGraphicsContext = 0L;
						fPeer.fCurrentContext = 0L;
					}
				}
			} catch (Throwable e) {
				e.printStackTrace();
			} finally {
				canceled = false;
				doingDispatch = false;
				fDispatched = true;
				notifyAll();
			}
		}
	}

	// Don't wait more than about 100 milliseconds for a PeerPaintEvent to be
	// dispatched -- this reduces (or eliminates) the number of hangs in
	// paintFromCocoaDrawRect() waiting for this method to return.
	protected synchronized boolean waitCorrectlyForDispatch() {
		try {
			EventFactoryProxy.setCurrentEvent(this);
			for (int i = 1; i <= 3; ++i) {
				wait(30);
				Object currentEvent = EventFactoryProxy.getCurrentEvent();
				if ((currentEvent == null) || fDispatched)
					break;
			}
			EventFactoryProxy.setCurrentEvent(null);
			if (!fDispatched) {
				canceled = true;
				if (doingDispatch)
					MyEventQueue.getOurDispatchThread().interrupt();
			}
			return false;
		} catch (InterruptedException e) {
			EventFactoryProxy.setCurrentEvent(null);
			e.printStackTrace();
			return false;
		}
	}

	public void paintFromCocoaDrawRect(int x, int y, int width, int height,
		boolean liveResize, long nsGraphicsContext)
	{
		canceled = false;
		fDispatched = false;
		doingDispatch = false;
		this.nsGraphicsContext = nsGraphicsContext;
		fLiveResize = liveResize;
		fPeer.addDamagedArea(x, y, width, height);
		postToQueue();
		waitCorrectlyForDispatch();
		synchronized (fInPaintLOCK) {
			nsGraphicsContext = 0L;
			fDispatched = true;
			fLiveResize = false;
		}
	}

}

