//Copyright (c) 2007, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// This class exists to work around an Apple bug (or what seems to be an Apple
// bug) that effects OS X 10.3.X:  You sometimes see the following message
// displayed in the system console -- "Please Attach Debugger : No Autorelease
// pool exists" (this message is displayed by AskUserToAttachDebugger() in
// libawt.jnilib).  This happens especially often in Camino on OS X 10.3.X
// (which may indicate that Camino is also partially at fault).  I work around
// the problem by making sure an autorelease pool is in place before any AWT
// event is dispatched.  This fix isn't needed in later OS versions, but I'm
// using it anyway on general principles (since it also helps to make sure
// items in the autorelease pool are cleaned out periodically).

package jep;

import java.awt.*;
import java.util.*;

public class MySimpleEventQueue extends EventQueue {
	private static MySimpleEventQueue topQueue = null;

	public MySimpleEventQueue() {
	}

	protected void dispatchEvent(AWTEvent event) {
		long pool = AppletHandlerFactory.allocAutoreleasePool();
		try {
			Object source = event.getSource();
			Component ancestor = (Component) source;
			for (;;) {
				Component next = (Component) ancestor.getParent();
				if (next == null)
					break;
				ancestor = next;
			}
			if (ancestor instanceof AppletFrame)
				AppletFrame.setCurrentAppletFrame((AppletFrame) ancestor);
		} catch (Throwable e) {
		}
		super.dispatchEvent(event);
		AppletFrame.setCurrentAppletFrame(null);
		AppletHandlerFactory.releaseAutoreleasePool(pool);
	}

	// Don't let an applet "pop()" back above us.
	public synchronized void push(EventQueue newEventQueue) {
		if (topQueue == null)
			topQueue = this;
		super.push(newEventQueue);
	}

	protected void pop() throws EmptyStackException {
		if ((topQueue == null) || (topQueue == this))
			throw new EmptyStackException();
		super.pop();
	}

}
