//Copyright (c) 2006, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// This is a replacement for a class that comes with Apple's JVM (licensed
// from Sun).  It's used when running Java 1.4.X.  In AppletLoaderFactory.
// initProperties() we ensure that this class gets loaded before the Apple
// class that it replaces.  We need it to load our own "browser service",
// which in turn loads our own versions of "handlers" like the "cookie
// handler" (which can access the browser's cookie store in other browsers
// than Safari).

package sun.plugin.services;

import jep.AppletHandlerFactory;

public class ServiceProvider {
	private static BrowserService service = null;

	public ServiceProvider() {
	}

	public static synchronized BrowserService getService() {
		if (service == null) {
			try {
				Class c = Class.forName("sun.plugin.services.DefaultBrowserService");
				service = (BrowserService) c.newInstance();
			} catch (Throwable e) {
				System.err.println("ServiceProvider.getService(): exception: " +
					e.getMessage());
				e.printStackTrace();
			}
		}
		return service;
	}

	public static synchronized void setService(int browserType) {
		Class c = null;
		try {
			switch (browserType) {
				case 5:
					c = Class.forName("sun.plugin.services.AxBridgeBrowserService");
					break;
				case 17:
					c = Class.forName("sun.plugin.services.MacOSXBrowserService");
					break;
				case 21:
					c = Class.forName("sun.plugin.services.WebKitBrowserService");
					break;
				case 257:
					c = Class.forName("sun.plugin.services.WIExplorerBrowserService");
					break;
				case 258:
					c = Class.forName("sun.plugin.services.WNetscape4BrowserService");
					break;
				case 259:
					c = Class.forName("sun.plugin.services.WNetscape6BrowserService");
					break;
				case 666:
					if (AppletHandlerFactory.isJava142()) {
						c = Class.forName("sun.plugin.services.JEPBrowserService142");
					} else {
						c = Class.forName("sun.plugin.services.JEPBrowserService141");
					}
					break;
				case 4098:
					c = Class.forName("sun.plugin.services.MNetscape4BrowserService");
					break;
				case 4099:
					c = Class.forName("sun.plugin.services.MNetscape6BrowserService");
					break;
				default:
					c = Class.forName("sun.plugin.services.DefaultBrowserService");
					break;
			}
			service = (BrowserService) c.newInstance();
		} catch (Throwable e) {
			System.err.println("ServiceProvider.setService(): exception: " +
				e.getMessage());
			e.printStackTrace();
		}
    }

}
