/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the MRJ Carbon OJI Plugin.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corp.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Patrick C. Beard <beard@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

/*
    CSecureEnv.h
    
    Rewritten for use with MRJ plugin by Patrick C. Beard.
 */

#ifndef CSecureJNI2_h___
#define CSecureJNI_h___

#include <MacTypes.h>
#include "nsISecureEnv.h"
#include "nsjvmFixes.h"
#include "nsIScriptSecurityManagerOld.h"
#include "nsIScriptSecurityManagerNew.h"
#include "nsIScriptSecurityManagerNewer.h"
#include "nsIScriptSecurityManagerNewest.h"

/* See https://bugzilla.mozilla.org/show_bug.cgi?id=293973 for the latest
 * on this issue -- proliferating nsIScriptSecurityManager IIDs.
 */
#define NS_ISCRIPTSECURITYMANAGERNEW_IID \
  {0x18158378, 0xb632, 0x436a, \
    { 0x9c, 0xed, 0xc5, 0x2d, 0xed, 0x35, 0x0b, 0xde }}
#define NS_ISCRIPTSECURITYMANAGERNEWALT1_IID \
  {0x82bff35d, 0xebb7, 0x4cb9, \
    { 0xbb, 0x47, 0x4f, 0x9e, 0xbd, 0xf1, 0x1f, 0x77 }}
#define NS_ISCRIPTSECURITYMANAGERNEWALT2_IID \
  {0x463eb1fa, 0x9dac, 0x4ca7, \
    { 0x82, 0x6f, 0x1f, 0xc9, 0x21, 0x97, 0x1d, 0x3a }}
#define NS_ISCRIPTSECURITYMANAGERNEWALT3_IID \
  {0xf4d74511, 0x2b2d, 0x4a14, \
    { 0xa3, 0xe4, 0xa3, 0x92, 0xac, 0x5a, 0xc3, 0xff }}
#define NS_ISCRIPTSECURITYMANAGERNEWALT4_IID \
  {0x6c86a0d8, 0x7a19, 0x42a4, \
    { 0xbe, 0x18, 0xd2, 0x82, 0xa2, 0x63, 0x88, 0xfd }}
#define NS_ISCRIPTSECURITYMANAGERNEWERALT1_IID \
  {0x5a46e611, 0xcf4a, 0x4407, \
    { 0xa5, 0xb4, 0x00, 0x2f, 0xcf, 0x05, 0x11, 0x20 }}
#define NS_ISCRIPTSECURITYMANAGERNEWERALT2_IID \
  {0x5ed0b9b8, 0xe7f4, 0x4b55, \
    { 0x98, 0x05, 0x7a, 0xb8, 0x57, 0xb3, 0x35, 0xa1 }}
#define NS_ISCRIPTSECURITYMANAGERNEWERALT3_IID \
  {0xb9024f88, 0x9ad2, 0x491c, \
    { 0x97, 0x77, 0xce, 0x51, 0x91, 0x03, 0x3d, 0x3d }}
#define NS_ISCRIPTSECURITYMANAGERNEWERALT4_IID \
  {0xc61d3ad1, 0x00aa, 0x407c, \
    { 0xb7, 0xd7, 0xde, 0x48, 0xb3, 0xf1, 0x86, 0x62 }}
#define NS_ISCRIPTSECURITYMANAGERNEWERALT5_IID \
  {0x0b8a9b32, 0x713f, 0x4c39, \
    { 0xbe, 0xa0, 0x6c, 0xac, 0xec, 0x46, 0xf3, 0x85 }}
#define NS_ISCRIPTSECURITYMANAGERNEWERALT6_IID \
  {0xce216cf7, 0x3bcb, 0x48ab, \
    { 0x9f, 0xf8, 0xd0, 0x3a, 0x24, 0xf1, 0x9c, 0xa5 }}
#define NS_ISCRIPTSECURITYMANAGERNEWERALT7_IID \
  {0x3fffd8e8, 0x3fea, 0x442e, \
    { 0xa0, 0xed, 0x2b, 0xa8, 0x1a, 0xe1, 0x97, 0xd5 }}
#define NS_ISCRIPTSECURITYMANAGERNEWESTALT1_IID \
  {0xe74487fd, 0x98ef, 0x48d1, \
    { 0xb9, 0x73, 0x3f, 0x29, 0x38, 0xf0, 0x4e, 0x8e }}
#ifndef NS_ISCRIPTSECURITYMANAGEROBSOLETE_IID
#define NS_ISCRIPTSECURITYMANAGEROBSOLETE_IID \
  {0x58df5780, 0x8006, 0x11d2, \
    { 0xbd, 0x91, 0x00, 0x80, 0x5f, 0x8a, 0xe3, 0xf4 }}
#endif

class MRJPlugin;
class MRJSession;

class Monitor;
class nsIJVMManager;
class nsIScriptSecurityManager;
class nsIPrincipal;

class JavaMessage;
class JavaMessageQueue;

class CSecureEnv : public nsISecureEnv, public nsIRunnable {
public:
    NS_DECL_ISUPPORTS

    static NS_METHOD Create(MRJPlugin* plugin, JNIEnv* proxyEnv, const nsIID& aIID, void* *aInstancePtr);

    ////////////////////////////////////////////////////////////////////////////
    // from nsISecureJNI2:


    /**
     * Create new Java object in LiveConnect.
     *
     * @param clazz      -- Java Class object.
     * @param methodID   -- Method id
     * @param args       -- arguments for invoking the constructor.
     * @param result     -- return new Java object.
     * @param ctx        -- security context 
     */
    NS_IMETHOD NewObject(/*[in]*/  jclass clazz, 
                         /*[in]*/  jmethodID methodID,
                         /*[in]*/  jvalue *args, 
                         /*[out]*/ jobject* result,
                         /*[in]*/  nsISecurityContext* ctx = NULL);
   
    /**
     * Invoke method on Java object in LiveConnect.
     *
     * @param type       -- Return type
     * @param obj        -- Java object.
     * @param methodID   -- method id
     * @param args       -- arguments for invoking the constructor.
     * @param result     -- return result of invocation.
     * @param ctx        -- security context 
     */
    NS_IMETHOD CallMethod(/*[in]*/  jni_type type,
                          /*[in]*/  jobject obj, 
                          /*[in]*/  jmethodID methodID,
                          /*[in]*/  jvalue *args, 
                          /*[out]*/ jvalue* result,
                          /*[in]*/  nsISecurityContext* ctx = NULL);

    /**
     * Invoke non-virtual method on Java object in LiveConnect.
     *
     * @param type       -- Return type
     * @param obj        -- Java object.
     * @param clazz      -- Class object
     * @param methodID   -- method id
     * @param args       -- arguments for invoking the constructor.
     * @param ctx        -- security context 
     * @param result     -- return result of invocation.
     */
    NS_IMETHOD CallNonvirtualMethod(/*[in]*/  jni_type type,
                                    /*[in]*/  jobject obj, 
                                    /*[in]*/  jclass clazz,
                                    /*[in]*/  jmethodID methodID,
                                    /*[in]*/  jvalue *args, 
                                    /*[out]*/ jvalue* result,
                                    /*[in]*/  nsISecurityContext* ctx = NULL);

    /**
     * Get a field on Java object in LiveConnect.
     *
     * @param type       -- Return type
     * @param obj        -- Java object.
     * @param fieldID    -- field id
     * @param result     -- return field value
     * @param ctx        -- security context 
     */
    NS_IMETHOD GetField(/*[in]*/  jni_type type,
                        /*[in]*/  jobject obj, 
                        /*[in]*/  jfieldID fieldID,
                        /*[out]*/ jvalue* result,
                        /*[in]*/  nsISecurityContext* ctx = NULL);

    /**
     * Set a field on Java object in LiveConnect.
     *
     * @param type       -- Return type
     * @param obj        -- Java object.
     * @param fieldID    -- field id
     * @param val        -- field value to set
     * @param ctx        -- security context 
     */
    NS_IMETHOD SetField(/*[in]*/ jni_type type,
                        /*[in]*/ jobject obj, 
                        /*[in]*/ jfieldID fieldID,
                        /*[in]*/ jvalue val,
                        /*[in]*/ nsISecurityContext* ctx = NULL);

    /**
     * Invoke static method on Java object in LiveConnect.
     *
     * @param type       -- Return type
     * @param clazz      -- Class object.
     * @param methodID   -- method id
     * @param args       -- arguments for invoking the constructor.
     * @param result     -- return result of invocation.
     * @param ctx        -- security context 
     */
    NS_IMETHOD CallStaticMethod(/*[in]*/  jni_type type,
                                /*[in]*/  jclass clazz,
                                /*[in]*/  jmethodID methodID,
                                /*[in]*/  jvalue *args, 
                                /*[out]*/ jvalue* result,
                                /*[in]*/  nsISecurityContext* ctx = NULL);

    /**
     * Get a static field on Java object in LiveConnect.
     *
     * @param type       -- Return type
     * @param clazz      -- Class object.
     * @param fieldID    -- field id
     * @param result     -- return field value
     * @param ctx        -- security context 
     */
    NS_IMETHOD GetStaticField(/*[in]*/  jni_type type,
                              /*[in]*/  jclass clazz, 
                              /*[in]*/  jfieldID fieldID, 
                              /*[out]*/ jvalue* result,
                              /*[in]*/  nsISecurityContext* ctx = NULL);


    /**
     * Set a static field on Java object in LiveConnect.
     *
     * @param type       -- Return type
     * @param clazz      -- Class object.
     * @param fieldID    -- field id
     * @param val        -- field value to set
     * @param ctx        -- security context 
     */
    NS_IMETHOD SetStaticField(/*[in]*/ jni_type type,
                              /*[in]*/ jclass clazz, 
                              /*[in]*/ jfieldID fieldID,
                              /*[in]*/ jvalue val,
                              /*[in]*/ nsISecurityContext* ctx = NULL);


    NS_IMETHOD GetVersion(/*[out]*/ jint* version);

    NS_IMETHOD DefineClass(/*[in]*/  const char* name,
                           /*[in]*/  jobject loader,
                           /*[in]*/  const jbyte *buf,
                           /*[in]*/  jsize len,
                           /*[out]*/ jclass* clazz);

    NS_IMETHOD FindClass(/*[in]*/  const char* name, 
                         /*[out]*/ jclass* clazz);

    NS_IMETHOD GetSuperclass(/*[in]*/  jclass sub,
                             /*[out]*/ jclass* super);

    NS_IMETHOD IsAssignableFrom(/*[in]*/  jclass sub,
                                /*[in]*/  jclass super,
                                /*[out]*/ jboolean* result);

    NS_IMETHOD Throw(/*[in]*/  jthrowable obj,
                     /*[out]*/ jint* result);

    NS_IMETHOD ThrowNew(/*[in]*/  jclass clazz,
                        /*[in]*/  const char *msg,
                        /*[out]*/ jint* result);

    NS_IMETHOD ExceptionOccurred(/*[out]*/ jthrowable* result);

    NS_IMETHOD ExceptionDescribe(void);

    NS_IMETHOD ExceptionClear(void);

    NS_IMETHOD FatalError(/*[in]*/ const char* msg);

    NS_IMETHOD NewGlobalRef(/*[in]*/  jobject lobj, 
                            /*[out]*/ jobject* result);

    NS_IMETHOD DeleteGlobalRef(/*[in]*/ jobject gref);

    NS_IMETHOD DeleteLocalRef(/*[in]*/ jobject obj);

    NS_IMETHOD IsSameObject(/*[in]*/  jobject obj1,
                            /*[in]*/  jobject obj2,
                            /*[out]*/ jboolean* result);

    NS_IMETHOD AllocObject(/*[in]*/  jclass clazz,
                           /*[out]*/ jobject* result);

    NS_IMETHOD GetObjectClass(/*[in]*/  jobject obj,
                              /*[out]*/ jclass* result);

    NS_IMETHOD IsInstanceOf(/*[in]*/  jobject obj,
                            /*[in]*/  jclass clazz,
                            /*[out]*/ jboolean* result);

    NS_IMETHOD GetMethodID(/*[in]*/  jclass clazz, 
                           /*[in]*/  const char* name,
                           /*[in]*/  const char* sig,
                           /*[out]*/ jmethodID* id);

    NS_IMETHOD GetFieldID(/*[in]*/  jclass clazz, 
                          /*[in]*/  const char* name,
                          /*[in]*/  const char* sig,
                          /*[out]*/ jfieldID* id);

    NS_IMETHOD GetStaticMethodID(/*[in]*/  jclass clazz, 
                                 /*[in]*/  const char* name,
                                 /*[in]*/  const char* sig,
                                 /*[out]*/ jmethodID* id);

    NS_IMETHOD GetStaticFieldID(/*[in]*/  jclass clazz, 
                                /*[in]*/  const char* name,
                                /*[in]*/  const char* sig,
                                /*[out]*/ jfieldID* id);

    NS_IMETHOD NewString(/*[in]*/  const jchar* unicode,
                         /*[in]*/  jsize len,
                         /*[out]*/ jstring* result);

    NS_IMETHOD GetStringLength(/*[in]*/  jstring str,
                               /*[out]*/ jsize* result);
    
    NS_IMETHOD GetStringChars(/*[in]*/  jstring str,
                              /*[in]*/  jboolean *isCopy,
                              /*[out]*/ const jchar** result);

    NS_IMETHOD ReleaseStringChars(/*[in]*/  jstring str,
                                  /*[in]*/  const jchar *chars);

    NS_IMETHOD NewStringUTF(/*[in]*/  const char *utf,
                            /*[out]*/ jstring* result);

    NS_IMETHOD GetStringUTFLength(/*[in]*/  jstring str,
                                  /*[out]*/ jsize* result);
    
    NS_IMETHOD GetStringUTFChars(/*[in]*/  jstring str,
                                 /*[in]*/  jboolean *isCopy,
                                 /*[out]*/ const char** result);

    NS_IMETHOD ReleaseStringUTFChars(/*[in]*/  jstring str,
                                     /*[in]*/  const char *chars);

    NS_IMETHOD GetArrayLength(/*[in]*/  jarray array,
                              /*[out]*/ jsize* result);

    NS_IMETHOD NewObjectArray(/*[in]*/  jsize len,
                        /*[in]*/  jclass clazz,
                        /*[in]*/  jobject init,
                        /*[out]*/ jobjectArray* result);

    NS_IMETHOD GetObjectArrayElement(/*[in]*/  jobjectArray array,
                                     /*[in]*/  jsize index,
                                     /*[out]*/ jobject* result);

    NS_IMETHOD SetObjectArrayElement(/*[in]*/  jobjectArray array,
                                     /*[in]*/  jsize index,
                                     /*[in]*/  jobject val);

    NS_IMETHOD NewArray(/*[in]*/ jni_type element_type,
                        /*[in]*/  jsize len,
                        /*[out]*/ jarray* result);

    NS_IMETHOD GetArrayElements(/*[in]*/  jni_type type,
                                /*[in]*/  jarray array,
                                /*[in]*/  jboolean *isCopy,
                                /*[out]*/ void* result);

    NS_IMETHOD ReleaseArrayElements(/*[in]*/ jni_type type,
                                    /*[in]*/ jarray array,
                                    /*[in]*/ void *elems,
                                    /*[in]*/ jint mode);

    NS_IMETHOD GetArrayRegion(/*[in]*/  jni_type type,
                              /*[in]*/  jarray array,
                              /*[in]*/  jsize start,
                              /*[in]*/  jsize len,
                              /*[out]*/ void* buf);

    NS_IMETHOD SetArrayRegion(/*[in]*/  jni_type type,
                              /*[in]*/  jarray array,
                              /*[in]*/  jsize start,
                              /*[in]*/  jsize len,
                              /*[in]*/  void* buf);

    NS_IMETHOD RegisterNatives(/*[in]*/  jclass clazz,
                               /*[in]*/  const JNINativeMethod *methods,
                               /*[in]*/  jint nMethods,
                               /*[out]*/ jint* result);

    NS_IMETHOD UnregisterNatives(/*[in]*/  jclass clazz,
                                 /*[out]*/ jint* result);

    NS_IMETHOD MonitorEnter(/*[in]*/  jobject obj,
                            /*[out]*/ jint* result);

    NS_IMETHOD MonitorExit(/*[in]*/  jobject obj,
                           /*[out]*/ jint* result);

    NS_IMETHOD GetJavaVM(/*[in]*/  JavaVM **vm,
                         /*[out]*/ jint* result);
                         
    ////////////////////////////////////////////////////////////////////////////
    // from nsIRunnable:
    
    NS_IMETHOD Run();

    CSecureEnv(MRJPlugin* plugin, JNIEnv* proxyEnv, JNIEnv* javaEnv = NULL);
    virtual ~CSecureEnv(void);

    /**
     * Called by the native run method, to connect the
     * thread and the secure env.
     */
    void initialize(JNIEnv* javaEnv, jboolean* isRunning, JavaMessageQueue* javaQueue, JavaMessageQueue* nativeQueue);
    
    jboolean isInitialized() { return mJavaQueue != NULL; }

    void setProxyEnv(JNIEnv* proxyEnv) { mProxyEnv = proxyEnv; }
    JNIEnv* getProxyEnv() { return mProxyEnv; }
    
    void setJavaEnv(JNIEnv* javaEnv) { mJavaEnv = javaEnv; }
    JNIEnv* getJavaEnv() { return mJavaEnv; }
    
    MRJSession* getSession() { return mSession; }
    nsIJVMThreadManager* getThreadManager() { return mThreadManager; }
    
    void getMessageQueues(JavaMessageQueue*& javaQueue, JavaMessageQueue*& nativeQueue)
    {
        javaQueue = mJavaQueue;
        nativeQueue = mNativeQueue;
    }
    
    void sendMessageToJava(JavaMessage* msg);
    void sendMessageFromJava(JNIEnv* javaEnv, JavaMessage* msg, Boolean busyWaiting = false);
    void messageLoop(JNIEnv* env, JavaMessage* msgToSend, JavaMessageQueue* sendQueue, JavaMessageQueue* receiveQueue, Boolean busyWaiting = false);

    void savePendingException(JNIEnv* env);
    jthrowable getPendingException(JNIEnv* env);
    void clearPendingException(JNIEnv* env);
    static void suspendExceptions() { ++mSuspendExceptionsLevel; }
    static void resumeExceptions() { --mSuspendExceptionsLevel; }

	void resetLiveConnectProxy(JNIEnv *env);

protected:
    MRJPlugin*                  mPlugin;
    JNIEnv*                     mProxyEnv;
    JNIEnv*                     mJavaEnv;
    MRJSession*                 mSession;
    nsIJVMThreadManager*        mThreadManager;
    
    jboolean*                   mIsRunning;
    JavaMessageQueue*           mJavaQueue;
    JavaMessageQueue*           mNativeQueue;

    jthrowable                  mPendingException;
    static long                 mSuspendExceptionsLevel;
    
    nsIScriptSecurityManager*       mScriptSecurityManager;
    nsIScriptSecurityManagerNew*    mScriptSecurityManagerNew;
    nsIScriptSecurityManagerNewer*  mScriptSecurityManagerNewer;
    nsIScriptSecurityManagerNewest* mScriptSecurityManagerNewest;
    nsIPrincipal*                   mScriptPrincipal;
    jclass                          mLiveConnectProxy;
};

/**
 * Returns the secure JNI associated with the current thread (if any).
 */
CSecureEnv* GetSecureJNI(JNIEnv* env, jobject thread);

#endif // CSecureJNI2_h___
