/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the MRJ Carbon OJI Plugin.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corp.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Patrick C. Beard <beard@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

/*
    JavaEmbeddingFramework.cpp

    Dynamically loads the JavaEmbeddingFramework.  We need this because the
	JavaEmbeddingFramework isn't present on Mac OS X for Intel.

    by Steven Michaud (modeled on JavaVMFramework.cpp by Patrick C. Beard).
 */

#include <jni.h>
#include <CarbonEvents.h>
#include <JavaApplet.h>
#include <JavaControl.h>

#include <CFURL.h>
#include <CFBundle.h>
#include <CFString.h>

static CFBundleRef getBundle(CFStringRef frameworkPath)
{
	CFBundleRef bundle = NULL;

	//	Make a CFURLRef from the CFString representation of the bundle's path.
	//	See the Core Foundation URL Services chapter for details.
	CFURLRef bundleURL = CFURLCreateWithFileSystemPath(NULL, frameworkPath, kCFURLPOSIXPathStyle, true);
	if (bundleURL != NULL) {
		bundle = CFBundleCreate(NULL, bundleURL);
		if (bundle != NULL)
			CFBundleLoadExecutable(bundle);
		CFRelease(bundleURL);
	}

	return bundle;
}

static void* getJavaEmbeddingFunction(CFStringRef functionName)
{
// The #ifdef __ppc__ is from Mark Mentovai's patch at
// https://bugzilla.mozilla.org/show_bug.cgi?id=315163.
#ifdef __ppc__
	static CFBundleRef javaBundle = getBundle(CFSTR("/System/Library/Frameworks/JavaEmbedding.framework"));
	if (javaBundle) return CFBundleGetFunctionPointerForName(javaBundle, functionName);
#endif
	return NULL;
}

typedef OSStatus (*CreateJavaApplet_proc_ptr) (JNIEnv *env, AppletDescriptor applet, Boolean trusted, AppletArena arena, jobject *outJavaFrame);
static CreateJavaApplet_proc_ptr _CreateJavaApplet = (CreateJavaApplet_proc_ptr) getJavaEmbeddingFunction(CFSTR("CreateJavaApplet"));

OSStatus CreateJavaApplet(JNIEnv *env, AppletDescriptor applet,
	Boolean trusted, AppletArena arena, jobject *outJavaFrame)
{
	if (_CreateJavaApplet) return _CreateJavaApplet(env, applet, trusted, arena, outJavaFrame);
	return -1;
}

Boolean CheckJavaEmbeddingFramework(void)
{
	static Boolean frameworkChecked = FALSE;
	static Boolean frameworkStatus = FALSE;
	if (frameworkChecked) return frameworkStatus;
	if (_CreateJavaApplet)
		frameworkStatus = TRUE;
	frameworkChecked = TRUE;
	return frameworkStatus;
}

typedef OSStatus (*CreateJavaControl_proc_ptr) (JNIEnv *env, WindowRef inNativeWindow, const Rect *inBounds, jobject inAppletFrame, Boolean inVisible, ControlRef *outControl);
static CreateJavaControl_proc_ptr _CreateJavaControl = (CreateJavaControl_proc_ptr) getJavaEmbeddingFunction(CFSTR("CreateJavaControl"));

OSStatus CreateJavaControl(JNIEnv *env, WindowRef inNativeWindow,
	const Rect *inBounds, jobject inAppletFrame, Boolean inVisible,
	ControlRef *outControl)
{
	if (_CreateJavaControl) return _CreateJavaControl(env, inNativeWindow, inBounds, inAppletFrame, inVisible, outControl);
	return -1;
}

typedef OSStatus (*DrawJavaControl_proc_ptr) (JNIEnv *env, ControlRef theControl);
static DrawJavaControl_proc_ptr _DrawJavaControl = (DrawJavaControl_proc_ptr) getJavaEmbeddingFunction(CFSTR("DrawJavaControl"));

OSStatus DrawJavaControl(JNIEnv *env, ControlRef theControl)
{
	if (_DrawJavaControl) return _DrawJavaControl(env, theControl);
	return -1;
}

typedef OSStatus (*MoveAndClipJavaControl_proc_ptr) (JNIEnv *env, ControlRef theControl, int posX, int posY, int clipX, int clipY, int clipWidth, int clipHeight);
static MoveAndClipJavaControl_proc_ptr _MoveAndClipJavaControl = (MoveAndClipJavaControl_proc_ptr) getJavaEmbeddingFunction(CFSTR("MoveAndClipJavaControl"));

OSStatus MoveAndClipJavaControl(JNIEnv *env, ControlRef theControl,
	int posX, int posY, int clipX, int clipY, int clipWidth, int clipHeight)
{
	if (_MoveAndClipJavaControl) return _MoveAndClipJavaControl(env, theControl, posX, posY, clipX, clipY, clipWidth, clipHeight);
	return -1;
}

typedef OSStatus (*RegisterShowDocumentCallback_proc_ptr) (JNIEnv *env, jobject inJavaFrame, JE_ShowDocumentCallback showDocumentFunction, void *userData);
static RegisterShowDocumentCallback_proc_ptr _RegisterShowDocumentCallback = (RegisterShowDocumentCallback_proc_ptr) getJavaEmbeddingFunction(CFSTR("RegisterShowDocumentCallback"));

OSStatus RegisterShowDocumentCallback(JNIEnv *env, jobject inJavaFrame,
	JE_ShowDocumentCallback showDocumentFunction, void *userData)
{
	if (_RegisterShowDocumentCallback) return _RegisterShowDocumentCallback(env, inJavaFrame, showDocumentFunction, userData);
	return -1;
}

typedef OSStatus (*RegisterStatusCallback_proc_ptr) (JNIEnv *env, jobject inJavaFrame, JE_SetStatusCallback showStatusFunction, void *userData);
static RegisterStatusCallback_proc_ptr _RegisterStatusCallback = (RegisterStatusCallback_proc_ptr) getJavaEmbeddingFunction(CFSTR("RegisterStatusCallback"));

OSStatus RegisterStatusCallback(JNIEnv *env, jobject inJavaFrame,
	JE_SetStatusCallback showStatusFunction, void *userData)
{
	if (_RegisterStatusCallback) return _RegisterStatusCallback(env, inJavaFrame, showStatusFunction, userData);
	return -1;
}

typedef OSStatus (*RestartJavaControlAsyncDrawing_proc_ptr) (JNIEnv *env, ControlRef theControl);
static RestartJavaControlAsyncDrawing_proc_ptr _RestartJavaControlAsyncDrawing = (RestartJavaControlAsyncDrawing_proc_ptr) getJavaEmbeddingFunction(CFSTR("RestartJavaControlAsyncDrawing"));

OSStatus RestartJavaControlAsyncDrawing(JNIEnv *env, ControlRef theControl)
{
	if (_RestartJavaControlAsyncDrawing) return _RestartJavaControlAsyncDrawing(env, theControl);
	return -1;
}

typedef OSStatus (*SetJavaAppletState_proc_ptr) (JNIEnv *env, jobject inAppletFrame, AppletState inNewState);
static SetJavaAppletState_proc_ptr _SetJavaAppletState = (SetJavaAppletState_proc_ptr) getJavaEmbeddingFunction(CFSTR("SetJavaAppletState"));

OSStatus SetJavaAppletState(JNIEnv *env, jobject inAppletFrame, AppletState inNewState)
{
	if (_SetJavaAppletState) return _SetJavaAppletState(env, inAppletFrame, inNewState);
	return -1;
}

typedef OSStatus (*ShowHideJavaControl_proc_ptr) (JNIEnv *env, ControlRef theControl, Boolean visible);
static ShowHideJavaControl_proc_ptr _ShowHideJavaControl = (ShowHideJavaControl_proc_ptr) getJavaEmbeddingFunction(CFSTR("ShowHideJavaControl"));

OSStatus ShowHideJavaControl(JNIEnv *env, ControlRef theControl, Boolean visible)
{
	if (_ShowHideJavaControl) return _ShowHideJavaControl(env, theControl, visible);
	return -1;
}

typedef OSStatus (*SizeJavaControl_proc_ptr) (JNIEnv *env, ControlRef theControl, int width, int height);
static SizeJavaControl_proc_ptr _SizeJavaControl = (SizeJavaControl_proc_ptr) getJavaEmbeddingFunction(CFSTR("SizeJavaControl"));

OSStatus SizeJavaControl(JNIEnv *env, ControlRef theControl, int width, int height)
{
	if (_SizeJavaControl) return _SizeJavaControl(env, theControl, width, height);
	return -1;
}

typedef OSStatus (*StopJavaControlAsyncDrawing_proc_ptr) (JNIEnv *env, ControlRef theControl);
static StopJavaControlAsyncDrawing_proc_ptr _StopJavaControlAsyncDrawing = (StopJavaControlAsyncDrawing_proc_ptr) getJavaEmbeddingFunction(CFSTR("StopJavaControlAsyncDrawing"));

OSStatus StopJavaControlAsyncDrawing(JNIEnv *env, ControlRef theControl)
{
	if (_StopJavaControlAsyncDrawing) return _StopJavaControlAsyncDrawing(env, theControl);
	return -1;
}

