/*
 * Copyright (c) 2004, Steven Michaud, All Rights Reserved
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#import "JavaEmbeddingPlugin.h"
extern FSSpec thePluginSpec;
extern CFBundleRef thePluginBundle;
CFBundleRef getPluginBundle();

#ifdef __cplusplus
extern "C" {
#endif

/* This routine must not be called more than once. */
static CFBundleRef getJEPBundle(void)
{
	int i, j;
	OSStatus retval;
	FSRef pluginsFolderRef;
	CFURLRef bundleURL, bundleBaseURL;
	CFBundleRef bundle = NULL;

	thePluginBundle = getPluginBundle();

	FSRef thePluginRef;
	char path[1024] = {0};
	char *lastSlash;
	for (i = 1, retval = -1; i <= 4; ++i) {
		switch (i) {
			/* First look in our (MRJPlugin.plugin's) current directory */
			case 1:
				retval = FSpMakeFSRef(&thePluginSpec, &thePluginRef);
				if (retval != noErr)
					break;
				retval = FSRefMakePath(&thePluginRef, (UInt8 *) path,
					sizeof(path));
				if (retval != noErr)
					break;
				for (j = 1; j <= 4; ++j) {
					lastSlash = strrchr(path, '/');
					if (lastSlash) {
						*lastSlash = 0;
					} else {
						break;
					}
				}
				retval = FSPathMakeRef((const UInt8 *) path,
					&pluginsFolderRef, NULL);
				break;
			/* Then (if need be) look in our parent directory */
			case 2:
				lastSlash = strrchr(path, '/');
				if (lastSlash)
					*lastSlash = 0;
				retval = FSPathMakeRef((const UInt8 *) path,
					&pluginsFolderRef, NULL);
				break;
			/* Then (if need be) look in the user's Plug-Ins directory */
			case 3:
				retval = FSFindFolder(kUserDomain, kInternetPlugInFolderType,
					kDontCreateFolder, &pluginsFolderRef);
				break;
			/* Then (if need be) look in the system Plug-Ins directory */
			case 4:
				retval = FSFindFolder(kLocalDomain, kInternetPlugInFolderType,
					kDontCreateFolder, &pluginsFolderRef);
				break;
			default:
				break;
		}
		if (retval != noErr)
			continue;
		bundleBaseURL = CFURLCreateFromFSRef(kCFAllocatorSystemDefault,
			&pluginsFolderRef);
		bundleURL = CFURLCreateWithFileSystemPathRelativeToBase(
			kCFAllocatorSystemDefault, CFSTR("JavaEmbeddingPlugin.bundle"),
			kCFURLPOSIXPathStyle, true, bundleBaseURL);
		if (bundleURL) {
			bundle = CFBundleCreate(kCFAllocatorDefault, bundleURL);
			if (bundle) {
				if (!CFBundleLoadExecutable(bundle)) {
					CFRelease(bundle);
					bundle = NULL;
				}
			}
			CFRelease(bundleURL);
		}
		if (bundleBaseURL)
			CFRelease(bundleBaseURL);
		if (bundle)
			break;
	}

	return bundle;
}

void *getJEPFunction(CFStringRef functionName)
{
	static CFBundleRef bundle = getJEPBundle();
	if (bundle) return CFBundleGetFunctionPointerForName(bundle, functionName);
	return NULL;
}

typedef Boolean (*JEPCheckBundleWithVersion_proc_ptr) (CFStringRef version);
static JEPCheckBundleWithVersion_proc_ptr _JEPCheckBundleWithVersion = (JEPCheckBundleWithVersion_proc_ptr) getJEPFunction(CFSTR("JEPCheckBundleWithVersion"));

Boolean JEPCheckBundle(void)
{
	static Boolean JEPBundleChecked = FALSE;
	static Boolean JEPBundleStatus = FALSE;
	if (JEPBundleChecked) return JEPBundleStatus;
	if (_JEPCheckBundleWithVersion)
		JEPBundleStatus = _JEPCheckBundleWithVersion(CFSTR(JEP_VERSION));
	JEPBundleChecked = TRUE;
	return JEPBundleStatus;
}

typedef CFStringRef (*JEPGetInstalledPath_proc_ptr) (void);
static JEPGetInstalledPath_proc_ptr _JEPGetInstalledPath = (JEPGetInstalledPath_proc_ptr) getJEPFunction(CFSTR("JEPGetInstalledPath"));

CFStringRef JEPGetInstalledPath(void)
{
    if (_JEPGetInstalledPath) return _JEPGetInstalledPath();
    return NULL;
}

typedef OSStatus (*JEPCreateJavaVM_proc_ptr) (JavaVMOption *jvmExtraOpts, int jvmExtraOptsNum, JavaVM **vm, void **env);
static JEPCreateJavaVM_proc_ptr _JEPCreateJavaVM = (JEPCreateJavaVM_proc_ptr) getJEPFunction(CFSTR("JEPCreateJavaVM"));

OSStatus JEPCreateJavaVM(JavaVMOption *jvmExtraOpts, int jvmExtraOptsNum, JavaVM **vm,  void **env)
{
    if (_JEPCreateJavaVM) return _JEPCreateJavaVM(jvmExtraOpts, jvmExtraOptsNum, vm, env);
    return -1;
}

typedef OSStatus (*JEPCreateJavaVMAndInitAWT_proc_ptr) (JavaVMOption *jvmExtraOpts, int jvmExtraOptsNum, JavaVM **vm, void **env, void (*runAfterInitAWT) (void *), void *runAfterInitAWTInfo);
static JEPCreateJavaVMAndInitAWT_proc_ptr _JEPCreateJavaVMAndInitAWT = (JEPCreateJavaVMAndInitAWT_proc_ptr) getJEPFunction(CFSTR("JEPCreateJavaVMAndInitAWT"));

OSStatus JEPCreateJavaVMAndInitAWT(JavaVMOption *jvmExtraOpts, int jvmExtraOptsNum, JavaVM **vm,  void **env, void (*runAfterInitAWT) (void *), void *runAfterInitAWTInfo)
{
    if (_JEPCreateJavaVMAndInitAWT) return _JEPCreateJavaVMAndInitAWT(jvmExtraOpts, jvmExtraOptsNum, vm, env, runAfterInitAWT, runAfterInitAWTInfo);
    return -1;
}

typedef OSStatus (*JEPCreateJavaApplet_proc_ptr) (CFURLRef docbase, CFDictionaryRef attributes, WindowRef parentWindow, JEPOptions *options, void **handle);
static JEPCreateJavaApplet_proc_ptr _JEPCreateJavaApplet = (JEPCreateJavaApplet_proc_ptr) getJEPFunction(CFSTR("JEPCreateJavaApplet"));

OSStatus JEPCreateJavaApplet(CFURLRef docbase, CFDictionaryRef attributes, WindowRef parentWindow, JEPOptions *options, void **handle)
{
    if (_JEPCreateJavaApplet) return _JEPCreateJavaApplet(docbase, attributes, parentWindow, options, handle);
    return -1;
}

typedef OSStatus (*JEPGetJavaApplet_proc_ptr) (void *handle, jobject *applet);
static JEPGetJavaApplet_proc_ptr _JEPGetJavaApplet = (JEPGetJavaApplet_proc_ptr) getJEPFunction(CFSTR("JEPGetJavaApplet"));

OSStatus JEPGetJavaApplet(void *handle, jobject *applet)
{
    if (_JEPGetJavaApplet) return _JEPGetJavaApplet(handle, applet);
    return -1;
}

typedef OSStatus (*JEPGetJavaVM_proc_ptr) (void *handle, JavaVM **vm);
static JEPGetJavaVM_proc_ptr _JEPGetJavaVM = (JEPGetJavaVM_proc_ptr) getJEPFunction(CFSTR("JEPGetJavaVM"));

OSStatus JEPGetJavaVM(void *handle, JavaVM **vm)
{
    if (_JEPGetJavaVM) return _JEPGetJavaVM(handle, vm);
    return -1;
}

typedef OSStatus (*JEPSetJavaAppletState_proc_ptr) (void *handle, AppletStateJEP newState);
static JEPSetJavaAppletState_proc_ptr _JEPSetJavaAppletState = (JEPSetJavaAppletState_proc_ptr) getJEPFunction(CFSTR("JEPSetJavaAppletState"));

OSStatus JEPSetJavaAppletState(void *handle, AppletStateJEP newState)
{
    if (_JEPSetJavaAppletState) return _JEPSetJavaAppletState(handle, newState);
    return -1;
}

typedef OSStatus (*JEPGetJavaAppletState_proc_ptr) (void *handle, AppletStateJEP *state);
static JEPGetJavaAppletState_proc_ptr _JEPGetJavaAppletState = (JEPGetJavaAppletState_proc_ptr) getJEPFunction(CFSTR("JEPGetJavaAppletState"));

OSStatus JEPGetJavaAppletState(void *handle, AppletStateJEP *state)
{
    if (_JEPGetJavaAppletState) return _JEPGetJavaAppletState(handle, state);
    return -1;
}

typedef OSStatus (*JEPDrawJavaApplet_proc_ptr) (void *handle);
static JEPDrawJavaApplet_proc_ptr _JEPDrawJavaApplet = (JEPDrawJavaApplet_proc_ptr) getJEPFunction(CFSTR("JEPDrawJavaApplet"));

OSStatus JEPDrawJavaApplet(void *handle)
{
    if (_JEPDrawJavaApplet) return _JEPDrawJavaApplet(handle);
    return -1;
}

typedef OSStatus (*JEPMaybeDrawJavaApplet_proc_ptr) (void *handle);
static JEPMaybeDrawJavaApplet_proc_ptr _JEPMaybeDrawJavaApplet = (JEPMaybeDrawJavaApplet_proc_ptr) getJEPFunction(CFSTR("JEPMaybeDrawJavaApplet"));

OSStatus JEPMaybeDrawJavaApplet(void *handle)
{
    if (_JEPMaybeDrawJavaApplet) return _JEPMaybeDrawJavaApplet(handle);
    return -1;
}

typedef OSStatus (*JEPSuspendJavaAppletUpdates_proc_ptr) (void *handle);
static JEPSuspendJavaAppletUpdates_proc_ptr _JEPSuspendJavaAppletUpdates = (JEPSuspendJavaAppletUpdates_proc_ptr) getJEPFunction(CFSTR("JEPSuspendJavaAppletUpdates"));

OSStatus JEPSuspendJavaAppletUpdates(void *handle)
{
    if (_JEPSuspendJavaAppletUpdates) return _JEPSuspendJavaAppletUpdates(handle);
    return -1;
}

typedef OSStatus (*JEPResumeJavaAppletUpdates_proc_ptr) (void *handle);
static JEPResumeJavaAppletUpdates_proc_ptr _JEPResumeJavaAppletUpdates = (JEPResumeJavaAppletUpdates_proc_ptr) getJEPFunction(CFSTR("JEPResumeJavaAppletUpdates"));

OSStatus JEPResumeJavaAppletUpdates(void *handle)
{
    if (_JEPResumeJavaAppletUpdates) return _JEPResumeJavaAppletUpdates(handle);
    return -1;
}

typedef OSStatus (*JEPSizeJavaApplet_proc_ptr) (void *handle, int width, int height);
static JEPSizeJavaApplet_proc_ptr _JEPSizeJavaApplet = (JEPSizeJavaApplet_proc_ptr) getJEPFunction(CFSTR("JEPSizeJavaApplet"));

OSStatus JEPSizeJavaApplet(void *handle, int width, int height)
{
    if (_JEPSizeJavaApplet) return _JEPSizeJavaApplet(handle, width, height);
    return -1;
}

typedef OSStatus (*JEPPositionJavaApplet_proc_ptr) (void *handle, int posX, int posY);
static JEPPositionJavaApplet_proc_ptr _JEPPositionJavaApplet = (JEPPositionJavaApplet_proc_ptr) getJEPFunction(CFSTR("JEPPositionJavaApplet"));

OSStatus JEPPositionJavaApplet(void *handle, int posX, int posY)
{
    if (_JEPPositionJavaApplet) return _JEPPositionJavaApplet(handle, posX, posY);
    return -1;
}

typedef OSStatus (*JEPMoveAndClipJavaApplet_proc_ptr) (void *handle, int posX, int posY, int clipX, int clipY, int clipWidth, int clipHeight);
static JEPMoveAndClipJavaApplet_proc_ptr _JEPMoveAndClipJavaApplet = (JEPMoveAndClipJavaApplet_proc_ptr) getJEPFunction(CFSTR("JEPMoveAndClipJavaApplet"));

OSStatus JEPMoveAndClipJavaApplet(void *handle, int posX, int posY, int clipX, int clipY, int clipWidth, int clipHeight)
{
    if (_JEPMoveAndClipJavaApplet) return _JEPMoveAndClipJavaApplet(handle, posX, posY, clipX, clipY, clipWidth, clipHeight);
    return -1;
}

typedef OSStatus (*JEPShowHideJavaApplet_proc_ptr) (void *handle, Boolean visible);
static JEPShowHideJavaApplet_proc_ptr _JEPShowHideJavaApplet = (JEPShowHideJavaApplet_proc_ptr) getJEPFunction(CFSTR("JEPShowHideJavaApplet"));

OSStatus JEPShowHideJavaApplet(void *handle, Boolean visible)
{
    if (_JEPShowHideJavaApplet) return _JEPShowHideJavaApplet(handle, visible);
    return -1;
}

typedef OSStatus (*JEPRegisterStatusCallback_proc_ptr) (void *handle, JEPShowStatusCallback showStatusFunction, void *userData);
static JEPRegisterStatusCallback_proc_ptr _JEPRegisterStatusCallback = (JEPRegisterStatusCallback_proc_ptr) getJEPFunction(CFSTR("JEPRegisterStatusCallback"));

OSStatus JEPRegisterStatusCallback(void *handle, JEPShowStatusCallback showStatusFunction, void *userData)
{
    if (_JEPRegisterStatusCallback) return _JEPRegisterStatusCallback(handle, showStatusFunction, userData);
    return -1;
}

typedef OSStatus (*JEPRegisterShowDocumentCallback_proc_ptr) (void *handle, JEPShowDocumentCallback showDocumentFunction, void *userData);
static JEPRegisterShowDocumentCallback_proc_ptr _JEPRegisterShowDocumentCallback = (JEPRegisterShowDocumentCallback_proc_ptr) getJEPFunction(CFSTR("JEPRegisterShowDocumentCallback"));

OSStatus JEPRegisterShowDocumentCallback(void *handle, JEPShowDocumentCallback showDocumentFunction, void *userData)
{
    if (_JEPRegisterShowDocumentCallback) return _JEPRegisterShowDocumentCallback(handle, showDocumentFunction, userData);
    return -1;
}

typedef Boolean (*JEPGetJavaRuntimeParameterBoolean_proc_ptr) (CFStringRef key);
static JEPGetJavaRuntimeParameterBoolean_proc_ptr _JEPGetJavaRuntimeParameterBoolean = (JEPGetJavaRuntimeParameterBoolean_proc_ptr) getJEPFunction(CFSTR("JEPGetJavaRuntimeParameterBoolean"));

Boolean JEPGetJavaRuntimeParameterBoolean(CFStringRef key)
{
    if (_JEPGetJavaRuntimeParameterBoolean) return _JEPGetJavaRuntimeParameterBoolean(key);
    return false;
}

typedef CFStringRef (*JEPGetJavaRuntimeParameterString_proc_ptr) (CFStringRef key);
static JEPGetJavaRuntimeParameterString_proc_ptr _JEPGetJavaRuntimeParameterString = (JEPGetJavaRuntimeParameterString_proc_ptr) getJEPFunction(CFSTR("JEPGetJavaRuntimeParameterString"));

CFStringRef JEPGetJavaRuntimeParameterString(CFStringRef key)
{
    if (_JEPGetJavaRuntimeParameterString) return _JEPGetJavaRuntimeParameterString(key);
    return NULL;
}

typedef OSStatus (*JEPLogJavaMessage_proc_ptr) (CFStringRef message);
static JEPLogJavaMessage_proc_ptr _JEPLogJavaMessage = (JEPLogJavaMessage_proc_ptr) getJEPFunction(CFSTR("JEPLogJavaMessage"));

OSStatus JEPLogJavaMessage(CFStringRef message)
{
    if (_JEPLogJavaMessage) return _JEPLogJavaMessage(message);
    return -1;
}

typedef void (*JEPDoLiveConnectProxy_proc_ptr) (jclass proxyClass, jobject runnable, jobject applet, jboolean isCallMethodCall);
JEPDoLiveConnectProxy_proc_ptr _JEPDoLiveConnectProxy = (JEPDoLiveConnectProxy_proc_ptr) getJEPFunction(CFSTR("JEPDoLiveConnectProxy"));

void JEPDoLiveConnectProxy(jclass proxyClass, jobject runnable, jobject applet, jboolean isCallMethodCall)
{
    if (_JEPDoLiveConnectProxy) return _JEPDoLiveConnectProxy(proxyClass, runnable, applet, isCallMethodCall);
    return;
}

typedef void (*JEPRunOnMainThread_proc_ptr) (void (*runnable) (void *), void *info, Boolean wait);
JEPRunOnMainThread_proc_ptr _JEPRunOnMainThread = (JEPRunOnMainThread_proc_ptr) getJEPFunction(CFSTR("JEPRunOnMainThread"));

void JEPRunOnMainThread(void (*runnable) (void *), void *info, Boolean wait)
{
    if (_JEPRunOnMainThread) return _JEPRunOnMainThread(runnable, info, wait);
    return;
}

typedef jthrowable (*JEPRunJRunnableOnMainThread_proc_ptr) (jobject jrunnable);
JEPRunJRunnableOnMainThread_proc_ptr _JEPRunJRunnableOnMainThread = (JEPRunJRunnableOnMainThread_proc_ptr) getJEPFunction(CFSTR("JEPRunJRunnableOnMainThread"));

jthrowable JEPRunJRunnableOnMainThread(jobject jrunnable)
{
    if (_JEPRunJRunnableOnMainThread) return _JEPRunJRunnableOnMainThread(jrunnable);
    return NULL;
}

#ifdef __cplusplus
}
#endif
