/*
 * Copyright (c) 2004, Steven Michaud, All Rights Reserved
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* The API declared in this file is modeled on Apple's old Java Embedding
 * API declared and documented in JavaApplet.h and JavaControl.h in
 * /System/Library/Frameworks/JavaEmbedding.framework/Headers.  Some of
 * the differences from the original API are required by changes in the
 * newer JVMs (1.4.1 and 1.4.2).  Others are intended to fill holes in
 * the original API, or to make improvements upon it.
 *
 * The biggest difference from the original Java Embedding API is that
 * applets must now be created in a separate thread.  The calling app gets
 * notified by callbacks of progress (or failure) at several stages in
 * creating the applet.  As of version 0.8.7, the JEP defaults to doing part
 * of this synchronously (creation and at least part of initialization).
 * This makes it possible for the applet to be visible in a JavaScript onLoad
 * handler in a browser that (like all of the Mozilla family) assumes applet
 * creation and initialization will be synchronous.  To go back to the
 * original, fully asynchronous behavior, add "-Djep.asyncinit=true" to the
 * Java Control Panel's "runtime parameters".)
 */

#ifdef __cplusplus
extern "C" {
#endif

#define JEP_VERSION "0.9.6.4"
#define JEP_VERSION_OPTION "jep.version"

/* As described above */
#define JEP_ASYNCINIT_OPTION "jep.asyncinit"
/* As of JEP 0.8.7, JSObject's methods (the Java-to-JavaScript LiveConnect
 * calls) time out after a fairly long interval (currently 2000 milliseconds).
 * This was how versions 0.8.3 through 0.8.5 behaved.  Version 0.8.6 waited
 * indefinitely for these calls to return.  This sometimes allows poorly
 * written applets to hang the browser, but timing out the calls can also
 * cause trouble.  Adding "-Djep.jsobjectnotimeout=true" to the Java Control
 * Panel's "runtime parameters" restores JEP 0.8.6's behavior.
 */
#define JEP_JSOBJECT_NOTIMEOUT_OPTION "jep.jsobjectnotimeout"
/* Setting this to "true" makes the JEP log applet visibility and clipping
 * information to System.out.
 */
#define JEP_DEBUG_VISIBILITY_OPTION "jep.debug.visibility"
/* Setting this to "true" makes the JEP log the various stages of releasing
 * applet resources to System.out.
 */
#define JEP_DEBUG_RELEASE_OPTION "jep.debug.release"
/* Setting this to "true" makes the JEP (actually the MRJ Plugin JEP) log
 * to System.out calls (such as setWindow() and drawApplet()) that result in
 * whole-applet updates.
 */
#define JEP_DEBUG_UPDATES_OPTION "jep.debug.updates"
/* Setting this to "true" makes the JEP (actually the MRJ Plugin JEP) log
 * to System.out CAPS interface calls used when doing JavaScript-to-Java
 * LiveConnect (and other "ProxyJNI" stuff).  _Very_ noisy and slow.
 */
#define JEP_DEBUG_PROXYJNI_OPTION "jep.debug.proxyjni"
#define JEP_DEBUG_PRINTING_OPTION "jep.debug.printing"

#import <Carbon/Carbon.h>
#import <JavaVM/jni.h>

#ifndef JNI_VERSION_1_4
#define JNI_VERSION_1_4 0x00010004
#endif

enum AppletStateJEP {
	kAppletCreateJEP	= 1,
	kAppletInitJEP		= 2,
	kAppletStartJEP		= 3,
	kAppletStopJEP		= 4,
	kAppletDestroyJEP	= 5,
	kAppletErrorJEP		= 7
};
typedef enum AppletStateJEP AppletStateJEP;

typedef CALLBACK_API_C( void , JEPShowDocumentCallback )(jobject applet, CFURLRef url, CFStringRef windowName, void *userData);
typedef CALLBACK_API_C( void , JEPShowStatusCallback )(jobject applet, CFStringRef statusMessage, void *userData);
typedef CALLBACK_API_C( void , JEPAppletCreatedCallback )(void *handle, JavaVM *vm, jobject applet, WindowRef window, AppletStateJEP status, void *userData);
typedef CALLBACK_API_C( void , JEPAppletInitializedCallback )(void *handle, JavaVM *vm, jobject applet, WindowRef window, AppletStateJEP status, void *userData);
typedef CALLBACK_API_C( void , JEPAppletStartedCallback )(void *handle, JavaVM *vm, jobject applet, WindowRef window, AppletStateJEP status, void *userData);

typedef struct {
	Rect appletClip;		// In parentWindow's coordinate system
	int appletPosX;			// In parentWindow's coordinate system, upper-left origin
	int appletPosY;			// In parentWindow's coordinate system, upper-left origin
	unsigned appletWidth;	// May be overridden by applet's "WIDTH" tag
	unsigned appletHeight;	// May be overridden by applet's "HEIGHT" tag

	Boolean createdInvisible;
	Boolean erasePreviousFrame;
	Boolean resizeAppletDirectly;

	JavaVMOption *jvmExtraOpts;
	int jvmExtraOptsNum;

	JEPShowDocumentCallback showDocumentCallback;
	void *showDocumentUserData;
	JEPShowStatusCallback showStatusCallback;
	void *showStatusUserData;
	/* appletCreatedCallback, if specified, will _always_ (we hope) be
	 * called if JEPCreateJavaApplet() returned no error -- even if the
	 * applet ultimately failed to load or if it was destroyed while
	 * loading.  At this point the applet has been created (or it failed
	 * to load, or was created and destroyed), but it hasn't necessarily
	 * either been initialized or started (and it might never be).  The
	 * callback might happen before JEPCreateJavaApplet() returns.  It
	 * might be called on a secondary thread (not the main thread) -- as
	 * of JEP 0.8.7 it always is.
	 */
	JEPAppletCreatedCallback appletCreatedCallback;
	void *appletCreatedUserData;
	/* appletInitializedCallback is called just after the applet has been
	 * initialized.  The callback might happen before JEPCreateJavaApplet()
	 * returns.  It might be called on a secondary thread (not the main
	 * thread).
	 */
	JEPAppletInitializedCallback appletInitializedCallback;
	void *appletInitializedUserData;
	/* appletStartedCallback is called just after the applet has been started.
	 * The callback might happen before JEPCreateJavaApplet() returns.  It
	 * might be called on a secondary thread (not the main thread) -- though
	 * as of JEP 0.8.7 it's always called on the main thread.
	 */
	JEPAppletStartedCallback appletStartedCallback;
	void *appletStartedUserData;
	void *instance; // For Mozilla.org browsers, a MRJPluginInstance*.
} JEPOptions;

/* JEPCheckBundleWithVersion() checks whether the Java Embedding Plugin
 * matches the specified version string and has everything that it needs.
 * Always call this first.  The version check prevents version mismatches
 * between JavaEmbeddingPlugin.bundle and the MRJ Plugin JEP.  (As of JEP
 * 0.8.7, this function replaces JEPCheckBundle(), which is no longer
 * supported.)
 */
Boolean JEPCheckBundleWithVersion(CFStringRef version);
/* JEPGetInstalledPath() returns the full path to the Java Embedding Plugin's
 * location.
 */
CFStringRef JEPGetInstalledPath(void);
/* Call JEPCreateJavaVM() instead of calling JNI_CreateJavaVM() directly --
 * otherwise the JVM you create may not have all the options we need (such as
 * an appropriate xbootclasspath).
 */
OSStatus JEPCreateJavaVM(JavaVMOption *jvmExtraOpts, int jvmExtraOptsNum,
	JavaVM **vm, void **env);
OSStatus JEPCreateJavaVMAndInitAWT(JavaVMOption *jvmExtraOpts,
	int jvmExtraOptsNum, JavaVM **vm, void **env,
	void (*runAfterInitAWT) (void *), void *runAfterInitAWTInfo);
/* Given input from an HTML tag in either of the two standard forms
 * (< ... "name"="value" ... > or <param name="name" value="value">),
 * an entry in "attributes" should have "name" as its key and "value"
 * as its value.  Only returns an error if appletCreatedCallback will
 * _not_ be called.
 */
OSStatus JEPCreateJavaApplet(CFURLRef docbase, CFDictionaryRef attributes,
	WindowRef parentWindow, JEPOptions *options, void **handle);
OSStatus JEPGetJavaApplet(void *handle, jobject *applet);
OSStatus JEPGetJavaVM(void *handle, JavaVM **vm);
OSStatus JEPSetJavaAppletState(void *handle, AppletStateJEP newState);
OSStatus JEPGetJavaAppletState(void *handle, AppletStateJEP *state);
/* A call to JEPDrawJavaApplet() tells the JEP to redraw the (entire) applet.
 * A call to JEPMaybeDrawJavaApplet() tells the JEP to redraw or not redraw
 * the applet, at its own discretion -- a client makes this call if it thinks
 * a given update might be excessive or unnecessary, but isn't completely
 * sure.  For now I use JEPMaybeDrawJavaApplet() to deal with excessive
 * updates from the Mozilla family of browsers, particularly recent versions
 * of Camino (e.g. 0.9a1).
 */
OSStatus JEPDrawJavaApplet(void *handle);
OSStatus JEPMaybeDrawJavaApplet(void *handle);
/* Unlike their counterparts in the original Java Embedding API,
 * JEPSuspendJavaAppletUpdates() and JEPResumeJavaAppletUpdates() suspend and
 * resume _all_ updates -- not just asynchronous ones.  And calls to them
 * may be nested.
 */
OSStatus JEPSuspendJavaAppletUpdates(void *handle);
OSStatus JEPResumeJavaAppletUpdates(void *handle);
OSStatus JEPSizeJavaApplet(void *handle, int width, int height);
/* posX and posY are in JEPCreateJavaApplet's parentWindow's coordinate
 * system.
 */
OSStatus JEPPositionJavaApplet(void *handle, int posX, int posY);
/* posX and posY are in JEPCreateJavaApplet's parentWindow's coordinate
 * system.
 */
OSStatus JEPMoveAndClipJavaApplet(void *handle, int posX, int posY,
	int clipX, int clipY, int clipWidth, int clipHeight);
OSStatus JEPShowHideJavaApplet(void *handle, Boolean visible);
OSStatus JEPRegisterStatusCallback(void *handle,
	JEPShowStatusCallback showStatusFunction, void *userData);
OSStatus JEPRegisterShowDocumentCallback(void *handle,
	JEPShowDocumentCallback showDocumentFunction, void * userData);
/* Convenience functions to access Java Console "runtime parameters" from
 * outside the JVM, possibly before it has been started.
 */
Boolean JEPGetJavaRuntimeParameterBoolean(CFStringRef key);
CFStringRef JEPGetJavaRuntimeParameterString(CFStringRef key);
/* Convenience function to log messages to System.out.
 */
OSStatus JEPLogJavaMessage(CFStringRef message);
/* In Tiger (both Java 1.4.2_07 and 1.5.0), LiveConnect calls that trigger
 * Java drawing can no longer be made on the main thread (hangs will result).
 * But all the Mozilla-family's LiveConnect calls are made on the main
 * thread, so we need a way to call some of them on another thread and wait
 * for the results.  At the same time we can ensure that an autorelease pool
 * is always present (a problem we used to use hookJNIRunnable in AppletView.m
 * to fix).
 */
void JEPDoLiveConnectProxy(jclass proxyClass, jobject runnable, jobject applet,
	jboolean isCallMethodCall);
void JEPRunOnMainThread(void (*runnable) (void *), void *info, Boolean wait);
/* JEPRunJRunnableOnMainThread is a hack, unused as of JEP 0.8.6, that might
 * not stay in the Java Embedding Plugin API.
 */
jthrowable JEPRunJRunnableOnMainThread(jobject jrunnable);

#ifdef __cplusplus
}
#endif
