/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the MRJ Carbon OJI Plugin.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corp.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Patrick C. Beard <beard@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

/*
	JavaMessageQueue.h
 */

#pragma once

#ifndef JNI_H
#include "jni.h"
#endif

#include "MRJPlugin.h"

class Monitor;

class JavaMessage : public nsISupports {
public:
	JavaMessage();
	JavaMessage(MRJPluginInstance *pluginInstance);
	virtual ~JavaMessage();

	NS_DECL_ISUPPORTS

	void setNext(JavaMessage* next) { mNext = next; }
	JavaMessage* getNext() { return mNext; }
	MRJPluginInstance *getPluginInstance() { return mPluginInstance; }
	Boolean isFinalizeMessage() { return mIsFinalizeMessage; }
	Boolean waitTilDone(long long millis = 0);
	Boolean waitAWhile(long long millis = 2000);
	void setIsDone(Boolean release = true);
	Boolean timedOut(Boolean release = true);
	/* In Tiger, we need to avoid making LiveConnect calls on the main thread
	 * that might trigger Java drawing (using JEPDoLiveConnectProxy()).
	 * Fortunately (since avoiding the main thread is very costly), only a
	 * few calls (the CallMethod ones) need this treatment.
	 */
	void setCallMethodCall() { mCallMethodCall = true; }
	Boolean isCallMethodCall() { return mCallMethodCall; }

	virtual void execute(JNIEnv* env) = 0;

private:
	Boolean mDone;
	Boolean mCallMethodCall;
	JavaMessage* mNext;
	Monitor *mMessageMonitor;
protected:
	MRJPluginInstance *mPluginInstance;
	Boolean mIsFinalizeMessage;
};

class JavaMessageQueue {
public:
	JavaMessageQueue(Monitor* monitor);

	void putMessage(JavaMessage* msg);
	JavaMessage* getMessage();

	void enter();
	void exit();

	void wait();
	void wait(long long millis);
	void notify();

private:
	// Message queue.
	JavaMessage* mFirst;
	JavaMessage* mLast;
	Monitor* mMonitor;
};

/*
 * Wraps a JavaMessage in an nsIRunnable form, so that it runs on the correct
 * native thread.
 */
class MessageRunnable : public JavaMessage, public nsIRunnable {
public:
    MessageRunnable(nsPluginThreadPtr threadID, JavaMessage* msg);
    virtual ~MessageRunnable();

    NS_DECL_ISUPPORTS_INHERITED

    virtual void execute(JNIEnv* env);

    NS_IMETHOD Run();
    static void RunAlt(void *info);

private:
    nsPluginThreadPtr mThreadID;
    JavaMessage* mMessage;
};

