/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the MRJ Carbon OJI Plugin.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corp.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Patrick C. Beard <beard@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

/*
    LiveConnectNativeMethods.cpp
 */

#include <memory>

#include "LiveConnectNativeMethods.h"

/* Two and a half years ago the MRJ Plugin changed from using
 * nsIComponentManager to using nsIComponentManagerObsolete, to preserve
 * binary compatibility with older Mozilla (pre 0.9.4?) versions after
 * the original nsIComponentManager interface was "obsoleted"
 * (https://bugzilla.mozilla.org/show_bug.cgi?id=129250#c10 and
 * https://bugzilla.mozilla.org/show_bug.cgi?id=98553).  Now we need to
 * undo this change, because Mozilla.org is about to dump support for the
 * nsIComponentManagerObsolete interface
 * (https://bugzilla.mozilla.org/show_bug.cgi?id=267040).  Binary
 * compatibility with pre-1.0 versions should no longer be a concern.
 */
#include "nsIComponentManager.h"
#include "nsIPluginManager.h"
#include "nsIJVMManager.h"
#include "nsILiveconnect.h"
#include "nsISecureLiveconnect.h"
#include "nsIPluginInstancePeer2.h"

#ifndef NS_COMPONENTMANAGER_CID
#define NS_COMPONENTMANAGER_CID                      \
{ /* 91775d60-d5dc-11d2-92fb-00e09805570f */         \
    0x91775d60,                                      \
    0xd5dc,                                          \
    0x11d2,                                          \
    {0x92, 0xfb, 0x00, 0xe0, 0x98, 0x05, 0x57, 0x0f} \
}
#endif

#include "MRJPlugin.h"
#include "MRJContext.h"
#include "MRJSession.h"
#include "MRJSecurityContext.h"
#include "CSecureEnv.h"
#include "JavaMessageQueue.h"
#include "MRJMonitor.h"
#include "NativeMonitor.h"
#include "StringUtils.h"

#include "netscape_javascript_JSObject.h"   /* javah-generated headers */

#include "jsapi.h"
#import <mach-o/dyld.h>
#import "JavaEmbeddingPlugin.h"
extern "C" Boolean JEPCheckBundle();
static unsigned jsObjectTimeout = 2000;  /* In milliseconds */

extern nsIPluginManager* thePluginManager;
extern Boolean IsNested(void);
extern void ProcessMainThreadEvents(void);

extern MRJPlugin *theJVMPlugin;
static nsILiveconnect* theLiveConnectManager = NULL;
static nsIComponentManager* theComponentManager = NULL;
nsPluginThreadPtr theJavaScriptThread = 0;
Boolean liveConnectInited = false;

static jclass netscape_javascript_JSObject = NULL;
static jmethodID netscape_javascript_JSObject_JSObject_applet = NULL;
static jmethodID netscape_javascript_JSObject_JSObject_instance = NULL;
static jmethodID netscape_javascript_JSObject_getApplet = NULL;
static jmethodID netscape_javascript_JSObject_getPluginInstance = NULL;
static jfieldID netscape_javascript_JSObject_internal = NULL;

static jclass netscape_oji_JNIUtils = NULL;
static jmethodID netscape_oji_JNIUtils_NewLocalRef = NULL;
static jmethodID netscape_oji_JNIUtils_GetCurrentThread = NULL;
static jmethodID netscape_oji_JNIUtils_GetCurrentClassLoader = NULL;
static jmethodID netscape_oji_JNIUtils_GetObjectClassLoader = NULL;
static jmethodID netscape_oji_JNIUtils_PushLCInfo = NULL;
static jmethodID netscape_oji_JNIUtils_PopLCInfo = NULL;
static jmethodID netscape_oji_JNIUtils_CurrentLCPluginInstance = NULL;

static NS_DEFINE_IID(kLiveConnectCID, NS_CLIVECONNECT_CID);
static NS_DEFINE_IID(kComponentManagerCID, NS_COMPONENTMANAGER_CID);

static nsresult getGlobalComponentManager(nsIComponentManager* *result)
{
    return MRJPlugin::GetService(kComponentManagerCID, NS_GET_IID(nsIComponentManager), (void**)result);
}

static MRJMonitor *callEvalMonitor = NULL;

static MRJSecurityContext *staticSecurityContext = NULL;

// From js/src/liveconnect/jsj_private.h
struct JSObjectHandle {
    JSObject *js_obj;
    void *rt;	// Originally JSRuntime *rt
};
typedef struct JSObjectHandle JSObjectHandle;

//JSRuntime *JS_GetRuntime(JSContext *cx);
//void *JS_malloc(JSContext *cx, size_t nbytes);
//void JS_ClearPendingException(JSContext *cx);
typedef void * (*JSGETRUNTIME) (JSContext *);
typedef void * (*JSMALLOC) (JSContext *, size_t);
typedef void (*JSCLEAREXCEPTION) (JSContext *);
static JSGETRUNTIME JS_GetRuntimePtr = NULL;
static JSMALLOC JS_mallocPtr = NULL;
static JSCLEAREXCEPTION JS_ClearPendingExceptionPtr = NULL;

static Boolean findJSUtils(void)
{
	unsigned long i, count;
	NSSymbol symbol = nil;
	const struct mach_header *image = nil;
	count = _dyld_image_count();
	if (JS_GetRuntimePtr && JS_mallocPtr && JS_ClearPendingExceptionPtr)
		return true;
	for (i = 1; i <= count; ++i) {
		image = (const struct mach_header *) _dyld_get_image_header(i - 1);
		if (!JS_GetRuntimePtr) {
			symbol = NSLookupSymbolInImage(image, "_JS_GetRuntime",
				NSLOOKUPSYMBOLINIMAGE_OPTION_RETURN_ON_ERROR);
			if (symbol)
				JS_GetRuntimePtr = (JSGETRUNTIME) NSAddressOfSymbol(symbol);
		}
		if (!JS_mallocPtr) {
			symbol = NSLookupSymbolInImage(image, "_JS_malloc",
				NSLOOKUPSYMBOLINIMAGE_OPTION_RETURN_ON_ERROR);
			if (symbol)
				JS_mallocPtr = (JSMALLOC) NSAddressOfSymbol(symbol);
		}
		if (!JS_ClearPendingExceptionPtr) {
			symbol = NSLookupSymbolInImage(image, "_JS_ClearPendingException",
				NSLOOKUPSYMBOLINIMAGE_OPTION_RETURN_ON_ERROR);
			if (symbol) {
				JS_ClearPendingExceptionPtr = (JSCLEAREXCEPTION)
					NSAddressOfSymbol(symbol);
			}
		}
		if (JS_GetRuntimePtr && JS_mallocPtr && JS_ClearPendingExceptionPtr)
			return true;
	}
	return false;
}

nsresult InitLiveConnectSupport(MRJPlugin *jvmPlugin)
{
    if (!jvmPlugin)
        return NS_ERROR_FAILURE;

    findJSUtils();

    /* See JavaEmbeddingPlugin.h for info */
    if (JEPGetJavaRuntimeParameterBoolean(CFSTR(JEP_JSOBJECT_NOTIMEOUT_OPTION)))
       jsObjectTimeout = 0;

    getGlobalComponentManager(&theComponentManager);

    nsresult result = MRJPlugin::GetService(kLiveConnectCID, NS_GET_IID(nsILiveconnect),
                                            (void**)&theLiveConnectManager);
    if (result != NS_OK)
        return result;

    /* This needs to be created as early as possible -- if it's created on
     * demand, we may end up with more than one of it (which makes it useless
     * as a "monitor").
     */
    callEvalMonitor = new MRJMonitor(jvmPlugin->getSession());

    staticSecurityContext = new MRJSecurityContext(NULL);
    NS_IF_ADDREF(staticSecurityContext);

    // Manually load the required native methods.
    static JNINativeMethod nativeMethods[] = {
        "getMember", "(Ljava/lang/String;)Ljava/lang/Object;", (void*)&Java_netscape_javascript_JSObject_getMember,
        "getSlot", "(I)Ljava/lang/Object;", (void*)&Java_netscape_javascript_JSObject_getSlot,
        "setMember", "(Ljava/lang/String;Ljava/lang/Object;)V", (void*)&Java_netscape_javascript_JSObject_setMember,
        "setSlot", "(ILjava/lang/Object;)V", (void*)&Java_netscape_javascript_JSObject_setSlot,
        "removeMember", "(Ljava/lang/String;)V", (void*)&Java_netscape_javascript_JSObject_removeMember,
        "call", "(Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/Object;", (void*)&Java_netscape_javascript_JSObject_call,
        "eval", "(Ljava/lang/String;)Ljava/lang/Object;", (void*)&Java_netscape_javascript_JSObject_eval,
        "toString", "()Ljava/lang/String;", (void*)&Java_netscape_javascript_JSObject_toString,
        "getWindow", "(Ljava/applet/Applet;)Lnetscape/javascript/JSObject;", (void*)&Java_netscape_javascript_JSObject_getWindow,
        "finalize", "()V", (void*)&Java_netscape_javascript_JSObject_finalize,
    };
    
    MRJSession* session = jvmPlugin->getSession();
    JNIEnv* env = session->getCurrentEnv();
    if (env != NULL) {
        // load netscape.oji.JNIUtils class.
        jclass classJNIUtils = env->FindClass("netscape/oji/JNIUtils");
        if (classJNIUtils != NULL) {
            netscape_oji_JNIUtils = (jclass) env->NewGlobalRef(classJNIUtils);
            env->DeleteLocalRef(classJNIUtils);
            netscape_oji_JNIUtils_NewLocalRef = env->GetStaticMethodID(netscape_oji_JNIUtils, "NewLocalRef", "(Ljava/lang/Object;)Ljava/lang/Object;");
            netscape_oji_JNIUtils_GetCurrentThread = env->GetStaticMethodID(netscape_oji_JNIUtils, "GetCurrentThread", "()Ljava/lang/Object;");
            netscape_oji_JNIUtils_GetCurrentClassLoader = env->GetStaticMethodID(netscape_oji_JNIUtils, "GetCurrentClassLoader", "()Ljava/lang/Object;");
            netscape_oji_JNIUtils_GetObjectClassLoader = env->GetStaticMethodID(netscape_oji_JNIUtils, "GetObjectClassLoader", "(Ljava/lang/Object;)Ljava/lang/Object;");
            netscape_oji_JNIUtils_PushLCInfo = env->GetStaticMethodID(netscape_oji_JNIUtils, "PushLCInfo", "(I)V");
            netscape_oji_JNIUtils_PopLCInfo = env->GetStaticMethodID(netscape_oji_JNIUtils, "PopLCInfo", "()V");
            netscape_oji_JNIUtils_CurrentLCPluginInstance = env->GetStaticMethodID(netscape_oji_JNIUtils, "CurrentLCPluginInstance", "()I");
        }

        jclass classJSObject = env->FindClass("netscape/javascript/JSObject");
        if (classJSObject != NULL) {
            // register LiveConnect native methods.
            netscape_javascript_JSObject = (jclass) env->NewGlobalRef(classJSObject);
            env->DeleteLocalRef(classJSObject);
            
            netscape_javascript_JSObject_JSObject_applet = env->GetMethodID(netscape_javascript_JSObject, "<init>", "(ILjava/applet/Applet;)V");
            if (env->ExceptionCheck()) {
                env->ExceptionClear();
                result = NS_ERROR_FAILURE;
            }

            netscape_javascript_JSObject_JSObject_instance = env->GetMethodID(netscape_javascript_JSObject, "<init>", "(II)V");
            if (env->ExceptionCheck()) {
                env->ExceptionClear();
                result = NS_ERROR_FAILURE;
            }

            netscape_javascript_JSObject_getApplet = env->GetMethodID(netscape_javascript_JSObject, "getApplet", "()Ljava/applet/Applet;");
            if (env->ExceptionCheck()) {
                env->ExceptionClear();
                result = NS_ERROR_FAILURE;
            }

            netscape_javascript_JSObject_getPluginInstance = env->GetMethodID(netscape_javascript_JSObject, "getPluginInstance", "()I");
            if (env->ExceptionCheck()) {
                env->ExceptionClear();
                result = NS_ERROR_FAILURE;
            }

            netscape_javascript_JSObject_internal = env->GetFieldID(netscape_javascript_JSObject, "internal", "I");
            if (env->ExceptionCheck()) {
                env->ExceptionClear();
                result = NS_ERROR_FAILURE;
            }

            env->RegisterNatives(netscape_javascript_JSObject, nativeMethods, sizeof(nativeMethods) / sizeof(JNINativeMethod));
            if (env->ExceptionCheck()) {
                env->ExceptionClear();
                result = NS_ERROR_FAILURE;
            }
        }
    }

    return result;
}

nsresult PostInitLiveConnectSupport(MRJPlugin *jvmPlugin)
{
    if (!jvmPlugin)
        return NS_ERROR_FAILURE;
    theJVMPlugin = jvmPlugin;
    liveConnectInited = true;
    return noErr;
}

nsresult ShutdownLiveConnectSupport()
{
    theJVMPlugin = NULL;
    liveConnectInited = false;
    
    NS_IF_RELEASE(staticSecurityContext);

    NS_IF_RELEASE(theLiveConnectManager);
    NS_IF_RELEASE(theComponentManager);
    
    return NS_OK;
}

static jobject ToGlobalRef(JNIEnv* env, jobject localRef);

jobject Wrap_JSObject(JNIEnv *env, jsobject js_obj, jobject applet)
{
    jobject retval = env->NewObject(netscape_javascript_JSObject,
        netscape_javascript_JSObject_JSObject_applet, js_obj, applet);
    if (retval)
        retval = ToGlobalRef(env, retval);
    return retval;
}

jobject Wrap_JSObject(JNIEnv *env, jsobject js_obj, MRJPluginInstance *instance)
{
    jobject retval = env->NewObject(netscape_javascript_JSObject,
        netscape_javascript_JSObject_JSObject_instance, js_obj, (jint) instance);
    if (retval)
        retval = ToGlobalRef(env, retval);
    return retval;
}

jobject GetAppletForWrapper(JNIEnv *env, jobject java_wrapper_obj)
{
    jobject retval = env->CallObjectMethod(java_wrapper_obj,
        netscape_javascript_JSObject_getApplet);
    if (retval)
        retval = ToGlobalRef(env, retval);
    return retval;
}

MRJPluginInstance *GetInstanceForWrapper(JNIEnv *env, jobject java_wrapper_obj)
{
    return (MRJPluginInstance *) env->CallObjectMethod(java_wrapper_obj,
        netscape_javascript_JSObject_getPluginInstance);
}

/* Wrap_JSObject() and Unwrap_JSObject() may, like ToGlobalRef() below, be
 * called either from a "Java thread" or a "native thread".  We need the
 * same fix here that we use in ToGlobalRef().  See that routine for more
 * info.
 */
jobject Wrap_JSObject(JNIEnv* env, jsobject js_obj)
{
    if (!theJVMPlugin)
        return (jobject) NULL;
    MRJSession* session = theJVMPlugin->getSession();
    env = session->getCurrentEnv();
    jmethodID constructorID = env->GetMethodID(netscape_javascript_JSObject, "<init>", "(I)V");
    jobject retval = NULL;
    jobject obj = env->NewObject(netscape_javascript_JSObject, constructorID, js_obj);
    if (obj) {
        retval = env->NewGlobalRef(obj);
        env->DeleteLocalRef(obj);
    }
    return retval;
}

jsobject Unwrap_JSObject(JNIEnv* env, jobject java_wrapper_obj)
{
    if (!theJVMPlugin)
        return (jsobject) NULL;
    MRJSession* session = theJVMPlugin->getSession();
    env = session->getCurrentEnv();
    return env->GetIntField(java_wrapper_obj, netscape_javascript_JSObject_internal);
}

static jobject NewLocalRef(JNIEnv* env, jobject global_ref)
{
    return env->CallStaticObjectMethod(netscape_oji_JNIUtils, netscape_oji_JNIUtils_NewLocalRef, global_ref);
}

/* This routine may be called either from a "Java thread" (with 'env' set to a
 * pointer to an ordinary JNIEnv structure) or from a "native thread" (with
 * 'env' set to a pointer to a ProxyJNIEnv class).  The calls from Java
 * threads work fine.  And so should the calls from native threads -- the
 * ProxyJNIEnv class is just a subclass of the JNIEnv structure, and similar
 * calls work fine inside the browser.  But inside the MRJ Plugin, for some
 * reason the compiler "forgets" (or never knows) that it's dealing with a
 * ProxyJNIEnv pointer, which leads to crashes when this routine tries to
 * call NewGlobalRef() (presumably what the compiler thinks is the function
 * pointer really is something else).  The easiest fix for this is to use
 * MRJSession.cpp's getCurrentEnv() to make sure that we're always dealing
 * with an ordinary JNIEnv pointer.
 */
static jobject ToGlobalRef(JNIEnv* env, jobject localRef)
{
    if (!theJVMPlugin || !localRef)
        return NULL;
    MRJSession* session = theJVMPlugin->getSession();
    env = session->getCurrentEnv();
    jobject globalRef = env->NewGlobalRef(localRef);
    env->DeleteLocalRef(localRef);
    return globalRef;
}

static jobject ToLocalRef(JNIEnv* env, jobject globalRef)
{
    if (!globalRef)
        return NULL;
    jobject localRef = NewLocalRef(env, globalRef);
    env->DeleteGlobalRef(globalRef);
    return localRef;
}

#if 0
static jobject GetCurrentThread(JNIEnv* env)
{
    return env->CallStaticObjectMethod(netscape_oji_JNIUtils, netscape_oji_JNIUtils_GetCurrentThread);
}
#endif

static JSContext *getJSContext(MRJPluginInstance *pluginInstance)
{
    JSContext *context = NULL;
    if (MRJPluginInstance::isValidInstance(pluginInstance)) {
        nsIPluginInstancePeer *peer;
        if (pluginInstance->GetPeer(&peer) == NS_OK) {
            nsIPluginInstancePeer2 *peer2 = NULL;
            if (peer->QueryInterface(NS_GET_IID(nsIPluginInstancePeer2), (void**) &peer2) == NS_OK) {
                if (peer2->GetJSContext(&context) != NS_OK)
                    context = NULL;
                NS_RELEASE(peer2);
            }
            NS_RELEASE(peer);
        }
    }
    return context;
}

/**
 * Security Considerations.
 */

MRJSecurityContext::MRJSecurityContext(const char* location)
    :   mLocation(nsnull), mConnection(nsnull)
{
    if (location)
        mLocation = ::strdup(location);
    if (mLocation) {
        // find the 3rd slash of the URL, hopefully it's in canonical form.
        char* colon = ::strchr(mLocation, ':');
        if (colon) {
            if (colon[1] == '/' && colon[2] == '/') {
                char* slash = ::strchr(colon + 3, '/');
                if (slash) *slash = '\0';
            }
        }
    }

    if (theComponentManager) {
        theComponentManager->CreateInstance(kLiveConnectCID, nsnull, NS_GET_IID(nsILiveconnect),
                                            (void**)&mConnection);
    } else {
        mConnection = theLiveConnectManager;
        NS_IF_ADDREF(mConnection);
    }
}

MRJSecurityContext::~MRJSecurityContext()
{
    delete[] mLocation;
    NS_IF_RELEASE(mConnection);
}

// work around a bug in Metrowerks pre-processor.
NS_IMPL_ISUPPORTS1(MRJSecurityContext, nsISecurityContext)

NS_METHOD MRJSecurityContext::Implies(const char* target, const char* action, PRBool *bAllowedAccess)
{
    *bAllowedAccess = (target != NULL && action == NULL);
    return NS_OK;
}

NS_METHOD 
MRJSecurityContext::GetOrigin(char* buf, int len)
{
    if (mLocation) {
        ::strncpy(buf, mLocation, len);
        return NS_OK;
    }
    return NS_ERROR_FAILURE;
}

NS_METHOD 
MRJSecurityContext::GetCertificateID(char* buf, int len)
{
    // ACTION: Implement me.
    return NS_ERROR_NOT_IMPLEMENTED;
}

// NOTE:  this a weak reference to the MRJSecurityContext associated with this
// plugin instance. The MRJSecurityContext is owned by the MRJContext.

static MRJSecurityContext* getSecurityContext(MRJPluginInstance* pluginInstance)
{
    if (MRJPluginInstance::isValidInstance(pluginInstance)) {
        MRJContext* context = pluginInstance->getContext();
        if (context == NULL)
            return NULL;
        MRJSecurityContext* securityContext = context->getSecurityContext();
        if (securityContext == NULL) {
            securityContext = new MRJSecurityContext(context->getDocumentBase());
            context->setSecurityContext(securityContext);
        }
        return securityContext;
    }
    return NULL;
}

inline nsILiveconnect* getLiveconnectInstance(MRJSecurityContext* securityContext)
{
    if (!theJVMPlugin)
        return NULL;
    return (securityContext ? securityContext->getConnection() : theLiveConnectManager);
}

static jobject GetCurrentClassLoader(JNIEnv* env)
{
    return env->CallStaticObjectMethod(netscape_oji_JNIUtils, netscape_oji_JNIUtils_GetCurrentClassLoader);
}

static jobject GetObjectClassLoader(JNIEnv* env, jobject object)
{
    return env->CallStaticObjectMethod(netscape_oji_JNIUtils, netscape_oji_JNIUtils_GetObjectClassLoader, object);
}

static void PushLCInfo(JNIEnv* env, jint pluginInstance)
{
	env->CallStaticVoidMethod(netscape_oji_JNIUtils, netscape_oji_JNIUtils_PushLCInfo, pluginInstance);
}

static void PopLCInfo(JNIEnv* env)
{
	env->CallStaticVoidMethod(netscape_oji_JNIUtils, netscape_oji_JNIUtils_PopLCInfo);
}

static MRJPluginInstance *CurrentLCPluginInstance(JNIEnv* env)
{
	return (MRJPluginInstance *)
		env->CallStaticIntMethod(netscape_oji_JNIUtils, netscape_oji_JNIUtils_CurrentLCPluginInstance);
}

/* Returns a plugin instance in which an applet has been created (if at least
 * one exists), without being too picky about which one gets returned.  If
 * we're running in a non-privileged context (i.e. if GetCurrentClassLoader()
 * doesn't return NULL), we return any plugin instance whose applet was loaded
 * by the same class loader (applets that have the same docbase will share the
 * same class loader).  GetAppletInstance() and GetUniqueInstance() are more
 * picky (and therefore more secure), but need more information than just the
 * current JNIEnv.  GetAnInstance() is useful when we really should be running
 * in some applet's context (i.e. when we're in a JSObject.call() or eval(),
 * or a JavaScript-to-Java CallMethod call, either of which may call Java).
 */
static MRJPluginInstance *GetAnInstance(JNIEnv* env)
{
    MRJPluginInstance* pluginInstance = NULL;
    jobject classLoader = GetCurrentClassLoader(env);
    if (classLoader != NULL) {
        pluginInstance = MRJPluginInstance::getInstances();
        while (pluginInstance && !pluginInstance->isDestroying()) {
            jobject applet = NULL;
            pluginInstance->GetJavaObject(&applet);
            if (applet) {
                jobject appletClassLoader = GetObjectClassLoader(env, applet);
                jboolean sameClassLoader = env->IsSameObject(appletClassLoader, classLoader);
                env->DeleteLocalRef(appletClassLoader);
                if (sameClassLoader)
                    break;
            }
            pluginInstance = pluginInstance->getNextInstance();
        }
        env->DeleteLocalRef(classLoader);
    }
    return pluginInstance;
}

/**
 * Maps the given applet to a given plugin instance by walking the PluginInstance list looking
 * for an instance which contains the current applet. 
 * Note:  the reference count of the plugin instance isn't affected by this call.
 */
MRJPluginInstance* GetAppletInstance(JNIEnv* env, jobject applet)
{
    if (!applet || !env->IsInstanceOf(applet, env->FindClass("java/applet/Applet")))
       return (MRJPluginInstance *) NULL;
    MRJPluginInstance* pluginInstance = MRJPluginInstance::getInstances();
    while (pluginInstance && !pluginInstance->isDestroying()) {
        jobject object = NULL;
        pluginInstance->GetJavaObject(&object);
        if (object && env->IsSameObject(applet, object))
            return pluginInstance;
        pluginInstance = pluginInstance->getNextInstance();
    }
    return NULL;
}

/* Returns a plugin instance in which an applet has been created, whose code
 * base is the same as that of our nsISecurityContext parameter.  More than
 * one applet may share the same code base.  GetSecContextInstance() is useful
 * when we really should be running in some applet's context (i.e. when we're
 * in a JSObject.call() or eval(), or a JavaScript-to-Java CallMethod call,
 * either of which may call Java).
 */
MRJPluginInstance *GetSecContextInstance(nsISecurityContext *context)
{
    if (!context)
        return (MRJPluginInstance *) NULL;
    char codebase[4096] = {0};
    if (context->GetOrigin(codebase, sizeof(codebase)-1) != NS_OK)
        return (MRJPluginInstance *) NULL;
    MRJPluginInstance *pluginInstance = MRJPluginInstance::getInstances();
    while (pluginInstance && !pluginInstance->isDestroying()) {
        jobject object = NULL;
        pluginInstance->GetJavaObject(&object);
        if (!object) {
            pluginInstance = pluginInstance->getNextInstance();
            continue;
        }
        nsISecurityContext *aContext =
            (nsISecurityContext *) getSecurityContext(pluginInstance);
        char aCodebase[4096] = {0};
        if (aContext->GetOrigin(aCodebase, sizeof(aCodebase)-1) == NS_OK) {
            if (strncmp(codebase, aCodebase, sizeof(codebase)-1) == 0) {
                return pluginInstance;
            }
        }
        pluginInstance = pluginInstance->getNextInstance();
    }
    return NULL;
}

/* Use this instead of GetAnInstance(JNIEnv *) above when calling from one
 * of the non-static JSObject "native" methods.  This routine guarantees a
 * one-to-one correspondence between a JSObject object and a plugin instance.
 */
static MRJPluginInstance* GetUniqueInstance(JNIEnv* env, jobject java_wrapper_obj)
{
    MRJPluginInstance *retval = GetInstanceForWrapper(env, java_wrapper_obj);
    if (!retval) {
        jobject applet = GetAppletForWrapper(env, java_wrapper_obj);
        if (applet) {
            retval = GetAppletInstance(env, applet);
            env->DeleteGlobalRef(applet);
        }
    }
    return retval;
}

/* The extra initial AddRef() gives us more control over when we'll be
 * released.  (It stops us from being released after the next pair of calls
 * to AddRef() and Release(), but requires us to call Release() on ourselves
 * at the appropriate time.  As a result, we'll be released after the end of
 * the _last_ pair of calls (including our own pair) to AddRef() and
 * Release().)
 */
MessageRunnable::MessageRunnable(nsPluginThreadPtr threadID, JavaMessage* msg)
    : mThreadID(threadID), mMessage(msg)
{
	AddRef();
	NS_ADDREF(mMessage);
	mPluginInstance = mMessage->getPluginInstance();
	NS_IF_ADDREF(mPluginInstance);
}

MessageRunnable::~MessageRunnable()
{
	NS_RELEASE(mMessage);
	NS_IF_RELEASE(mPluginInstance);
}

NS_IMPL_ADDREF_INHERITED(MessageRunnable, JavaMessage)
NS_IMPL_RELEASE_INHERITED(MessageRunnable, JavaMessage)

NS_METHOD MessageRunnable::QueryInterface(const nsIID& aIID, void** aInstancePtr)
{
    if (aInstancePtr == NULL) {
        return NS_ERROR_NULL_POINTER;
    }
    *aInstancePtr = NULL;
    if (aIID.Equals(NS_GET_IID(nsIRunnable))) {
        *aInstancePtr = (void *) (nsIRunnable *) this;
        AddRef();
        return NS_OK;
    }
    return JavaMessage::QueryInterface(aIID, aInstancePtr);
}

void MessageRunnable::execute(JNIEnv* env)
{
    if (mMessage->timedOut()) {
        Release();
        return;
    }
    // because a spontaneous Java thread called us, we have to switch to the JavaScript thread
    // to handle this request.
    nsIJVMThreadManager* threadManager = NULL;
    if (MRJPlugin::GetService(nsIJVMManager::GetCID(), nsIJVMThreadManagerIID, (void**)&threadManager) == NS_OK) {
        /* As of JEP 0.8.7 we post our event asynchrously, instead of
         * synchronously as we used to.  This helps to avoid hangs as
         * we're waiting for JavaScript methods (which always run on
         * the main thread) to return.
         */
        threadManager->PostEvent(mThreadID, this, PR_TRUE);
        NS_RELEASE(threadManager);
    }
}

NS_IMETHODIMP MessageRunnable::Run()
{
    if (mMessage->timedOut()) {
        Release();
        return NS_OK;
    }
    nsIJVMManager* javaManager = NULL;
    if (MRJPlugin::GetService(nsIJVMManager::GetCID(), NS_GET_IID(nsIJVMManager), (void**)&javaManager) == NS_OK) {
        JNIEnv* proxyEnv = NULL;
        JNIEnv *javaEnv = theJVMPlugin->getSession()->getCurrentEnv();
        if (javaEnv && (javaManager->GetProxyJNI(&proxyEnv) == NS_OK) && proxyEnv) {
            PushLCInfo(javaEnv, (unsigned) mPluginInstance);
            mMessage->execute(proxyEnv);
            PopLCInfo(javaEnv);
        }
        NS_RELEASE(javaManager);
    }
    Release();
    return NS_OK;
}

void MessageRunnable::RunAlt(void *info)
{
    if (!info)
        return;
    MessageRunnable *self = (MessageRunnable *) info;
    self->Run();
}

static void sendMessage(JNIEnv* env, JavaMessage* msg)
{
    if (!theJVMPlugin)
        return;

    /* This routine is now _much_ simpler (and faster).  The original MRJ
     * Plugin Carbon ran CSecureEnv::Run() on a number of threads to process
     * messages sent from here.  But that's completely unnecessary (and when
     * we came through here on the main thread it actually caused hangs).
     */

    if (env == theJVMPlugin->getSession()->getMainEnv()) {
        MessageRunnable* runnableMsg =
            new MessageRunnable(theJavaScriptThread, msg);
        if (runnableMsg)
            runnableMsg->Run();
    } else if (JEPCheckBundle()) {
        MessageRunnable *runnableMsg =
            new MessageRunnable(theJavaScriptThread, msg);
        if (runnableMsg)
            JEPRunOnMainThread(runnableMsg->RunAlt, runnableMsg, false);
    } else if (theJavaScriptThread) {
        MessageRunnable *runnableMsg =
            new MessageRunnable(theJavaScriptThread, msg);
        if (runnableMsg)
            runnableMsg->execute(NULL);
    }
}

static nsIPrincipal* newCodebasePrincipal(const char* codebaseURL)
{
    nsIPrincipal* principal = NULL;
#if 0
    nsICapsManager* capsManager = NULL;
    static NS_DEFINE_IID(kICapsManagerIID, NS_ICAPSMANAGER_IID);
    if (thePluginManager->QueryInterface(kICapsManagerIID, &capsManager) == NS_OK) {
        if (capsManager->CreateCodebasePrincipal(codebaseURL, &principal) != NS_OK)
            principal = NULL;
        capsManager->Release();
    }
#endif
    return principal;
}

/****************** Implementation of methods of JSObject *******************/

/*
 * Class:     netscape_javascript_JSObject
 * Method:    getMember
 * Signature: (Ljava/lang/String;)Ljava/lang/Object;
 */

class GetMemberMessage : public JavaMessage {
    jsobject mObject;
    const jchar* mPropertyName;
    jsize mLength;
    jstring mJPropertyName;
    jobject* mResultObject;
public:
    GetMemberMessage(MRJPluginInstance* pluginInstance, jsobject js_obj,
                     const jchar* propertyName, jsize nameLength,
                     jstring jPropertyName, jobject* member)
        :   JavaMessage(pluginInstance), mObject(js_obj), mPropertyName(propertyName),
            mLength(nameLength), mJPropertyName(jPropertyName), mResultObject(member)
    {
    }

    virtual void execute(JNIEnv* env)
    {
        if (timedOut())
            return;
        MRJSecurityContext* securityContext = NULL;
        // mPluginInstance will be NULL if we were called from static code in
        // an applet.
        if (mPluginInstance) {
            securityContext = getSecurityContext(mPluginInstance);
        } else {
            securityContext = staticSecurityContext;
        }
        if (securityContext == NULL) {
            setIsDone();
            return;
        }
        nsILiveconnect* connection = getLiveconnectInstance(securityContext);
        if (connection == NULL) {
            setIsDone();
            return;
        }
        if (timedOut())
            return;
        jobject member;
        nsresult result = connection->GetMember(env, mObject, mPropertyName, mLength, NULL, 0, securityContext, &member);
        JNIEnv *javaEnv = theJVMPlugin->getSession()->getCurrentEnv();
        if (javaEnv)
            javaEnv->ReleaseStringChars(mJPropertyName, mPropertyName);
        if (timedOut())
            return;
        if (result == NS_OK) {
            // convert reference to a global reference, in case we're switching threads.
            *mResultObject = ToGlobalRef(env, member);
        }
        setIsDone();
    }
};

JNIEXPORT jobject JNICALL
Java_netscape_javascript_JSObject_getMember(JNIEnv* env,
                                            jobject java_wrapper_obj,
                                            jstring property_name_jstr)
{
    if (property_name_jstr == NULL) {
        env->ThrowNew(env->FindClass("java/lang/NullPointerException"), "illegal null member name");
        return NULL;
    }

    MRJPluginInstance* pluginInstance = GetUniqueInstance(env, java_wrapper_obj);
    if (!pluginInstance)
        pluginInstance = CurrentLCPluginInstance(env);
    if (pluginInstance && !MRJPluginInstance::isValidInstance(pluginInstance))
        return NULL;
    // pluginInstance will be NULL if we were called from static code in
    // an applet.
    NS_IF_ADDREF(pluginInstance);

    /* Get the Unicode string for the JS property name */
    jboolean is_copy;
    const jchar* property_name_ucs2 = env->GetStringChars(property_name_jstr, &is_copy);
    jsize property_name_len = env->GetStringLength(property_name_jstr);

    jsobject js_obj = Unwrap_JSObject(env, java_wrapper_obj);
    jobject member = NULL;

    GetMemberMessage *msg = new GetMemberMessage(pluginInstance, js_obj,
        property_name_ucs2, property_name_len, property_name_jstr, &member);
    if (msg) {
        sendMessage(env, msg);
        msg->waitTilDone(jsObjectTimeout);
    }

    // convert the resulting reference back to a local reference.
    if (member != NULL)
        member = ToLocalRef(env, member);
    
    NS_IF_RELEASE(pluginInstance);
    return member;
}

/*
 * Class:     netscape_javascript_JSObject
 * Method:    getSlot
 * Signature: (I)Ljava/lang/Object;
 */

class GetSlotMessage : public JavaMessage {
    jsobject mObject;
    jint mSlot;
    jobject* mResultObject;
public:
    GetSlotMessage(MRJPluginInstance* pluginInstance, jsobject js_obj,
                   jint slot, jobject* member)
        :   JavaMessage(pluginInstance), mObject(js_obj), mSlot(slot), mResultObject(member)
    {
    }

    virtual void execute(JNIEnv* env)
    {
        if (timedOut())
            return;
        MRJSecurityContext* securityContext = NULL;
        // mPluginInstance will be NULL if we were called from static code in
        // an applet.
        if (mPluginInstance) {
            securityContext = getSecurityContext(mPluginInstance);
        } else {
            securityContext = staticSecurityContext;
        }
        if (securityContext == NULL) {
            setIsDone();
            return;
        }
        nsILiveconnect* connection = getLiveconnectInstance(securityContext);
        if (connection == NULL) {
            setIsDone();
            return;
        }
        if (timedOut())
            return;
        jobject member;
        nsresult result = connection->GetSlot(env, mObject, mSlot, NULL, 0, securityContext, &member);
        if (timedOut())
            return;
        if (result == NS_OK) {
            // convert reference to a global reference, in case we're switching threads.
            *mResultObject = ToGlobalRef(env, member);
        }
        setIsDone();
    }
};

JNIEXPORT jobject JNICALL
Java_netscape_javascript_JSObject_getSlot(JNIEnv* env,
                                          jobject java_wrapper_obj,
                                          jint slot)
{
    MRJPluginInstance* pluginInstance = GetUniqueInstance(env, java_wrapper_obj);
    if (!pluginInstance)
        pluginInstance = CurrentLCPluginInstance(env);

    if (pluginInstance && !MRJPluginInstance::isValidInstance(pluginInstance))
        return NULL;
    // pluginInstance will be NULL if we were called from static code in
    // an applet.
    NS_IF_ADDREF(pluginInstance);

    jsobject js_obj = Unwrap_JSObject(env, java_wrapper_obj);
    jobject member = NULL;

    GetSlotMessage *msg = new GetSlotMessage(pluginInstance, js_obj, slot, &member);
    if (msg) {
        sendMessage(env, msg);
        msg->waitTilDone(jsObjectTimeout);
    }

    // convert the resulting reference back to a local reference.
    if (member != NULL)
        member = ToLocalRef(env, member);

    NS_IF_RELEASE(pluginInstance);
    return member;
}

/*
 * Class:     netscape_javascript_JSObject
 * Method:    setMember
 * Signature: (Ljava/lang/String;Ljava/lang/Object;)V
 */

class SetMemberMessage : public JavaMessage {
    jsobject mObject;
    const jchar* mPropertyName;
    jsize mLength;
    jstring mJPropertyName;
    jobject mJavaObject;
public:
    SetMemberMessage(MRJPluginInstance* pluginInstance, jsobject js_obj,
                     const jchar* propertyName, jsize nameLength,
                     jstring jPropertyName, jobject java_obj)
        :   JavaMessage(pluginInstance), mObject(js_obj), mPropertyName(propertyName),
            mLength(nameLength), mJPropertyName(jPropertyName), mJavaObject(java_obj)
    {
    }

    virtual void execute(JNIEnv* env)
    {
        if (!theJVMPlugin)
            return;
        MRJSecurityContext* securityContext = NULL;
        // mPluginInstance will be NULL if we were called from static code in
        // an applet.
        if (mPluginInstance) {
            securityContext = getSecurityContext(mPluginInstance);
        } else {
            securityContext = staticSecurityContext;
        }
        if (securityContext == NULL) {
            Release();
            return;
        }
        nsILiveconnect* connection = getLiveconnectInstance(securityContext);
        if (connection == NULL) {
            Release();
            return;
        }
        connection->SetMember(env, mObject, mPropertyName, mLength, mJavaObject, 0, 0, securityContext);
        JNIEnv *javaEnv = theJVMPlugin->getSession()->getCurrentEnv();
        if (javaEnv) {
            javaEnv->ReleaseStringChars(mJPropertyName, mPropertyName);
            if (mJavaObject)
                javaEnv->DeleteGlobalRef(mJavaObject);
        }
        Release();
    }
};

JNIEXPORT void JNICALL
Java_netscape_javascript_JSObject_setMember(JNIEnv* env,
                                            jobject java_wrapper_obj,
                                            jstring property_name_jstr,
                                            jobject java_obj)
{
    if (property_name_jstr == NULL) {
        env->ThrowNew(env->FindClass("java/lang/NullPointerException"), "illegal null member name");
        return;
    }

    MRJPluginInstance* pluginInstance = GetUniqueInstance(env, java_wrapper_obj);
    if (!pluginInstance)
        pluginInstance = CurrentLCPluginInstance(env);
    if (pluginInstance && !MRJPluginInstance::isValidInstance(pluginInstance))
        return;
    // pluginInstance will be NULL if we were called from static code in
    // an applet.
    NS_IF_ADDREF(pluginInstance);

    /* Get the Unicode string for the JS property name */
    jboolean is_copy;
    const jchar* property_name_ucs2 = env->GetStringChars(property_name_jstr, &is_copy);
    jsize property_name_len = env->GetStringLength(property_name_jstr);

    /* Add an extra global ref for java_obj (to make sure it _is_ global) */
    if (java_obj)
        java_obj = (jobjectArray) env->NewGlobalRef(java_obj);
    
    jsobject js_obj = Unwrap_JSObject(env, java_wrapper_obj);

    SetMemberMessage *msg = new SetMemberMessage(pluginInstance, js_obj,
        property_name_ucs2, property_name_len, property_name_jstr, java_obj);
    if (msg) {
        sendMessage(env, msg);
    }

    NS_IF_RELEASE(pluginInstance);
}

/*
 * Class:     netscape_javascript_JSObject
 * Method:    setSlot
 * Signature: (ILjava/lang/Object;)V
 */

class SetSlotMessage : public JavaMessage {
    jsobject mObject;
    jint mSlot;
    jobject mJavaObject;
public:
    SetSlotMessage(MRJPluginInstance* pluginInstance, jsobject js_obj, jint slot, jobject java_obj)
        :   JavaMessage(pluginInstance), mObject(js_obj), mSlot(slot), mJavaObject(java_obj)
    {
    }

    virtual void execute(JNIEnv* env)
    {
        MRJSecurityContext* securityContext = NULL;
        // mPluginInstance will be NULL if we were called from static code in
        // an applet.
        if (mPluginInstance) {
            securityContext = getSecurityContext(mPluginInstance);
        } else {
            securityContext = staticSecurityContext;
        }
        if (securityContext == NULL) {
            Release();
            return;
        }
        nsILiveconnect* connection = getLiveconnectInstance(securityContext);
        if (connection == NULL) {
            Release();
            return;
        }
        connection->SetSlot(env, mObject, mSlot, mJavaObject, 0, 0, securityContext);
        if (mJavaObject) {
            JNIEnv *javaEnv = theJVMPlugin->getSession()->getCurrentEnv();
            if (javaEnv)
                javaEnv->DeleteGlobalRef(mJavaObject);
        }
		Release();
    }
};

JNIEXPORT void JNICALL
Java_netscape_javascript_JSObject_setSlot(JNIEnv* env,
                                          jobject java_wrapper_obj,
                                          jint slot,
                                          jobject java_obj)
{
    MRJPluginInstance* pluginInstance = GetUniqueInstance(env, java_wrapper_obj);
    if (!pluginInstance)
        pluginInstance = CurrentLCPluginInstance(env);

    if (pluginInstance && !MRJPluginInstance::isValidInstance(pluginInstance))
        return;
    // pluginInstance will be NULL if we were called from static code in
    // an applet.
    NS_IF_ADDREF(pluginInstance);

    /* Add an extra global ref for java_obj (to make sure it _is_ global) */
    if (java_obj)
        java_obj = (jobjectArray) env->NewGlobalRef(java_obj);

    jsobject js_obj = Unwrap_JSObject(env, java_wrapper_obj);
    
    SetSlotMessage *msg = new SetSlotMessage(pluginInstance, js_obj, slot, java_obj);
    if (msg) {
        sendMessage(env, msg);
    }

    NS_IF_RELEASE(pluginInstance);
}

/*
 * Class:     netscape_javascript_JSObject
 * Method:    removeMember
 * Signature: (Ljava/lang/String;)V
 */

class RemoveMemberMessage : public JavaMessage {
    jsobject mObject;
    const jchar* mPropertyName;
    jsize mLength;
    jstring mJPropertyName;
public:
    RemoveMemberMessage(MRJPluginInstance* pluginInstance, jsobject obj,
                        const jchar* propertyName, jsize length,
                        jstring jPropertyName)
        :   JavaMessage(pluginInstance), mObject(obj),
            mPropertyName(propertyName), mLength(length),
            mJPropertyName(jPropertyName)
    {
    }

    virtual void execute(JNIEnv* env)
    {
        if (!theJVMPlugin)
            return;
        MRJSecurityContext* securityContext = NULL;
        // mPluginInstance will be NULL if we were called from static code in
        // an applet.
        if (mPluginInstance) {
            securityContext = getSecurityContext(mPluginInstance);
        } else {
            securityContext = staticSecurityContext;
        }
        if (securityContext == NULL) {
            Release();
            return;
        }
        nsILiveconnect* connection = getLiveconnectInstance(securityContext);
        if (connection == NULL) {
            Release();
            return;
        }
        connection->RemoveMember(env, mObject, mPropertyName, mLength, NULL, 0, securityContext);
        JNIEnv *javaEnv = theJVMPlugin->getSession()->getCurrentEnv();
        if (javaEnv)
            javaEnv->ReleaseStringChars(mJPropertyName, mPropertyName);
        Release();
    }
};

JNIEXPORT void JNICALL
Java_netscape_javascript_JSObject_removeMember(JNIEnv* env,
                                               jobject java_wrapper_obj,
                                               jstring property_name_jstr)
{
    if (property_name_jstr == NULL) {
        env->ThrowNew(env->FindClass("java/lang/NullPointerException"), "illegal null member name");
        return;
    }

    MRJPluginInstance* pluginInstance = GetUniqueInstance(env, java_wrapper_obj);
    if (!pluginInstance)
        pluginInstance = CurrentLCPluginInstance(env);
    if (pluginInstance && !MRJPluginInstance::isValidInstance(pluginInstance))
        return;
    // pluginInstance will be NULL if we were called from static code in
    // an applet.
    NS_IF_ADDREF(pluginInstance);

    /* Get the Unicode string for the JS property name */
    jboolean is_copy;
    const jchar* property_name_ucs2 = env->GetStringChars(property_name_jstr, &is_copy);
    jsize property_name_len = env->GetStringLength(property_name_jstr);

    jsobject js_obj = Unwrap_JSObject(env, java_wrapper_obj);
    RemoveMemberMessage *msg = new RemoveMemberMessage(pluginInstance, js_obj,
        property_name_ucs2, property_name_len, property_name_jstr);
    if (msg) {
        sendMessage(env, msg);
    }

    NS_IF_RELEASE(pluginInstance);
}

/* The Mozilla-family browsers have trouble handling several calls to
 * nsILiveconnect::Call() or nsILiveconnect::Eval() in quick succession --
 * particularly if user interaction is involved (e.g. a JavaScript alert()
 * box gets displayed), which holds open the call that spawned it (e.g. while
 * the alert() box is waiting for the user to press OK), and seems to cause
 * Call() or Eval() to be interrupted by subsequent Calls or Evals, which can
 * lead to crashes on exiting.  The problem is worst on Tiger (probably thanks
 * to its new Java "threading model") -- only Mozilla's Carbon-based browsers
 * (Firefox and "SeaMonkey") seem to have this problem on earlier OSes.  A
 * good solution is to use the following monitor to space out calls to Call()
 * or Eval():  As long as we don't exceed the two-second timeout, only one
 * call to Call() or Eval() may be active at any one time.  And if we do
 * exceed the timeout, there's a two-second gap between the start of one call
 * the start of the next -- which allows us (on Carbon browsers) to check for
 * a popup window after the timeout has expired, and (if necessary) extend it
 * indefinitely.
 */

/*
 * Class:     netscape_javascript_JSObject
 * Method:    call
 * Signature: (Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/Object;
 */
class CallMessage : public JavaMessage {
    jsobject mObject;
    const jchar* mFunctionName;
    jsize mLength;
    jstring mJFunctionName;
    jobjectArray mJavaArgs;
    jobject* mJavaResult;
public:
    CallMessage(MRJPluginInstance* pluginInstance, jsobject obj,
                const jchar* functionName, jsize length, jstring jFunctionName,
                jobjectArray javaArgs, jobject* javaResult)
        :   JavaMessage(pluginInstance), mObject(obj), mFunctionName(functionName),
            mLength(length), mJFunctionName(jFunctionName), mJavaArgs(javaArgs),
            mJavaResult(javaResult)
    {
    }

    virtual void execute(JNIEnv* env)
    {
        if (timedOut())
            return;
        MRJSecurityContext* securityContext = NULL;
        // mPluginInstance will be NULL if we were called from static code in
        // an applet.
        if (mPluginInstance) {
            securityContext = getSecurityContext(mPluginInstance);
        } else {
            securityContext = staticSecurityContext;
        }
        if (securityContext == NULL) {
            setIsDone();
            return;
        }
        nsILiveconnect* connection = getLiveconnectInstance(securityContext);
        if (connection == NULL) {
            setIsDone();
            return;
        }
        if (timedOut())
            return;
        jobject jresult = NULL;
        nsresult result = connection->Call(env, mObject, mFunctionName, mLength, mJavaArgs, NULL, 0, securityContext, &jresult);
        JNIEnv *javaEnv = theJVMPlugin->getSession()->getCurrentEnv();
        if (javaEnv) {
            javaEnv->ReleaseStringChars(mJFunctionName, mFunctionName);
            if (mJavaArgs)
                javaEnv->DeleteGlobalRef(mJavaArgs);
        }
        if (timedOut())
            return;
        if (result == NS_OK)
            *mJavaResult = ToGlobalRef(env, jresult);
        setIsDone();
    }
};

JNIEXPORT jobject JNICALL
Java_netscape_javascript_JSObject_call(JNIEnv* env, jobject java_wrapper_obj,
                                       jstring function_name_jstr, jobjectArray java_args)
{
    if (function_name_jstr == NULL) {
        env->ThrowNew(env->FindClass("java/lang/NullPointerException"), "illegal null function name");
        return NULL;
    }

    if (callEvalMonitor == NULL)
        return NULL;

    /* Try to determine which plugin instance is responsible for this thread. */
    MRJPluginInstance* pluginInstance = GetUniqueInstance(env, java_wrapper_obj);
    if (!pluginInstance)
        pluginInstance = CurrentLCPluginInstance(env);
    if (!pluginInstance)
        pluginInstance = GetAnInstance(env);
    if (pluginInstance && !MRJPluginInstance::isValidInstance(pluginInstance))
        return NULL;
    // pluginInstance will be NULL if we were called from static code in
    // an applet.
    NS_IF_ADDREF(pluginInstance);

    /* Get the Unicode string for the JS function name */
    jboolean is_copy;
    const jchar* function_name_ucs2 = env->GetStringChars(function_name_jstr, &is_copy);
    jsize function_name_len = env->GetStringLength(function_name_jstr);

    /* Add an extra global ref for java_args (to make sure it _is_ global)
	 * Doing this clears up the following crash bug --
	 * http://sourceforge.net/tracker/index.php?func=detail&aid=1638030&group_id=107955&atid=649116.
	 * As far as I can tell we don't need to do the same for the objects
	 * contained in java_args.
	 */
    if (java_args)
        java_args = (jobjectArray) env->NewGlobalRef(java_args);

    jsobject js_obj = Unwrap_JSObject(env, java_wrapper_obj);
    jobject jresult = NULL;

    CallMessage *msg = new CallMessage(pluginInstance, js_obj, function_name_ucs2,
        function_name_len, function_name_jstr, java_args, &jresult);
    if (msg) {
        callEvalMonitor->enter();
        sendMessage(env, msg);
        Boolean done = msg->waitAWhile(jsObjectTimeout);
        /* If our "message" hasn't finished executing by the end of the normal
         * timeout, check for popup windows that have been spawned as children
         * of the current browser window.  If any are found we extend the
         * timeout indefinitely -- on the assumption that the reason our
         * "message" hasn't finished is that it's waiting for user interaction
         * in the popped-up window (which is presumably a JavaScript alert()
         * box or some other kind of dialog).  (This solution doesn't work for
         * Cocoa browsers like Camino -- their JavaScript dialogs seem to be
         * "pure" Cocoa windows, which aren't included in the Carbon "window
         * list".  But that's not a big deal, since orphaning a call to
         * nsILiveconnect::Call() doesn't seem to cause crashes on exit in
         * Cocoa browsers (if we use callEvalMonitor).)
         */
        if (!done) {
            long long timeout = 100;
            WindowRef browserWindow = NULL;
            WindowRef activeWindow = GetUserFocusWindow();
            nsPluginWindow *pluginWindow = NULL;
            if (pluginInstance)
                pluginWindow = pluginInstance->getWindow();
            if (pluginWindow) {
                browserWindow = GetWindowFromPort(pluginWindow->window->port);
            } else {
                browserWindow = ActiveNonFloatingWindow();
            }
            if (activeWindow && browserWindow &&
                (activeWindow != browserWindow))
            {
                WindowRef prevWindow = GetPreviousWindow(browserWindow);
                while (prevWindow) {
                    if (prevWindow == activeWindow) {
                        timeout = 0;
                        break;
                    }
                    prevWindow = GetPreviousWindow(prevWindow);
                }
            }
            msg->waitTilDone(timeout);
        }
        callEvalMonitor->exit();
    }

    if (jresult != NULL)
        jresult = ToLocalRef(env, jresult);

    NS_IF_RELEASE(pluginInstance);
    return jresult;
}

/*
 * Class:     netscape_javascript_JSObject
 * Method:    eval
 * Signature: (Ljava/lang/String;)Ljava/lang/Object;
 */
class EvalMessage : public JavaMessage {
    jsobject mObject;
    const jchar* mScript;
    jsize mLength;
    jstring mJScript;
    jobject* mJavaResult;
public:
    EvalMessage(MRJPluginInstance* pluginInstance, jsobject obj, const jchar* script,
                jsize length, jstring jScript, jobject* javaResult)
        :   JavaMessage(pluginInstance), mObject(obj), mScript(script),
            mLength(length), mJScript(jScript), mJavaResult(javaResult)
    {
    }

    virtual void execute(JNIEnv* env)
    {
        if (timedOut())
            return;
        jobject jresult = NULL;
        MRJSecurityContext* securityContext = NULL;
        // mPluginInstance will be NULL if we were called from static code in
        // an applet.
        if (mPluginInstance) {
            securityContext = getSecurityContext(mPluginInstance);
        } else {
            securityContext = staticSecurityContext;
        }
        if (securityContext == NULL) {
            setIsDone();
            return;
        }
        nsILiveconnect* connection = getLiveconnectInstance(securityContext);
        if (connection == NULL) {
            setIsDone();
            return;
        }
        if (timedOut())
            return;
        nsresult result = connection->Eval(env, mObject, mScript, mLength, NULL, 0, securityContext, &jresult);
        JNIEnv *javaEnv = theJVMPlugin->getSession()->getCurrentEnv();
        if (javaEnv)
            javaEnv->ReleaseStringChars(mJScript, mScript);
        if (timedOut())
            return;
        if ((result == NS_OK) && (jresult != NULL))
            *mJavaResult = ToGlobalRef(env, jresult);
        setIsDone();
    }
};

JNIEXPORT jobject JNICALL
Java_netscape_javascript_JSObject_eval(JNIEnv* env,
                                       jobject java_wrapper_obj,
                                       jstring script_jstr)
{
    /* Get the Unicode string for the JS function name */
    if (script_jstr == NULL) {
        env->ThrowNew(env->FindClass("java/lang/NullPointerException"), "illegal null script string");
        return NULL;
    }

    if (callEvalMonitor == NULL)
        return NULL;

    jboolean is_copy;
    const jchar* script_ucs2 = env->GetStringChars(script_jstr, &is_copy);
    jsize script_len = env->GetStringLength(script_jstr);

    /* unwrap the JS object from the Java object. */
    jsobject js_obj = Unwrap_JSObject(env, java_wrapper_obj);
    jobject jresult = NULL;
    
    /* determine the plugin instance so we can obtain its codebase. */
    // beard: should file a bug with Apple that JMJNIToAWTContext doesn't work.
    MRJPluginInstance* pluginInstance = GetUniqueInstance(env, java_wrapper_obj);
    if (!pluginInstance)
        pluginInstance = CurrentLCPluginInstance(env);
    if (!pluginInstance)
        pluginInstance = GetAnInstance(env);
    if (pluginInstance && !MRJPluginInstance::isValidInstance(pluginInstance))
        return NULL;
    // pluginInstance will be NULL if we were called from static code in
    // an applet.
    NS_IF_ADDREF(pluginInstance);

    EvalMessage *msg = new EvalMessage(pluginInstance, js_obj, script_ucs2,
        script_len, script_jstr, &jresult);
    if (msg) {
        callEvalMonitor->enter();
        sendMessage(env, msg);
        Boolean done = msg->waitAWhile(jsObjectTimeout);
        /* If our "message" hasn't finished executing by the end of the normal
         * timeout, check for popup windows that have been spawned as children
         * of the current browser window.  If any are found we extend the
         * timeout indefinitely -- on the assumption that the reason our
         * "message" hasn't finished is that it's waiting for user interaction
         * in the popped-up window (which is presumably a JavaScript alert()
         * box or some other kind of dialog).  (This solution doesn't work for
         * Cocoa browsers like Camino -- their JavaScript dialogs seem to be
         * "pure" Cocoa windows, which aren't included in the Carbon "window
         * list".  But that's not a big deal, since orphaning a call to
         * nsILiveconnect::Eval() doesn't seem to cause crashes on exit in
         * Cocoa browsers (if we use callEvalMonitor).)
         */
        if (!done) {
            long long timeout = 100;
            WindowRef browserWindow = NULL;
            WindowRef activeWindow = GetUserFocusWindow();
            nsPluginWindow *pluginWindow = NULL;
            if (pluginInstance)
                pluginWindow = pluginInstance->getWindow();
            if (pluginWindow) {
                browserWindow = GetWindowFromPort(pluginWindow->window->port);
            } else {
                browserWindow = ActiveNonFloatingWindow();
            }
            if (activeWindow && browserWindow &&
                (activeWindow != browserWindow))
            {
                WindowRef prevWindow = GetPreviousWindow(browserWindow);
                while (prevWindow) {
                    if (prevWindow == activeWindow) {
                        timeout = 0;
                        break;
                    }
                    prevWindow = GetPreviousWindow(prevWindow);
                }
            }
            msg->waitTilDone(timeout);
        }
        callEvalMonitor->exit();
    }

    if (jresult != NULL)
        jresult = ToLocalRef(env, jresult);

    NS_IF_RELEASE(pluginInstance);
    return jresult;
}

/*
 * Class:     netscape_javascript_JSObject
 * Method:    toString
 * Signature: ()Ljava/lang/String;
 */
class ToStringMessage : public JavaMessage {
    jsobject mObject;
    jstring* mStringResult;
public:
    ToStringMessage(MRJPluginInstance* pluginInstance, jsobject js_obj, jstring* stringResult)
        :   JavaMessage(pluginInstance), mObject(js_obj), mStringResult(stringResult)
    {
    }

    virtual void execute(JNIEnv* env)
    {
        if (timedOut())
            return;
        MRJSecurityContext* securityContext = NULL;
        // mPluginInstance will be NULL if we were called from static code in
        // an applet.
        if (mPluginInstance) {
            securityContext = getSecurityContext(mPluginInstance);
        } else {
            securityContext = staticSecurityContext;
        }
        if (securityContext == NULL) {
            setIsDone();
            return;
        }
        nsILiveconnect* connection = getLiveconnectInstance(securityContext);
        if (connection == NULL) {
            setIsDone();
            return;
        }
        if (timedOut())
            return;
        jstring jresult = NULL;
        nsresult status = connection->ToString(env, mObject, &jresult);
        if (timedOut())
            return;
        if ((status == NS_OK) && (jresult != NULL))
            *mStringResult = (jstring) ToGlobalRef(env, jresult);
        setIsDone();
    }
};

JNIEXPORT jstring JNICALL
Java_netscape_javascript_JSObject_toString(JNIEnv* env, jobject java_wrapper_obj)
{
    /* unwrap the JS object from the Java object. */
    jstring jresult = NULL;
    jsobject js_obj = Unwrap_JSObject(env, java_wrapper_obj);

    MRJPluginInstance* pluginInstance = GetUniqueInstance(env, java_wrapper_obj);
    if (!pluginInstance)
        pluginInstance = CurrentLCPluginInstance(env);
    if (pluginInstance && !MRJPluginInstance::isValidInstance(pluginInstance))
        return NULL;
    // pluginInstance will be NULL if we were called from static code in
    // an applet.
    NS_IF_ADDREF(pluginInstance);

    ToStringMessage *msg = new ToStringMessage(pluginInstance, js_obj, &jresult);
    if (msg) {
        sendMessage(env, msg);
        msg->waitTilDone(jsObjectTimeout);
    }

    if (jresult != NULL)
        jresult = (jstring) ToLocalRef(env, jresult);
    
    NS_IF_RELEASE(pluginInstance);
    return jresult;
}

/*
 * Class:     netscape_javascript_JSObject
 * Method:    getWindow
 * Signature: (Ljava/applet/Applet;)Lnetscape/javascript/JSObject;
 */
class GetWindowMessage : public JavaMessage {
    jsobject* mWindowResult;
public:
    GetWindowMessage(MRJPluginInstance* pluginInstance, jsobject* windowResult)
        :   JavaMessage(pluginInstance), mWindowResult(windowResult)
    {
    }
    
    virtual void execute(JNIEnv* env)
    {
        if (timedOut())
            return;
        jsobject jresult = 0;
        MRJSecurityContext* securityContext = getSecurityContext(mPluginInstance);
        if (securityContext == NULL) {
            setIsDone();
            return;
        }
        nsILiveconnect* connection = getLiveconnectInstance(securityContext);
        if (connection == NULL) {
            setIsDone();
            return;
        }
        if (timedOut())
            return;
        nsresult status = connection->GetWindow(env, mPluginInstance, NULL, 0, securityContext, &jresult);
        if (timedOut())
            return;
        if ((status == NS_OK) && (jresult != 0))
            *mWindowResult = jresult;
        setIsDone();
    }
};

JNIEXPORT jobject JNICALL
Java_netscape_javascript_JSObject_getWindow(JNIEnv* env,
                                            jclass js_object_class,
                                            jobject java_applet_obj)
{
    MRJPluginInstance* pluginInstance = GetAppletInstance(env, java_applet_obj);
    if (MRJPluginInstance::isValidInstance(pluginInstance)) {
        NS_ADDREF(pluginInstance);
        jsobject jswindow = 0;
        GetWindowMessage *msg = new GetWindowMessage(pluginInstance, &jswindow);
        if (msg) {
            sendMessage(env, msg);
            msg->waitTilDone(jsObjectTimeout);
        }

        jobject retval = NULL;
        if (jswindow != 0)
            retval = Wrap_JSObject(env, jswindow, java_applet_obj);
        if (retval != NULL)
            retval = ToLocalRef(env, retval);
        NS_RELEASE(pluginInstance);
        return retval;
    }
    return NULL;
}

/* Used only by MRJGetJSObject() below, which needs to be able to obtain a
 * JSObject for an MRJPluginInstance that can be used in Java-to-JavaScript
 * LiveConnect (so that, for example, the JVM's proxy support will work
 * properly), without the MAYSCRIPT property having been set in one of that
 * plugin instance's tags.  The Mozilla.org browsers don't seem to have
 * anticipated the need for this, so I've had to hack out a solution myself.
 * (The JVM calls the FindProxyForURL() JavaScript method (from
 * AutoProxyHandler.getProxyInfo()) using Java-to-JavaScript LiveConnect,
 * just after an MRJPluginInstance has been created (and before its applet has
 * been loaded/created).)
 */
class GetWindowAltMessage : public JavaMessage {
    jsobject* mWindowResult;
public:
    GetWindowAltMessage(MRJPluginInstance* pluginInstance, jsobject* windowResult)
        :   JavaMessage(pluginInstance), mWindowResult(windowResult)
    {
    }
    
    virtual void execute(JNIEnv* env)
    {
        if (timedOut())
            return;
        jsobject jresult = 0;
        MRJSecurityContext* securityContext = getSecurityContext(mPluginInstance);
        if (securityContext == NULL) {
            setIsDone();
            return;
        }
        nsILiveconnect* connection = getLiveconnectInstance(securityContext);
        if (connection == NULL) {
            setIsDone();
            return;
        }
        if (timedOut())
            return;
        JSContext *jsContext = getJSContext(mPluginInstance);
        if (jsContext == NULL) {
            setIsDone();
            return;
        }
        if (timedOut())
            return;
        /* nsCLiveconnect::GetWindow() won't work if MAYSCRIPT isn't set, but
         * we still need to call it, because it initializes mJavaClient.  And
         * we need to squelch the exception that gets thrown when GetWindow()
         * fails:  JS_ClearPendingException() works fine in Firefox 2.0 and
         * on the trunk, but not in earlier Firefox versions (presumably
         * because they don't yet support JavaScript exceptions) -- those
         * earlier versions immediately throw a Java exception (in
         * jsj_exit_js()).  But fortunately the JNI Throw call goes through
         * our CSecureEnv class, so we're able to intervene to stop it from
         * having any effect.
         */
        CSecureEnv::suspendExceptions();
        nsresult status = connection->GetWindow(env, mPluginInstance, NULL, 0, securityContext, &jresult);
        JS_ClearPendingExceptionPtr(jsContext);
        CSecureEnv::resumeExceptions();
        if (timedOut())
            return;
        if ((status != NS_OK) || !jresult) {
            /* If GetWindow() fails because MAYSCRIPT isn't set, do here what
             * GetWindow() would normally do if it _had_ been set (and the
             * call to map_java_object_to_js_object() hadn't failed).
             */
            jresult = 0;
            JSObject *window = NULL;
            nsIPluginInstancePeer *peer;
            if (mPluginInstance->GetPeer(&peer) == NS_OK) {
                nsIPluginInstancePeer2 *peer2 = NULL;
                if (peer->QueryInterface(NS_GET_IID(nsIPluginInstancePeer2),
                    (void **) &peer2) == NS_OK)
                {
                    if (peer2->GetJSWindow(&window) != NS_OK)
                        window = NULL;
                    NS_RELEASE(peer2);
                }
                NS_RELEASE(peer);
            }
            if (window) {
                JSObjectHandle *handle = (JSObjectHandle *)
                    JS_mallocPtr(jsContext, sizeof(JSObjectHandle));
                if (handle) {
                    handle->js_obj = window;
                    handle->rt = JS_GetRuntimePtr(jsContext);
                    jresult = (jsobject) handle;
                }
            }
        }
        if (timedOut())
            return;
        *mWindowResult = jresult;
        setIsDone();
    }
};

#pragma export on

extern "C"
jobject MRJGetJSObject(JNIEnv *env, void *instance)
{
    MRJPluginInstance *pluginInstance = (MRJPluginInstance *) instance;
    if (!liveConnectInited || !MRJPluginInstance::isValidInstance(pluginInstance))
        return NULL;
    NS_ADDREF(pluginInstance);
    jsobject jswindow = 0;
    GetWindowAltMessage *msg = new GetWindowAltMessage(pluginInstance, &jswindow);
    if (msg) {
        sendMessage(env, msg);
        msg->waitTilDone(jsObjectTimeout);
    }
    jobject retval = NULL;
    if (jswindow)
        retval = Wrap_JSObject(env, jswindow, pluginInstance);
    NS_RELEASE(pluginInstance);
    return retval;
}

#pragma export off

/*
 * Class:     netscape_javascript_JSObject
 * Method:    finalize
 * Signature: ()V
 */
class FinalizeMessage : public JavaMessage {
    jsobject m_jsobj;
public:
	/* The extra initial AddRef() gives us more control over when we'll be
	 * released.  (It stops us from being released after the next pair of
	 * calls to AddRef() and Release(), but requires us to call Release() on
	 * ourselves at the appropriate time.  As a result, we'll be released
	 * after the end of the _last_ pair of calls (including our own pair) to
	 * AddRef() and Release().)
	 */
    FinalizeMessage(jsobject jsobj)
        :   JavaMessage(), m_jsobj(jsobj)
    {
        AddRef();
        mIsFinalizeMessage = true;
    }

    virtual void execute(JNIEnv* env)
    {
        theLiveConnectManager->FinalizeJSObject(env, m_jsobj);
        Release();
    }
};

JNIEXPORT void JNICALL
Java_netscape_javascript_JSObject_finalize(JNIEnv* env, jobject java_wrapper_obj)
{
    jsobject jsobj = Unwrap_JSObject(env, java_wrapper_obj);

    FinalizeMessage *msg = new FinalizeMessage(jsobj);
    if (msg) {
        sendMessage(env, msg);
    }
}
