/*
 * Copyright (c) 2004, Steven Michaud, All Rights Reserved
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#import "AppletView.h"

/* This routine and showDocumentLock are needed to stop Carbon browsers from
 * displaying new URLs too rapidly when earlier ones have popup windows (like
 * JavaScript alert boxes) awaiting user input.  For some reason this isn't
 * a problem with Cocoa browsers (like Camino).  Apparently, as of JEP 0.9.5+h
 * this problem no longer exists on OS X Tiger and Panther -- only on Jaguar
 * (OS X 10.2.8).
 */
static void waitForActivePopupClose(AppletView *appletView)
{
	WindowRef window;
	WindowRef browserWindow = nil;
	WindowRef activeWindow = nil;
	WindowRef unavailableWindow = nil;
	WindowModality carbonModal = 0;
	NSWindow *activeCocoaWindow = nil;
	Boolean popupFound, visible;
	WindowLatentVisibility latent;
	if ((osxVersion >= 0x1030) || !isCarbonApp || !appletView ||
		!appletView->myController || !appletView->myController->myWindowCarbon ||
		[appletView destroying])
		return;
	browserWindow = appletView->myController->myWindowCarbon;
	activeWindow = GetUserFocusWindow();
	if (!activeWindow || (activeWindow == browserWindow))
		return;
	popupFound = false;
	window = GetPreviousWindow(browserWindow);
	while (window) {
		if (window == activeWindow) {
			popupFound = true;
			break;
		}
		window = GetPreviousWindow(window);
	}
	if (!popupFound)
		return;
	/* If the active window has a corresponding NSWindow object, it can't be
	 * an alert box (or generally any child window popped up by the Carbon-
	 * based browser).  This stops us from freezing when showDocument is
	 * called from a JVM popup window.  And it clears up bugzilla.mozilla.org
	 * bug 311843.
	 */
	if ((activeCocoaWindow =
		[JEPController nsWindowFromCarbonWindow:activeWindow]) != nil)
		return;
	/* If we're not running Carbon-modal then we can't be in an alert box.
	 * This change clears up bugzilla.mozilla.org bug 356866.
	 */
	GetWindowModality(activeWindow, &carbonModal, &unavailableWindow);
	if (!carbonModal)
		return;
	/* We can't directly detect when activeWindow has closed.  And closing it
	 * doesn't (groan) even remove it from the window list (at least before
	 * AppletFrameContext_showDocument() below has returned).  Moreover
	 * IsWindowVisible(), IsWindowLatentVisible() and IsWindowActive() all
	 * sometimes return false negatives.  The following is the best we can
	 * manage, under the circumstances.  The "latent" stuff is to distinguish
	 * between activeWindow becoming invisible because the browser has been
	 * hidden or minimized, and because activeWindow has been closed.
	 */
	while (![appletView destroying]) {
		window = GetUserFocusWindow();
		visible = IsWindowLatentVisible(activeWindow, &latent);
		if ((window != activeWindow) && !visible && !latent)
			break;
		[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:.05]];
	}
}

static NSLock *showDocumentLock = nil;

/* "Native" implementation of showDocument() in AppletFrameContext.java */
JNIEXPORT void JNICALL Java_AppletFrameContext_showDocument(JNIEnv *env,
	jobject self, jlong cocoaParentView, jstring url, jstring windowName)
{
	const jchar *urlUTF16 = nil;
	const jchar *windowNameUTF16 = nil;
	jint urlLength, windowNameLength;
	NSString *urlStrCocoa = nil;
	NSURL *urlCocoa = nil;
	NSString *windowNameCocoa = nil;
	NSThread *currentThread;
	BOOL isMainThread = false;
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSAutoreleasePool *localPool;
	if (!mainThread || !appletView || !appletView->myController || !url ||
		[appletView destroying])
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	if (isCarbonApp && (osxVersion < 0x1030) && !showDocumentLock) {
		if ((showDocumentLock = [[NSLock alloc] init]) == nil) {
			[localPool release];
			return;
		}
	}
	currentThread = [NSThread currentThread];
	if ([mainThread isEqual:currentThread])
		isMainThread = true;
	urlLength = (*env)->GetStringLength(env, url);
	urlUTF16 = (*env)->GetStringChars(env, url, nil);
	if (windowName) {
		windowNameLength = (*env)->GetStringLength(env, windowName);
		windowNameUTF16 = (*env)->GetStringChars(env, windowName, nil);
	}
	if ([appletView checkJNIException:env]) {
		[localPool release];
		return;
	}
	if (urlUTF16) {
		urlStrCocoa = [NSString stringWithCharacters:urlUTF16
			length:urlLength];
		if (urlStrCocoa) {
			urlCocoa = [[NSURL alloc] initWithString:urlStrCocoa];
		}
		(*env)->ReleaseStringChars(env, url, urlUTF16);
		[appletView checkJNIException:env];
	}
	if (windowNameUTF16) {
		windowNameCocoa = [NSString stringWithCharacters:windowNameUTF16
			length:windowNameLength];
		(*env)->ReleaseStringChars(env, windowName, windowNameUTF16);
		[appletView checkJNIException:env];
	}
	if (isCarbonApp && (osxVersion < 0x1030) && !isMainThread) {
		[showDocumentLock lock];
		waitForActivePopupClose(appletView);
	}
	[appletView->myController performSelector:@selector(showURL:inFrame:)
		withObject:urlCocoa withObject:windowNameCocoa];
	if (isCarbonApp && (osxVersion < 0x1030) && !isMainThread) {
		/* We need to wait awhile here -- otherwise waitForActivePopupClose
		 * will get called (the next time through) before an alert box could
		 * have been popped up by this call to showDocument().
		 */
		[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:.6]];
		[showDocumentLock unlock];
	}
	[localPool release];
}

/* "Native" implementation of showStatus() in AppletFrameContext.java */
JNIEXPORT void JNICALL Java_AppletFrameContext_showStatus(JNIEnv *env,
	jobject self, jlong cocoaParentView, jstring statusMessage)
{
	const jchar *statusMessageUTF16 = nil;
	jint statusMessageLength;
	NSString *statusMessageCocoa = nil;
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSAutoreleasePool *localPool;
	if (!appletView || !appletView->myController || !statusMessage || [appletView destroying])
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	statusMessageLength = (*env)->GetStringLength(env, statusMessage);
	statusMessageUTF16 = (*env)->GetStringChars(env, statusMessage, nil);
	if ([appletView checkJNIException:env]) {
		[localPool release];
		return;
	}
	if (statusMessageUTF16) {
		statusMessageCocoa = [NSString stringWithCharacters:statusMessageUTF16
			length:statusMessageLength];
		(*env)->ReleaseStringChars(env, statusMessage, statusMessageUTF16);
		[appletView checkJNIException:env];
	}
	[appletView->myController performSelector:@selector(showStatus:)
		withObject:statusMessageCocoa];
	[localPool release];
}

typedef jobject (*MRJGETJSOBJECT) (JNIEnv *, void *);

static MRJGETJSOBJECT MRJGetJSObjectPtr = nil;

static BOOL findGetJSObject(void)
{
	unsigned long i, count;
	NSSymbol symbol = nil;
	const struct mach_header *image = nil;
	count = _dyld_image_count();
	if (MRJGetJSObjectPtr)
		return true;
	for (i = 1; i <= count; ++i) {
		image = (const struct mach_header *) _dyld_get_image_header(i - 1);
		symbol = NSLookupSymbolInImage(image, "_MRJGetJSObject",
			NSLOOKUPSYMBOLINIMAGE_OPTION_RETURN_ON_ERROR);
		if (symbol)
			MRJGetJSObjectPtr = (MRJGETJSOBJECT) NSAddressOfSymbol(symbol);
		if (MRJGetJSObjectPtr)
			return true;
	}
	return false;
}

JNIEXPORT jobject JNICALL Java_AppletFrameContext_nativeGetJSObject(JNIEnv *env,
	jobject self, jint pluginInstance)
{
	jobject retval = nil;
	JEPController *controller = nil;
	NSAutoreleasePool *localPool;
	if (!findGetJSObject() || !pluginInstance)
		return nil;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return nil;
	controller = [JEPController
		controllerFromPluginInstance:(void *)pluginInstance
		timeoutType:lockTimeoutNone];
	if (controller && [controller destroying]) {
		[localPool release];
		return nil;
	}
	if (controller && controller->myAppletView &&
		controller->myAppletView->myJSObject)
	{
		retval = controller->myAppletView->myJSObject;
	} else {
		retval = MRJGetJSObjectPtr(env, (void *) pluginInstance);
		if (retval && controller && controller->myAppletView)
			controller->myAppletView->myJSObject = retval;
	}
	[localPool release];
	return retval;
}

typedef unsigned (*MRJSETCOOKIE) (const char *, void *, unsigned);
typedef unsigned (*MRJGETCOOKIE) (const char *, void *, unsigned *);
#define COOKIE_MAX 8192

static MRJSETCOOKIE MRJSetCookiePtr = nil;
static MRJGETCOOKIE MRJGetCookiePtr = nil;

static BOOL findSetCookie(void)
{
	unsigned long i, count;
	NSSymbol symbol = nil;
	const struct mach_header *image = nil;
	count = _dyld_image_count();
	if (MRJSetCookiePtr)
		return true;
	for (i = 1; i <= count; ++i) {
		image = (const struct mach_header *) _dyld_get_image_header(i - 1);
		symbol = NSLookupSymbolInImage(image, "_MRJSetCookie",
			NSLOOKUPSYMBOLINIMAGE_OPTION_RETURN_ON_ERROR);
		if (symbol)
			MRJSetCookiePtr = (MRJSETCOOKIE) NSAddressOfSymbol(symbol);
		if (MRJSetCookiePtr)
			return true;
	}
	return false;
}

static BOOL findGetCookie(void)
{
	unsigned long i, count;
	NSSymbol symbol = nil;
	const struct mach_header *image = nil;
	count = _dyld_image_count();
	if (MRJGetCookiePtr)
		return true;
	for (i = 1; i <= count; ++i) {
		image = (const struct mach_header *) _dyld_get_image_header(i - 1);
		symbol = NSLookupSymbolInImage(image, "_MRJGetCookie",
			NSLOOKUPSYMBOLINIMAGE_OPTION_RETURN_ON_ERROR);
		if (symbol)
			MRJGetCookiePtr = (MRJGETCOOKIE) NSAddressOfSymbol(symbol);
		if (MRJGetCookiePtr)
			return true;
	}
	return false;
}

JNIEXPORT jboolean JNICALL Java_JEPCookieHandler_cookiesSupported(JNIEnv *env,
	jobject self)
{
	if (!findSetCookie() || !findGetCookie()) {
		return false;
	} else {
		return true;
	}
}

JNIEXPORT void JNICALL Java_JEPCookieHandler_setCookie(JNIEnv *env,
	jobject self, jstring url, jstring value)
{
	const char *urlUTF8 = nil;
	const char *valueUTF8 = nil;
	unsigned valueSize = 0;
	if (!findSetCookie() || !url)
		return;
	urlUTF8 = (*env)->GetStringUTFChars(env, url, nil);
	if (value) {
		valueUTF8 = (*env)->GetStringUTFChars(env, value, nil);
		valueSize = strlen(valueUTF8);
	}
	MRJSetCookiePtr(urlUTF8, (void *) valueUTF8, valueSize);
	if (urlUTF8)
		(*env)->ReleaseStringUTFChars(env, url, urlUTF8);
	if (valueUTF8)
		(*env)->ReleaseStringUTFChars(env, value, valueUTF8);
}

JNIEXPORT jstring JNICALL Java_JEPCookieHandler_getCookie(JNIEnv *env,
	jobject self, jstring url)
{
	jstring retval = nil;
	const char *urlUTF8 = nil;
	char valueUTF8[COOKIE_MAX] = {0};
	unsigned valueSize = COOKIE_MAX;
	if (!findGetCookie() || !url)
		return nil;
	urlUTF8 = (*env)->GetStringUTFChars(env, url, nil);
	if (MRJGetCookiePtr(urlUTF8, (void *) valueUTF8, &valueSize) == 0) {
		if (valueSize && (valueUTF8[0] != 0))
			retval = (*env)->NewStringUTF(env, valueUTF8);
	}
	if (urlUTF8)
		(*env)->ReleaseStringUTFChars(env, url, urlUTF8);
	return retval;
}

/* "Native" implementation of getNSColor() in AppletHandlerFactory.java */
JNIEXPORT jlong JNICALL Java_AppletHandlerFactory_getNSColor(JNIEnv *env,
	jclass clazz, jint red, jint green, jint blue)
{
	NSColor *retval = nil;
	float denom = 255;
	float fRed, fGreen, fBlue;
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return 0;
	if (red < 0) red = 0;
	else if (red > 255) red = 255;
	if (green < 0) green = 0;
	else if (green > 255) green = 255;
	if (blue < 0) blue = 0;
	else if (blue > 255) blue = 255;
	fRed = (float) red / denom;
	fGreen = (float) green / denom;
	fBlue = (float) blue / denom;
	retval = [[NSColor colorWithCalibratedRed:fRed green:fGreen blue:fBlue
		alpha:1.0] retain];
	[localPool release];
#ifdef SUPPORT_64BITS
	return (jlong) retval;
#else
	return (jlong) ((unsigned) retval);
#endif
}

/* "Native" implementation of getNSSystemColor() in AppletHandlerFactory.java */
JNIEXPORT jlong JNICALL Java_AppletHandlerFactory_getNSSystemColor(JNIEnv *env,
	jclass clazz, jstring colorName, jintArray rgb)
{
	id retval = nil;
	int arrayLen;
	float fRed = 0, fGreen = 0, fBlue = 0;
	jint intArray[3] = {0};
	const jchar *colorNameUTF16 = nil;
	jint colorNameLength;
	NSString *colorNameCocoa = nil;
	NSAutoreleasePool *localPool;

	if (!colorName || !rgb)
		return 0;

	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return 0;

	colorNameLength = (*env)->GetStringLength(env, colorName);
	colorNameUTF16 = (*env)->GetStringChars(env, colorName, nil);
	colorNameCocoa = [NSString stringWithCharacters:colorNameUTF16
		length:colorNameLength];
	(*env)->ReleaseStringChars(env, colorName, colorNameUTF16);

	arrayLen = (*env)->GetArrayLength(env, rgb);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		[localPool release];
		return 0;
	}
	if (arrayLen < 3) {
		[localPool release];
		return 0;
	}

	retval = [[NSColor performSelector:sel_registerName([colorNameCocoa UTF8String])] retain];
	if (retval == nil) {
		[localPool release];
		return 0;
	}

	[[retval colorUsingColorSpaceName:NSCalibratedRGBColorSpace]
		getRed:&fRed green:&fGreen blue:&fBlue alpha:nil];
	if (fRed < 0) fRed = 0;
	else if (fRed > 1.0) fRed = 1.0;
	if (fGreen < 0) fGreen = 0;
	else if (fGreen > 1.0) fGreen = 1.0;
	if (fBlue < 0) fBlue = 0;
	else if (fBlue > 1.0) fBlue = 1.0;
	intArray[0] = (jint) rint(fRed * 255);
	intArray[1] = (jint) rint(fGreen * 255);
	intArray[2] = (jint) rint(fBlue * 255);
	(*env)->SetIntArrayRegion(env, rgb, 0, 3, intArray);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		[localPool release];
		return 0;
	}

	[localPool release];
#ifdef SUPPORT_64BITS
	return (jlong) retval;
#else
	return (jlong) ((unsigned) retval);
#endif
}

JNIEXPORT jint JNICALL Java_AppletHandlerFactory_getCurrentPThreadID(
	JNIEnv *env, jclass clazz)
{
	return (jint) pthread_self();
}

JNIEXPORT jint JNICALL Java_AppletHandlerFactory_getOSXVersion(
	JNIEnv *env, jclass clazz)
{
	return (jint) osxVersion;
}

JNIEXPORT void JNICALL Java_AppletHandlerFactory_disableScreenUpdates(
	JNIEnv *env, jclass clazz)
{
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	[JEPController disableScreenUpdates];
	[localPool release];
}

JNIEXPORT void JNICALL Java_AppletHandlerFactory_enableScreenUpdates(
	JNIEnv *env, jclass clazz)
{
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	[JEPController enableScreenUpdates];
	[localPool release];
}

JNIEXPORT jlong JNICALL Java_AppletHandlerFactory_allocAutoreleasePool(
	JNIEnv *env, jclass clazz)
{
	NSAutoreleasePool *localPool = [[NSAutoreleasePool alloc] init];
#ifdef SUPPORT_64BITS
	return (jlong) localPool;
#else
	return (jlong) ((unsigned) localPool);
#endif
}

JNIEXPORT void JNICALL Java_AppletHandlerFactory_releaseAutoreleasePool(
	JNIEnv *env, jclass clazz, jlong pool)
{
#ifdef SUPPORT_64BITS
	NSAutoreleasePool *localPool = (NSAutoreleasePool *) pool;
#else
	NSAutoreleasePool *localPool = (NSAutoreleasePool *) ((unsigned) pool);
#endif
	if (localPool) [localPool release];
}

JNIEXPORT jboolean JNICALL Java_AppletHandlerFactory_isAppletViewDisplayable(
	JNIEnv *env, jclass clazz, jlong cocoaParentView)
{
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	jboolean retval = false;
	NSAutoreleasePool *localPool;
	if (!appletView)
		return false;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return false;
	retval = [appletView isDisplayable];
	[localPool release];
	return retval;
}

/* Called from AppletFramePanel.appletResize(), which is itself called
 * whenever Applet.resize() is called.  As of JEP 0.9.1 we make a call to
 * Applet.resize() trigger our support for "natively" resizing applets.
 * (Other ways of triggering it are the resizeAppletDirectly option to
 * JEPCreateJavaApplet() and using percentages (instead of absolute
 * numbers of pixels) to specify an applet's "width" or "height" in HTML
 * code).
 */
JNIEXPORT void JNICALL Java_AppletHandlerFactory_resetHybridView(JNIEnv *env,
	jclass clazz, jlong cocoaParentView)
{
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSThread *currentThread;
	NSAutoreleasePool *localPool;
	if (!mainThread || !appletView || !appletView->myController ||
		!nsAppShadow || [appletView destroying])
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	appletView->myController->resizeAppletDirectly = true;
	currentThread = [NSThread currentThread];
	[appletView->myController suspendAppletUpdates];
	if ([mainThread isEqual:currentThread]) {
		[nsAppShadow resetHybridView:appletView->myController];
	} else {
		[nsAppShadow performSelectorOnMainThread:@selector(resetHybridView:)
			withObject:appletView->myController waitUntilDone:false];
	}
	[localPool release];
}

JNIEXPORT void JNICALL Java_AppletHandlerFactory_suspendAppletUpdates(
	JNIEnv *env, jclass clazz, jlong cocoaParentView)
{
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSAutoreleasePool *localPool;
	if (!appletView || !appletView->myController || [appletView destroying])
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	[appletView->myController suspendAppletUpdates];
	[localPool release];
}

JNIEXPORT void JNICALL Java_AppletHandlerFactory_resumeAppletUpdates(
	JNIEnv *env, jclass clazz, jlong cocoaParentView)
{
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSAutoreleasePool *localPool;
	if (!appletView || !appletView->myController || [appletView destroying])
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	[appletView->myController resumeAppletUpdates];
	[localPool release];
}

JNIEXPORT void JNICALL Java_AppletHandlerFactory_cocoaDrawApplet(
	JNIEnv *env, jclass clazz, jlong cocoaParentView)
{
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSAutoreleasePool *localPool;
	if (!appletView || !appletView->myController || [appletView destroying])
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	appletView->myController->previousFrame =
		appletView->myController->currentFrame;
	[appletView->myController drawApplet];
	[localPool release];
}

/* If we don't delete our global refs soon enough (presumably before the applet
 * is destroyed), the Java garbage collector won't ever free our AppletFrame
 * and AppletPanel objects.  So we need to call this at the appropriate time
 * from AppletFrame, instead of waiting for our AppletView object to be
 * deallocated (as we used to before JEP 0.9.5).
 */
JNIEXPORT void JNICALL Java_AppletHandlerFactory_deleteAppletJNIGlobalRefs(
	JNIEnv *env, jclass clazz, jlong cocoaParentView)
{
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	if (!appletView || ![JEPController controllerFromAppletView:appletView
		timeoutType:lockTimeoutNormal])
		return;
	[appletView deleteJNIGlobalRefs:env];
	[localPool release];
}

/* Handwaving to stop all our AppletHandler objects from being leaked by the
 * Java garbage collector:  If we return an AppletHandler (or AppletHolder)
 * object from a function called from JNI (as AppletHolderFactory.
 * createAppletHolder() did before JEP 0.9.5), the garbage collector never
 * frees it.  But if (as we do here) we pass our newly created AppletHandler
 * object as a parameter to a JNI native function, the garbage collector does
 * what it's supposed to do.
 */
JNIEXPORT void JNICALL Java_AppletHandlerFactory_tellHandlerCreated(JNIEnv *env,
	jobject self, jlong cocoaParentView, jobject handler)
{
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	if (!appletView || !handler)
		return;
	appletView->myAppletHandler = (*env)->NewGlobalRef(env, handler);
}

/* "Native" implementation of tellAppletCreated() in AppletFramePanel.java */
JNIEXPORT void JNICALL Java_AppletFramePanel_tellAppletCreated(JNIEnv *env,
	jobject self, jlong cocoaParentView, jobject applet, jstring info, jstring name)
{
	NSString *appletInfo = nil;
	unsigned appletInfoLength = 0;
	unsigned char *appletInfoUTF = nil;
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSAutoreleasePool *localPool;
	if (!appletView || !appletView->myController)
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	appletView->myApplet = (*env)->NewGlobalRef(env, applet);
	appletView->myAppletName = [[AppletView toString:name javaEnv:env] retain];
	[appletView performSelector:@selector(returnCreatedInfo)];
	/* Log that we're creating an applet.  This may help people who can't
	 * figure out whether or not they're using applets!  Because the applet's
	 * URL in appletInfo may contain '%', don't use NSLog (or any printf-
	 * like function) to log it.  Instead get a UTF8 string from appletInfo
	 * and use puts() on that.  This should resolve the bug reported at
	 * http://sourceforge.net/forum/forum.php?thread_id=1700130&forum_id=371125.
	 */
	if ((appletInfo = [AppletView toString:info javaEnv:env]) != nil) {
		appletInfoLength = CFStringGetLength((CFStringRef) appletInfo);
		if ((appletInfoUTF = (unsigned char *)
			calloc(appletInfoLength + 1, sizeof(unichar))) != nil)
		{
			CFStringGetBytes((CFStringRef) appletInfo,
				CFRangeMake(0, appletInfoLength), kCFStringEncodingUTF8, '?',
				false, appletInfoUTF, appletInfoLength * sizeof(unichar), nil);
			puts((char *) appletInfoUTF);
			free(appletInfoUTF);
		}
	}
	[localPool release];
}

/* As of JEP 0.9.6.1 we use a "jobID" to identify each of the "jobs" that's
 * currently using Java_processMainThreadEvents().  This allows us to re-enter
 * Java_processMainThreadEvents() with impunity:  Each "job" that interrupts
 * previous one(s) will be processed to completion; then the previous "job"
 * will resume.  These "jobs" might not run (on other threads) in exactly this
 * order, but we can guarantee that nested calls to processMainThreadEvents()
 * will return in the order expected, and only after their respective "jobs"
 * have completed (on other threads).  Also, our previous method (using a
 * processingLevel counter) sometimes didn't work right -- sometimes (though
 * very rarely) processMainThreadEvents() would hang waiting for a job to
 * complete that had already completed.
 */

NSMutableArray *Java_runningJobs = nil;
NSLock *Java_runningJobs_lock = nil;

static BOOL IsJobRunning(jlong jobID)
{
	BOOL retval = false;
	NSNumber *aJob;
	unsigned count, i;
	if (!Java_runningJobs || !Java_runningJobs_lock)
		return false;
	[Java_runningJobs_lock lock];
	count = [Java_runningJobs count];
	for (i = 1; i <= count; ++i) {
		aJob = [Java_runningJobs objectAtIndex:i - 1];
		if ([aJob unsignedLongLongValue] == jobID) {
			retval = true;
			break;
		}
	}
	[Java_runningJobs_lock unlock];
	return retval;
}

static void AddJob(jlong jobID)
{
	NSNumber *aJob;
	unsigned count, i;
	BOOL alreadyRunning = false;
	if (!Java_runningJobs || !Java_runningJobs_lock)
		return;
	[Java_runningJobs_lock lock];
	count = [Java_runningJobs count];
	for (i = 1; i <= count; ++i) {
		aJob = [Java_runningJobs objectAtIndex:i - 1];
		if ([aJob unsignedLongLongValue] == jobID) {
			alreadyRunning = true;
			break;
		}
	}
	if (!alreadyRunning) {
		[Java_runningJobs addObject:
			[NSNumber numberWithUnsignedLongLong:jobID]];
	}
	[Java_runningJobs_lock unlock];
}

static void RemoveJob(jlong jobID)
{
	NSNumber *aJob;
	unsigned count, i;
	if (!Java_runningJobs || !Java_runningJobs_lock)
		return;
	[Java_runningJobs_lock lock];
	count = [Java_runningJobs count];
	for (i = 1; i <= count; ++i) {
		aJob = [Java_runningJobs objectAtIndex:i - 1];
		if ([aJob unsignedLongLongValue] == jobID) {
			[Java_runningJobs removeObjectAtIndex:i - 1];
			break;
		}
	}
	[Java_runningJobs_lock unlock];
}

JNIEXPORT void JNICALL Java_processMainThreadEvents(JNIEnv *env,
	jobject clazz, jlong millis, jlong jobID)
{
	NSDate *start;
	NSTimeInterval total, remaining;
	NSThread *currentThread = nil;
	NSAutoreleasePool *localPool;
	if (!mainThread || !Java_runningJobs || !Java_runningJobs_lock)
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	if (millis != 0) {
		total = (NSTimeInterval) millis / (NSTimeInterval) 1000;
	} else {
		total = .05;
	}
	currentThread = [NSThread currentThread];
	if (![mainThread isEqual:currentThread]) {
		[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:total]];
		[localPool release];
		return;
	}
	if (millis != 0) {
		remaining = total = (NSTimeInterval) millis / (NSTimeInterval) 1000;
		start = [NSDate date];
	} else {
		remaining = total = 0;
	}
	[AppletView postFakeEvent];
	for (;;) {
		[NSApp nextEventMatchingMask:NSAnyEventMask
			untilDate:nil inMode:NSDefaultRunLoopMode dequeue:false];
		if (!IsJobRunning(jobID))
			break;
		[AppletView processCustomEventSpool:false];
		[AppletView processCEventSpool:false];
		if (millis != 0) {
			if ((remaining = total + [start timeIntervalSinceNow]) <= 0)
				break;
		}
		usleep(20000);
	}
	[localPool release];
}

JNIEXPORT void JNICALL Java_processQueuedMainThreadEvents(JNIEnv *env,
	jobject clazz)
{
	NSEvent *anEvent;
	NSThread *currentThread = nil;
	NSAutoreleasePool *localPool;
	if (!mainThread)
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	currentThread = [NSThread currentThread];
	if (![mainThread isEqual:currentThread]) {
		[localPool release];
		return;
	}
	[AppletView postFakeEvent];
	for (;;) {
		anEvent = [NSApp nextEventMatchingMask:NSAnyEventMask untilDate:nil
			inMode:@"AWTRunLoopMode" dequeue:true];
		if (!anEvent) {
			anEvent = [NSApp nextEventMatchingMask:NSAnyEventMask
				untilDate:nil inMode:NSDefaultRunLoopMode dequeue:true];
		}
		if (!anEvent)
			break;
		[NSApp _setCurrentEvent:anEvent];
		[NSApp sendEvent:anEvent];
		[AppletView processCustomEventSpool:false];
		[AppletView processCEventSpool:false];
		usleep(20000);
	}
	[localPool release];
}

JNIEXPORT jlong JNICALL Java_beforeProcessing(JNIEnv *env, jobject clazz,
	jobject key)
{
	jlong retval;
	struct timeval time;// Current time in seconds/microseconds since 1970
	jlong javaTime;		// Java event timestamp in milliseconds since 1970
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return 0;
	retval = (jlong) ((unsigned) key) << 32;
	gettimeofday(&time, nil);
	javaTime = ((jlong) time.tv_sec) * 1000 + (jlong) (time.tv_usec / 1000);
	retval |= (javaTime & 0xFFFFFFFF);
	AddJob(retval);
	[localPool release];
	return retval;
}

JNIEXPORT void JNICALL Java_stopProcessing(JNIEnv *env, jobject clazz, jlong jobID)
{
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	RemoveJob(jobID);
	[localPool release];
}

/* As of Java 1.4.2 Update 1, the native methods that used to be in
 * libAppletsInCocoa.jnilib (loaded for earlier Java versions by
 * AppletHandlerFactory.java's loadAppleNatives() method) are now in
 * JavaPluginCocoa.bundle.  It'd be nice if we could use NSBundle to
 * load it -- but this would run Apple's initialization code, which
 * is incompatible with the JEP.  However, we _can_ run the bundle's
 * registerNatives() method, as long as we do so on a secondary thread.
 * (Prior to JEP 0.9.5+f we registered each native method individually,
 * using low-level JNI calls.  But that's only necessary if we're
 * working on the main thread.)
 */
JNIEXPORT void JNICALL Java_callRegisterNatives(JNIEnv *env, jobject clazz)
{
	NSString *path;
	NSString *subpath = @"/Contents/MacOS/JavaPluginCocoa";
	NSObjectFileImage image = nil;
	NSModule module = nil;
	NSSymbol symbol = nil;
	void (*registerNativesPtr) (JNIEnv *) = nil;
	NSAutoreleasePool *localPool;
	if (!isJava142Update1)
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	path = [javaPluginCocoaPath stringByAppendingPathComponent:subpath];
	if (NSCreateObjectFileImageFromFile([path UTF8String], &image)
		!= NSObjectFileImageSuccess)
		image = nil;
	if (image) {
		module = NSLinkModule(image, [path UTF8String],
			NSADDIMAGE_OPTION_RETURN_ON_ERROR);
	}
	if (!module) {
		NSLog(@"Java_registerNatives(): Couldn't load %@", path);
		[localPool release];
		return;
	}
	if ((symbol = NSLookupSymbolInModule(module, "_registerNatives")) != nil)
		registerNativesPtr = (void (*) (JNIEnv *)) NSAddressOfSymbol(symbol);
	if (registerNativesPtr) {
		registerNativesPtr(env);
	} else {
		NSLog(@"Java_registerNatives(): Couldn't find JavaPluginCocoa.bundle's registerNatives()");
	}
	[localPool release];
}

typedef JNIEXPORT jlong JNICALL (*CAPPLETFRAME_CREATE) (JNIEnv *, jobject, \
	jlong, jboolean, jboolean, jobject);
typedef JNIEXPORT jlong JNICALL (*CAPPLETFRAME_CREATE160) (JNIEnv *, jobject, \
	jlong, jboolean, jboolean, jboolean, jobject);

JNIEXPORT jlong JNICALL Java_MyCAppletFrame_createNativeEmbedded(JNIEnv *env,
	jobject self, jlong parentView, jboolean isEnabled, jboolean isVisible,
	jobject byteParms)
{
	jlong retval = 0;
	NSSymbol symbol = NULL;
	static CAPPLETFRAME_CREATE create = nil;
	if (!create) {
		if (symbol = NSLookupAndBindSymbol("_Java_apple_awt_CEmbedded_createNativeEmbedded"))
			create = (CAPPLETFRAME_CREATE) NSAddressOfSymbol(symbol);
	}
	if (create) {
		retval = create(env, self, parentView, isEnabled, isVisible, byteParms);
	}
	return retval;
}

JNIEXPORT jlong JNICALL Java_MyCAppletFrame_createNativeEmbedded160(JNIEnv *env,
	jobject self, jlong parentView, jboolean isEnabled, jboolean isVisible,
	jboolean isFocusableWindow, jobject byteParms)
{
	jlong retval = 0;
	NSSymbol symbol = NULL;
	static CAPPLETFRAME_CREATE160 create = nil;
	if (!create) {
		if (symbol = NSLookupAndBindSymbol("_Java_apple_awt_CEmbedded_createNativeEmbedded"))
			create = (CAPPLETFRAME_CREATE160) NSAddressOfSymbol(symbol);
	}
	if (create) {
		retval = create(env, self, parentView, isEnabled, isVisible,
			isFocusableWindow, byteParms);
	}
	return retval;
}

/* I've occasionally seen crashlogs that report this routine crashing when
 * it's been called at some wildly inappropriate time (for example as the
 * JVM is being created).  I don't know what causes this, and I can't
 * reproduce any of these crashes.  But I figure the following (new as of
 * JEP 0.9.6) is a safer way to do what needs to be done, since it doesn't
 * use any Java/JNI calls and therefore doesn't risk crashing because the
 * JVM hasn't yet been fully initialized.
 */
JNIEXPORT jlong JNICALL Java_MyCAppletFrame_getViewPtr(JNIEnv *env,
	jobject self, jlong model)
{
	jlong retval = 0;
	id cWindow = nil;
	JEPController *controller;
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return 0;
#ifdef SUPPORT_64BITS
	cWindow = (id) model;
#else
	cWindow = (id) ((unsigned) model);
#endif
	controller = [JEPController controllerFromCWindow:cWindow
		timeoutType:lockTimeoutNormal];
	if (!controller || [controller destroying]) {
		retval = 0;
	} else {
#ifdef SUPPORT_64BITS
		retval = (jlong) controller->myAppletView;
#else
		retval = (jlong) ((unsigned) controller->myAppletView);
#endif
	}
	[localPool release];
	return retval;
}

JNIEXPORT void JNICALL Java_AppletHandler_maybeDoAppletStart(JNIEnv *env,
	jobject clazz, jlong cocoaParentView)
{
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSAutoreleasePool *localPool;
	if (!appletView || !appletView->myController || [appletView destroying])
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	[appletView performSelector:@selector(maybeDoAppletStart)];
	[localPool release];
}

JNIEXPORT void JNICALL Java_AppletHandler_reportAppletError(JNIEnv *env,
	jobject clazz, jlong cocoaParentView)
{
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSAutoreleasePool *localPool;
	if (!appletView)
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	if (!appletView->myController || [appletView destroying])
		appletView->appletState = kAppletDestroyJEP;
	if (appletView->appletState != kAppletDestroyJEP)
		appletView->appletState = kAppletErrorJEP;
	[appletView performSelector:@selector(returnCreatedInfo)];
	[localPool release];
}

JNIEXPORT void JNICALL Java_AppletHandler_updateStatusFromJava(JNIEnv *env,
	jobject clazz, jlong cocoaParentView, jint status)
{
#ifdef SUPPORT_64BITS
	AppletView *appletView = (AppletView *) cocoaParentView;
#else
	AppletView *appletView = (AppletView *) ((unsigned) cocoaParentView);
#endif
	NSAutoreleasePool *localPool;
	if (!appletView)
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	if (!appletView->myController || [appletView destroying])
		appletView->appletState = kAppletDestroyJEP;
	if ((appletView->appletState != kAppletErrorJEP) &&
		(appletView->appletState != kAppletDestroyJEP))
	{
		appletView->appletState = status;
	}
	[localPool release];
}

/* Java_CPanel_createNativePanel() is needed to work around an Apple bug on
 * OS X 10.3 and 10.4:  Now that the applet's start() method is being called
 * earlier (for which see startApplet below), strange things can happen if the
 * Applet object (the child of the AppletFramePanel object and the grandchild
 * of the AppletFrame object) is created invisible (e.g. by loading it in
 * another tab) -- parts of the applet might never be created (on OS X 10.3),
 * or can disappear after being displayed just once (on OS X 10.4).  (As of
 * JEP 0.9.5+e we have (as needed) two versions of Java_CPanel_createNativePanel(),
 * since Tiger's Java 1.4.2 (strangely) adds an extra parameter.  This gets rid
 * of a (largely benign) java.lang.NoSuchMethodError error that appears as the
 * JVM starts up on Java 1.4.2 on OS X Tiger.)
 */

typedef JNIEXPORT jlong JNICALL (*CPANEL_CREATE142T) (JNIEnv *, jobject, \
	jlong, jlong, jint, jint, jint, jint, jboolean, jboolean, jobject, \
	jobject, jboolean, jobject, jboolean, jobject);
typedef JNIEXPORT jlong JNICALL (*CPANEL_CREATE) (JNIEnv *, jobject, \
	jlong, jint, jint, jint, jint, jboolean, jboolean, jobject, \
	jobject, jboolean, jobject, jboolean, jobject);

JNIEXPORT jlong JNICALL Java_CPanel_createNativePanel142T(JNIEnv *env, jobject self,
	jlong parentModel, jlong aboveModel, jint x, jint y, jint width, jint height,
	jboolean isEnabled, jboolean isVisible, jobject font, jobject bgColor,
	jboolean bgSet, jobject fgColor, jboolean fgSet, jobject cursor)
{
	jlong retval = 0;
	JNIEnv *envHolder = nil;
#ifdef SUPPORT_64BITS
	id parentModelLocal = (id) parentModel;
#else
	id parentModelLocal = (id) ((unsigned) parentModel);
#endif
	id gParentModel = nil;
	NSSymbol symbol = NULL;
	static CPANEL_CREATE142T create = nil;
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return 0;
	if ([AppletView getJavaEnvIfAttached:&envHolder] != true) {
		NSLog(@"Java_CPanel_createNativePanel142T() called with JVM detached (Apple bug)!");
		[localPool release];
		return 0;
	}
	if (parentModelLocal)
		gParentModel = [parentModelLocal getParent];
	if (gParentModel && [gParentModel isKindOfClass:appletWindowClass])
		isVisible = true;
	if (!create) {
		if (symbol = NSLookupAndBindSymbol("_Java_apple_awt_CPanel_createNativePanel"))
			create = (CPANEL_CREATE142T) NSAddressOfSymbol(symbol);
	}
	if (create) {
		retval = create(env, self, parentModel, aboveModel, x, y, width, height,
			isEnabled, isVisible, font, bgColor, bgSet, fgColor, fgSet, cursor);
	}
	[localPool release];
	return retval;
}

JNIEXPORT jlong JNICALL Java_CPanel_createNativePanel(JNIEnv *env, jobject self,
	jlong parentModel, jint x, jint y, jint width, jint height, jboolean isEnabled,
	jboolean isVisible, jobject font, jobject bgColor, jboolean bgSet,
	jobject fgColor, jboolean fgSet, jobject cursor)
{
	jlong retval = 0;
	JNIEnv *envHolder = nil;
#ifdef SUPPORT_64BITS
	id parentModelLocal = (id) parentModel;
#else
	id parentModelLocal = (id) ((unsigned) parentModel);
#endif
	id gParentModel = nil;
	NSSymbol symbol = NULL;
	static CPANEL_CREATE create = nil;
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return 0;
	if ([AppletView getJavaEnvIfAttached:&envHolder] != true) {
		NSLog(@"Java_CPanel_createNativePanel() called with JVM detached (Apple bug)!");
		[localPool release];
		return 0;
	}
	if (parentModelLocal)
		gParentModel = [parentModelLocal getParent];
	if (gParentModel && [gParentModel isKindOfClass:appletWindowClass])
		isVisible = true;
	if (!create) {
		if (symbol = NSLookupAndBindSymbol("_Java_apple_awt_CPanel_createNativePanel"))
			create = (CPANEL_CREATE) NSAddressOfSymbol(symbol);
	}
	if (create) {
		retval = create(env, self, parentModel, x, y, width, height, isEnabled,
			isVisible, font, bgColor, bgSet, fgColor, fgSet, cursor);
	}
	[localPool release];
	return retval;
}

/* Java 1.4.2 in Apple's J2SE 5.0 Release 4 has a serious bug that prevents
 * use of any of the graphics "primitives" in sun.java2d.loops.*:  SurfaceData
 * objects' "ops" aren't initialized properly -- many of the function pointers
 * in the SurfaceDataOps structure (which are used by the graphics
 * "primitives") are left NULL.  So when the JVM tries to perform one of these
 * operations (graphics primitives), you get NULL-pointer deferences and
 * crashes.  To see an example -- load one of the puzzles at
 * http://www.jigzone.com and choose "Shuffle Pieces".
 * It's not possible to fix this completely, but the following seems to work
 * well enough:  Hook calls to a SurfaceData object's initRaster() and
 * initOps() methods.  When these fail to initialize the appropriate function
 * pointers, we initialize them ourselves, using bare-bones "default" versions
 * of these functions.
 */

typedef struct {
    jint x1;
    jint y1;
    jint x2;
    jint y2;
} SurfaceDataBounds;

#define SD_RASINFO_PRIVATE_SIZE 64

typedef struct {
    SurfaceDataBounds   bounds;                 /* bounds of raster array */
    void                *rasBase;               /* Pointer to (0, 0) pixel */
    unsigned int        pixelStride;            /* bytes to next X pixel */
    unsigned int        scanStride;             /* bytes to next Y pixel */
    jint                *lutBase;               /* Pointer to colormap[0] */
    unsigned int        lutSize;                /* # colors in colormap */
    unsigned char       *invColorTable;         /* Inverse color table */
    char                *redErrTable;           /* Red ordered dither table */
    char                *grnErrTable;           /* Green ordered dither table */
    char                *bluErrTable;           /* Blue ordered dither table */
    int                 *invGrayTable;          /* Inverse gray table */
    char                priv[SD_RASINFO_PRIVATE_SIZE];
} SurfaceDataRasInfo;

typedef struct _SurfaceDataOps SurfaceDataOps;

typedef jint LockFunc(JNIEnv *env,
                      SurfaceDataOps *ops,
                      SurfaceDataRasInfo *rasInfo,
                      jint lockflags);

typedef void GetRasInfoFunc(JNIEnv *env,
                            SurfaceDataOps *ops,
                            SurfaceDataRasInfo *pRasInfo);

typedef void ReleaseFunc(JNIEnv *env,
                         SurfaceDataOps *ops,
                         SurfaceDataRasInfo *pRasInfo);

typedef void UnlockFunc(JNIEnv *env,
                        SurfaceDataOps *ops,
                        SurfaceDataRasInfo *pRasInfo);

typedef void SetupFunc(JNIEnv *env,
                       SurfaceDataOps *ops);

typedef void DisposeFunc(JNIEnv *env,
                         SurfaceDataOps *ops);

struct _SurfaceDataOps {
    LockFunc            *Lock;
    GetRasInfoFunc      *GetRasInfo;
    ReleaseFunc         *Release;
    UnlockFunc          *Unlock;
    SetupFunc           *Setup;
    DisposeFunc         *Dispose;
    int                 dirty;
    jobject             sdObject;
};

static jint DefaultSD_Lock(JNIEnv *env, SurfaceDataOps *ops,
	SurfaceDataRasInfo *pRasInfo, jint lockflags)
{
	return 0;
}

static void DefaultSD_GetRasInfo(JNIEnv *env, SurfaceDataOps *ops,
	SurfaceDataRasInfo *pRasInfo)
{
	pRasInfo->rasBase = NULL;
	pRasInfo->pixelStride = 0;
	pRasInfo->scanStride = 0;
	pRasInfo->lutBase = NULL;
	pRasInfo->lutSize = 0;
	pRasInfo->invColorTable = NULL;
	pRasInfo->redErrTable = NULL;
	pRasInfo->grnErrTable = NULL;
	pRasInfo->bluErrTable = NULL;
	pRasInfo->invGrayTable = NULL;
}

static void DefaultSD_Release(JNIEnv *env, SurfaceDataOps *ops,
	SurfaceDataRasInfo *pRasInfo)
{
}

static void DefaultSD_Unlock(JNIEnv *env, SurfaceDataOps *ops,
	SurfaceDataRasInfo *pRasInfo)
{
}

typedef JNIEXPORT SurfaceDataOps * JNICALL (*SDGETOPS) (JNIEnv *, jobject);
static SDGETOPS SurfaceData_GetOpsPtr = nil;

typedef JNIEXPORT void JNICALL (*OSXOFFSCREENSD_INITRASTER) (JNIEnv *, jobject, \
	jobject, jint, jint, jint, jint, jint, jobject, jint, jobject, jobject, jobject);

JNIEXPORT void JNICALL Java_apple_awt_OSXOffScreenSurfaceData_initRaster_redirect(
	JNIEnv *env, jobject self, jobject data, jint dataOffset, jint width, jint height,
	jint pixelStride, jint scanlineStride, jobject colorModel, jint imageType,
	jobject graphicsStates, jobject graphicsStatesObject, jobject imageInfo)
{
	NSSymbol symbol = NULL;
	static OSXOFFSCREENSD_INITRASTER initRaster = nil;
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	if (!SurfaceData_GetOpsPtr) {
		if (symbol = NSLookupAndBindSymbol("_SurfaceData_GetOps"))
			SurfaceData_GetOpsPtr = (SDGETOPS) NSAddressOfSymbol(symbol);
	}
	if (!initRaster) {
		if (symbol = NSLookupAndBindSymbol("_Java_apple_awt_OSXOffScreenSurfaceData_initRaster"))
			initRaster = (OSXOFFSCREENSD_INITRASTER) NSAddressOfSymbol(symbol);
	}
	if (initRaster) {
		initRaster(env, self, data, dataOffset, width, height, pixelStride,
			scanlineStride, colorModel, imageType, graphicsStates,
			graphicsStatesObject, imageInfo);
	}
	if (SurfaceData_GetOpsPtr) {
		SurfaceDataOps *ops = SurfaceData_GetOpsPtr(env, self);
		if (!ops->Lock)
			ops->Lock = DefaultSD_Lock;
		if (!ops->GetRasInfo)
			ops->GetRasInfo = DefaultSD_GetRasInfo;
		if (!ops->Release)
			ops->Release = DefaultSD_Release;
		if (!ops->Unlock)
			ops->Unlock = DefaultSD_Unlock;
	}
	[localPool release];
}

typedef JNIEXPORT void JNICALL (*OSXOFFSCREENSD_INITCUSTOMRASTER) (JNIEnv *, jobject, \
	jobject, jint, jint, jobject, jobject, jobject);

JNIEXPORT void JNICALL Java_apple_awt_OSXOffScreenSurfaceData_initCustomRaster_redirect(
	JNIEnv *env, jobject self, jobject buffer, jint width, jint height,
	jobject graphicsStates, jobject graphicsStatesObject, jobject imageInfo)
{
	NSSymbol symbol = NULL;
	static OSXOFFSCREENSD_INITCUSTOMRASTER initRaster = nil;
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	if (!SurfaceData_GetOpsPtr) {
		if (symbol = NSLookupAndBindSymbol("_SurfaceData_GetOps"))
			SurfaceData_GetOpsPtr = (SDGETOPS) NSAddressOfSymbol(symbol);
	}
	if (!initRaster) {
		if (symbol = NSLookupAndBindSymbol("_Java_apple_awt_OSXOffScreenSurfaceData_initCustomRaster"))
			initRaster = (OSXOFFSCREENSD_INITCUSTOMRASTER) NSAddressOfSymbol(symbol);
	}
	if (initRaster) {
		initRaster(env, self, buffer, width, height, graphicsStates, graphicsStatesObject, imageInfo);
	}
	if (SurfaceData_GetOpsPtr) {
		SurfaceDataOps *ops = SurfaceData_GetOpsPtr(env, self);
		if (!ops->Lock)
			ops->Lock = DefaultSD_Lock;
		if (!ops->GetRasInfo)
			ops->GetRasInfo = DefaultSD_GetRasInfo;
		if (!ops->Release)
			ops->Release = DefaultSD_Release;
		if (!ops->Unlock)
			ops->Unlock = DefaultSD_Unlock;
	}
	[localPool release];
}

typedef JNIEXPORT void JNICALL (*CPEERSD_INITOPS) (JNIEnv *, jobject, \
	jlong, jobject, jobjectArray);

JNIEXPORT void JNICALL Java_apple_awt_CPeerSurfaceData_initOps_redirect(
	JNIEnv *env, jobject self, jlong modelPtr, jobject graphicsStates,
	jobjectArray graphicsStatesObject)
{
	NSSymbol symbol = NULL;
	static CPEERSD_INITOPS initOps = nil;
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	if (!SurfaceData_GetOpsPtr) {
		if (symbol = NSLookupAndBindSymbol("_SurfaceData_GetOps"))
			SurfaceData_GetOpsPtr = (SDGETOPS) NSAddressOfSymbol(symbol);
	}
	if (!initOps) {
		if (symbol = NSLookupAndBindSymbol("_Java_apple_awt_CPeerSurfaceData_initOps"))
			initOps = (CPEERSD_INITOPS) NSAddressOfSymbol(symbol);
	}
	if (initOps) {
		initOps(env, self, modelPtr, graphicsStates, graphicsStatesObject);
	}
	if (SurfaceData_GetOpsPtr) {
		SurfaceDataOps *ops = SurfaceData_GetOpsPtr(env, self);
		if (!ops->Lock)
			ops->Lock = DefaultSD_Lock;
		if (!ops->GetRasInfo)
			ops->GetRasInfo = DefaultSD_GetRasInfo;
		if (!ops->Release)
			ops->Release = DefaultSD_Release;
		if (!ops->Unlock)
			ops->Unlock = DefaultSD_Unlock;
	}
	[localPool release];
}

//#define TEST_GIF_OVERFLOW 1
#define FILL_BYTE 0xDA
#define TEST_ARRAY_SIZE 2000000

/* All of Apple's 1.4.X JVMs, and their 1.5 JVMs that pre-date J2SE 5.0
 * Release 5, suffer from a heap-overflow triggered by processing an other-
 * wise valid GIF file whose 'image width' (from its 'image descriptor') has
 * been set to 0.  One way to work around this is to swallow all calls to
 * sun.awt.image.GifImageDecoder.parseImage() that would pass a 0 "width"
 * parameter.  For good measure I also swallow calls that would pass a 0
 * "height" parameter -- since I can't see how trying to process it would do
 * any good.  When swallowing a call I return "true", to avoid repetitions of
 * the invalid call.  For more information see:
 * http://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-CVE-2007-0243
 * http://lists.apple.com/archives/Java-dev/2007/Feb/msg00270.html
 * http://lists.apple.com/archives/Java-dev/2007/Feb/msg00277.html
 */
typedef JNIEXPORT jboolean JNICALL (*GIFDECODER_PARSEIMAGE) (JNIEnv *, jobject, \
	jint, jint, jint, jint, jboolean, jint, jbyteArray, jbyteArray, jobject);

JNIEXPORT jboolean JNICALL Java_sun_awt_image_GifImageDecoder_parseImage_redirect(
	JNIEnv *env, jobject self, jint x, jint y, jint width, jint height,
	jboolean interlace, jint initCodeSize, jbyteArray block, jbyteArray rasline,
	jobject colorModel)
{
	jboolean retval = true;
	NSSymbol symbol = NULL;
	static GIFDECODER_PARSEIMAGE parseImage = nil;
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return false;
	if (!parseImage) {
		if (symbol = NSLookupAndBindSymbol("_Java_sun_awt_image_GifImageDecoder_parseImage"))
			parseImage = (GIFDECODER_PARSEIMAGE) NSAddressOfSymbol(symbol);
	}
	if (parseImage) {
#ifdef TEST_GIF_OVERFLOW
		jboolean isCopy = false;
		jbyte *contents = nil;
		jbyteArray array = nil;
		if ((array = (*env)->NewByteArray(env, TEST_ARRAY_SIZE)) != nil) {
			contents = (*env)->GetPrimitiveArrayCritical(env, array, &isCopy);
			if (contents) {
				if (!isCopy) {
					memset(contents, FILL_BYTE, TEST_ARRAY_SIZE);
					rasline = array;
				}
				(*env)->ReleasePrimitiveArrayCritical(env, array, contents, 0);
			}
		}
		NSLog(@"GifImageDecoder_parseImage(): width %i, height %i, array 0x%.8X, contents 0x%.8X, isCopy %i",
			width, height, (unsigned) array, (unsigned) contents, isCopy);
#else
		if (width && height) {
#endif
			retval = parseImage(env, self, x, y, width, height, interlace,
				initCodeSize, block, rasline, colorModel);
#ifdef TEST_GIF_OVERFLOW
		if (array) {
			if (!isCopy) {
				contents = (*env)->GetPrimitiveArrayCritical(env, array, nil);
				if (contents) {
					long offset;
					for (offset = TEST_ARRAY_SIZE - 1; offset >= 0; --offset) {
						if (((unsigned char) contents[offset]) != FILL_BYTE)
							break;
					}
					if (offset == -1)
						offset = 0;
					(*env)->ReleasePrimitiveArrayCritical(env, array, contents, 0);
					NSLog(@"GifImageDecoder_parseImage(): %u bytes used in rasline, returning %i",
						offset, retval);
				}
			}
			(*env)->DeleteLocalRef(env, array);
		}
#else
		}
#endif
	}
	[localPool release];
	return retval;
}

/* RunnableWrapper_poseAsThread() makes one Thread object "pose as" (or take
 * the place of) another.  The "poser" is the thread that actually runs.  But
 * "target" is the thread that (to its creator) appears to be running.  This
 * way the "poser" can override one or more of the "target"'s methods.  Before
 * the "poser" thread quits, the two threads' identities need to be swapped
 * back (in RunnableWrapper_unposeThreads()).  "Poser" must already be running
 * when RunnableWrapper_poseAsThread() is called.
 * Underneath the hierarchy of Java objects, Sun's JVM has several (more or
 * less) parallel hierarchies of C++ objects (including Handle objects,
 * CHeapObj objects, oopDesc objects and Klass objects).  In it a JavaThread
 * object corresponds to a java.lang.Thread object (JavaThread is defined in
 * hotspot/src/share/vm/runtime/thread.hpp and thread.cpp, and is an indirect
 * subclass of CHeapObj).  Once a thread is running, the VM stores a pointer
 * to the thread's JavaThread object in its java.lang.Thread.eetop field.  The
 * JavaThread object in turn stores a pointer to its java.lang.Thread object
 * (or more precisely to the corresponding "oop") in its _threadObj field.  An
 * oop is a pointer to an oopDesc object.  The oop (or oopDesc) and Klass
 * hierarchies are described in hotspot/src/share/vm/oops/oopsHierarchy.hpp.
 * A jobject is a pointer to a Handle object whose first field (_handle) is a
 * pointer to its corresponding oop.  Each Java object has exactly one oopDesc
 * object, but (potentially) more than one Handle (so each Handle/jobject is a
 * "reference" (global or local) to the single oopDesc).  The Handle object is
 * defined in hotspot/src/share/vm/runtime/handles.hpp and handles.cpp.
 * Since the oopDesc object more-or-less _is_ the Java object (and not just a
 * reference to it), putting "target"'s oop in "poser"'s JavaThread object
 * makes it seem to "target"'s creators that "target" is the thread that's
 * running (and not "poser").
 */
JNIEXPORT void JNICALL Java_RunnableWrapper_poseAsThread(
	JNIEnv *env, jobject clazz, jobject poser, jobject target)
{
	unsigned offset = 0;
	jlong poserEetop = 0;
	void *poserJavaThread = nil;      // JavaThread *
	void *targetThreadObj = nil;      // oop
	void *poserThreadObj = nil;       // oop
	void **poserThreadObjAddr = nil;  // oop *
	static jfieldID stillbornID = nil;
	static jfieldID eetopID = nil;
	NSAutoreleasePool *localPool;

	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;

	if (!eetopID) {
		jclass threadClass = (*env)->FindClass(env, "java/lang/Thread");
		stillbornID = (*env)->GetFieldID(env, threadClass, "stillborn", "Z");
		eetopID = (*env)->GetFieldID(env, threadClass, "eetop", "J");
		if ((*env)->ExceptionCheck(env)) {
			(*env)->ExceptionClear(env);
			eetopID = nil;
		}
	}

	if (eetopID && poser && target) {
		poserEetop = (*env)->GetLongField(env, poser, eetopID);
		(*env)->SetLongField(env, target, eetopID, poserEetop);

		if ((*env)->ExceptionCheck(env)) {
			(*env)->ExceptionClear(env);
		} else {
			poserJavaThread = (void *) ((unsigned) (poserEetop >> 32));

			targetThreadObj = *((void **) target);
			poserThreadObj = *((void **) poser);

			/* The _threadObj field of a JavaThread object is offset 23
			 * 32-bit "pointers" from the beginning of the object in both OS X
			 * 10.2.8's Java 1.4.1 JVM and OS X 10.3.9's Java 1.4.2 JVM.
			 * But for safety's sake I search for this field in its general
			 * location, and verify that it's the correct field by comparing
			 * its contents against the oop derived from "poser".
			 */
			for (offset = 20; offset < 30; ++offset) {
				void **threadObjAddr = nil;
				threadObjAddr = (void **)
					((unsigned) poserJavaThread + offset * sizeof(void *));
				if (*threadObjAddr == poserThreadObj) {
					poserThreadObjAddr = threadObjAddr;
					break;
				}
			}

			/* Set "poser"'s JavaThread's _threadObj field to "target"'s oop.
			 */
			if (poserThreadObjAddr)
				*poserThreadObjAddr = targetThreadObj;

			/* We need to (temporarily) set "poser"'s stillborn field and zero
			 * its eetop field -- otherwise "poser" will also show up in lists
			 * of threads belonging to "target"'s ThreadGroup.
			 */
			(*env)->SetBooleanField(env, poser, stillbornID, true);
			(*env)->SetLongField(env, poser, eetopID, 0);
		}
	}

	//NSLog(@"RunnableWrapper_poseAsThread(): poserEetop 0x%.16qX, poserJavaThread 0x%.8X",
	//	poserEetop, (unsigned) poserJavaThread);
	//NSLog(@"RunnableWrapper_poseAsThread(): poserThreadObj 0x%.8X, targetThreadObj 0x%.8X",
	//	(unsigned) poserThreadObj, (unsigned) targetThreadObj);
	//NSLog(@"RunnableWrapper_poseAsThread(): offset %u, poserThreadObjAddr 0x%.8X",
	//	offset, (unsigned) poserThreadObjAddr);

	[localPool release];
}

JNIEXPORT void JNICALL Java_RunnableWrapper_unposeThreads(
	JNIEnv *env, jobject clazz, jobject poser, jobject target)
{
	unsigned offset = 0;
	jlong poserEetop = 0;
	void *poserJavaThread = nil;      // JavaThread *
	void *targetThreadObj = nil;      // oop
	void *poserThreadObj = nil;       // oop
	void **poserThreadObjAddr = nil;  // oop *
	static jmethodID exit = nil;
	static jfieldID stillbornID = nil;
	static jfieldID eetopID = nil;
	NSAutoreleasePool *localPool;

	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;

	if (!eetopID) {
		jclass threadClass = (*env)->FindClass(env, "java/lang/Thread");
		exit = (*env)->GetMethodID(env, threadClass, "exit", "()V");
		stillbornID = (*env)->GetFieldID(env, threadClass, "stillborn", "Z");
		eetopID = (*env)->GetFieldID(env, threadClass, "eetop", "J");
		if ((*env)->ExceptionCheck(env)) {
			(*env)->ExceptionClear(env);
			eetopID = nil;
		}
	}

	if (eetopID && poser && target) {
		poserEetop = (*env)->GetLongField(env, target, eetopID);

		if ((*env)->ExceptionCheck(env)) {
			(*env)->ExceptionClear(env);
		} else {
			poserJavaThread = (void *) ((unsigned) (poserEetop >> 32));

			targetThreadObj = *((void **) target);
			poserThreadObj = *((void **) poser);

			/* The _threadObj field of a JavaThread object is offset 23
			 * 32-bit "pointers" from the beginning of the object in both OS X
			 * 10.2.8's Java 1.4.1 JVM and OS X 10.3.9's Java 1.4.2 JVM.
			 * But for safety's sake I search for this field in its general
			 * location, and verify that it's the correct field by comparing
			 * its contents against the oop derived from "target".
			 */
			for (offset = 20; offset < 30; ++offset) {
				void **threadObjAddr = nil;
				threadObjAddr = (void **)
					((unsigned) poserJavaThread + offset * sizeof(void *));
				if (*threadObjAddr == targetThreadObj) {
					poserThreadObjAddr = threadObjAddr;
					break;
				}
			}

			/* Set "poser"'s JavaThread's _threadObj field back to "poser"'s
			 * oop.
			 */
			if (poserThreadObjAddr)
				*poserThreadObjAddr = poserThreadObj;

			/* Undo the changes we made to these fields in
			 * RunnableWrapper_poseAsThread().
			 */
			(*env)->SetBooleanField(env, poser, stillbornID, false);
			(*env)->SetLongField(env, poser, eetopID, poserEetop);

			/* Make it seem to "target"'s owners that it has finished running,
			 * and make sure that it gets removed from its ThreadGroup (by
			 * the call to Thread.exit()).  See JavaThread::exit() and
			 * ensure_join() (called from the first) in hotspot/src/share/vm/
			 * runtime/thread.cpp.
			 */
			(*env)->SetBooleanField(env, target, stillbornID, true);
			(*env)->SetLongField(env, target, eetopID, 0);
			(*env)->CallVoidMethod(env, target, exit);
		}
	}

	//NSLog(@"RunnableWrapper_unposeThreads(): poserEetop 0x%.16qX, poserJavaThread 0x%.8X",
	//	poserEetop, (unsigned) poserJavaThread);
	//NSLog(@"RunnableWrapper_unposeThreads(): poserThreadObj 0x%.8X, targetThreadObj 0x%.8X",
	//	(unsigned) poserThreadObj, (unsigned) targetThreadObj);
	//NSLog(@"RunnableWrapper_unposeThreads(): offset %u, poserThreadObjAddr 0x%.8X",
	//	offset, (unsigned) poserThreadObjAddr);

	[localPool release];
}

typedef JNIEXPORT void JNICALL (*THREAD_START) (JNIEnv *, jobject);

/* Under certain conditions substitute a RunnableWrapper for a Thread object's
 * original Runnable, or run a RunnableWrapper.WrapperThread object in place
 * of a Thread subclass object.  For information see RunnableWrapper.java and
 * Java_RunnableWrapper_poseAsThread() and Java_RunnableWrapper_unposeThreads()
 * (above).
 */
JNIEXPORT void JNICALL Java_java_lang_Thread_start_redirect(
	JNIEnv *env, jobject self)
{
	NSSymbol symbol = NULL;
	static jclass threadClass = nil;
	static jfieldID targetID = nil;
	static jfieldID classLoaderID = nil;
	static jfieldID stillbornID = nil;
	static jclass wrapperClass = nil;
	static jclass wrapperThreadClass = nil;
	static jmethodID newWrapperWithRunnable = nil;
	static jmethodID newWrapperWithThread = nil;
	static jmethodID getNewThread = nil;
	static THREAD_START start = nil;
	NSAutoreleasePool *localPool;

	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;

	if (!start) {
		jclass threadClassHolder = nil;
		jclass wrapperClassHolder = nil;
		jclass wrapperThreadClassHolder = nil;

		if (symbol = NSLookupAndBindSymbol("_JVM_StartThread"))
			start = (THREAD_START) NSAddressOfSymbol(symbol);

		threadClassHolder = (*env)->FindClass(env, "java/lang/Thread");
		wrapperClassHolder = (*env)->FindClass(env, "jep/RunnableWrapper");
		wrapperThreadClassHolder =
			(*env)->FindClass(env, "jep/RunnableWrapper$WrapperThread");

		if ((*env)->ExceptionCheck(env)) {
			(*env)->ExceptionClear(env);
			start = nil;
		} else {
			// Thread
			targetID = (*env)->GetFieldID(env, threadClassHolder, "target",
				"Ljava/lang/Runnable;");
			classLoaderID = (*env)->GetFieldID(env, threadClassHolder,
				"contextClassLoader", "Ljava/lang/ClassLoader;");
			stillbornID = (*env)->GetFieldID(env, threadClassHolder,
				"stillborn", "Z");
			threadClass = (*env)->NewGlobalRef(env, threadClassHolder);
			(*env)->DeleteLocalRef(env, threadClassHolder);
			// RunnableWrapper and WrapperThread
			newWrapperWithRunnable = (*env)->GetStaticMethodID(
				env, wrapperClassHolder, "newWrapper",
				"(Ljava/lang/Runnable;)Ljep/RunnableWrapper;");
			newWrapperWithThread = (*env)->GetStaticMethodID(
				env, wrapperClassHolder, "newWrapper",
				"(Ljava/lang/Thread;Ljava/lang/Runnable;)Ljep/RunnableWrapper;");
			getNewThread = (*env)->GetMethodID(env, wrapperClassHolder,
				"getNewThread", "()Ljava/lang/Thread;");
			wrapperClass = (*env)->NewGlobalRef(env, wrapperClassHolder);
			wrapperThreadClass =
				(*env)->NewGlobalRef(env, wrapperThreadClassHolder);
			(*env)->DeleteLocalRef(env, wrapperClassHolder);
			(*env)->DeleteLocalRef(env, wrapperThreadClassHolder);
			if ((*env)->ExceptionCheck(env)) {
				(*env)->ExceptionClear(env);
				start = nil;
			}
		}
	}

	if (start) {
		jobject threadToStart = self;

		if ((*env)->IsInstanceOf(env, self, wrapperThreadClass)) {
			(*env)->ThrowNew(env,
				(*env)->FindClass(env, "java/lang/IllegalThreadStateException"),
				"Thread already started");
			[localPool release];
			return;
		}

		/* If we start running replacement Threads too early, Apple's
		 * Java Console stops working.
		 */		
		if (!awtLibrariesLoaded) {
			start(env, self);
			[localPool release];
			return;
		}

		if (!(*env)->GetBooleanField(env, self, stillbornID)) {
			jobject wrapper = nil;
			jclass ourClass = (*env)->GetObjectClass(env, self);
			jobject target = (*env)->GetObjectField(env, self, targetID);

			if ((*env)->IsSameObject(env, threadClass, ourClass)) {
				if (target) {
					if ((*env)->IsInstanceOf(env, target, wrapperClass)) {
						(*env)->ThrowNew(env,
							(*env)->FindClass(env, "java/lang/IllegalThreadStateException"),
							"Thread already started");
						[localPool release];
						return;
					}
					wrapper = (*env)->CallStaticObjectMethod(env, wrapperClass,
						newWrapperWithRunnable, target);
					if ((*env)->ExceptionCheck(env)) {
						(*env)->ExceptionClear(env);
					} else {
						(*env)->SetObjectField(env, self, targetID, wrapper);
					}
				}
			} else {
				wrapper = (*env)->CallStaticObjectMethod(env, wrapperClass,
					newWrapperWithThread, self, target);
				if ((*env)->ExceptionCheck(env)) {
					(*env)->ExceptionClear(env);
				} else {
					jobject classLoader;
					jobject newThread = nil;
					classLoader = (*env)->GetObjectField(env, self,
						classLoaderID);
					newThread = (*env)->CallObjectMethod(env, wrapper,
						getNewThread);
					if ((*env)->ExceptionCheck(env)) {
						(*env)->ExceptionClear(env);
					} else {
						(*env)->SetObjectField(env, self, targetID, wrapper);
						(*env)->SetObjectField(env, newThread, classLoaderID,
							classLoader);
						if ((*env)->ExceptionCheck(env)) {
							(*env)->ExceptionClear(env);
						} else {
							threadToStart = newThread;
							(*env)->MonitorEnter(env, threadToStart);
						}
					}
				}
			}

			if (wrapper)
				(*env)->DeleteLocalRef(env, wrapper);
			if (ourClass)
				(*env)->DeleteLocalRef(env, ourClass);
			if (target)
				(*env)->DeleteLocalRef(env, target);
		}

		start(env, threadToStart);
		if (threadToStart != self)
			(*env)->MonitorExit(env, threadToStart);
	}

	[localPool release];
}

jclass eventQueueClass = nil;
jclass myEventQueueClass = nil;
jclass eventMungerClass = nil;

typedef JNIEXPORT void JNICALL (*OBJECT_WAIT) (JNIEnv *, jobject, jlong);

/* Needed (in OS X 10.2.8) to set the value of MyEventQueue.safeToLock.
 * safeToLock is true if and only if our queue (MyEventQueue) is "wait"ing on
 * its EventMunger thread.  See MyEventQueue.java.
 */
JNIEXPORT void JNICALL Java_java_lang_Object_wait_redirect(
	JNIEnv *env, jobject self, jlong millis)
{
	NSSymbol symbol = NULL;
	static jclass threadClass = nil;
	static jmethodID currentThreadID = nil;
	static jfieldID safeToLockID = nil;
	static OBJECT_WAIT wait = nil;
	NSAutoreleasePool *localPool;

	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;

	if (!wait) {
		if (symbol = NSLookupAndBindSymbol("_JVM_MonitorWait"))
			wait = (OBJECT_WAIT) NSAddressOfSymbol(symbol);
	}

	if (!safeToLockID && myEventQueueClass) {
		jclass threadClassHolder = (*env)->FindClass(env, "java/lang/Thread");

		if ((*env)->ExceptionCheck(env)) {
			(*env)->ExceptionClear(env);
			wait = nil;
		} else {
			currentThreadID = (*env)->GetStaticMethodID(env, threadClassHolder,
				"currentThread", "()Ljava/lang/Thread;");
			threadClass = (*env)->NewGlobalRef(env, threadClassHolder);
			(*env)->DeleteLocalRef(env, threadClassHolder);
			safeToLockID = (*env)->GetFieldID(env, myEventQueueClass,
				"safeToLock", "Z");
			if ((*env)->ExceptionCheck(env)) {
				(*env)->ExceptionClear(env);
				safeToLockID = nil;
			}
		}
	}

	if (wait) {
		jboolean shouldSetSafeToLock = false;

		if (safeToLockID) {
			jobject currentThread = (*env)->CallStaticObjectMethod(env,
				threadClass, currentThreadID);
			if ((*env)->ExceptionCheck(env)) {
				(*env)->ExceptionClear(env);
			} else {
				if ((*env)->IsInstanceOf(env, self, myEventQueueClass) &&
					(*env)->IsInstanceOf(env, currentThread, eventMungerClass))
				{
					shouldSetSafeToLock = true;
					(*env)->SetBooleanField(env, self, safeToLockID, true);
				}
			}
		}

		wait(env, self, millis);

		if (shouldSetSafeToLock)
			(*env)->SetBooleanField(env, self, safeToLockID, false);

	}

	[localPool release];
}

/* I've already taken many precautions to avoid hangs in JNI_OnLoad() while
 * loading the AWT libraries (loading them on a secondary thread, waiting on
 * the main thread (and processing events through the run loop) while they're
 * being loaded).  But these hangs _still_ happen (occasionally) on recent
 * trunk and 1.8 branch versions of Firefox and Camino (probably beginning
 * with the change to loading the OJI plugin (MRJPlugin.plugin) on demand (see
 * https://bugzilla.mozilla.org/show_bug.cgi?id=128366)).  So (as of JEP
 * 0.9.4) I've now adopted more aggressive measures.  One sign of the problem
 * (which must be caused by one or more Apple bugs of very long standing) is
 * hangs in NSObject performSelectorOnMainThread, waiting for a "condition"
 * that never happens, while the AWT libraries (or even applets) are being
 * loaded.  I've gotten around the problem by using my own custom events and
 * event loop to do the equivalent of performSelectorOnMainThread while the
 * AWT libraries and applets are being loaded, intervening just before the
 * hangs would normally happen -- in start_AWTST(), performOnMainThread_TU()
 * and beginApp_SU() in Handlers.m.  This set of workarounds clears up
 * Bugzilla bug 306872.
 */
@implementation CustomEvent

- (id)initWithParams:(id)aTarget aSelector:(SEL)aSelector
	anObject:(id)anObject waitUntilDone:(BOOL)waitUntilDone
	modes:(NSArray *)modesArray
{
	if ((self = [super init]) == nil)
		return nil;
	if (aTarget) {
		target = [aTarget retain];
	} else {
		target = nil;
	}
	selector = aSelector;
	if (anObject) {
		object = [anObject retain];
	} else {
		object = nil;
	}
	wait = waitUntilDone;
	if (modesArray) {
		modes = [modesArray retain];
	} else {
		modes = nil;
	}
	done = false;
	return self;
}

- (id)getTarget
{
	return target;
}

- (SEL)getSelector
{
	return selector;
}

- (id)getObject
{
	return object;
}

- (BOOL)getWaitUntilDone
{
	return wait;
}

- (NSArray *)getModes
{
	return modes;
}

- (BOOL)getIsDone
{
	return done;
}

- (void)dispatch
{
	if (target && selector)
		[target performSelector:selector withObject:object];
	done = true;
}

- (void)dealloc
{
	JNIEnv *env = nil;
	BOOL jvmAttached = [AppletView getJavaEnvIfAttached:&env];
	if (target) {
		if ([target isKindOfClass:cPerformerClass]) {
			CPerformer *performer = (CPerformer *) target;
			if (performer->fTarget) {
				[performer->fTarget release];
				performer->fTarget = nil;
			}
			// [CustomEvent dealloc] is sometimes called as a thread is
			// shutting down, after the JVM has been detached -- which can
			// lead to crashes releasing CPerformer->fArg.  We try to avoid
			// this situation by releasing CustomEvents as soon as possible
			// after they've been "dispatched" (e.g.. by using "release"
			// instead of "autorelease").  But we'll leak CPerformer->fArg
			// if we have to.
			if (performer->fArg) {
				if (jvmAttached) {
					[performer->fArg release];
				} else {
					NSLog(@"[CustomEvent dealloc] called with JVM detached (Apple bug)!  Leaking CPerformer->fArg");
				}
				performer->fArg = nil;
			}
			[performer release];
		} else {
			[target release];
		}
	}
	if (object)
		[object release];
	if (modes)
		[modes release];
	[super dealloc];
}

@end

@implementation CEvent

- (id)initWithParams:(void (*) (void *))aRunnable userInfo:(void *)userInfo
	waitUntilDone:(BOOL)waitUntilDone
{
	if ((self = [super init]) == nil)
		return nil;
	runnable = aRunnable;
	info = userInfo;
	wait = waitUntilDone;
	done = false;
	return self;
}

- (void (*) (void *))getRunnable
{
	return runnable;
}

- (void *)getInfo
{
	return info;
}

- (BOOL)getWaitUntilDone
{
	return wait;
}

- (BOOL)getIsDone
{
	return done;
}

- (void)dispatch
{
	if (runnable)
		runnable(info);
	done = true;
}

@end

@implementation AppletView

/* Previous versions (before 0.8.2) naively assumed that every view to be
 * clipped uses the same coordinate system as the top-level applet view
 * (myAppletView) ... but that isn't always true.
 */
+ (void)clipWithTranslation:(NSView *)drawRectView controller:(JEPController *)controller
{
	NSRect viewBounds = NSZeroRect;
	NSPoint viewLocation = NSZeroPoint;
	NSRect translatedClip = NSZeroRect;
	if (!drawRectView || !controller || !controller->myAppletView ||
		controller->myAppletView->isInvisible ||
		![controller->myAppletView isAppletVisible])
	{
		[NSBezierPath clipRect:NSZeroRect];
		return;
	}
	translatedClip = controller->myAppletView->currentBounds;
	if ((translatedClip.size.width <= 0) || (translatedClip.size.height <= 0)) {
		[NSBezierPath clipRect:NSZeroRect];
		return;
	}
	viewBounds = [drawRectView bounds];
	viewLocation = [controller->myAppletView convertPoint:viewBounds.origin
		fromView:drawRectView];
	translatedClip.origin.x -= viewLocation.x;
	translatedClip.origin.y -= viewLocation.y;
	[NSBezierPath clipRect:translatedClip];
}

+ (BOOL)isJava142:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jboolean retval = false;
	jclass appletHandlerFactoryClass;
	jmethodID isJava142;

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return false;
	}

	appletHandlerFactoryClass =
		(*env)->FindClass(env, "jep/AppletHandlerFactory");
	isJava142 = (*env)->GetStaticMethodID(env, appletHandlerFactoryClass,
		"isJava142", "()Z");
	retval = (*env)->CallStaticBooleanMethod(env, appletHandlerFactoryClass,
		isJava142);
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return false;
	}

	return retval;
}

+ (BOOL)isJava142Update1:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jboolean retval = false;
	jclass appletHandlerFactoryClass;
	jmethodID isJava142Update1;

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return false;
	}

	appletHandlerFactoryClass =
		(*env)->FindClass(env, "jep/AppletHandlerFactory");
	isJava142Update1 = (*env)->GetStaticMethodID(env, appletHandlerFactoryClass,
		"isJava142Update1", "()Z");
	retval = (*env)->CallStaticBooleanMethod(env, appletHandlerFactoryClass,
		isJava142Update1);
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return false;
	}

	return retval;
}

+ (BOOL)isJava142Update2:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jboolean retval = false;
	jclass appletHandlerFactoryClass;
	jmethodID isJava142Update2;

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return false;
	}

	appletHandlerFactoryClass =
		(*env)->FindClass(env, "jep/AppletHandlerFactory");
	isJava142Update2 = (*env)->GetStaticMethodID(env, appletHandlerFactoryClass,
		"isJava142Update2", "()Z");
	retval = (*env)->CallStaticBooleanMethod(env, appletHandlerFactoryClass,
		isJava142Update2);
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return false;
	}

	return retval;
}

+ (BOOL)isJava142Tiger:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jboolean retval = false;
	jclass appletHandlerFactoryClass;
	jmethodID isJava142Tiger;

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return false;
	}

	appletHandlerFactoryClass =
		(*env)->FindClass(env, "jep/AppletHandlerFactory");
	isJava142Tiger = (*env)->GetStaticMethodID(env, appletHandlerFactoryClass,
		"isJava142Tiger", "()Z");
	retval = (*env)->CallStaticBooleanMethod(env, appletHandlerFactoryClass,
		isJava142Tiger);
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return false;
	}

	return retval;
}

+ (BOOL)isJava150:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jboolean retval = false;
	jclass appletHandlerFactoryClass;
	jmethodID isJava150;

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return false;
	}

	appletHandlerFactoryClass =
		(*env)->FindClass(env, "jep/AppletHandlerFactory");
	isJava150 = (*env)->GetStaticMethodID(env, appletHandlerFactoryClass,
		"isJava150", "()Z");
	retval = (*env)->CallStaticBooleanMethod(env, appletHandlerFactoryClass,
		isJava150);
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return false;
	}

	return retval;
}

+ (BOOL)isJava160:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jboolean retval = false;
	jclass appletHandlerFactoryClass;
	jmethodID isJava160;

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return false;
	}

	appletHandlerFactoryClass =
		(*env)->FindClass(env, "jep/AppletHandlerFactory");
	isJava160 = (*env)->GetStaticMethodID(env, appletHandlerFactoryClass,
		"isJava160", "()Z");
	retval = (*env)->CallStaticBooleanMethod(env, appletHandlerFactoryClass,
		isJava160);
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return false;
	}

	return retval;
}

+ (NSString *)getJavaVersionString:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jstring retvalJava = nil;
	const jchar *retvalUTF16 = nil;
	jint retvalLength;
	NSString *retvalCocoa = nil;
	jclass appletHandlerFactoryClass;
	jmethodID getJavaVersionString;

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return nil;
	}

	appletHandlerFactoryClass =
		(*env)->FindClass(env, "jep/AppletHandlerFactory");
	getJavaVersionString = (*env)->GetStaticMethodID(env, appletHandlerFactoryClass,
		"getJavaVersionString", "()Ljava/lang/String;");
	retvalJava = (*env)->CallStaticObjectMethod(env, appletHandlerFactoryClass,
		getJavaVersionString);
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return nil;
	}
	if (retvalJava) {
		retvalLength = (*env)->GetStringLength(env, retvalJava);
		retvalUTF16 = (*env)->GetStringChars(env, retvalJava, nil);
	}
	if (retvalUTF16) {
		retvalCocoa = [[NSString stringWithCharacters:retvalUTF16
			length:retvalLength] autorelease];
		(*env)->ReleaseStringChars(env, retvalJava, retvalUTF16);
	}
	if (retvalJava)
		(*env)->DeleteLocalRef(env, retvalJava);
	if ((*env)->ExceptionCheck(env))
		(*env)->ExceptionClear(env);

	return retvalCocoa;
}

+ (OSStatus)maybeCreateJavaVM:(JavaVMOption *)jvmExtraOpts
	extraOptsNum:(int)jvmExtraOptsNum vmPtr:(JavaVM **)vmPtr
	envPtr:(void **)envPtr initAWT:(BOOL)initAWT
{
	OSStatus retval = noErr;
	const char *homeDirectoryUTF8 = nil;
	const char *installedPathUTF8 = nil;
	unsigned homeDirectoryLen = 0;
	unsigned installedPathLen = 0;
	char *classpathDefault;
	char *classpathDefault141, *classpathDefault142, *classpathDefault15X;
	char *classpathOpt = nil;
	char *ourJarOpt = nil;
	char *ourPluginHome = nil;
	JavaVM *vmBuf[50] = {0};
	jsize numVMs = 0;
	int i, numOpts, len;
	unsigned j, count;
	NSMutableArray *runtimeParams = nil;
	NSString *parm;
	JavaVMInitArgs initArgs = {0};
	JavaVMOption *opts = nil;
	JNIEnv *env;
	JavaVM *vm;
	NSThread *currentThread = [NSThread currentThread];
	BOOL isMainThread;

	if (vmPtr)
		*vmPtr = nil;
	if (envPtr)
		*envPtr = nil;

	/* If mainThread or nsAppShadow are nil, JEPController checkBundle hasn't
	 * yet been called.
	 */
	if (!mainThread || !nsAppShadow)
		return EINVAL;
	isMainThread = [mainThread isEqual:currentThread];

	if (installedPath) {
		installedPathUTF8 = [installedPath UTF8String];
		installedPathLen = strlen(installedPathUTF8);
	}
	if (!installedPathLen)
		return ENOMEM;

	/* JNI_GetCreatedJavaVMs can return bogus JVMs if there has ever been a
	 * failed call to JNI_CreateJavaVM ... though perhaps only when running
	 * under Rosetta on an Intel Mac.  Apple's 1.4.X and 5.0 JVMs don't run
	 * under Rosetta, and so calls to JNI_CreateJavaVM and
	 * JNI_GetCreatedJavaVMs should both always fail.  This is an Apple bug,
	 * which has been given radar number 4434041.
	 * We can't know if other code (outside the JEP) has tried (and failed)
	 * to create a JVM ... but that doesn't seem to happen in the normal
	 * course of events.  This patch (used here and at all the rest of the
	 * JEP's calls to JNI_GetCreatedJavaVMs) is expanded from one by Mark
	 * Mentovai.  See https://bugzilla.mozilla.org/show_bug.cgi?id=325084.
	 */
	if (everCreatedJVM &&
		!JNI_GetCreatedJavaVMs(vmBuf, sizeof(vmBuf)/sizeof(JavaVM *), &numVMs))
	{
		for (i = 1; i <= numVMs; ++i) {
			/* Return with error EALREADY if a VM already exists that is
			 * or can be connected to the current thread, and fill vmPtr
			 * and envPtr (if non-NULL) with values for the first VM we
			 * encounter.
			 */
			if ((retval = (*vmBuf[i-1])->GetEnv(vmBuf[i-1], (void **) &env,
				JNI_VERSION_1_4)) == JNI_OK)
			{
				retval = EALREADY;
				break;
			}
			if (retval == JNI_EDETACHED) {
				if (!(*vmBuf[i-1])->AttachCurrentThread(vmBuf[i-1],
					(void **) &env, nil))
				{
					retval = EALREADY;
					break;
				}
			}
			retval = noErr;
		}
	}

	if (retval == EALREADY) {
		if (vmPtr)
			*vmPtr = vmBuf[i-1];
		if (envPtr)
			*envPtr = (void *) env;
		[AppletView registerJEPFactoryNatives:env];
		isJava142 = [AppletView isJava142:env];
		isJava142Update1 = [AppletView isJava142Update1:env];
		isJava142Update2 = [AppletView isJava142Update2:env];
		isJava142Tiger = [AppletView isJava142Tiger:env];
		isJava150 = [AppletView isJava150:env];
		isJava160 = [AppletView isJava160:env];
		if (initAWT && !awtLibrariesLoaded) {
			[NSThread detachNewThreadSelector:@selector(loadAWTLibraries)
				toTarget:[AppletView class] withObject:nil];
		}
		if (!awtLibrariesLoaded) {
			/* Wait for the AWT libraries to finish loading.  This makes it
			 * possible to go back to loading the JVM on demand even if
			 * LiveConnect is being used outside any applet.  It also ensures
			 * that our security is initialized by the time this method
			 * returns (needed by our fix (in the MRJ Plugin JEP) for the
			 * "Arbitrary Package Access" security hole).
			 */
			/* As of JEP 0.9.5+h we always wait here for the AWT libraries
			 * to be loaded (if they haven't been already), whether or not
			 * initAWT is true.  This is because, since the new
			 * nsIThreadManager was landed on the trunk (bmo bug 326273,
			 * 2006-05-10 10:30 PDT), it's possible for a second applet to
			 * begin loading before the first one has finished loading ...
			 * or even before the AWT libraries have finished loading.  In
			 * earlier JEP versions this causes wierd UnsatisfiedLinkError
			 * messages for native methods that are definitely present, but
			 * which haven't yet been loaded by registerJEPNatives.  So far
			 * I've only seen this with Camino, but in principle it should
			 * happen with any trunk build.  To see the problem in Camino,
			 * make the first page you load (containing Java applets) be a
			 * page that contains multiple applets which load rather slowly.
			 * A good example is http://gemal.dk/browserspy/java.html.
			 * Since maybeCreateJavaVM is always called from
			 * JEPCreateJavaApplet, always waiting here will ensure that no
			 * applet is loaded before the AWT libraries have finished
			 * loading.  (The following loop may be re-entered -- but that
			 * shouldn't cause any problems.)
			 */
			if (isMainThread) {
				[AppletView postFakeEvent];
				while (!awtLibrariesLoaded) {
					if (osxVersion >= 0x1030) {
						[NSApp nextEventMatchingMask:NSAnyEventMask untilDate:nil
							inMode:NSDefaultRunLoopMode dequeue:false];
						[AppletView processCustomEventSpool:false];
						[AppletView processCEventSpool:false];
						usleep(20000);
					} else {
						[NSApp nextEventMatchingMask:NSAnyEventMask
							untilDate:[NSDate dateWithTimeIntervalSinceNow:.05]
							inMode:NSDefaultRunLoopMode dequeue:false];
						[AppletView processCustomEventSpool:false];
						[AppletView processCEventSpool:false];
					}
				}
			} else {
				while (!awtLibrariesLoaded)
					[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:.1]];
			}
		}
		[nsAppShadow performSelectorOnMainThread:@selector(setMainJavaThread)
			withObject:nil waitUntilDone:true];
		return retval;
	}

	/* For some reason, the following avoids several different kinds of
	 * crashes in Camino (from at least 0.8.4 to the latest versions,
	 * including 1.0) on OS X 10.4.4 and 10.4.5 (which is the latest version).
	 * Some of the crashes actually take place in preferences code (e.g. in
	 * methods of the NSUserDefaults class).  Others are (apparently) caused
	 * by strange things that happen in preferences code (e.g. many of the
	 * crashes I reported at Bugzilla bug 326195).  Still others seem (on the
	 * surface) to be completely unrelated (e.g. this fix seems to reduce or
	 * eliminate Bugzilla bug 312062 crashes on OS X 10.2.8 -- ones that
	 * aren't covered by my workaround that actually addressed that bug).
	 * It's difficult to tell why this fix works.  The problems it fixes are
	 * deeply hidden in Apple code (and possibly also Mozilla.org code) -- in
	 * the JVM, in recent versions of OS X 10.4's CoreText framework, and
	 * possibly also in Camino.  But this much is clear:  The "AppleLanguages"
	 * preferences (the individual language identifier strings like "en", but
	 * not the array that contains them) are "released" while they're still in
	 * use -- which makes subsequent attempts to access them lead to crashes.
	 * What the fix actually does is to delete the "standard user defaults"
	 * object, so that it gets re-created on a subsequent call to the
	 * standardUserDefaults method -- one made from somewhere in the JVM.
	 * This call doesn't cause any changes to be made to Camino's (or other
	 * browsers') preferences -- either ones that were made in the current
	 * session, or ones that were made (and saved) before the last time the
	 * browser was started.
	 */
	[NSUserDefaults resetStandardUserDefaults];

	classpathDefault141 =
		"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.1/Home/lib/jaws.jar:"
		"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.1/Home/lib/netscape.jar:"
		"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.1/Home/lib/ext/apple_provider.jar:"
		"/Library/Internet Plug-Ins/JavaPluginCocoa.bundle/Contents/Resources/AppletsInCocoa.jar";
	classpathDefault142 =
		"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.2/Home/lib/jaws.jar:"
		"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.2/Home/lib/netscape.jar:"
		"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.2/Home/lib/ext/apple_provider.jar:"
		"/Library/Internet Plug-Ins/JavaPluginCocoa.bundle/Contents/Resources/AppletsInCocoa.jar";
	classpathDefault15X =
		"/System/Library/Frameworks/JavaVM.framework/Versions/1.5/Home/lib/deploy.jar:"
		"/System/Library/Frameworks/JavaVM.framework/Versions/1.5/Home/lib/plugin.jar:"
		"/System/Library/Frameworks/JavaVM.framework/Versions/1.5/Home/lib/ext/apple_provider.jar";
	if (isJava150) {
		classpathDefault = classpathDefault15X;
	} else if (isJava142) {
		classpathDefault = classpathDefault142;
	} else {
		classpathDefault = classpathDefault141;
	}
	len = strlen("-Xbootclasspath/a:") + strlen(classpathDefault) + 1;
	if (homeDirectory) {
		homeDirectoryUTF8 = [homeDirectory UTF8String];
		homeDirectoryLen = strlen(homeDirectoryUTF8);
	}
	if (homeDirectoryLen && !isJava142Update1) {
		len += homeDirectoryLen +
			strlen("/Library/Internet Plug-Ins/JavaPluginCocoa.bundle/Contents/Resources/AppletsInCocoa.jar")
			+ 1;
	}
	if ((classpathOpt = (char *) malloc(len*sizeof(char))) == nil)
		return ENOMEM;
	if (homeDirectoryLen && !isJava142Update1) {
		if (isJava142) {
			sprintf(classpathOpt, "-Xbootclasspath/a:"
				"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.2/Home/lib/jaws.jar:"
				"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.2/Home/lib/netscape.jar:"
				"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.2/Home/lib/ext/apple_provider.jar:"
				"%s/Library/Internet Plug-Ins/JavaPluginCocoa.bundle/Contents/Resources/AppletsInCocoa.jar:"
				"/Library/Internet Plug-Ins/JavaPluginCocoa.bundle/Contents/Resources/AppletsInCocoa.jar",
				homeDirectoryUTF8);
		} else {
			sprintf(classpathOpt, "-Xbootclasspath/a:"
				"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.1/Home/lib/jaws.jar:"
				"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.1/Home/lib/netscape.jar:"
				"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.1/Home/lib/ext/apple_provider.jar:"
				"%s/Library/Internet Plug-Ins/JavaPluginCocoa.bundle/Contents/Resources/AppletsInCocoa.jar:"
				"/Library/Internet Plug-Ins/JavaPluginCocoa.bundle/Contents/Resources/AppletsInCocoa.jar",
				homeDirectoryUTF8);
		}
	} else {
		sprintf(classpathOpt, "-Xbootclasspath/a:%s", classpathDefault);
	}

	len = strlen("-Xbootclasspath/p:") + installedPathLen +
		strlen("/Contents/Resources/Java/JavaEmbeddingPlugin.jar") + 1;
	if ((ourJarOpt = (char *) malloc(len*sizeof(char))) == nil) {
		free(classpathOpt);
		return ENOMEM;
	}
	sprintf(ourJarOpt,
		"-Xbootclasspath/p:%s/Contents/Resources/Java/JavaEmbeddingPlugin.jar",
		installedPathUTF8);

	len = strlen("-Djep.pluginhome=") + installedPathLen + 1;
	if ((ourPluginHome = (char *) malloc(len*sizeof(char))) == nil) {
		free(classpathOpt);
		free(ourJarOpt);
		return ENOMEM;
	}
	sprintf(ourPluginHome, "-Djep.pluginhome=%s", installedPathUTF8);

	numOpts = 16;
	if (jvmExtraOpts)
		numOpts += jvmExtraOptsNum;
	if ((runtimeParams = [AppletView getRuntimeParams]) != nil)
		numOpts += [runtimeParams count];
	if ((opts = (JavaVMOption *) malloc(numOpts*sizeof(JavaVMOption))) == nil) {
		free(classpathOpt);
		free(ourJarOpt);
		free(ourPluginHome);
		return ENOMEM;
	}
	memset(opts, 0, numOpts*sizeof(JavaVMOption));
	i = 1;
	if (jvmExtraOpts) {
		for (; i <= jvmExtraOptsNum; ++i) {
			opts[i-1].optionString = jvmExtraOpts[i-1].optionString;
			opts[i-1].extraInfo = jvmExtraOpts[i-1].extraInfo;
		}
	}
	/* As of "Java for Mac OS X 10.4 Release 5", Apple now sets the maximum
	 * heap size to 96 MB.  But it doesn't do this in a way that's accessible
	 * to the JEP, so we need to do it ourselves.  I've chosen to simply set
	 * the maximum heap size to 96 MB whenever we run on OS X 10.4 or higher.
	 * Since Java Control Panel settings are processed later, they can
	 * override this setting.
	 * In JEP 0.9.6.1 I tried to accomplish this by setting the javaplugin.
	 * maxHeapSize property in AppletHandlerFactory.initProperties().  But
	 * that didn't work, so now (as of JEP 0.9.6.4) I do it as follows.
	 * This resolves
	 * http://sourceforge.net/forum/forum.php?thread_id=1696451&forum_id=371125 and
	 * http://sourceforge.net/tracker/index.php?func=detail&aid=1796738&group_id=107955&atid=649116.
	 * See http://developer.apple.com/releasenotes/Java/Java104R5RN/index.html.
	 */
	if (osxVersion >= 0x1040) {
		opts[i++-1].optionString = "-Xmx96m";
		opts[i++-1].optionString = "-Djavaplugin.maxHeapSize=96m";
    }
	if (runtimeParams) {
		count = [runtimeParams count];
		for (j = 1; j <= count; ++j) {
			parm = [runtimeParams objectAtIndex:j-1];
			opts[i++-1].optionString = (char *) [parm UTF8String];
		}
	}
	opts[i++-1].optionString = classpathOpt;
	opts[i++-1].optionString = ourJarOpt;
	opts[i++-1].optionString = ourPluginHome;
	opts[i++-1].optionString = "-D" JEP_VERSION_OPTION "=" JEP_VERSION;
	opts[i++-1].optionString = "-D" JEP_VERSION_OPTION ".applet=true";
	opts[i++-1].optionString = "-D" JEP_DEBUG_VISIBILITY_OPTION ".applet=true";
	opts[i++-1].optionString = "-D" JEP_DEBUG_RELEASE_OPTION ".applet=true";
	opts[i++-1].optionString = "-D" JEP_DEBUG_UPDATES_OPTION ".applet=true";
#ifdef DEBUG_JVM_EVENTS
	opts[i++-1].optionString = "-enableassertions";
	opts[i++-1].optionString = "-enablesystemassertions";
#endif
#ifdef DEBUG_APPLET_HOLDER
	opts[i++-1].optionString = "-Djep.debuginfo=true";
#endif
	opts[i++-1].optionString = "-DtrustProxy=true";
	opts[i-1].optionString = "vfprintf";
	opts[i-1].extraInfo = &java_vfprintf;
	initArgs.version = JNI_VERSION_1_4;
	initArgs.nOptions = i;
	initArgs.options = opts;
	initArgs.ignoreUnrecognized = JNI_TRUE;
	if (JNI_CreateJavaVM(&vm, (void **) &env, &initArgs))
		retval = ENOMEM;

	free(opts);
	free(classpathOpt);
	free(ourJarOpt);
	free(ourPluginHome);
	[runtimeParams removeAllObjects];
	[runtimeParams release];
	if (retval == noErr) {
		everCreatedJVM = true;
		if (vmPtr)
			*vmPtr = vm;
		if (envPtr)
			*envPtr = (void *) env;
		[AppletView registerJEPFactoryNatives:env];
		isJava142 = [AppletView isJava142:env];
		isJava142Update1 = [AppletView isJava142Update1:env];
		isJava142Update2 = [AppletView isJava142Update2:env];
		isJava142Tiger = [AppletView isJava142Tiger:env];
		isJava150 = [AppletView isJava150:env];
		isJava160 = [AppletView isJava160:env];
		if (initAWT && !awtLibrariesLoaded) {
			[NSThread detachNewThreadSelector:@selector(loadAWTLibraries)
				toTarget:[AppletView class] withObject:nil];
			/* Wait for the AWT libraries to finish loading.  This makes it
			 * possible to go back to loading the JVM on demand even if
			 * LiveConnect is being used outside any applet.  It also ensures
			 * that our security is initialized by the time this method
			 * returns (needed by our fix (in the MRJ Plugin JEP) for the
			 * "Arbitrary Package Access" security hole).
			 */
			if (isMainThread) {
				[AppletView postFakeEvent];
				while (!awtLibrariesLoaded) {
					if (osxVersion >= 0x1030) {
						[NSApp nextEventMatchingMask:NSAnyEventMask untilDate:nil
							inMode:NSDefaultRunLoopMode dequeue:false];
						[AppletView processCustomEventSpool:false];
						[AppletView processCEventSpool:false];
						usleep(20000);
					} else {
						[NSApp nextEventMatchingMask:NSAnyEventMask
							untilDate:[NSDate dateWithTimeIntervalSinceNow:.05]
							inMode:NSDefaultRunLoopMode dequeue:false];
						[AppletView processCustomEventSpool:false];
						[AppletView processCEventSpool:false];
					}
				}
			} else {
				while (!awtLibrariesLoaded)
					[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:.1]];
			}
		}
		[nsAppShadow performSelectorOnMainThread:@selector(setMainJavaThread)
			withObject:nil waitUntilDone:true];
	}
	return retval;
}

/* Here we safely load the AWT libraries on a secondary thread (to avoid the
 * Apple bug that causes a hang on JNI_OnLoad() in libawt.jnilib when we try
 * to do this on the same thread on which we just started the JVM).  This is
 * similar to what we do in AppletView.m's DoInitialization -- but here we're
 * not creating an applet.  It turns out that JavaScript uses Java to do
 * certain kinds of LiveConnect things (e.g. many of the LiveConnect tests at
 * http://www.mozilla.org/quality/browser/front-end/testcases/oji/liveconnecttest.html)
 * without ever loading an applet.  Before we started making maybeCreateJavaVM
 * wait for the AWT libraries to finish loading when called with initAWT set
 * to TRUE, it was unsafe to do this except as the browser was starting up.
 * But now (as of JEP 0.8.9) we've gone back to starting the JVM on demand,
 * even when using LiveConnect outside of any applet.  See CreateJVM() in the
 * MRJ Plugin JEP's MRJSession.cpp for more information.  (As of JEP 0.9.4 we
 * load the "java/awt/Frame" class (which triggers loading the AWT libraries)
 * in AppletHandlerFactory.initProperties(), instead of doing it here.)
 */
+ (void)loadAWTLibraries
{
	JNIEnv *env = nil;
	JavaVM *vm = nil;
	NSAutoreleasePool *localPool;
	jclass appletHandlerFactoryClass;
	jmethodID initProperties;
	jclass eventQueueClassLocal = nil;
	jclass myEventQueueClassLocal = nil;
	jclass eventMungerClassLocal = nil;

	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;

	if (![AppletView getJavaEnv:&env vmPtr:&vm]) {
		[localPool release];
		return;
	}

	appletHandlerFactoryClass =
		(*env)->FindClass(env, "jep/AppletHandlerFactory");
	initProperties = (*env)->GetStaticMethodID(env,
		appletHandlerFactoryClass, "initProperties", "()V");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		[localPool release];
		(*vm)->DetachCurrentThread(vm);
		return;
	}
	(*env)->CallStaticVoidMethod(env, appletHandlerFactoryClass, initProperties);
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		[localPool release];
		(*vm)->DetachCurrentThread(vm);
		return;
	}

	[AppletView registerJEPNatives:env];

	eventQueueClassLocal =
		(*env)->FindClass(env, "java/awt/EventQueue");
	myEventQueueClassLocal =
		(*env)->FindClass(env, "apple/awt/MyEventQueue");
	eventMungerClassLocal =
		(*env)->FindClass(env, "apple/awt/MyEventQueue$EventMunger");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
	} else {
		eventQueueClass =
			(*env)->NewGlobalRef(env, eventQueueClassLocal);
		myEventQueueClass =
			(*env)->NewGlobalRef(env, myEventQueueClassLocal);
		eventMungerClass =
			(*env)->NewGlobalRef(env, eventMungerClassLocal);
		(*env)->DeleteLocalRef(env, eventQueueClassLocal);
		(*env)->DeleteLocalRef(env, myEventQueueClassLocal);
		(*env)->DeleteLocalRef(env, eventMungerClassLocal);
		if ((*env)->ExceptionCheck(env)) {
			(*env)->ExceptionClear(env);
			eventQueueClass = nil;
			myEventQueueClass = nil;
			eventMungerClass = nil;
		}
	}

	if (afterInitAWT)
		afterInitAWT(afterInitAWTInfo);

	awtLibrariesLoaded = true;

	[localPool release];

	/* This thread will become a zombie thread if we don't call
	 * DetachCurrentThread().
	 */
	(*vm)->DetachCurrentThread(vm);

	return;
}

/* Some of these might be used in AppletHandlerFactory.initProperties(), so
 * they need to be registered before that's called (in loadAWTLibraries).
 * None of this should code should trigger loading the AWT libraries
 * (prematurely).
 */
+ (void)registerJEPFactoryNatives:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jclass appletHandlerFactoryClass;
	jclass jepCookieHandlerClass = nil;
	jclass runnableWrapperClass = nil;
	jclass threadClass = nil;
	jclass objectClass = nil;
	JNINativeMethod AppletHandlerFactoryMethods[] = {
		{"getNSColor", "(III)J", &Java_AppletHandlerFactory_getNSColor},
		{"getNSSystemColor", "(Ljava/lang/String;[I)J",
			&Java_AppletHandlerFactory_getNSSystemColor},
		{"getCurrentPThreadID", "()I",
			&Java_AppletHandlerFactory_getCurrentPThreadID},
		{"getOSXVersion", "()I",
			&Java_AppletHandlerFactory_getOSXVersion},
		{"disableScreenUpdates", "()V",
			&Java_AppletHandlerFactory_disableScreenUpdates},
		{"enableScreenUpdates", "()V",
			&Java_AppletHandlerFactory_enableScreenUpdates},
		{"allocAutoreleasePool", "()J",
			&Java_AppletHandlerFactory_allocAutoreleasePool},
		{"releaseAutoreleasePool", "(J)V",
			&Java_AppletHandlerFactory_releaseAutoreleasePool},
		{"isAppletViewDisplayable", "(J)Z",
			&Java_AppletHandlerFactory_isAppletViewDisplayable},
		{"resetHybridView", "(J)V",
			&Java_AppletHandlerFactory_resetHybridView},
		{"suspendAppletUpdates", "(J)V",
			&Java_AppletHandlerFactory_suspendAppletUpdates},
		{"resumeAppletUpdates", "(J)V",
			&Java_AppletHandlerFactory_resumeAppletUpdates},
		{"cocoaDrawApplet", "(J)V",
			&Java_AppletHandlerFactory_cocoaDrawApplet},
		{"deleteAppletJNIGlobalRefs", "(J)V",
			&Java_AppletHandlerFactory_deleteAppletJNIGlobalRefs},
		{"tellHandlerCreated", "(JLjep/AppletHandler;)V",
			&Java_AppletHandlerFactory_tellHandlerCreated},
		{"processMainThreadEvents", "(JJ)V",
			&Java_processMainThreadEvents},
		{"processQueuedMainThreadEvents", "()V",
			&Java_processQueuedMainThreadEvents},
		{"beforeProcessing", "(Ljava/lang/Object;)J",
			&Java_beforeProcessing},
		{"stopProcessing", "(J)V",
			&Java_stopProcessing},
		{"callRegisterNatives", "()V",
			&Java_callRegisterNatives}
		};
	JNINativeMethod JEPCookieHandlerMethods[] = {
		{"cookiesSupported", "()Z",
			&Java_JEPCookieHandler_cookiesSupported},
		{"setCookie", "(Ljava/lang/String;Ljava/lang/String;)V",
			&Java_JEPCookieHandler_setCookie},
		{"getCookie", "(Ljava/lang/String;)Ljava/lang/String;",
			&Java_JEPCookieHandler_getCookie}
		};
	JNINativeMethod RunnableWrapperMethods[] = {
		{"poseAsThread", "(Ljava/lang/Thread;Ljava/lang/Thread;)V",
			&Java_RunnableWrapper_poseAsThread},
		{"unposeThreads", "(Ljava/lang/Thread;Ljava/lang/Thread;)V",
			&Java_RunnableWrapper_unposeThreads}
		};
	JNINativeMethod ThreadMethods[] = {
		{"start", "()V",
			&Java_java_lang_Thread_start_redirect}
		};
	JNINativeMethod ObjectMethods[] = {
		{"wait", "(J)V",
			&Java_java_lang_Object_wait_redirect}
		};
	static Boolean done = false;
	if (done)
		return;
	done = true;
	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return;
	}
	appletHandlerFactoryClass =
		(*env)->FindClass(env, "jep/AppletHandlerFactory");
	runnableWrapperClass =
		(*env)->FindClass(env, "jep/RunnableWrapper");
	threadClass =
		(*env)->FindClass(env, "java/lang/Thread");
	objectClass =
		(*env)->FindClass(env, "java/lang/Object");
	if (isJava150) {
		jepCookieHandlerClass =
			(*env)->FindClass(env, "com/sun/deploy/net/cookie/JEPCookieHandler150");
	} else {
		jepCookieHandlerClass =
			(*env)->FindClass(env, "sun/plugin/net/cookie/JEPCookieHandler14X");
	}
	if ((*env)->ExceptionCheck(env))
		return;
	(*env)->RegisterNatives(env, appletHandlerFactoryClass,
		AppletHandlerFactoryMethods,
		sizeof(AppletHandlerFactoryMethods)/sizeof(JNINativeMethod));
	(*env)->RegisterNatives(env, jepCookieHandlerClass,
		JEPCookieHandlerMethods,
		sizeof(JEPCookieHandlerMethods)/sizeof(JNINativeMethod));
	(*env)->RegisterNatives(env, runnableWrapperClass,
		RunnableWrapperMethods,
		sizeof(RunnableWrapperMethods)/sizeof(JNINativeMethod));
	if (osxVersion < 0x1040) {
		(*env)->RegisterNatives(env, threadClass, ThreadMethods,
			sizeof(ThreadMethods)/sizeof(JNINativeMethod));
	}
	if (!isJava142Update1) {
		(*env)->RegisterNatives(env, objectClass, ObjectMethods,
			sizeof(ObjectMethods)/sizeof(JNINativeMethod));
	}
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	(*env)->DeleteLocalRef(env, jepCookieHandlerClass);
	(*env)->DeleteLocalRef(env, runnableWrapperClass);
	(*env)->DeleteLocalRef(env, threadClass);
	(*env)->DeleteLocalRef(env, objectClass);
}

+ (void)registerJEPNatives:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jclass appletHandlerClass;
	jclass appletFrameClass;
	jclass appletFrameContextClass;
	jclass appletFramePanelClass;
	jclass myCAppletFrameClass = nil;
	jclass cPanelClass = nil;
	jclass osxOffScreenSDClass = nil;
	jclass cPeerSDClass = nil;
	jclass gifDecoderClass = nil;
	JNINativeMethod AppletFrameContextMethods[] = {
		{"showDocument", "(JLjava/lang/String;Ljava/lang/String;)V",
			&Java_AppletFrameContext_showDocument},
		{"showStatus", "(JLjava/lang/String;)V",
			&Java_AppletFrameContext_showStatus},
		{"nativeGetJSObject", "(I)Lnetscape/javascript/JSObject;",
			&Java_AppletFrameContext_nativeGetJSObject}
		};
	JNINativeMethod AppletFramePanelMethods[] = {
		{"tellAppletCreated", "(JLjava/applet/Applet;Ljava/lang/String;Ljava/lang/String;)V",
			&Java_AppletFramePanel_tellAppletCreated}
		};
	JNINativeMethod MyCAppletFrameMethods[] = {
		{"createNativeEmbedded", "(JZZLjava/nio/ByteBuffer;)J",
			&Java_MyCAppletFrame_createNativeEmbedded},
		{"_getViewPtr", "(J)J",
			&Java_MyCAppletFrame_getViewPtr}
		};
	JNINativeMethod MyCAppletFrameMethods160[] = {
		{"createNativeEmbedded", "(JZZZLjava/nio/ByteBuffer;)J",
			&Java_MyCAppletFrame_createNativeEmbedded160},
		{"_getViewPtr", "(J)J",
			&Java_MyCAppletFrame_getViewPtr}
		};
	JNINativeMethod AppletHandlerMethods[] = {
		{"maybeDoAppletStart", "(J)V",
			&Java_AppletHandler_maybeDoAppletStart},
		{"reportAppletError", "(J)V",
			&Java_AppletHandler_reportAppletError},
		{"updateStatusFromJava", "(JI)V",
			&Java_AppletHandler_updateStatusFromJava}
		};
	JNINativeMethod CPanelMethods142T[] = {
		{"createNativePanel", "(JJIIIIZZLjava/awt/Font;Ljava/awt/Color;ZLjava/awt/Color;ZLjava/awt/Cursor;)J",
			&Java_CPanel_createNativePanel142T}
		};
	JNINativeMethod CPanelMethods[] = {
		{"createNativePanel", "(JIIIIZZLjava/awt/Font;Ljava/awt/Color;ZLjava/awt/Color;ZLjava/awt/Cursor;)J",
			&Java_CPanel_createNativePanel}
		};
	JNINativeMethod OSXOffScreenSDMethods[] = {
		{"initRaster", "(Ljava/lang/Object;IIIIILjava/awt/image/IndexColorModel;ILjava/nio/ByteBuffer;Ljava/lang/Object;Ljava/nio/ByteBuffer;)V",
			&Java_apple_awt_OSXOffScreenSurfaceData_initRaster_redirect},
		{"initCustomRaster", "(Ljava/nio/IntBuffer;IILjava/nio/ByteBuffer;Ljava/lang/Object;Ljava/nio/ByteBuffer;)V",
			&Java_apple_awt_OSXOffScreenSurfaceData_initCustomRaster_redirect}
		};
	JNINativeMethod CPeerSDMethods[] = {
		{"initOps", "(JLjava/nio/ByteBuffer;[Ljava/lang/Object;)V",
			&Java_apple_awt_CPeerSurfaceData_initOps_redirect}
		};
	JNINativeMethod GIFDecoderMethods[] = {
		{"parseImage", "(IIIIZI[B[BLjava/awt/image/IndexColorModel;)Z",
			&Java_sun_awt_image_GifImageDecoder_parseImage_redirect}
		};
	static Boolean done = false;
	if (done)
		return;
	done = true;
	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return;
	}
	appletHandlerClass =
		(*env)->FindClass(env, "jep/AppletHandler");
	appletFrameClass =
		(*env)->FindClass(env, "jep/AppletFrame");
	appletFrameContextClass =
		(*env)->FindClass(env, "jep/AppletFrameContext");
	appletFramePanelClass =
		(*env)->FindClass(env, "jep/AppletFramePanel");
	if (osxVersion >= 0x1040) {
		myCAppletFrameClass =
			(*env)->FindClass(env, "apple/awt/MyCAppletFrame142Tiger");
		osxOffScreenSDClass =
			(*env)->FindClass(env, "apple/awt/OSXOffScreenSurfaceData");
		cPeerSDClass =
			(*env)->FindClass(env, "apple/awt/CPeerSurfaceData");
	}
	cPanelClass =
		(*env)->FindClass(env, "apple/awt/CPanel");
	gifDecoderClass =
		(*env)->FindClass(env, "sun/awt/image/GifImageDecoder");
	if ((*env)->ExceptionCheck(env))
		return;
	(*env)->RegisterNatives(env, appletFrameContextClass,
		AppletFrameContextMethods,
		sizeof(AppletFrameContextMethods)/sizeof(JNINativeMethod));
	(*env)->RegisterNatives(env, appletFramePanelClass,
		AppletFramePanelMethods,
		sizeof(AppletFramePanelMethods)/sizeof(JNINativeMethod));
	(*env)->RegisterNatives(env, appletHandlerClass,
		AppletHandlerMethods,
		sizeof(AppletHandlerMethods)/sizeof(JNINativeMethod));
	if (osxVersion >= 0x1040) {
		if (isJava160) {
			(*env)->RegisterNatives(env, myCAppletFrameClass,
				MyCAppletFrameMethods160,
				sizeof(MyCAppletFrameMethods160)/sizeof(JNINativeMethod));
		} else {
			(*env)->RegisterNatives(env, myCAppletFrameClass,
				MyCAppletFrameMethods,
				sizeof(MyCAppletFrameMethods)/sizeof(JNINativeMethod));
		}
		(*env)->RegisterNatives(env, osxOffScreenSDClass,
			OSXOffScreenSDMethods,
			sizeof(OSXOffScreenSDMethods)/sizeof(JNINativeMethod));
		(*env)->RegisterNatives(env, cPeerSDClass, CPeerSDMethods,
			sizeof(CPeerSDMethods)/sizeof(JNINativeMethod));
	}
	if (isJava142) {
		if ((osxVersion >= 0x1040) && !isJava150) {
			(*env)->RegisterNatives(env, cPanelClass,
				CPanelMethods142T,
				sizeof(CPanelMethods142T)/sizeof(JNINativeMethod));
		} else {
			(*env)->RegisterNatives(env, cPanelClass,
				CPanelMethods,
				sizeof(CPanelMethods)/sizeof(JNINativeMethod));
		}
	}
	(*env)->RegisterNatives(env, gifDecoderClass,
		GIFDecoderMethods,
		sizeof(GIFDecoderMethods)/sizeof(JNINativeMethod));
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	(*env)->DeleteLocalRef(env, appletFrameClass);
	(*env)->DeleteLocalRef(env, appletFrameContextClass);
	(*env)->DeleteLocalRef(env, appletFramePanelClass);
	(*env)->DeleteLocalRef(env, cPanelClass);
	(*env)->DeleteLocalRef(env, gifDecoderClass);
	if (osxVersion >= 0x1040) {
		(*env)->DeleteLocalRef(env, myCAppletFrameClass);
		(*env)->DeleteLocalRef(env, osxOffScreenSDClass);
		(*env)->DeleteLocalRef(env, cPeerSDClass);
	}
}

+ (void)setMainJavaThread:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jclass appletHandlerFactoryClass = nil;
	jmethodID setMainThread = nil;
	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return;
	}
	/* It's not possible to look up a class from a string except on the main
	 * thread -- possibly because the routines that do this aren't thread-
	 * safe.  (On other threads the result is always NULL.)
	 */
	cPerformerClass = NSClassFromString(@"CPerformer");
	cScrollPaneClass = NSClassFromString(@"CScrollPane");
	appletHandlerFactoryClass = (*env)->FindClass(env, "jep/AppletHandlerFactory");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return;
	}
	setMainThread = (*env)->GetStaticMethodID(env, appletHandlerFactoryClass,
		"setMainThread", "()V");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return;
	}
	(*env)->CallStaticVoidMethod(env, appletHandlerFactoryClass, setMainThread);
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	if ((*env)->ExceptionCheck(env))
		(*env)->ExceptionClear(env);
}

/* There's no API we can use to read Apple's Java Control Panel's "runtime
 * parameters" setting before the JVM starts up.  So we've got to read the
 * settings file directly.
 */
+ (NSMutableArray *)getRuntimeParams
{
	NSString *subpath;
	NSString *settingsFilePath;
	FILE *settingsFile = nil;
	char line[PATH_MAX+1] = {0};
	char params[PATH_MAX+1] = {0};
	char *key = nil;
	char *key14X = "javaplugin.jre.params=";
	char *key150 = "deployment.javapi.jre.1.5.0.args=";
	char *key160 = "deployment.javapi.jre.1.6.0.args=";
	int keyLength = 0;
	NSMutableArray *parmArray;
	char *escape, *parm;

	if (!homeDirectory)
		return nil;

	if (isJava160) {
		key = key160;
		subpath = @"%@/Library/Caches/Java/deployment.properties";
	} else if (isJava150) {
		key = key150;
		subpath = @"%@/Library/Caches/Java/deployment.properties";
	} else if (isJava142Update1) {
		key = key14X;
		subpath = @"%@/Library/Caches/Java Applets/deployment.properties";
	} else if (isJava142) {
		key = key14X;
		subpath = @"%@/.java/deployment/deployment.properties";
	} else {
		key = key14X;
		subpath = @"%@/.java/properties141_01";
	}
	keyLength = strlen(key);
	settingsFilePath = [NSString stringWithFormat:subpath, homeDirectory];
	if ((settingsFile = fopen([settingsFilePath UTF8String], "r")) == nil)
		return nil;
	while (fgets(line, sizeof(line)-1, settingsFile)) {
		if (strstr(line, key) != line)
			continue;
		strcpy(params, line + keyLength);
		params[strcspn(params, "\n\r")] = 0;
	}
	fclose(settingsFile);
	if (params[0] == 0)
		return nil;

	/* Get rid of any escape characters that may have been included in the
	 * runtime parameters string.
	 */
	escape = strstr(params, "\\");
	while (escape) {
		memmove(escape, escape+1, strlen(escape));
		escape = strstr(escape+1, "\\");
	}

	if ((parmArray = [[NSMutableArray alloc] init]) == nil)
		return nil;
	parm = strtok(params, " \t");
	while (parm) {
		[parmArray addObject:[NSString stringWithCString:(const char *)parm]];
		parm = strtok(nil, " \t");
	}

	return parmArray;
}

+ (NSMutableDictionary *)getRuntimeParamsAsDictionary
{
	NSMutableDictionary *retval = nil;
	NSMutableArray *runtimeParams = nil;
	NSString *parm, *key, *value;
	NSArray *elements;
	unsigned i, count;
	if ((runtimeParams = [AppletView getRuntimeParams]) == nil)
		return nil;
	if ((retval = [[NSMutableDictionary alloc] init]) == nil)
		return nil;
	count = [runtimeParams count];
	for (i = 1; i <= count; ++i) {
		parm = [runtimeParams objectAtIndex:i-1];
		if ((elements = [parm componentsSeparatedByString:@"="]) == nil)
			continue;
		if ([elements count] < 2)
			continue;
		key = (NSString *) [elements objectAtIndex:0];
		if ([key length] <= 2)
			continue;
		if (![[key substringToIndex:2] isEqualToString:@"-D"])
			continue;
		key = [[key substringFromIndex:2] lowercaseString];
		value = [parm substringFromIndex:[key length] + 3];
		[retval setObject:value forKey:key];
	}
	[runtimeParams removeAllObjects];
	[runtimeParams release];
	return retval;
}

+ (BOOL)getRuntimeParamBoolean:(NSString *)name
{
	BOOL retval = false;
	NSString *value;
	NSMutableDictionary *runtimeParams = nil;
	if (!name)
		return false;
	if ((runtimeParams = [AppletView getRuntimeParamsAsDictionary]) == nil)
		return false;
	value = (NSString *) [runtimeParams objectForKey:[name lowercaseString]];
	if (!value) {
		[runtimeParams removeAllObjects];
		return false;
	}
	value = [value lowercaseString];
	if ([value isEqualToString:@"true"] || [value isEqualToString:@"yes"] ||
		[value isEqualToString:@"on"])
		retval = true;
	[runtimeParams removeAllObjects];
	[runtimeParams release];
	return retval;
}

+ (NSString *)getRuntimeParamString:(NSString *)name
{
	NSString *value;
	NSMutableDictionary *runtimeParams = nil;
	if (!name)
		return nil;
	if ((runtimeParams = [AppletView getRuntimeParamsAsDictionary]) == nil)
		return nil;
	value = (NSString *) [runtimeParams objectForKey:[name lowercaseString]];
	if (value)
		[[value retain] autorelease];
	[runtimeParams removeAllObjects];
	[runtimeParams release];
	return value;
}

/* General purpose method (not just for an AppletView object) that will fail
 * only if the JVM has not yet been created (by us or by someone else).
 */
+ (BOOL)getJavaEnv:(JNIEnv **)envPtr
{
	int i;
	OSStatus retval;
	JavaVM *vm = nil;
	JNIEnv *env = nil;
	JavaVM *vmBuf[50] = {0};
	jsize numVMs = 0;

	if (!envPtr)
		return false;
	*envPtr = nil;

	if (everCreatedJVM &&
		!JNI_GetCreatedJavaVMs(vmBuf, sizeof(vmBuf)/sizeof(JavaVM *), &numVMs))
	{
		for (i = 1; i <= numVMs; ++i) {
			if ((retval = (*vmBuf[i-1])->GetEnv(vmBuf[i-1], (void **) &env,
				JNI_VERSION_1_4)) == JNI_OK)
			{
				vm = vmBuf[i-1];
				break;
			} else if (retval == JNI_EDETACHED) {
				if (!(*vmBuf[i-1])->AttachCurrentThread(vmBuf[i-1],
					(void **) &env, nil))
				{
					vm = vmBuf[i-1];
					break;
				}
			}
		}
	}
	if (!vm || !env) {
		return false;
	}

	*envPtr = env;
	return true;
}

/* Use this to get the Java environment in a thread from which we're loading
 * an applet or some important part of the JVM (like the AWT libraries).
 * To make sure the thread only exists as long as we need it, we must call
 * DetachCurrentThread() just before we exit (for which we also need a
 * JavaVM *).  (If we don't call DetachCurrentThread(), the JavaVM won't be
 * able to release its resources, and it will become a zombie thread.)
 */
+ (BOOL)getJavaEnv:(JNIEnv **)envPtr vmPtr:(JavaVM **)vmPtr
{
	int i;
	OSStatus retval;
	JavaVM *vm = nil;
	JNIEnv *env = nil;
	JavaVM *vmBuf[50] = {0};
	jsize numVMs = 0;

	if (!envPtr || !vmPtr)
		return false;
	*envPtr = nil;
	*vmPtr = nil;

	if (everCreatedJVM &&
		!JNI_GetCreatedJavaVMs(vmBuf, sizeof(vmBuf)/sizeof(JavaVM *), &numVMs))
	{
		for (i = 1; i <= numVMs; ++i) {
			if ((retval = (*vmBuf[i-1])->GetEnv(vmBuf[i-1], (void **) &env,
				JNI_VERSION_1_4)) == JNI_OK)
			{
				vm = vmBuf[i-1];
				break;
			} else if (retval == JNI_EDETACHED) {
				if (!(*vmBuf[i-1])->AttachCurrentThread(vmBuf[i-1],
					(void **) &env, nil))
				{
					vm = vmBuf[i-1];
					break;
				}
			}
		}
	}
	if (!vm || !env) {
		return false;
	}

	*envPtr = env;
	*vmPtr = vm;
	return true;
}

+ (BOOL)getJavaEnvIfAttached:(JNIEnv **)envPtr
{
	int i;
	OSStatus retval;
	JavaVM *vm = nil;
	JNIEnv *env = nil;
	JavaVM *vmBuf[50] = {0};
	jsize numVMs = 0;

	if (!envPtr)
		return false;
	*envPtr = nil;

	if (everCreatedJVM &&
		!JNI_GetCreatedJavaVMs(vmBuf, sizeof(vmBuf)/sizeof(JavaVM *), &numVMs))
	{
		for (i = 1; i <= numVMs; ++i) {
			if ((retval = (*vmBuf[i-1])->GetEnv(vmBuf[i-1], (void **) &env,
				JNI_VERSION_1_4)) == JNI_OK)
			{
				vm = vmBuf[i-1];
				break;
			}
		}
	}
	if (!vm || !env) {
		return false;
	}

	*envPtr = env;
	return true;
}

+ (NSString *)toString:(jobject)javaThing javaEnv:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jclass objectClass;
	jmethodID toString;
	jobject description;
	const jchar *descriptionUTF16;
	jint descriptionLength;
	NSString *descriptionCocoa = nil;

	if (!javaThing)
		return nil;

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return nil;
	}

	objectClass = (*env)->FindClass(env, "java/lang/Object");
	toString = (*env)->GetMethodID(env, objectClass, "toString",
		"()Ljava/lang/String;");
	description = (*env)->CallObjectMethod(env, javaThing, toString);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return nil;
	}
	descriptionLength = (*env)->GetStringLength(env, description);
	descriptionUTF16 = (*env)->GetStringChars(env, description, nil);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return nil;
	}
	if (descriptionUTF16) {
		descriptionCocoa = [NSString stringWithCharacters:descriptionUTF16
			length:descriptionLength];
		(*env)->ReleaseStringChars(env, description, descriptionUTF16);
	}
	(*env)->DeleteLocalRef(env, objectClass);
	if (description)
		(*env)->DeleteLocalRef(env, description);
	if ((*env)->ExceptionCheck(env))
		(*env)->ExceptionClear(env);

	return descriptionCocoa;
}

+ (NSString *)getJavaThread:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jclass myCToolkitClass;
	jmethodID getJavaThread;
	jobject description;
	const jchar *descriptionUTF16;
	jint descriptionLength;
	NSString *descriptionCocoa = nil;

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return nil;
	}

	myCToolkitClass = (*env)->FindClass(env, "apple/awt/MyCToolkit");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return nil;
	}
	getJavaThread = (*env)->GetStaticMethodID(env, myCToolkitClass,
		"getJavaThread", "()Ljava/lang/String;");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return nil;
	}
	description = (*env)->CallStaticObjectMethod(env, myCToolkitClass,
		getJavaThread);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return nil;
	}
	descriptionLength = (*env)->GetStringLength(env, description);
	descriptionUTF16 = (*env)->GetStringChars(env, description, nil);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return nil;
	}
	if (descriptionUTF16) {
		descriptionCocoa = [NSString stringWithCharacters:descriptionUTF16
			length:descriptionLength];
		(*env)->ReleaseStringChars(env, description, descriptionUTF16);
	}
	(*env)->DeleteLocalRef(env, myCToolkitClass);
	if (description)
		(*env)->DeleteLocalRef(env, description);
	if ((*env)->ExceptionCheck(env))
		(*env)->ExceptionClear(env);

	return descriptionCocoa;
}

+ (OSStatus)printMessage:(NSString *)message javaEnv:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jclass systemClass, printStreamClass;
	jfieldID sysoutID;
	jobject sysout;
	jmethodID printString;
	jstring messageJava;
	unichar *messageUTF16 = nil;
	unsigned messageLength;
	OSStatus retval = noErr;

	if (!message)
		return EINVAL;

	messageLength = [message length];
	messageUTF16 = calloc(messageLength + 1, sizeof(unichar));
	if (!messageUTF16)
		return ENOMEM;
	[message getCharacters:messageUTF16];

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return ENOENT;
	}

	systemClass = (*env)->FindClass(env, "java/lang/System");
	printStreamClass = (*env)->FindClass(env, "java/io/PrintStream");
    sysoutID = (*env)->GetStaticFieldID(env, systemClass, "out",
		"Ljava/io/PrintStream;");
	sysout = (*env)->GetStaticObjectField(env, systemClass, sysoutID);
	printString = (*env)->GetMethodID(env, printStreamClass, "println",
		"(Ljava/lang/String;)V");
	messageJava = (*env)->NewString(env, messageUTF16, messageLength);
	free(messageUTF16);
	(*env)->CallVoidMethod(env, sysout, printString, messageJava);

	(*env)->DeleteLocalRef(env, systemClass);
	(*env)->DeleteLocalRef(env, printStreamClass);
	(*env)->DeleteLocalRef(env, sysout);
	(*env)->DeleteLocalRef(env, messageJava);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		retval = ENOENT;
	}

	return retval;
}

+ (OSStatus)printStackTrace:(JNIEnv *)javaEnv
{
	JNIEnv *env = nil;
	jclass myCToolkitClass = nil;
	jmethodID printStackTrace = nil;
	OSStatus retval = noErr;
	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return ENOENT;
	}
	myCToolkitClass = (*env)->FindClass(env, "apple/awt/MyCToolkit");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return ENOENT;
	}
	printStackTrace = (*env)->GetStaticMethodID(env, myCToolkitClass,
		"printStackTrace", "()V");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return ENOENT;
	}
	(*env)->CallStaticVoidMethod(env, myCToolkitClass, printStackTrace);
	(*env)->DeleteLocalRef(env, myCToolkitClass);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		retval = ENOENT;
	}
	return retval;
}

/* Used to initialize and wake up event loops that we run ourselves, so that
 * we can process (Cocoa) events on the main thread (and thereby avoid
 * blocking it) while we're waiting for something to finish on a secondary
 * thread.  If we don't explicitly wake up the loop (from the secondary
 * thread) when the task we're waiting for is finished, we might wait too
 * long (or even forever).  I'm not sure if any problems result from our
 * not initializing our event loops.  But it seems like a good idea (in
 * principle) to reset the "current event" (to our fake event) as soon as
 * possible, so that NSApplication getCurrentEvent doesn't return stale
 * information.  Using postFakeEvent to initialize our event loops is new as
 * of JEP 0.9.4.
 */
+ (void)postFakeEvent
{
	NSEvent *fakeEvent = nil;
	[NSRunLoop currentRunLoop];
	/* Before JEP 0.9.4 we used a fake mouse-moved event here.  But it's safer
	 * (less likely to confuse AWT or the browser) to use an entirely custom
	 * event.
	 */
	fakeEvent = [NSEvent otherEventWithType:NSApplicationDefined
		location:NSMakePoint(0, 0)
		modifierFlags:0
		timestamp:0
		windowNumber:0
		context:nil
		subtype:0
		data1:0
		data2:0];
	if (fakeEvent)
		[NSApp postEvent:fakeEvent atStart:false];
}

/* See the comment on the CustomEvent class (above) for more information on
 * why this custom event spool and its associated methods are necessary.
 */

NSLock *customEventSpoolLock = nil;
NSMutableArray *customEventSpool = nil;

/* Call this on the main thread. */
+ (void)processCustomEventSpool:(BOOL)justOneEvent
{
	int i, count;
	BOOL moreEvents = false;
	NSArray *modes;
	CustomEvent *event;
	NSMutableArray *willDispatch;
	NSThread *currentThread = [NSThread currentThread];
	if (!mainThread || ![mainThread isEqual:currentThread])
		return;
	if (!customEventSpoolLock || !customEventSpool)
		return;
	for (;;) {
		[customEventSpoolLock lock];
		if ((count = [customEventSpool count]) == 0) {
			[customEventSpoolLock unlock];
			return;
		}
		if ((willDispatch = [[NSMutableArray alloc] init]) == nil) {
			[customEventSpoolLock unlock];
			return;
		}
		if (justOneEvent) {
			if (count > 1)
				moreEvents = true;
			count = 1;
		}
		for (i = 1; i <= count; ++i) {
			event = [customEventSpool objectAtIndex:i-1];
			[willDispatch addObject:event];
		}
		[customEventSpool removeObjectsInArray:willDispatch];
		[customEventSpoolLock unlock];
		count = [willDispatch count];
		for (i = 1; i <= count; ++i) {
			event = [willDispatch objectAtIndex:i-1];
			/* Don't actually dispatch any events while the browser is
			 * terminating -- otherwise we may crash on OS X Tiger and
			 * Leopard.
			 */
			if (!nsAppTerminating) {
				modes = [event getModes];
				if (modes) {
					[event performSelectorOnMainThread:@selector(dispatch)
						withObject:nil waitUntilDone:true modes:modes];
				} else {
					[event performSelectorOnMainThread:@selector(dispatch)
						withObject:nil waitUntilDone:true];
				}
			}
			// CustomEvents should be released as soon as possible (i.e. not
			// autoreleased) -- otherwise they tend to get released as a thread
			// is shutting down, after the JVM has been detached (which causes
			// crashes, at least on OS X 10.4).
			if (![event getWaitUntilDone])
				[event release];
		}
		[willDispatch removeAllObjects];
		[willDispatch release];
		if (justOneEvent) {
			//if (moreEvents)
			//	CFRunLoopSourceSignal(jepRunLoopSource);
 			break;
		}
	}
}

+ (void)performOnMainThread:(SEL)aSelector onTarget:(id)target
	withObject:(id)anObject waitUntilDone:(BOOL)wait
{
	[AppletView performOnMainThread:aSelector onTarget:target
		withObject:anObject waitUntilDone:wait inModes:nil];
}

+ (void)performOnMainThread:(SEL)aSelector onTarget:(id)target
	withObject:(id)anObject waitUntilDone:(BOOL)wait inModes:(NSArray *)modes
{
	CustomEvent *event;
	NSThread *currentThread = [NSThread currentThread];
	BOOL isMainThread = !mainThread || [mainThread isEqual:currentThread];
	if (!aSelector || !target || nsAppTerminating)
		return;
	[NSRunLoop currentRunLoop];
	event = [[CustomEvent alloc] initWithParams:target aSelector:aSelector
		anObject:anObject waitUntilDone:wait modes:modes];
	if (!event)
		return;
	if (isMainThread) {
		[event performSelectorOnMainThread:@selector(dispatch)
			withObject:nil waitUntilDone:true modes:modes];
		return;
	}
	if (!customEventSpoolLock || !customEventSpool || nsAppTerminating)
		return;
	[customEventSpoolLock lock];
	[customEventSpool addObject:event];
	[customEventSpoolLock unlock];
	CFRunLoopSourceSignal(jepRunLoopSource);
	CFRunLoopWakeUp(mainRunLoop);
	if (!wait)
		return;
	/* If we don't break out of our waiting loop when the browser is
	 * terminating, we can sometimes hang on exit in OS X Tiger and Leopard.
	 * Apparently we need to use usleep() here instead of NSThread
	 * sleepUntilDate -- otherwise we can (rarely, in OS X Tiger and Leopard)
	 * crash in _class_lookupMethodAndLoadCache on two threads at once.  This
	 * is more than a little bizarre, but it may be that sleepUntilDate does
	 * some event-loop and/or timer spinning that usleep doesn't do.
	 */
	while (![event getIsDone] && !nsAppTerminating)
		usleep(10000);
	/* CustomEvents should be released as soon as possible (i.e. not
	 * autoreleased) -- otherwise they tend to get released as a thread is
	 * shutting down, after the JVM has been detached (which causes crashes,
	 * at least on OS X 10.4).  But sometimes we crash if we release events
	 * here while lots of threads are waiting for CustomEvents to be handled
	 * on the main thread, so now (as of JEP 0.9.6.3) we also pass [event
	 * release] off to the main thread (without waiting for it).  (Here's one
	 * way to make these crashes happen on JEP versions prior to 0.9.6.3 (I've
	 * only got it to work on OS X 10.4.X on an Intel Mac):  Load a bunch of
	 * applets in different tabs.  Then load http://securitywizardry.com/radar.htm
	 * in another tab, and move the mouse rapidly over the news-scrolling
	 * applets (on the left) as they're loading.)
	 */
	[event performSelectorOnMainThread:@selector(release)
		withObject:nil waitUntilDone:false];
}

NSLock *cEventSpoolLock = nil;
NSMutableArray *cEventSpool = nil;

+ (void)processCEventSpool:(BOOL)justOneEvent
{
	int i, count;
	BOOL moreEvents = false;
	CEvent *event;
	NSMutableArray *willDispatch;
	NSThread *currentThread = [NSThread currentThread];
	if (!mainThread || ![mainThread isEqual:currentThread])
		return;
	if (!cEventSpoolLock || !cEventSpool)
		return;
	for (;;) {
		[cEventSpoolLock lock];
		if ((count = [cEventSpool count]) == 0) {
			[cEventSpoolLock unlock];
			return;
		}
		if ((willDispatch = [[NSMutableArray alloc] init]) == nil) {
			[cEventSpoolLock unlock];
			return;
		}
		if (justOneEvent) {
			if (count > 1)
				moreEvents = true;
			count = 1;
		}
		for (i = 1; i <= count; ++i) {
			event = [cEventSpool objectAtIndex:i-1];
			[willDispatch addObject:event];
		}
		[cEventSpool removeObjectsInArray:willDispatch];
		[cEventSpoolLock unlock];
		count = [willDispatch count];
		for (i = 1; i <= count; ++i) {
			event = [willDispatch objectAtIndex:i-1];
			/* Don't actually dispatch any events while the browser is
			 * terminating -- otherwise we may crash on OS X Tiger and
			 * Leopard.
			 */
			if (!nsAppTerminating)
				[event performSelector:@selector(dispatch)];
			if (![event getWaitUntilDone])
				[event release];
		}
		[willDispatch removeAllObjects];
		[willDispatch release];
		if (justOneEvent) {
			//if (moreEvents)
			//	CFRunLoopSourceSignal(jepCRunLoopSource);
			break;
		}
	}
}

+ (void)performCEventOnMainThread:(void (*) (void *))runnable
	userInfo:(void *)userInfo waitUntilDone:(BOOL)wait
{
	CEvent *event;
	NSThread *currentThread = [NSThread currentThread];
	BOOL isMainThread = !mainThread || [mainThread isEqual:currentThread];
	if (!runnable || nsAppTerminating)
		return;
	event = [[CEvent alloc] initWithParams:runnable userInfo:userInfo
		waitUntilDone:wait];
	if (!event)
		return;
	if (isMainThread) {
		[event performSelector:@selector(dispatch)];
		return;
	}
	if (!cEventSpoolLock || !cEventSpool || nsAppTerminating)
		return;
	[cEventSpoolLock lock];
	[cEventSpool addObject:event];
	[cEventSpoolLock unlock];
	CFRunLoopSourceSignal(jepCRunLoopSource);
	CFRunLoopWakeUp(mainRunLoop);
	if (!wait)
		return;
	/* If we don't break out of our waiting loop when the browser is
	 * terminating, we can sometimes hang on exit in OS X Tiger and Leopard.
	 * Apparently we need to use usleep() here instead of NSThread
	 * sleepUntilDate -- otherwise we can (rarely, in OS X Tiger and Leopard)
	 * crash in _class_lookupMethodAndLoadCache on two threads at once.  This
	 * is more than a little bizarre, but it may be that sleepUntilDate does
	 * some event-loop and/or timer spinning that usleep doesn't do.
	 */
	while (![event getIsDone] && !nsAppTerminating)
		usleep(10000);
	[event performSelectorOnMainThread:@selector(release)
		withObject:nil waitUntilDone:false];
}

/* The mozilla.org browsers' frame code is very complex and fragile, and in
 * particular doesn't like to be reentered.  At various points the JEP needs
 * to process events (or cause them to be processed) on the main thread, while
 * it's waiting for some operation to finish on one or more secondary threads.
 * At these points the JEP has always used NSApplication nextEventMatchingMask
 * with "mask" set to NSAnyEventMask.  It also (before JEP 0.9.6) used to
 * always pass these events along to NSApp by using its sendEvent method.  But
 * using NSApplication sendEvent can cause the mozilla.org frame code to be
 * reentered, leading to crashes (and presumably to other nasty things).  See
 *   https://bugzilla.mozilla.org/show_bug.cgi?id=312062
 *   https://bugzilla.mozilla.org/show_bug.cgi?id=315111#c13
 * By lucky chance, I've discovered way around this problem:  Make sure the
 * NSApplication event queue has at least one event in it, then keep calling
 * nextEventMatchingMask (at short intervals) with "mask" set to NSAnyEventMask
 * and "dequeue" set to false.  This somehow allows OS code to continue
 * processing events that the JEP needs to have processed, without allowing
 * the mozilla.org browsers' own event queue(s) to be processed.  This doesn't
 * always work on OS X 10.2.8 -- for example when the JVM might put up a modal
 * dialog while the JEP is processing events.  So I've had to keep using
 * sendEvent in initWithParms on OS X 10.2.8 (and I've confirmed that this can
 * cause the bugzilla bug 312062 test case to crash).  But OS X 10.2.8 is old
 * and (mostly) unsupported by Apple, so I'm not too concerned.  (The
 * mozilla.org event queue that gets processed when NSApplication sendEvent
 * is being used is their CFRunLoop "source", _md_EventReceiverProc(), in
 * xpcom/threads/plevent.c.)
 * The following two nofications are something that mozilla.org's Simon Fraser
 * requested before I came up with the solution I've just described.  The idea
 * is for Camino to be able to disable frame code while the JEP is processing
 * events.  This is being worked on at:
 *   https://bugzilla.mozilla.org/show_bug.cgi?id=324281
 * My solution has made this largely unnecessary.  But I _am_ sending these
 * noficiations in initWithParams on OS X 10.2.8, where I still need to use
 * sendEvent.
 */
+ (void)postJEPSendingEventsStartedNotification
{
	//NSLog(@"Posting JEP_EventLoopStart notification");
	[[NSNotificationCenter defaultCenter]
		postNotificationName:@"JEP_EventLoopStart" object:nil];
}

+ (void)postJEPSendingEventsEndedNotification
{
	//NSLog(@"Posting JEP_EventLoopStart notification");
	[[NSNotificationCenter defaultCenter]
		postNotificationName:@"JEP_EventLoopEnd" object:nil];
}

/* We need the following to make keyboard and mouse events work with applets
 * in tabs in Cocoa browsers.  This (together with the code in Handlers.m and
 * AppletFrame.java that stops the drawing of "undisplayable" applets views)
 * is basically the poor man's equivalent of the "hide" message, which is only
 * available in OS X 10.3.X.  Without it keyboard and mouse events keep being
 * sent to invisible applet views, and a visible applet view can be completely
 * hidden (stopped from receiving keyboard and mouse events) by an invisible
 * one.
 */
/* We also use hitTest do deal with a nasty problem in Camino:  When tracking
 * mouse events it fails to pay any attention to how an applet has been
 * clipped, so mouse events aren't sent to controls that overlap an applet's
 * unclipped boundaries.
 */
- (NSView *)hitTest:(NSPoint)aPoint
{
	NSPoint location;
	if (!isCarbonApp) {
		if (![self isDisplayable])
			return nil;
		location = [self convertPoint:aPoint fromView:[self superview]];
		if (![self mouse:location inRect:currentBounds])
			return nil;
	}
	return [super hitTest:aPoint];
}

/* We should normally return TRUE here (since we're in effect drawing into
 * another app's window).  And that's how we always behave in a Carbon app.
 * But it turns out that in a Cocoa app we need to return FALSE when our
 * own view is invisible or non-"dirty" -- otherwise the OS tries to redraw
 * parts of our view that merely overlay other "dirty" structures, which can
 * result in strange artifacts (either stray background-color rectangles if
 * we actually draw in these rectangles, or dark blobs if we fail to draw in
 * them).  This change (new as of JEP 0.9.4) is a fix for
 * https://bugzilla.mozilla.org/show_bug.cgi?id=302712.  These problems are
 * (ultimately) due to the fact that our view is positioned above all of the
 * browser's own views (as it needs to if we wish to avoid dependencies on
 * the behavior of particular browsers).  This situation also forces us to
 * play tricks in hitTest (above).
 */
- (BOOL)isOpaque
{
	NSRect dirty;
	if (isCarbonApp) {
		return TRUE;
	/* If we return FALSE merely when we're not visible, we'll crash in
	 * Java 1.4.1.  And though we need this for Panther and Tiger, we don't
	 * seem to need it for Jaguar.
	 */
	} else if (isJava142Update1 &&
		(!myController || myController->updatesOff || ![self isDisplayable]))
	{
		return FALSE;
	}
	dirty = [self _dirtyRect];
	if ((dirty.size.width <= 0) || (dirty.size.height <= 0))
		return FALSE;
	return TRUE;
}

/* Needed to shift the coordinate system's origin from the bottom left
 * (Cocoa's default) to the top left (Java's default).
 */
- (BOOL)isFlipped
{
	return TRUE;
}

/* Guarantees that our AppletView object really is "opaque" (as we claim
 * when we return TRUE from isOpaque above).  If our object is visible, we
 * erase it (fill it with the background color) after clipping.  What we do
 * here is equivalent to (and more efficient than) what we used to do (prior
 * to version 0.8.2) in NSCarbonWindowContentView's drawRect method (see
 * Handlers.m).
 */
- (void)drawRect:(NSRect)aRect
{
	if (!myController || myController->updatesOff || ![self isDisplayable]) {
		if (myController)
			myController->drawAppletPending = false;
		return;
	}
	if (debugUpdates) {
		NSString *info = [NSString stringWithFormat:
			@"%@ AppletView drawRect: Redrawing %@",
			[NSDate date], myAppletName];
		[AppletView printMessage:info javaEnv:nil];
	}
	[NSBezierPath clipRect:currentBounds];
	NSEraseRect(aRect);
	myController->drawAppletPending = false;
}

/* Thanks to the "fix" for Bugzilla bug 298961, a plugin is almost never made
 * invisible in ordinary usage (scrolling a page, switching tabs).  Instead,
 * an applet is zero-clipped when it needs to (temporarily) stop being
 * displayed.  For some reason, not making an applet view invisible on tab
 * switches does't seem to cause trouble in Carbon browsers, or even in Cocoa
 * browsers on all platforms.  But some updates from applets in other tabs
 * _do_ bleed through in Cocoa browsers on Mac OS X 10.4.2 (notably for the
 * NOAA weather applets' "ticker"), so we need to do something about it.
 * After many false starts and much time wasted, I discovered the following
 * solution.  (The MRJ Plugin has long relied on the browser calling
 * nsIPluginInstance::SetWindow() with pluginWindow == NULL as a signal to
 * make the applet view invisible.  I'm told that this should never have
 * happened -- instead the clip being zero should have been the signal that
 * an applet view needs to be made invisible.  But if we do this, we'll end up
 * changing an applet view's visibility several times a second every time the
 * browser page is scrolled, and the cost of doing this is simply too high.)
 * As of JEP 0.9.5+h, this has been moved from Controller.m (it makes much
 * better sense here).  And it appears this method doesn't get called on
 * Leopard (Mac OS X 10.5), so we need to (also) do other things to deal with
 * applets that have been zero-clipped but not made invisible -- see MyNSView
 * setNeedsDisplay, and the code for the views that are "ours" in MyNSView
 * setNeedsDisplayInRect and _drawRect (all in Controller.m).
 */
- (void)_propagateDirtyRectsToOpaqueAncestors
{
	struct objc_super target;
	if (!isCarbonApp) {
		/* As of JEP 0.9.5+h we no longer squelch an update if updatesOff is
		 * set -- that's not necessary, and is possibly too aggressive.
		 */
		if (![self isDisplayable]) {
			target.receiver = self;
			target.class = [MyNSView superclass];
			objc_msgSendSuper(&target, @selector(setNeedsDisplay:), false);
			if (debugUpdates) {
				NSString *info = [NSString stringWithFormat:
					@"%@ AppletView _propagateDirtyRectsToOpaqueAncestors: Stopping update to non-displayable applet %@",
					[NSDate date], myAppletName];
				[AppletView printMessage:info javaEnv:nil];
			}
			return;
		}
	}
	target.receiver = self;
	target.class = [AppletView superclass];
	objc_msgSendSuper(&target, _cmd);
}

/* We need to do extra work every time one of NSView's "setFrame" methods is
 * called:  We need to (perhaps temporarily) unset the clip.  And we need to
 * set the "bounds" back to what it was previously -- NSView's "setFrame"
 * methods change the "bounds" to match the "frame", but we resize an applet
 * view by setting its "frame" differently than its "bounds" (see sizeApplet
 * in Controller.m).  (As of JEP 0.8.3 we now also support resizing applets
 * directly or "natively", when that's appropriate.  See resizeApplet() in
 * AppletFrame.java for more info.)  (As of JEP 0.9.5+h, if we're currently
 * zero-clipped we make sure to stay that way.  Otherwise applets that have
 * been "made invisible" by zero-clipping can get displayed when they
 * shouldn't be.)
 */

- (void)setFrame:(NSRect)frameRect
{
	BOOL wasZeroClipped = [self isCurrentClipEmpty];
	if (myController) {
		[self unsetClip];
		myController->currentFrame = frameRect;
	}
	[super setFrame:frameRect];
	if (myController) {
		if (myController->resizeAppletDirectly) {
			// resizeApplet changes currentBounds
			[self resizeApplet:frameRect.size.width
				newHeight:frameRect.size.height];
		}
		[self setBoundsSize:currentBounds.size];
		if (wasZeroClipped)
			[self setClip:NSZeroRect];
	}
}

- (void)setFrameSize:(NSSize)newSize
{
	BOOL wasZeroClipped = [self isCurrentClipEmpty];
	if (myController) {
		[self unsetClip];
		myController->currentFrame.size = newSize;
	}
	[super setFrameSize:newSize];
	if (myController) {
		if (myController->resizeAppletDirectly) {
			// resizeApplet changes currentBounds
			[self resizeApplet:newSize.width newHeight:newSize.height];
		}
		[self setBoundsSize:currentBounds.size];
		if (wasZeroClipped)
			[self setClip:NSZeroRect];
	}
}

- (void)setFrameOrigin:(NSPoint)newOrigin
{
	BOOL wasZeroClipped = [self isCurrentClipEmpty];
	if (myController) {
		[self unsetClip];
		myController->currentFrame.origin = newOrigin;
	}
	[super setFrameOrigin:newOrigin];
	if (myController) {
		[self setBoundsSize:currentBounds.size];
		if (wasZeroClipped)
			[self setClip:NSZeroRect];
	}
}

- (void)setFrameRotation:(float)angle
{
	BOOL wasZeroClipped = [self isCurrentClipEmpty];
	if (myController)
		[self unsetClip];
	[super setFrameRotation:angle];
	if (myController) {
		[self setBoundsSize:currentBounds.size];
		if (wasZeroClipped)
			[self setClip:NSZeroRect];
	}
}

- (void)dealloc
{
	NSString *info;

	if (debugRelease) {
		info = [NSString stringWithFormat:@"%@ AppletView:dealloc for %@",
			[NSDate date], myAppletName];
		[AppletView printMessage:info javaEnv:nil];
	}

	if (!myLock)
		return;

	[myLock lock];

	if (myApplet)
		[self stopApplet];
	[self destroyApplet];
	myApplet = nil;

	[self deleteJNIGlobalRefs:nil];

	if (myAppletName) {
		[myAppletName release];
		myAppletName = nil;
	}

	myController = nil;
	javaVM = nil;

	[myLock unlock];
	[myLock release];
	myLock = nil;

	[super dealloc];
}

- (BOOL)destroying
{
	if (isDestroying)
		return true;
	if (![JEPController isControllerValid:myController])
		return true;
	return false;
}

- (void)deleteJNIGlobalRefs:(JNIEnv *)javaEnv
{
	JNIEnv *env;
	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![self getJavaEnv:&env])
			return;
	}
	if (!myLock)
		return;
	[myLock lock];
	if ([self getJavaEnv:&env]) {
		if (myApplet) {
			(*env)->DeleteGlobalRef(env, myApplet);
			[self checkJNIException:env];
		}
		if (myAppletHandler) {
			(*env)->DeleteGlobalRef(env, myAppletHandler);
			[self checkJNIException:env];
		}
		if (myJSObject) {
			(*env)->DeleteGlobalRef(env, myJSObject);
			[self checkJNIException:env];
		}
	}
	myApplet = nil;
	myAppletHandler = nil;
	myJSObject = nil;
	[myLock unlock];
}

/* BMO bug 312072 is caused by bugs and/or design flaws in Mozilla.org
 * browsers.  But it can be triggered by the JEP on Camino (though for some
 * reason not on Firefox) when the JEP processes events on the main thread
 * while waiting for an applet to load.  There's a workaround for this, which
 * the JEP has used since JEP 0.9.5+c -- call NSApp nextEventMatchingMask
 * without dequeueing any events.  This usually processes low level events
 * (e.g. keyboard and mouse events, from Version 1 run loop sources) without
 * processing any browser events (i.e. without processing any Version 0 run
 * loop sources).  But on OS X 10.3.X and 10.2.8, this doesn't work inside a
 * Java modal dialog -- keyboard events and most mouse events don't get
 * processed, and the result is a hang.  Previous JEP versions (prior to JEP
 * 0.9.6.1) disabled the workaround on OS X 10.2.8, which was most badly
 * effected by these hangs.  But the hangs can also happen on OS X 10.3.X.
 * So now I (temporarily) disable the workaround on both 10.2.8 and 10.3.X
 * while running in a Java modal dialog.  This gets rid of the hangs, and
 * also still manages to work around bug 312072 -- now on both 10.2.8 and
 * 10.3.X.  (Without this workaround, doing the following causes a hang on
 * OS X 10.3.9 in JEP 0.9.6.1, in either Camino or Firefox:  Visit
 * http://www.sferyx.com/htmleditorapplet/demo/htmleditordemo.php.  A "Demo
 * Version" alert box will come up, which lets you choose between "Buy it!"
 * and "Continue evaluation".  It won't be possible to dismiss this dialog,
 * using either the mouse or the keyboard.  The coding of this alert box is
 * very peculiar -- it gets displayed in a modal dialog from the applet's
 * constructor!)
 */
- (BOOL)shouldProcessAllEvents:(JNIEnv *)javaEnv
{
	BOOL retval = false;
	JNIEnv *env = nil;
	jclass focusManagerClass;
	jclass dialogClass;
	jmethodID getFocusManager;
	jobject focusManager = nil;
	jmethodID getGlobalActiveWindow;
	jmethodID isModal;
	jobject globalActiveWindow = nil;
	NSWindow *keyWindow = nil;

	if (javaEnv) {
		env = javaEnv;
	} else {
		if (![AppletView getJavaEnv:&env])
			return false;
	}

	focusManagerClass = (*env)->FindClass(env, "java/awt/KeyboardFocusManager");
	dialogClass = (*env)->FindClass(env, "java/awt/Dialog");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return false;
	}
	getFocusManager = (*env)->GetStaticMethodID(env, focusManagerClass,
		"getCurrentKeyboardFocusManager", "()Ljava/awt/KeyboardFocusManager;");
	getGlobalActiveWindow = (*env)->GetMethodID(env, focusManagerClass,
		"getGlobalActiveWindow", "()Ljava/awt/Window;");
	isModal = (*env)->GetMethodID(env, dialogClass, "isModal", "()Z");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return false;
	}
	focusManager = (*env)->CallStaticObjectMethod(env, focusManagerClass,
		getFocusManager);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return false;
	}

	globalActiveWindow = (*env)->CallObjectMethod(env, focusManager,
		getGlobalActiveWindow);
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		return false;
	}

	/* Sometimes, though a Java modal dialog is running, it isn't able to get
	 * the KeyboardFocusManager focus (it seems this can happen when you load
	 * the Sferyx applet is a second tab while a bunch of applets in the first
	 * tab are still loading (http://securitywizardry.com/radar.htm in the
	 * first tab will do this)).  So as a fallback we also return "true" if
	 * the (Cocoa) key window is a child window created by the JVM.
	 */
	if (globalActiveWindow) {
		if ((*env)->IsInstanceOf(env, globalActiveWindow, dialogClass)) {
			retval = (*env)->CallBooleanMethod(env, globalActiveWindow,
				isModal);
		}
	} else if ((keyWindow = [NSApp keyWindow]) != nil) {
		retval = [keyWindow isKindOfClass:NSClassFromString(@"CocoaAppWindow")];
	}

	(*env)->DeleteLocalRef(env, focusManagerClass);
	(*env)->DeleteLocalRef(env, dialogClass);
	(*env)->DeleteLocalRef(env, focusManager);
	if (globalActiveWindow)
		(*env)->DeleteLocalRef(env, globalActiveWindow);
	return retval;
}

/* Only returns NULL if appletCreatedCallback() will _not_ be called. */
- (AppletView *)initWithParams:(JEPController *)controller
	docbase:(NSURL *)docbase attributes:(NSDictionary *)attributes
{
	NSEvent *anEvent;
	NSDate *start;
	NSTimeInterval total;
	NSString *docbaseHolder;
	NSThread *currentThread = [NSThread currentThread];

	if (!mainThread)	// If mainThread is nil, we haven't yet called
		return nil;		// JEPController checkBundle.
	doingInitOnMainThread = [mainThread isEqual:currentThread];

	if ((self = [super init]) == nil)
		return nil;
	if ((myLock = [[NSRecursiveLock alloc] init]) == nil) {
		[self autorelease];
		return nil;
	}
	[myLock lock];
	if (!docbase || !controller || !controller->javaVM || !controller->myWindow) {
		[myLock unlock];
		[self autorelease];
		return nil;
	}
	if ((docbaseHolder = [docbase absoluteString]) == nil) {
		[myLock unlock];
		[self autorelease];
		return nil;
	}
	appletDocbase = [docbaseHolder copy];
	if (attributes) {
		appletAttrs = [attributes copy];
	} else {
		appletAttrs = nil;
	}

	javaVM = controller->javaVM;
	myController = nil;
	myAppletHandler = nil;
	myApplet = nil;
	myAppletName = nil;
	myJSObject = nil;
	appletState = 0;
	isDestroying = false;
	awaitingAppletWindowCleanup = false;
	visibilitySetExplicitly = false;
	returnedCreatedInfo = false;
	returnedInitializedInfo = false;
	returnedStartedInfo = false;
	isInvisible = controller->createdInvisible;

	initialBounds = controller->initialFrame;
	initialBounds.origin.x = initialBounds.origin.y = 0;
	currentBounds = controller->currentFrame;
	currentBounds.origin.x = currentBounds.origin.y = 0;

	if (doingInitOnMainThread) {
		[self performSelector:@selector(doInitialization1:)
			withObject:controller];
	} else {
		[self performSelectorOnMainThread:@selector(doInitialization1:)
			withObject:controller waitUntilDone:true];
	}
	if (!myController) {
		[myLock unlock];
		[self autorelease];
		return nil;
	}

	if (![self performSelector:@selector(doInitialization2)]) {
		[myLock unlock];
		return self;
	}

	if (osxVersion >= 0x1040) {
		[self initApplet];
		[myLock unlock];
		if (!doingInitOnMainThread)
			return self;
	} else {
		if (!doingInitOnMainThread) {
			[AppletView performSelector:@selector(doInitialization3:)
				withObject:self];
			[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:.3]];
			[myLock unlock];
			return self;
		}
		/* We used to lock ourselves one extra time here, then unlock ourselves
		 * from the thread running doInitialization3 (to make absolutely sure we
		 * weren't interrupted).  But we really shouldn't have been doing that,
		 * and now Leopard doesn't allow it (and displays an error message about
		 * the lock being unlocked from a thread that didn't lock it).
		 */
		[NSThread detachNewThreadSelector:@selector(doInitialization3:)
			toTarget:[AppletView class] withObject:self];
		//[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:.1]];
		[myLock unlock];
	}

	/* As of JEP 0.8.7 we default to creating the applet synchronously --
	 * though part (possibly) of its init() method and all of its start()
	 * method still occur after this method (initWithParams) has returned.
	 * (Some applets (like the NOAA weather applets) change their display
	 * significantly in their init() methods.)  We accomplish this by looping
	 * here until we know that the applet has been created, then adding a
	 * short delay.  We need to handle events ourselves while we're waiting
	 * -- otherwise we may get freezes while the applet is being created (for
	 * example when we access a signed applet and security dialogs needs to
	 * be displayed).  There's no way to get the browser's display to update
	 * while we're waiting here -- so we want to return as soon as possible
	 * after the applet has been created.  But we also want to allow enough
	 * time for a simple initialization to finish before we return -- then
	 * this applet's methods can be safely called in a JavaScript onLoad
	 * handler by a browser that (like all of the Mozilla family) assumes
	 * applets will be created and initialized synchronously.
	 */
	if (!asyncInitializeApplet) {
		if (osxVersion < 0x1030)
			[AppletView postJEPSendingEventsStartedNotification];
		[AppletView postFakeEvent];
		for (;;) {
			if ((appletState == kAppletErrorJEP) ||
				(appletState == kAppletDestroyJEP) || [self destroying])
			{
				break;
			}
			if (returnedCreatedInfo) {
				if (osxVersion >= 0x1030) {
					total = .3;
					start = [NSDate date];
					while (total + [start timeIntervalSinceNow] > 0) {
						[NSApp nextEventMatchingMask:NSAnyEventMask untilDate:nil
							inMode:NSDefaultRunLoopMode dequeue:false];
						[AppletView processCustomEventSpool:false];
						[AppletView processCEventSpool:false];
						usleep(20000);
					}
				} else {
					[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:.3]];
				}
				break;
			}
			if ((osxVersion >= 0x1040) || ![self shouldProcessAllEvents:nil]) {
				[NSApp nextEventMatchingMask:NSAnyEventMask untilDate:nil
					inMode:NSDefaultRunLoopMode dequeue:false];
				[AppletView processCustomEventSpool:false];
				[AppletView processCEventSpool:false];
				usleep(20000);
			} else {
				anEvent = [NSApp nextEventMatchingMask:NSAnyEventMask
					untilDate:[NSDate dateWithTimeIntervalSinceNow:.05]
					inMode:NSDefaultRunLoopMode dequeue:true];
				if (anEvent) {
					[NSApp _setCurrentEvent:anEvent];
					[NSApp sendEvent:anEvent];
				} else {
					[AppletView postFakeEvent];
				}
				[AppletView processCustomEventSpool:false];
				[AppletView processCEventSpool:false];
			}
		}
		if (osxVersion < 0x1030)
			[AppletView postJEPSendingEventsEndedNotification];
		/* Give the event loop an extra kick before we return.
		 */
		[AppletView postFakeEvent];
	}

	return self;
}

- (BOOL)getJavaEnv:(JNIEnv **)envPtr;
{
	jint retval;
	JNIEnv *env;

	if (!envPtr)
		return false;
	if (!javaVM) {
		*envPtr = nil;
		return false;
	}

	retval = (*javaVM)->GetEnv(javaVM, (void **) &env, JNI_VERSION_1_4);
	if (retval == JNI_EDETACHED) {
		if (!(*javaVM)->AttachCurrentThread(javaVM, (void **) &env, nil))
			retval = JNI_OK;
	}
	if (retval != JNI_OK) {
		*envPtr = nil;
		return false;
	}

	*envPtr = env;
	return true;
}

- (BOOL)checkJNIException:(JNIEnv *)env
{
	jobject throwable;
	if (!env)
		return true;
	if ((throwable = (*env)->ExceptionOccurred(env)) != nil) {
		[self handleJNIException:throwable javaEnv:env];
		(*env)->DeleteLocalRef(env, throwable);
		return true;
	}
	return false;
}

- (void)handleJNIException:(jobject)throwable javaEnv:(JNIEnv *)env
{
	jclass throwableClass;
	jmethodID printStackTrace;
	if (!env)
		return;
	(*env)->ExceptionClear(env);
	if (!throwable)
		return;
	throwableClass = (*env)->FindClass(env, "java/lang/Throwable");
	printStackTrace = (*env)->GetMethodID(env, throwableClass,
		"printStackTrace", "()V");
	(*env)->CallVoidMethod(env, throwable, printStackTrace);
	(*env)->DeleteLocalRef(env, throwableClass);
	if ((*env)->ExceptionCheck(env))
		(*env)->ExceptionClear(env);
}

- (void)doInitialization1:(JEPController *)controller
{
	NSView *topView;
	if (!controller)
		return;
	if ((topView = [controller->myWindow contentView]) == nil)
		return;
	controller->myAppletView = self;
	if (isCarbonApp)
		[topView setAutoresizesSubviews:false];
	[topView addSubview:self];
	[self setFrame:NSMakeRect(controller->initialFrame.origin.x,
		controller->initialFrame.origin.y,
		controller->initialFrame.size.width,
		controller->initialFrame.size.height)];
	/* If an applet gets created (i.e. if no exception occurred while it's
	 * being loaded), the user focus will initially be inside it -- so we
	 * should set the "Cocoa" TSM state after it initializes and starts.  But
	 * if the applet fails to load, the user focus may or may not be inside
	 * our (empty) applet view -- so we should set the "Carbon" TSM state
	 * and input context here, just in case.  Doing so can avoid keyboard
	 * problems -- for example, the Mozilla family "about:plugins" page always
	 * contains an (invisible) applet that fails to load, and if we always set
	 * the "Cocoa" TSM state (as we used to in versions prior to JEP 0.8.3),
	 * the keyboard will be unusable in the URL box.
	 */
	if (isCarbonApp) {
		[controller setCarbonTSMState];
		[controller activateCarbonInputContext];
	}
	myController = controller;
	return;
}

- (BOOL)doInitialization2
{
	JNIEnv *env;
	NSArray *attribKeys;
	unsigned i, count;
	id key, value;
#ifdef SUPPORT_64BITS
	jlong ourView = (jlong) self;
#else
	jlong ourView = (jlong) ((unsigned) self);
#endif
	jclass appletHandlerFactoryClass, hashMapClass;
	jmethodID newHashMap, putHashMap, createAppletHandler;
	jobject hashMap = nil;
	jobject docbaseString, keyString, valueString;
	unichar *keyUTF16 = nil, *valueUTF16 = nil;
	unsigned keyLength, valueLength;
	jboolean succeeded;

	if (![self getJavaEnv:&env])
		return false;

	appletHandlerFactoryClass =
		(*env)->FindClass(env, "jep/AppletHandlerFactory");
	hashMapClass = (*env)->FindClass(env, "java/util/HashMap");
	if ([self checkJNIException:env]) {
		appletState = kAppletErrorJEP;
		[self performSelector:@selector(returnCreatedInfo)];
		return false;
	}
	newHashMap = (*env)->GetMethodID(env, hashMapClass, "<init>", "()V");
	putHashMap = (*env)->GetMethodID(env, hashMapClass, "put",
		"(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;");
	createAppletHandler = (*env)->GetStaticMethodID(env,
		appletHandlerFactoryClass, "createAppletHandler",
		"(JILjava/lang/String;Ljava/util/HashMap;ZZ)Z");
	if ([self checkJNIException:env]) {
		appletState = kAppletErrorJEP;
		[self performSelector:@selector(returnCreatedInfo)];
		return false;
	}

	if (appletAttrs) {
		hashMap = (*env)->NewObject(env, hashMapClass, newHashMap);
		if (!hashMap) {
			[self checkJNIException:env];
			appletState = kAppletErrorJEP;
			[self performSelector:@selector(returnCreatedInfo)];
			return false;
		}
		attribKeys = [appletAttrs allKeys];
		count = [attribKeys count];
		for (i = 1; i <= count; ++i) {
			key = [attribKeys objectAtIndex:i-1];
			if ((value = [appletAttrs objectForKey:key]) == nil)
				continue;
			keyLength = [[key description] length];
			valueLength = [[value description] length];
			keyUTF16 = calloc(keyLength + 1, sizeof(unichar));
			valueUTF16 = calloc(valueLength + 1, sizeof(unichar));
			if (!keyUTF16 || !valueUTF16) {
				appletState = kAppletErrorJEP;
				[self performSelector:@selector(returnCreatedInfo)];
				return false;
			}
			[[key description] getCharacters:keyUTF16];
			[[value description] getCharacters:valueUTF16];
			keyString = (*env)->NewString(env, keyUTF16, keyLength);
			valueString = (*env)->NewString(env, valueUTF16, valueLength);
			free(keyUTF16);
			free(valueUTF16);
			(*env)->CallObjectMethod(env, hashMap, putHashMap,
				keyString, valueString);
			(*env)->DeleteLocalRef(env, keyString);
			(*env)->DeleteLocalRef(env, valueString);
			if ([self checkJNIException:env]) {
				appletState = kAppletErrorJEP;
				[self performSelector:@selector(returnCreatedInfo)];
				return false;
			}
		}
		//[attribKeys release];	// If we release this we crash on exit
		[appletAttrs release];
	}

	valueLength = [appletDocbase length];
	valueUTF16 = calloc(valueLength + 1, sizeof(unichar));
	if (!valueUTF16) {
		appletState = kAppletErrorJEP;
		[self performSelector:@selector(returnCreatedInfo)];
		return false;
	}
	[appletDocbase getCharacters:valueUTF16];
	docbaseString = (*env)->NewString(env, valueUTF16, valueLength);
	free(valueUTF16);
	[appletDocbase release];

	succeeded = (*env)->CallStaticBooleanMethod(env, appletHandlerFactoryClass,
		createAppletHandler, ourView, (jint) myController->pluginInstance,
		docbaseString, hashMap, isCarbonApp, isInvisible);
	(*env)->DeleteLocalRef(env, docbaseString);
	(*env)->DeleteLocalRef(env, hashMap);
	(*env)->DeleteLocalRef(env, appletHandlerFactoryClass);
	(*env)->DeleteLocalRef(env, hashMapClass);
	(*env)->DeleteLocalRef(env, docbaseString);
	if ([self checkJNIException:env])
		succeeded = false;
	if (!succeeded) {
		appletState = kAppletErrorJEP;
		[self performSelector:@selector(returnCreatedInfo)];
		return false;
	}

	if ([self destroying]) {
		[self performSelector:@selector(returnCreatedInfo)];
		return false;
	}

	if (osxVersion < 0x1040) {
		[self imposeClip:currentBounds];
		if (doingInitOnMainThread)
			[self activateFrame];
	}

	return true;
}

/* This method must be called on a secondary thread -- otherwise you end up
 * hanging on a (non-recursive) lock inside libawt.jnilib's JNI_OnLoad
 * routine.  So when the caller invokes initWithParams, we can't just wait
 * for the process to finish -- instead we use a callback to let the caller
 * know the results.  I don't know whether the JNI_OnLoad business is by
 * design or is a bug.  Beware that, though we must initialize applet creation
 * on a secondary thread, and though (as it happens) a lot of Java drawing
 * takes place on one or more secondary threads, any drawing that we do (or
 * trigger directly) from Cocoa code must take place on the main thread --
 * otherwise we get wierd crashes and hangs.  (What I call Cocoa (or Cocoa-
 * initiated) drawing should always take place on the main thread, while Java
 * (or Java-initiated) drawing should always take place on some other thread
 * -- usually the applet's thread.)
 */
/* As of JEP 0.8.7, by default we make as much as possible of the applet's
 * creation and initialization synchronous -- this makes it possible to
 * support the JavaScript onLoad handler in browsers that (like the entire
 * Mozilla family) assume applet creation and initialization will happen
 * synchronously.  But there's a problem with making the _entire_ process
 * synchronous:  Some applets (like the NOAA weather applets) change their
 * display significantly in their init() methods.  And while we're being
 * synchronous the display can't be updated.
 */
/* As of JEP 0.9.5h this is only used on OS X Panther and Jaguar (on Tiger
 * and Leopard we use a Java secondary thread).  For more information see
 * AppletHandler.java.
 */
+ (void)doInitialization3:(AppletView *)aView
{
	JNIEnv *env;
	JavaVM *vm = nil;
	NSDate *started;
	NSTimeInterval interval;
	NSAutoreleasePool *localPool;

	if (!aView)
		return;

	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;

	[aView->myLock lock];

	if (![AppletView getJavaEnv:&env vmPtr:&vm]) {
		aView->appletState = kAppletErrorJEP;
		[aView performSelector:@selector(returnCreatedInfo)];
		[aView->myLock unlock];
		[localPool release];
		return;
	}

	if (!aView->myController) {
		aView->appletState = kAppletErrorJEP;
		[aView->myLock unlock];
		[localPool release];
		if (aView->doingInitOnMainThread)
			(*vm)->DetachCurrentThread(vm);
		return;
	}

	if ([aView destroying]) {
		[aView performSelector:@selector(returnCreatedInfo)];
		[aView->myLock unlock];
		[localPool release];
		if (aView->doingInitOnMainThread)
			(*vm)->DetachCurrentThread(vm);
		return;
	}

	/* On pre-Tiger Java versions the activateFrame method must run on the
	 * main thread (or we get a Java null-pointer error).
	 */
	if (!aView->doingInitOnMainThread) {
		[AppletView performOnMainThread:@selector(activateFrame)
			onTarget:aView withObject:nil waitUntilDone:true];
	}
	[aView initApplet];

	started = [NSDate date];

	for (;;) {
		if (!aView->myAppletHandler)
			break;
		if ([aView destroying]) {
			[aView performSelector:@selector(returnCreatedInfo)];
			break;
		}
		[aView getStatusFromJava];
		if (aView->appletState == kAppletErrorJEP) {
			[aView performSelector:@selector(returnCreatedInfo)];
			break;
		}
		if (aView->myApplet && (aView->appletState == kAppletInitJEP)) {
			[aView performSelector:@selector(maybeDoAppletStart)];
			break;
		}
		/* Sleep for 1/20 second. */
		[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:.05]];
		/* Stop looping after 60 seconds (30 was too short for some applets). */
		if ((interval = [started timeIntervalSinceNow]) < 0)
			interval = -interval;
		if (interval > 60) {
			if (aView->myApplet && aView->myController) {
				[aView->myController performSelector:@selector(destroyApplet)];
				break;
			}
			if ([aView destroying]) {
				aView->appletState = kAppletDestroyJEP;
			} else {
				aView->appletState = kAppletErrorJEP;
			}
			[aView performSelector:@selector(returnCreatedInfo)];
			break;
		}
	}

	awtLibrariesLoaded = true;

	[aView->myLock unlock];
	[localPool release];

	/* This thread will become a zombie thread if we don't call
	 * DetachCurrentThread().
	 */
	if (aView->doingInitOnMainThread)
		(*vm)->DetachCurrentThread(vm);

	return;
}

/* Always call this, even if there's an error or our applet is destroyed while
 * it's being created (unless our window is being (or has been) destroyed).
 */
- (void)returnCreatedInfo
{
	if (!myController || !myController->myWindow || returnedCreatedInfo)
		return;
	returnedCreatedInfo = true;
#ifdef DEBUG_EVENTS
	NSLog(@"returnCreatedInfo: Main thread %@, current thread %@, applet %@",
		mainThread, [NSThread currentThread], myAppletName);
#endif
	if ([self destroying]) {
		[myController performSelector:@selector(setNeedsCleanup)];
		/* In case appletState got changed accidentally in doInitialization */
		appletState = kAppletDestroyJEP;
	} else {
		if (!myApplet) {	/* Should only happen if state == error */
			appletState = kAppletErrorJEP;
			myAppletName = [[self description] retain];
		}
		if (appletState == kAppletErrorJEP)
			myController->resizeAppletDirectly = false;
		if (!visibilitySetExplicitly && !myController->createdInvisible)
			[self setAppletVisible:true flagOnly:false wait:false];
	}
	if (myController->appletCreatedCallback) {
		(*myController->appletCreatedCallback)((void *) myController, javaVM,
			myApplet, myController->myWindowCarbon, appletState,
			myController->appletCreatedUserData);
	}
}

- (void)returnInitializedInfo
{
	if (!myApplet || !myController || returnedInitializedInfo || [self destroying])
		return;
	returnedInitializedInfo = true;
#ifdef DEBUG_EVENTS
	NSLog(@"returnInitializedInfo: Main thread %@, current thread %@",
		mainThread, [NSThread currentThread]);
#endif
	if (myController->appletInitializedCallback) {
		(*myController->appletInitializedCallback)((void *) myController,
			javaVM, myApplet, myController->myWindowCarbon, appletState,
			myController->appletInitializedUserData);
	}
}

- (void)returnStartedInfo
{
	if (!myApplet || !myController || returnedStartedInfo || [self destroying])
		return;
	returnedStartedInfo = true;
#ifdef DEBUG_EVENTS
	NSLog(@"returnStartedInfo: Main thread %@, current thread %@",
		mainThread, [NSThread currentThread]);
#endif
	if (myController->appletStartedCallback) {
		(*myController->appletStartedCallback)((void *) myController,
			javaVM, myApplet, myController->myWindowCarbon, appletState,
			myController->appletStartedUserData);
	}
}

- (void)maybeDoAppletStart
{
	NSThread *currentThread = nil;
	[self performSelector:@selector(returnInitializedInfo)];
	if (!myController || !mainThread || (appletState == kAppletStartJEP) ||
		[self destroying])
		return;
	currentThread = [NSThread currentThread];
	if ([mainThread isEqual:currentThread]) {
		[myController performSelector:@selector(maybeStartAppletOnMainThread)];
	} else {
		[myController performSelectorOnMainThread:@selector(maybeStartAppletOnMainThread)
			withObject:nil waitUntilDone:false];
	}
}

- (void)repaint
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID repaint;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	repaint = (*env)->GetMethodID(env, appletHandlerClass,
		"repaint", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, repaint);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

- (void)validate
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID validateApplet;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	validateApplet = (*env)->GetMethodID(env, appletHandlerClass,
		"validateApplet", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, validateApplet);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

- (void)invalidate
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID invalidate;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	invalidate = (*env)->GetMethodID(env, appletHandlerClass,
		"invalidateAndValidate", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, invalidate);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

- (BOOL)isValid
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID isValid;
	jboolean valid = false;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return false;

	if ([self checkJNIException:env])
		return false;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	isValid = (*env)->GetMethodID(env, appletHandlerClass,
		"isValid", "()Z");
	valid = (*env)->CallIntMethod(env, myAppletHandler, isValid);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	if ([self checkJNIException:env])
		return false;

	return valid;
}

- (void)validateClips
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID validateClips;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	validateClips = (*env)->GetMethodID(env, appletHandlerClass,
		"validateClips", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, validateClips);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

- (void)zeroClips
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID zeroClips;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	zeroClips = (*env)->GetMethodID(env, appletHandlerClass,
		"zeroClips", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, zeroClips);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

- (NSRect)translateCurrentClip:(NSView *)viewToClip
{
	NSRect viewBounds = NSZeroRect;
	NSPoint viewLocation = NSZeroPoint;
	NSRect translatedClip = NSZeroRect;
	if (!viewToClip || isInvisible || ![self isAppletVisible])
		return translatedClip;
	if ((currentBounds.size.width <= 0) || (currentBounds.size.height <= 0))
		return translatedClip;
	translatedClip = currentBounds;
	viewBounds = [viewToClip bounds];
	viewLocation = [self convertPoint:viewBounds.origin fromView:viewToClip];
	translatedClip.origin.x -= viewLocation.x;
	translatedClip.origin.y -= viewLocation.y;
	return translatedClip;
}

/* This method must run on the main thread (or we get a Java null-pointer
 * error).
 */
- (void)activateFrame
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID activateFrame;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	activateFrame = (*env)->GetMethodID(env, appletHandlerClass,
		"activateFrame", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, activateFrame);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

- (void)initApplet
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID initApplet;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	initApplet = (*env)->GetMethodID(env, appletHandlerClass,
		"initApplet", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, initApplet);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

/* We use startApplet to start the applet for the first time (just after
 * it's been initialized), and we use stopApplet to stop it for the last
 * time just before we destroy it.  But for starting and stopping the applet
 * at other times we use appletStartApplet and appletStopApplet.  These
 * last two methods seem to get along better with our clipping hacks.  See
 * AppletFrame.java for more information.
 * Java Embedding Plugin versions prior to 0.9.6 (0.9.5+d) had code that
 * postponed calling an applet's start() method if the applet wasn't yet
 * visible (the code was here and in AppletView setClip and JEPController
 * showHideAppletImpl).  For a time this was needed to stop the start() method
 * from changing the display when the applet was loading in a different tab.
 * An exception was made for an applet whose bounds had zero width or height
 * (and which therefore could never display).  But my clipping strategies have
 * gotten good enough that they can now squelch any applet's start() method.
 * And my old code sometimes prevented an applet's start() method from ever
 * running if the applet could never display, but didn't have zero width or
 * height (see https://bugzilla.mozilla.org/show_bug.cgi?id=328466 and
 * http://sourceforge.net/tracker/index.php?func=detail&aid=1438179&group_id=107955&atid=649116).
 * Causing the start() method to sometimes run earlier has inevitably shaken
 * out a number of Apple bugs, which I've had to work around.  For more info
 * see Java_CPanel_createNativePanel() above and the comment above
 * AppletFramePanel.addImpl().
 */
- (void)startApplet
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID startApplet;
	NSString *info = nil;

	if (!myApplet || !myAppletHandler || !myController ||
		![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	if (appletState < kAppletInitJEP)
		return;

	if (debugVisibility) {
		info = [NSString stringWithFormat:@"%@ AppletView:startApplet for %@",
			[NSDate date], myAppletName];
		[AppletView printMessage:info javaEnv:nil];
	}

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	startApplet = (*env)->GetMethodID(env, appletHandlerClass,
		"startApplet", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, startApplet);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	if ([self checkJNIException:env])
		return;
	appletState = kAppletStartJEP;
	[self performSelector:@selector(returnStartedInfo)];
}

- (void)appletStartApplet
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID startApplet;

	if (!myApplet || !myAppletHandler ||
		![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	if (appletState < kAppletInitJEP)
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	startApplet = (*env)->GetMethodID(env, appletHandlerClass,
		"appletStartApplet", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, startApplet);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	if ([self checkJNIException:env])
		return;
	appletState = kAppletStartJEP;
	[self setAppletVisible:true flagOnly:true wait:false];
}

- (void)stopApplet
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID stopApplet;

	if (!myApplet || !myAppletHandler ||
		![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	if (appletState != kAppletStartJEP)
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	stopApplet = (*env)->GetMethodID(env, appletHandlerClass,
		"stopApplet", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, stopApplet);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	if ([self checkJNIException:env])
		return;
	appletState = kAppletStopJEP;
	[self setAppletVisible:false flagOnly:true wait:false];
}

- (void)appletStopApplet
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID stopApplet;

	if (!myApplet || !myAppletHandler ||
		![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	if (appletState != kAppletStartJEP)
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	stopApplet = (*env)->GetMethodID(env, appletHandlerClass,
		"appletStopApplet", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, stopApplet);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	if ([self checkJNIException:env])
		return;
	appletState = kAppletStopJEP;
	[self setAppletVisible:false flagOnly:true wait:false];
}

- (void)destroyApplet
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID destroyApplet;

	if ([self destroying])
		return;
	isDestroying = true;

	if (!myAppletHandler) {
		isInvisible = true;
		appletState = kAppletDestroyJEP;
		myApplet = nil;
		return;
	}

	if (![self getJavaEnv:&env] || [self checkJNIException:env]) {
		isInvisible = true;
		appletState = kAppletDestroyJEP;
		myApplet = nil;
		return;
	}

	[self setAppletVisible:false flagOnly:true wait:false];

	appletState = kAppletDestroyJEP;

	if (myController && myController->appletAWTContext && awtViewContextClass &&
		[myController->appletAWTContext isKindOfClass:awtViewContextClass])
	{
		((NSViewAWTContext *)myController->appletAWTContext)->alive = false;
		((NSViewAWTContext *)myController->appletAWTContext)->enabled = false;
	}

	if (myController && myController->appletWindow)
		awaitingAppletWindowCleanup = true;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	destroyApplet = (*env)->GetMethodID(env, appletHandlerClass,
		"destroyApplet", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, destroyApplet);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

/* See resizeApplet() in AppletFrame.java */
- (void)resizeApplet:(int)newWidth newHeight:(int)newHeight
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID resizeApplet;

	if (!myController || !myAppletHandler || ![self getJavaEnv:&env] ||
		!newWidth || !newHeight || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	resizeApplet = (*env)->GetMethodID(env, appletHandlerClass,
		"resizeApplet", "(II)V");
	(*env)->CallVoidMethod(env, myAppletHandler, resizeApplet, newWidth, newHeight);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	if ([self checkJNIException:env])
		return;
	currentBounds.origin.x = currentBounds.origin.y = 0;
	currentBounds.size.width = newWidth;
	currentBounds.size.height = newHeight;
	myController->currentFrame.size.width = newWidth;
	myController->currentFrame.size.height = newHeight;
}

- (void)printComponentInfo
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID showAppletInfo;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	showAppletInfo = (*env)->GetMethodID(env, appletHandlerClass,
		"showAppletInfo", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, showAppletInfo);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

/* In Java our applet has a Panel object parent and a Frame object
 * grandparent.  For our clipping hacks to work correctly, we need to
 * replace their default layout managers with a very "simple" one.
 * See AppletFrame.java and SimpleLayout.java for more information.
 */
- (void)setSimpleLayout
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID setSimpleLayout;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	setSimpleLayout = (*env)->GetMethodID(env, appletHandlerClass,
		"setSimpleLayout", "()V");
	(*env)->CallVoidMethod(env, myAppletHandler, setSimpleLayout);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

/* We don't currently use this (it causes display problems and doesn't seem
 * to do anything useful).  I leave it in for future reference.
 */
- (void)setIgnoreRepaint:(BOOL)flag
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID setIgnoreRepaint;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	setIgnoreRepaint = (*env)->GetMethodID(env, appletHandlerClass,
		"setIgnoreRepaint", "(Z)V");
	(*env)->CallVoidMethod(env, myAppletHandler, setIgnoreRepaint, flag);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

- (void)setUpdatesOff:(BOOL)flag
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID setUpdates;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	setUpdates = (*env)->GetMethodID(env, appletHandlerClass,
		"setUpdatesOff", "(Z)V");
	(*env)->CallVoidMethod(env, myAppletHandler, setUpdates, flag);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

- (void)setAsyncUpdatesOff:(BOOL)flag
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID setAsyncUpdates;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	setAsyncUpdates = (*env)->GetMethodID(env, appletHandlerClass,
		"setAsyncUpdatesOff", "(Z)V");
	(*env)->CallVoidMethod(env, myAppletHandler, setAsyncUpdates, flag);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];
}

- (void)setAppletVisible:(BOOL)visible flagOnly:(BOOL)flagOnly wait:(BOOL)wait
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID setAppletVisible;
	BOOL visibilityChanged;

	if (!myAppletHandler || !myController ||
		![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	visibilityChanged = (isInvisible == visible);

	/* This is like what we're doing in Handlers.m's registerModelForUpdate_TU(),
	 * but works in Tiger.  (This isn't available on Jaguar (OS X 10.2.8).  And
	 * though it _is_ available on Panther (10.3), we don't need it there and so
	 * aren't using it.)
	 */
	if ((osxVersion >= 0x1040) && visibilityChanged && myController->appletAWTContext) {
		if (visible) {
			[myController->appletAWTContext enable];
		} else {
			[myController->appletAWTContext disable];
		}
	}

	isInvisible = !visible;	// Needed here by validateClips() when called from
							// AppletFrame.setVisible().
	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	setAppletVisible = (*env)->GetMethodID(env, appletHandlerClass,
		"setAppletVisible", "(ZZZ)V");
	(*env)->CallVoidMethod(env, myAppletHandler, setAppletVisible,
		visible, flagOnly, wait);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	if ([self checkJNIException:env])
		return;
	visibilitySetExplicitly = true;
	if (isCarbonApp && visible && visibilityChanged && (appletState == kAppletStartJEP))
		myController->displayNeedsUpdate = true;
}

- (BOOL)isAppletVisible
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID isAppletVisible;
	BOOL retval;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return false;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	isAppletVisible = (*env)->GetMethodID(env, appletHandlerClass,
		"isAppletVisible", "()Z");
	retval = (*env)->CallBooleanMethod(env, myAppletHandler, isAppletVisible);
	(*env)->DeleteLocalRef(env, appletHandlerClass);

	if ([self checkJNIException:env])
		return false;

	return retval;
}

- (jobject)getAppletFromJava
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID getApplet;
	jobject applet = nil;

	if (!myAppletHandler || !myController || !myLock ||
		![self getJavaEnv:&env] || [self destroying])
	{
		myApplet = nil;
		return nil;
	}

	if ([self checkJNIException:env]) {
		myApplet = nil;
		return nil;
	}

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	getApplet = (*env)->GetMethodID(env, appletHandlerClass,
		"getApplet", "()Ljava/applet/Applet;");
	applet = (*env)->CallObjectMethod(env, myAppletHandler, getApplet);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	if ([self checkJNIException:env])
		applet = nil;

	[myLock lock];
	if (myApplet)
		(*env)->DeleteGlobalRef(env, myApplet);
	if (applet) {
		myApplet = (*env)->NewGlobalRef(env, applet);
	} else {
		myApplet = nil;
	}
	[myLock unlock];

	return myApplet;
}

- (int)getStatusFromJava
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID getLoadingStatus;
	jint status = 0;

	if ([self destroying])
		return kAppletDestroyJEP;

	if (!myAppletHandler || ![self getJavaEnv:&env])
		return 0;

	if ([self checkJNIException:env])
		return 0;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	getLoadingStatus = (*env)->GetMethodID(env, appletHandlerClass,
		"getLoadingStatus", "()I");
	status = (*env)->CallIntMethod(env, myAppletHandler, getLoadingStatus);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	if ([self checkJNIException:env])
		return 0;
	appletState = status;

	return status;
}

- (NSRect)getAppletBounds
{
	JNIEnv *env;
	jclass appletHandlerClass, rectangleClass;
	jmethodID getAppletBounds;
	jfieldID heightID, widthID, xID, yID;
	jobject bounds = nil;
	int height = 0, width = 0, x = 0, y = 0;
	NSRect boundsCocoa = initialBounds;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return boundsCocoa;

	if ([self checkJNIException:env])
		return boundsCocoa;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	rectangleClass = (*env)->FindClass(env, "java/awt/Rectangle");
	getAppletBounds = (*env)->GetMethodID(env, appletHandlerClass,
		"getAppletBounds", "()Ljava/awt/Rectangle;");
	heightID = (*env)->GetFieldID(env, rectangleClass, "height", "I");
	widthID = (*env)->GetFieldID(env, rectangleClass, "width", "I");
	xID = (*env)->GetFieldID(env, rectangleClass, "x", "I");
	yID = (*env)->GetFieldID(env, rectangleClass, "y", "I");
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	(*env)->DeleteLocalRef(env, rectangleClass);
	if ([self checkJNIException:env])
		return boundsCocoa;
	bounds = (*env)->CallObjectMethod(env, myAppletHandler, getAppletBounds);
	if (!bounds) {
		[self checkJNIException:env];
		return boundsCocoa;
	}
	height = (*env)->GetIntField(env, bounds, heightID);
	width = (*env)->GetIntField(env, bounds, widthID);
	x = (*env)->GetIntField(env, bounds, xID);
	y = (*env)->GetIntField(env, bounds, yID);
	(*env)->DeleteLocalRef(env, bounds);
	if ([self checkJNIException:env])
		return boundsCocoa;
	boundsCocoa.origin.x = x;
	boundsCocoa.origin.y = y;
	boundsCocoa.size.width = width;
	boundsCocoa.size.height = height;

	return boundsCocoa;
}

/* Expects "clip" to be in whatever coordinate system our parent window uses:
 * The Carbon one (origin top-left) if we're in a Carbon app, the Cocoa one
 * (origin bottom-left) if we're in a Cocoa one.
 */
- (void)setClip:(NSRect)clip
{
	NSRect appletFrame = NSZeroRect;
	NSRect appletBounds = NSZeroRect;
	NSRect frameHolder = NSZeroRect;
	float hFactor, vFactor;
	int windowNumber;
	NSWindow *cacheWindow;
	NSView *topView, *superView;
	JEPController *cacheController = nil;
	NSString *info = nil;
	BOOL willZeroClip = false;
	BOOL myWindowVisible = false;
	BOOL appletVisible = false;

	if (!myController || !myController->myWindow ||
		!myController->myWindowCarbon || [self destroying])
		return;

	appletFrame = [self frame];
	appletBounds = [self getAppletBounds];
	if ((appletFrame.size.width == 0) || (appletFrame.size.height == 0)) {
		[self unsetClip];
		return;
	}
	if ((appletBounds.size.width == 0) || (appletBounds.size.height == 0)) {
		[self unsetClip];
		return;
	}

	/* Apple's Java 1.3.1 and Java 1.4.X browser plugins (interfaces for its
	 * JVM) all zero-clip an applet view when you try to make its clip width
	 * or clip height negative.  Some applets (notably the map24.com ones)
	 * actually rely on this.  As of JEP 0.8.7, we now behave the same way.
	 */
	willZeroClip = ((clip.size.width <= 0) || (clip.size.height <= 0) ||
		(clip.origin.x > appletFrame.origin.x+appletFrame.size.width) ||
		(clip.origin.x+clip.size.width < appletFrame.origin.x) ||
		(clip.origin.y > appletFrame.origin.y+appletFrame.size.height) ||
		(clip.origin.y+clip.size.height < appletFrame.origin.y));

	appletVisible = [self isAppletVisible];

	/* This block of code (new as of JEP 0.9.5+e) works around bmo bug 277067
	 * (whose effects have grown much more severe since the Firefox 1.0.X
	 * days, and are now visible in both Firefox and Camino).
	 */
	if (willZeroClip && ![self isCurrentClipEmpty] && !isInvisible &&
		appletVisible && nsAppShadow)
	{
		if (debugUpdates) {
			info = [NSString stringWithFormat:
				@"%@ AppletView:setClip zero-clipping %@, then updating old location",
				[NSDate date], myAppletName];
			[AppletView printMessage:info javaEnv:nil];
		}
		if (isCarbonApp) {
			CarbonUpdateParams *params;
			Rect *updateRect = (Rect *) calloc(1, sizeof(Rect));
			if (updateRect) {
				updateRect->left = myController->currentFrame.origin.x;
				updateRect->top = myController->currentFrame.origin.y;
				updateRect->right = updateRect->left +
					myController->currentFrame.size.width;
				updateRect->bottom = updateRect->top +
					myController->currentFrame.size.height;
				params = [[CarbonUpdateParams alloc]
					initWithParams:myController->myWindowCarbon
					updateRects:updateRect numUpdateRects:1];
				if (params) {
					[nsAppShadow performSelectorOnMainThread:
						@selector(postCarbonUpdateOnMainThread:)
						withObject:params waitUntilDone:false];
				}
			}
		} else {
			CocoaUpdateParams *params;
			NSView *contentView = [myController->myWindow contentView];
			NSRect *updateRect = (NSRect *) calloc(1, sizeof(NSRect));
			if (updateRect && contentView) {
				*updateRect = myController->currentFrame;
				updateRect->origin.y -= 22;
				updateRect->size.height += 44;
				params = [[CocoaUpdateParams alloc]
					initWithParams:contentView updateNSRects:updateRect
					numUpdateNSRects:1];
				if (params) {
					[nsAppShadow performSelectorOnMainThread:
						@selector(postCocoaUpdateOnMainThread:)
						withObject:params waitUntilDone:false];
				}
			}
		}
	}

	if ((osxVersion >= 0x1040) && !isInvisible && appletVisible &&
		myController->appletAWTContext)
	{
		if (willZeroClip) {
			[myController->appletAWTContext disable];
		} else {
			[myController->appletAWTContext enable];
		}
	}

	if (isCarbonApp) {
		myWindowVisible = [myController->myWindow isVisible];
		if (!isInvisible && !willZeroClip && !myWindowVisible &&
			(appletState == kAppletStartJEP) && appletVisible)
			myController->displayNeedsUpdate = true;
	}

	if (debugVisibility) {
		info = [NSString stringWithFormat:
			@"%@ AppletView:setClip clipping %@ to X (%f) Y (%f) Width (%f) Height (%f)",
			[NSDate date], myAppletName, clip.origin.x, clip.origin.y,
			clip.size.width, clip.size.height];
		[AppletView printMessage:info javaEnv:nil];
	}

	/* The following code helps deal with applets in multiple tabs in the
	 * same window:  When switching between tabs, Mozilla-family browsers use
	 * clipping or the "ShowHide" function to make the applet views visible or
	 * invisible.  The NSCarbonWindow objects of applet views that become
	 * invisible should be removed from the NSApplication window cache, and
	 * the NSCarbonWindow objects of applet views that become visible should
	 * be added to the window cache.
	 */
	if (isCarbonApp && !isInvisible && appletVisible &&
		myController->myLock && [myController->myLock lockBeforeDate:
		[NSDate dateWithTimeIntervalSinceNow:LOCK_TIMEOUT]])
	{
		/* The "willZeroClip" block makes setClip more closely parallel
		 * JEPController showHideApplet.  The bugzilla bug 298961 fix seems
		 * also to have made this code necessary.  Without this change,
		 * keyboard input sometimes gets disabled outside an applet when
		 * you switch away from a tab containing one or more applets.
		 */
		if (willZeroClip) {
			myController->hasKeyWindow = false;
			[myController setCarbonTSMState];
			[myController activateCarbonInputContext];
			if ([myController->myWindow isVisible]) {
				[myController->myWindow _setVisible:false];
#ifdef DEBUG_EVENTS
				NSLog(@"setClip: Made invisible (directly) window %@",
					myController->myWindow);
#endif
			}
			/* As of JEP 0.9.6, we now always remove our NSWindow from the
			 * window cache here (after we've made sure that it's invisible).
			 * This avoids complicated wierdness -- such as sometimes having
			 * the wrong window in the window cache, which can lead to our
			 * code being mistaken about which applet is focused (if we have
			 * applets in several different tabs), which in turn can sometimes
			 * break keyboard input in browser fields.
			 */
			[NSApp _removeWindowFromCache:myController->myWindow];
		} else if (myController->displayNeedsUpdate) {
			myController->displayNeedsUpdate = false;
			windowNumber = [myController->myWindow _realWindowNumber];
			cacheWindow = [NSApp windowWithWindowNumber:windowNumber];
			if (!cacheWindow || (cacheWindow != myController->myWindow)) {
				if (cacheWindow) {
					cacheController = [JEPController
						controllerFromNSWindow:cacheWindow
						timeoutType:lockTimeoutShort];
					if (cacheController && cacheController->myLock) {
						[cacheController->myLock lock];
						cacheController->hasKeyWindow = false;
						[cacheController setCarbonTSMState];
						[cacheController activateCarbonInputContext];
						if ([cacheWindow isVisible]) {
							[cacheWindow _setVisible:false];
#ifdef DEBUG_EVENTS
							NSLog(@"setClip: Made invisible (indirectly) window %@",
								cacheWindow);
#endif
						}
						[NSApp _removeWindowFromCache:cacheWindow];
						[cacheController->myLock unlock];
					}
				}
				[myController->myWindow _setWindowNumber:0xFFFFFFFF];
				[myController->myWindow _setWindowNumber:windowNumber];
				[NSApp _getWindowCache:myController->myWindow add:true];
#ifdef DEBUG_EVENTS
				NSLog(@"setClip: Reset window in cache from %@ to %@, first responder now %@",
					cacheWindow, myController->myWindow, [myController->myWindow firstResponder]);
#endif
			}
			if (!myWindowVisible) {
				/* _invalidateFocus helps avoid the error message
				 * "CGContextReplaceTopGState: invalid context"
				 */
				if ((topView = [myController->myWindow contentView]) != nil) {
					if ((superView = [topView superview]) != nil)
						topView = superView;
				}
				if (topView)
					[topView _invalidateFocus];
				[myController->myWindow _setVisible:true];
#ifdef DEBUG_EVENTS
				NSLog(@"setClip: Made visible window %@", myController->myWindow);
#endif
			}
		}
		[myController->myLock unlock];
		/* Our fix to registerModelForUpdate (in Handlers.m) doesn't work
		 * properly unless we sometimes "wake it up" (information on whether
		 * our applet is visible or invisible can sometimes be out of date).
		 */
		if (osxVersion < 0x1040) {
			if (myController->appletWindow && myApplet) {
				if (isJava142) {
					[NSClassFromString(@"AWTModelUpdater")
						registerModelForUpdate:myController->appletWindow];
				} else {
					[NSClassFromString(@"ThreadUtilities")
						registerModelForUpdate:myController->appletWindow];
				}
			}
		}
	}

	frameHolder = appletFrame;
	frameHolder.size.height = frameHolder.size.width = 0;

	if (willZeroClip) {
		currentBounds = NSZeroRect;
		myController->currentFrame = frameHolder;
		[self imposeClip:currentBounds];
		return;
	}

	if ((clip.origin.x == appletFrame.origin.x) &&
		(clip.origin.y == appletFrame.origin.y) &&
		(clip.size.width == appletFrame.size.width) &&
		(clip.size.height == appletFrame.size.height))
	{
		[self unsetClip];
		return;
	}

	if (clip.origin.x + clip.size.width >
		appletFrame.origin.x + appletFrame.size.width)
	{
		clip.size.width = appletFrame.origin.x +
			appletFrame.size.width - clip.origin.x;
	}
	if (clip.origin.y + clip.size.height >
		appletFrame.origin.y + appletFrame.size.height)
	{
		clip.size.height = appletFrame.origin.y +
			appletFrame.size.height - clip.origin.y;
	}
	if (clip.origin.x < appletFrame.origin.x)
		clip.origin.x = appletFrame.origin.x;
	if (clip.origin.y < appletFrame.origin.y)
		clip.origin.y = appletFrame.origin.y;

	frameHolder = clip;

	clip.origin.x -= appletFrame.origin.x;
	clip.origin.y -= appletFrame.origin.y;
	hFactor = appletBounds.size.width / appletFrame.size.width;
	vFactor = appletBounds.size.height / appletFrame.size.height;
	clip.origin.x = rint(clip.origin.x * hFactor);
	clip.origin.y = rint(clip.origin.y * vFactor);
	clip.size.width = rint(clip.size.width * hFactor);
	clip.size.height = rint(clip.size.height * vFactor);

	/* If we're in a Cocoa app we need to translate our clip from the Cocoa
	 * coordinate system (bottom-left origin) to the Java one (top-left
	 * origin) before calling imposeClip.
	 */
	if (!isCarbonApp) {
		clip.origin.y += clip.size.height;
		clip.origin = [JEPController flipViewPoint:clip.origin
			view:self];
	}

	currentBounds = clip;
	myController->currentFrame = frameHolder;
	[self imposeClip:currentBounds];

	return;
}

/* Expects "clip" to be in our applet view's coordinate system (i.e. to be
 * either its "bounds" or a subset thereof), and to have a top-left origin
 * (like all Java objects).
 */
- (void)imposeClip:(NSRect)clip
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID imposeClip;

	if (!myAppletHandler || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	imposeClip = (*env)->GetMethodID(env, appletHandlerClass,
		"imposeClip", "(IIII)V");
	(*env)->CallVoidMethod(env, myAppletHandler, imposeClip,
		(int) rint(clip.origin.x), (int) rint(clip.origin.y),
		(int) rint(clip.size.width), (int) rint(clip.size.height));
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	[self checkJNIException:env];

	return;
}

- (void)unsetClip
{
	JNIEnv *env;
	jclass appletHandlerClass;
	jmethodID removeClip;

	if (!myAppletHandler || !myController || ![self getJavaEnv:&env] || [self destroying])
		return;

	if ([self checkJNIException:env])
		return;

	appletHandlerClass = (*env)->GetObjectClass(env, myAppletHandler);
	removeClip = (*env)->GetMethodID(env, appletHandlerClass,
		"removeClip", "()V");
	if ([self checkJNIException:env])
		return;
	(*env)->CallVoidMethod(env, myAppletHandler, removeClip);
	(*env)->DeleteLocalRef(env, appletHandlerClass);
	if ([self checkJNIException:env])
		return;

	currentBounds = [self getAppletBounds];
	myController->currentFrame = [self frame];

	return;
}

- (BOOL)isCurrentClipEmpty
{
	if ((currentBounds.size.width <= 0) || (currentBounds.size.height <= 0))
		return true;
	return false;
}

- (BOOL)areAppletBoundsEmpty
{
	NSRect appletBounds = [self getAppletBounds];
	if ((appletBounds.size.width <= 0) || (appletBounds.size.height <= 0))
		return true;
	return false;
}

/* This is a more thorough check of whether or not our applet view is
 * displayable that just looking at "isInvisible" or calling
 * "isCurrentClipEmpty".  We should use it in code that has to decide
 * whether or not to display our applet view.  We need it to handle a design
 * flaw in Mozilla-family browsers:  They don't ensure that the visibility and
 * clipping are set correctly for an applet view that's entirely outside of
 * the region where applets can be displayed -- which can cause "off screen"
 * applets to be displayed (in the wrong location) while the user is resizing
 * our window.
 */
- (BOOL)isDisplayable
{
	Rect boundsCarbon = {0};
	NSRect boundsCocoa = NSZeroRect;
	NSRect currentFrame = NSZeroRect;
	if (!myController || !myController->myLock || !myController->myWindow ||
		!myController->myWindowCarbon || isInvisible)
		return false;
	if (![myController->myLock 
		lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:LOCK_TIMEOUT]])
		return false;
	if ([self destroying] || [self isCurrentClipEmpty] || ![self isAppletVisible]) {
		[myController->myLock unlock];
		return false;
	}
	/* Oddly, GetWindowBounds() sometimes crashes in Camino (on threads other
	 * than the main thread, when called from Java).  And the Cocoa method for
	 * getting the bounds of our window's content region doesn't seem to be
	 * inaccurate with a Cocoa browser (which it definitely can be with Carbon
	 * browsers).
	 */
	if (isCarbonApp) {
		GetWindowBounds(myController->myWindowCarbon, kWindowContentRgn,
			&boundsCarbon);
		boundsCocoa.origin.x = boundsCocoa.origin.y = 0;
		boundsCocoa.size.width = boundsCarbon.right - boundsCarbon.left;
		boundsCocoa.size.height = boundsCarbon.bottom - boundsCarbon.top;
	} else {
		boundsCocoa = [myController->myWindow frame];
		/* Before JEP 0.9.5 we used the class method instead of the instance
		 * method.  But that set the bounds incorrectly (it didn't take
		 * account of Camino's toolbar) -- so parts of the applet could leak
		 * into the bar at the bottom of the browser page.
		 */
		if (osxVersion >= 0x1030) {
			boundsCocoa = [myController->myWindow
				contentRectForFrameRect:boundsCocoa];
		} else {
			boundsCocoa = [myController->myWindow
				contentRectForFrameRect:boundsCocoa styleMask:0];
		}
		boundsCocoa.origin.x = boundsCocoa.origin.y = 0;
	}
	currentFrame = myController->currentFrame;
	if ((boundsCocoa.size.width <= 0) || (boundsCocoa.size.width <= 0)) {
		[myController->myLock unlock];
		return false;
	}
	/* We don't know (and can't know) exactly what part of our window's
	 * content region applets may be displayed into.  But we can (I hope)
	 * assume that if our view really is displayable, our current clip is
	 * set accurately -- which means that no part of it should ever stray
	 * outside of our window's content region.
	 */
	if ((currentFrame.origin.x < boundsCocoa.origin.x) ||
		(currentFrame.origin.y < boundsCocoa.origin.y) ||
		(currentFrame.origin.x + currentFrame.size.width >
			boundsCocoa.origin.x + boundsCocoa.size.width) ||
		(currentFrame.origin.y + currentFrame.size.height >
			boundsCocoa.origin.y + boundsCocoa.size.height))
	{
		[myController->myLock unlock];
		return false;
	}
	[myController->myLock unlock];
	return true;
}

/* This is used from HandleDispatcher() in Handlers.m as part of the fix for
 * Apple's "unmatched mouse down" bug.  For an (as yet) unmatched mouse-down
 * Carbon event in unmatchedMouseDown, it generates and sends a matching
 * mouse-up event.  As of JEP 0.9.6 this method is pretty hackish ... but it's
 * definitely less hackish than it was in JEP 0.9.4 and 0.9.5.
 */
- (void)doMatchingLeftMouseUp:(EventRef)unmatchedLeftMouseDown
{
	UInt32 clickCount;
	EventMouseButton button;
	NSPoint windowLocation;
	Point windowLocationCarbon;
	Rect frameCarbon = {0};
	NSEvent *mouseUp, *anEvent;
	NSDate *start, *finish;
	NSTimeInterval total, remaining;
	EventRef mouseUpCarbon = nil;
	Boolean boolTrue = true;
	NSThread *currentThread = [NSThread currentThread];

	if (!unmatchedLeftMouseDown || !myController || !myController->myWindow ||
		!myController->myWindowCarbon || !mainThread ||
		![mainThread isEqual:currentThread] || [self destroying])
		return;

	if (GetEventParameter(unmatchedLeftMouseDown, kEventParamMouseButton,
		typeMouseButton, NULL, sizeof(EventMouseButton), NULL, &button) != noErr)
		return;
	/* The matching fix isn't needed except for a left mouse-down */
	if (button != kEventMouseButtonPrimary)
		return;
	if (GetEventParameter(unmatchedLeftMouseDown, kEventParamClickCount,
		typeUInt32, NULL, sizeof(UInt32), NULL, &clickCount) != noErr)
		return;
	if (GetEventParameter(unmatchedLeftMouseDown, kEventParamMouseLocation,
		typeQDPoint, NULL, sizeof(Point), NULL, &windowLocationCarbon) != noErr)
		return;
	GetWindowBounds(myController->myWindowCarbon, kWindowContentRgn,
		&frameCarbon);
	windowLocationCarbon.v -= frameCarbon.top;
	windowLocationCarbon.h -= frameCarbon.left;
	windowLocation = [JEPController flipWindowPoint:
		NSMakePoint(windowLocationCarbon.h, windowLocationCarbon.v)
		window:myController->myWindow];

	/* Apple documents (in CarbonEvents.h) a number of properties that a
	 * mouse-up Carbon event is expected to have.  But clearly additional,
	 * undocumented fields are also required -- when I create an event and
	 * fill in all the documented fields, I get error messages when I send it.
	 * So I use the following round-about way to get a valid mouse-down
	 * Carbon event -- I create a Cocoa event, post it, pull it off the
	 * Cocoa event loop without "executing" it, and copy the Carbon equivalent
	 * that the AppKit (presumably) has generated for it.  The _eventRef
	 * selector doesn't work (it just returns NULL) unless the Cocoa event has
	 * actually been posted.
	 */
	mouseUp = [NSEvent mouseEventWithType:NSLeftMouseUp
		location:windowLocation
		modifierFlags:0
		timestamp:0
		windowNumber:[myController->myWindow windowNumber]
		context:nil
		eventNumber:0
		clickCount:clickCount
		pressure:0];
	if (!mouseUp)
		return;
	[NSApp postEvent:mouseUp atStart:true];

	remaining = total = .1;
	start = [NSDate date];
	for (;;) {
		finish = [NSDate dateWithTimeIntervalSinceNow:remaining];
		anEvent = [NSApp nextEventMatchingMask:NSLeftMouseUpMask
			untilDate:finish inMode:NSEventTrackingRunLoopMode dequeue:true];
		if (anEvent && ([anEvent type] == NSLeftMouseUp)) {
			mouseUpCarbon = CopyEvent([anEvent _eventRef]);
			break;
		}
		if ((remaining = total + [start timeIntervalSinceNow]) <= 0)
			break;
	}
	if (mouseUpCarbon) {
		/* We need to set a "not for Cocoa" parameter in this event, so that
		 * JEPController routeEvent will swallow the Carbon mouseUp event that
		 * gets created by Apple's Cocoa-Carbon interface -- otherwise the JVM
		 * can behave strangely.  For more info see HandleDispatcher() in
		 * Handlers.m.
		 */
		SetEventParameter(mouseUpCarbon, kEventParamJEPNotForCocoa,
			typeBoolean, sizeof(Boolean), &boolTrue);
		PostEventToQueue(GetMainEventQueue(), mouseUpCarbon,
			kEventPriorityStandard);
	}
}

@end

