/*
 * Copyright (c) 2004, Steven Michaud, All Rights Reserved
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#import "Controller.h"

static OSStatus CreateCocoaWindow(JEPController *controller,
	NSDictionary *attributes, WindowRef parentWindow, Rect appletClip,
	int appletPosX, int appletPosY, unsigned appletWidth, unsigned appletHeight);

BOOL nsAppInited = FALSE;
BOOL nsAppTerminating = FALSE;
BOOL isCarbonApp = TRUE;
unsigned long osxVersion = 0;
BOOL isJava142 = FALSE;
BOOL isJava142Update1 = FALSE;
BOOL isJava142Update2 = FALSE;
BOOL isJava142Tiger = FALSE;
BOOL isJava150 = FALSE;
BOOL isJava160 = FALSE;
BOOL awtLibrariesLoaded = FALSE;
BOOL asyncInitializeApplet = FALSE;
BOOL debugVisibility = FALSE;
BOOL debugRelease = FALSE;
BOOL debugUpdates = FALSE;
BOOL wasInEventTrackingMode = FALSE;
NSThread *mainThread = nil;
CFRunLoopRef mainRunLoop = nil;
CFRunLoopSourceRef jepRunLoopSource = nil;
CFRunLoopSourceRef jepCRunLoopSource = nil;
BOOL everCreatedJVM = false;
NSMutableArray *controllerArray = nil;
NSRecursiveLock *controllersLock = nil;
NSMutableArray *popupsArray = nil;
NSRecursiveLock *popupsLock = nil;
NSAppShadow *nsAppShadow = nil;
void (*afterInitAWT) (void *) = nil;
void *afterInitAWTInfo = nil;

NSString *installedPath = nil;
NSString *homeDirectory = nil;
NSString *javaPluginCocoaPath = nil;

/* CancelMenuTracking() is only available starting with OS X 10.3, so we look
 * it up dynamically before we use it.  Fortunately, it's also only needed
 * starting with OS X 10.3.
 */
OSStatus (*CancelMenuTrackingPtr) (MenuRef, Boolean, UInt32) = nil;

/* This class replicates methods that are already present in Handlers.m.  We
 * need it because we can't use PoseAsUndocumentedEvent() on NSTSMInputContext
 * on Tiger.  See JEPController checkBundle (below) for more information.
 * See the originals in Handlers.m for what these methods do and why we need
 * them.
 */
@implementation MyNSTSMInputContext

- (BOOL)_activate
{
	int retval;
	TSMDocumentID activeDoc;
	TSMDocumentID shouldActivateDoc;
	Boolean boolTrue = true;
	struct objc_super target;
	target.receiver = self;
	target.class = [MyNSTSMInputContext superclass];
	retval = (int) objc_msgSendSuper(&target, _cmd);
	/* When used in a Carbon browser, NSTSMInputContext _activate should
	 * always result in a "Cocoa-specific" TSMDocument being made active.
	 * So we mark the current active TSMDocument with a property that
	 * indicates this, and which we can check in other locations.
	 */
	if ((activeDoc = TSMGetActiveDocument()) != nil) {
		TSMSetDocumentProperty(activeDoc, kPropertyCocoaTSMDoc,
			sizeof(Boolean), &boolTrue);
	}
	shouldActivateDoc =
		[JEPController getWindowTSMDocIfShouldBeActive:GetUserFocusWindow()
		timeoutType:lockTimeoutNone];
	if (shouldActivateDoc)
		ActivateTSMDocument(shouldActivateDoc);
#ifdef DEBUG_EVENTS
	NSLog(@"NSTSMInputContext _activate: active TSM doc is 0x%.8X, shouldActivateDoc is 0x%.8X",
		activeDoc, shouldActivateDoc);
	activeDoc = TSMGetActiveDocument();
	NSLog(@"NSTSMInputContext _activate called on %@, active TSM doc is now 0x%.8X",
		self, activeDoc);
#endif
	return (BOOL) retval;
}

@end

static void keyDownBreak()
{
	return;
}
static void keyUpBreak()
{
	return;
}

@implementation MyNSApplication

/* When you call Toolkit.getPrintJob(), NSApplication calls the applet's
 * window's runModalForCarbonWindow method instead of its runModalForWindow
 * method (as it would for an ordinary NSWindow object).  This in turn calls
 * _realDoModalLoopForCarbonWindow, which is broken -- the modal window it
 * creates doesn't accept any keyboard or mouse input.  Fortunately we can fix
 * the problem by re-routing calls from this method to _realDoModalLoop.  This
 * is still a problem on Tiger.  Before JEP 0.9.6.1 we re-routed
 * runModalForCarbonWindow to runModalForWindow.  But on OS X 10.4.X this
 * causes the Page Setup dialog to crash on exit if you've previously loaded
 * at least one signed applet in your (Carbon-based) browser session!
 * http://sourceforge.net/tracker/index.php?func=detail&aid=1670831&group_id=107955&atid=649116
 */
- (void)_realDoModalLoopForCarbonWindow:(NSModalSession)session peek:(BOOL)flag
{
	struct objc_super target;
	target.receiver = self;
	target.class = [MyNSApplication superclass];
	objc_msgSendSuper(&target, @selector(_realDoModalLoop:peek:), session, flag);
}

- (void)sendEvent:(NSEvent *)theEvent
{
	struct objc_super target;
	NSEventType type = [theEvent type];

#ifdef DEBUG_EVENTS
	if ((type != NSMouseMoved) && (type != NSCursorUpdate))
		NSLog(@"NSApplication event %@, from %@, selector %@", theEvent, self, NSStringFromSelector(_cmd));
#endif

	if (debugUpdates) {
		NSString *info = nil;
		if (type == NSKeyDown) {
			info = [NSString stringWithFormat:
				@"%@ NSApplication sendEvent: Key down", [NSDate date]];
			[AppletView printMessage:info javaEnv:nil];
			keyDownBreak();
		} else if (type == NSKeyUp) {
			info = [NSString stringWithFormat:
				@"%@ NSApplication sendEvent: Key up", [NSDate date]];
			[AppletView printMessage:info javaEnv:nil];
			keyUpBreak();
		}
	}
	if (!isCarbonApp) {
		target.receiver = self;
		target.class = [MyNSApplication superclass];
		objc_msgSendSuper(&target, _cmd, theEvent);
		return;
	}

	/* The following is necessary to stop a bad interaction with CocoaGestures
	 * (http://www.bitart.com/CocoaGestures.html) -- it seems that when this
	 * program is installed (in /Library/InputManagers/), _all_ messages sent
	 * to this method use the cocoaGesturesSendEvent: selector, instead of the
	 * one they should be using.  If we do nothing, the browser will crash on
	 * startup with the error "-[NSApplication cocoaGesturesSendEvent:]:
	 * selector not recognized".  Or if we simply ignore the event, keyboard
	 * input will stop working in applets.
	 */
	if (_cmd == @selector(cocoaGesturesSendEvent:))
		_cmd = @selector(sendEvent:);

	target.receiver = self;
	target.class = [MyNSApplication superclass];
	objc_msgSendSuper(&target, _cmd, theEvent);
}

- (NSEvent *)nextEventMatchingMask:(unsigned int)mask
	untilDate:(NSDate *)expiration inMode:(NSString *)mode dequeue:(BOOL)flag
{
	NSEvent *retval;
	struct objc_super target;
	/* Code in Handlers.m's HandleDispatcher() needs to keep track of when a
	 * left-mouse-down event causes an NSEventTrackingRunLoopMode to be
	 * entered as a Cocoa event is being processed.
	 */
	if ([NSEventTrackingRunLoopMode isEqual:mode])
		wasInEventTrackingMode = true;
	target.receiver = self;
	target.class = [MyNSApplication superclass];
	retval = objc_msgSendSuper(&target, _cmd, mask, expiration, mode, flag);
	return retval;
}

@end

static void JEPRunLoopSource(void *info)
{
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	[AppletView processCustomEventSpool:false];
	[localPool release];
}

static void JEPCRunLoopSource(void *info)
{
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	[AppletView processCEventSpool:false];
	[localPool release];
}

/* Install a CFRunLoopSource to handle our custom event spool.  This is _much_
 * better than what we did prior to JEP 0.9.5+e.
 */
static void InstallRunLoopSources(void)
{
	CFRunLoopSourceContext context = {0};
	CFRunLoopSourceContext cContext = {0};
	context.perform = JEPRunLoopSource;
	cContext.perform = JEPCRunLoopSource;
	mainRunLoop = CFRunLoopGetCurrent();
	CFRetain(mainRunLoop);
	if ((jepRunLoopSource =
		CFRunLoopSourceCreate(kCFAllocatorDefault, 0, &context)) != nil)
	{
		CFRunLoopAddSource(mainRunLoop, jepRunLoopSource,
			kCFRunLoopCommonModes);
		/* If we don't also add our run loop source to the
		 * "AWTRunLoopMode", we occasionally (on OS X Tiger) hang in
		 * Java_apple_awt_CToolkit_doAWTRunLoop() (in libawt.jnilib).
		 */
		CFRunLoopAddSource(mainRunLoop, jepRunLoopSource,
			CFSTR("AWTRunLoopMode"));
		CFRunLoopWakeUp(mainRunLoop);
	}
	if ((jepCRunLoopSource =
		CFRunLoopSourceCreate(kCFAllocatorDefault, 0, &cContext)) != nil)
	{
		CFRunLoopAddSource(mainRunLoop, jepCRunLoopSource,
			kCFRunLoopCommonModes);
	}
}

@implementation MyNSWindow

- (void)sendEvent:(NSEvent *)theEvent
{
	NSEventType type = [theEvent type];
	JEPController *controller = nil;
	struct objc_super target;
#ifdef DEBUG_EVENTS
	if ((type != NSMouseMoved) && (type != NSCursorUpdate))
		NSLog(@"NSWindows event %@, from %@", theEvent, self);
#endif
	if (isCarbonApp && ((type == NSLeftMouseDown) ||
		(type == NSRightMouseDown) || (type == NSOtherMouseDown)))
	{
		/* See handleMouseDownEvent_CW() in Handlers.m for more info. */
		controller = [JEPController controllerFromNSWindow:self
			timeoutType:lockTimeoutNone];
		if (controller && controller->suspendMouseDown) {
#ifdef DEBUG_EVENTS
			NSLog(@"NSWindow sendEvent: Skipping mouse down event %@",
				theEvent);
#endif
			return;
		}
	}
	target.receiver = self;
	target.class = [MyNSWindow superclass];
	objc_msgSendSuper(&target, _cmd, theEvent);
}

/* In a Carbon browser, lockFocus can sometimes be called on an NSView object
 * that has already been at least partially destroyed, leading to a Cocoa
 * error and crash.  These inappropriate lockFocus calls are made from
 * NSView:displayIfNeeded, which is in turn called from NSWindow:displayIfNeeded.
 * Here we prevent NSWindow:displayIfNeeded being used on an NSCarbonWindow
 * object containing an AppletView that is being destroyed.
 */
- (void)displayIfNeeded
{
	WindowRef carbonWindow;
	JEPController *controller;
	struct objc_super target;
	if (isCarbonApp) {
		/* Also skip if browser is terminating.
		 */
		if (nsAppTerminating)
			return;
		/* On OS X 10.2.8 we must also check whether the Carbon window
		 * equivalent of our Cocoa window is still valid.
		 */
		if ((osxVersion < 0x1030) && [self isKindOfClass:carbonWindowClass]) {
			if ((carbonWindow = [MyNSWindow existingCarbonWindowFor:self]) != nil) {
				if (!IsValidWindowPtr(carbonWindow))
					return;
			}
		}
		controller = [JEPController controllerFromNSWindow:self
			timeoutType:lockTimeoutNone];
		if (controller && [controller destroying])
			return;
	}
	target.receiver = self;
	target.class = [MyNSWindow superclass];
	objc_msgSendSuper(&target, _cmd);
}

/* We need to hook the "zoom" method to stop parts of applets from being
 * displayed in the wrong place when the user "zooms" a window.
 */
- (void)zoom:(id)sender
{
	unsigned i, count;
	JEPController *controller;
	NSMutableArray *controllersFound = nil;
	struct objc_super target;
	if (!isCarbonApp &&
		((controllersFound = [JEPController controllersInWindow:
		[MyNSWindow existingCarbonWindowFor:self]
		timeoutType:lockTimeoutNormal]) != nil))
	{
		count = [controllersFound count];
		for (i = 1; i <= count; ++i) {
			controller = [controllersFound objectAtIndex:i-1];
			if ([controller destroying])
				continue;
			[controller suspendAppletUpdates];
		}
	}
	target.receiver = self;
	target.class = [MyNSWindow superclass];
	objc_msgSendSuper(&target, _cmd, sender);
	if (controllersFound) {
		count = [controllersFound count];
		for (i = 1; i <= count; ++i) {
			controller = [controllersFound objectAtIndex:i-1];
			if ([controller destroying])
				continue;
			[controller resumeAppletUpdates];
		}
		[controllersFound removeAllObjects];
		[controllersFound release];
	}
}

/* Needed because of a change in JEPController hookExistingObjects -- it no
 * longer checks NSView objects in invisible windows.  So we check them here
 * when/if a window is made visible.
 */
- (void)_setVisible:(BOOL)flag
{
	Class *classPtr;
	NSString *className;
	NSView *aView, *contentView;
	NSMutableArray *viewArray;
	unsigned i, viewCount;
	struct objc_super target;

	target.receiver = self;
	target.class = [MyNSWindow superclass];
	objc_msgSendSuper(&target, _cmd, flag);

	if (!flag)
		return;
	if ((contentView = [self contentView]) == nil)
		return;
	if ((viewArray = [[NSMutableArray alloc] init]) == nil)
		return;

	[JEPController disableScreenUpdates];
	[viewArray addObject:contentView];
	[JEPController allSubviews:contentView container:viewArray];
	viewCount = [viewArray count];
	for (i = 1; i <= viewCount; ++i) {
		aView = [viewArray objectAtIndex:i - 1];
		className = NSStringFromClass([aView class]);
		if ([className isEqualToString:@"%NSView"])
			className = @"NSView";
		classPtr = (Class *) ((void *) aView);
		*classPtr = NSClassFromString(className);
	}
	[viewArray removeAllObjects];
	[JEPController enableScreenUpdates];
	[viewArray release];
}

+ (WindowRef) existingCarbonWindowFor:(NSWindow *)cocoaWindow
{
	if (![cocoaWindow _hasWindowRef])
		return nil;
	return [cocoaWindow windowRef];
}

@end

@implementation MyNSView

/* The following is a hack (new as of JEP 0.9.3, substantially revised in
 * JEP 0.9.5+e) to deal with the gross excess of Cocoa-based screen updates
 * that recent versions of Camino (e.g. 0.9a1 through 1.0) make to Java
 * applets' screen space.  (It's a bit like what I've already done for the
 * Mozilla-family Carbon-based browsers in handleCarbonUpdate.)  If we ignore
 * these updates the applets often disappear from the screen.  If we don't
 * ignore them, applets can be rapidly (and annoyingly) updated many times in
 * succession.  Here we examine NSView objects "above" our applets (in the
 * NSView hierarchy) to see if we should either skip the entire update, or
 * skip the update and then repost those parts of it that don't overlap our
 * applets.  Before JEP 0.9.5+b we hooked _drawRect:clip: to accomplish this.
 * But I've now found that we get much better results (on OS X 10.3.X and
 * above) if we intervene earlier and hook setNeedsDisplayInRect: instead.
 * As with _drawRect, almost all NSView subclasses (except for NSScrollView
 * and NSTextView) use the NSView class method -- which makes it a convenient
 * place to intervene.  But using setNeedsDisplayInRect doesn't work on Jaguar
 * (OS X 10.2.8) -- it seems that Jaguar has code to consolidate update
 * rectangles (and thereby expand them), which gets invoked after this method
 * (so spurious updates can occur that don't get blocked here).  Even the
 * _drawRect code, though, is much improved over previous JEP versions (before
 * JEP 0.9.5+e) -- because it can repost parts of an update and (therefore) no
 * longer relies on (temporarily) suspending all updates.
 */
- (void)setNeedsDisplayInRect:(NSRect)invalidRect
{
	unsigned i, count;
	NSRect updateRect;
	NSView *contentView;
	NSWindow *cocoaWindow;
	WindowRef carbonWindow;
	JEPController *controller;
	NSMutableArray *controllersFound;
	BOOL checkUpdate = true;
	BOOL skipUpdate = false;
	struct objc_super target;

	if (!nsAppShadow || isCarbonApp || (osxVersion < 0x1030))
		checkUpdate = false;
	else if ((invalidRect.size.width <= 0) || (invalidRect.size.height <= 0))
		checkUpdate = false;
	else if ((cocoaWindow = [self window]) == nil)
		checkUpdate = false;
	else if ((carbonWindow = [MyNSWindow existingCarbonWindowFor:cocoaWindow]) == nil)
		checkUpdate = false;
	else if ((contentView = [cocoaWindow contentView]) == nil)
		checkUpdate = false;
	/* We sometimes get into trouble (needed updates are suppressed) if we
	 * try to skip updates in our window's contentView (which is all our
	 * AppletViews' superview).
	 */
	else if ([self isEqual:contentView])
		checkUpdate = false;
	/* Only check updates to views that aren't "ours".  But also stop "our"
	 * updates if they're being sent to a zero-clipped applet.  This helps
	 * get around problems introduced by the "fix" for Bugzilla bug 298961.
	 * For more info see AppletView _propagateDirtyRectsToOpaqueAncestors.
	 */
	else if ((controller = [JEPController
		controllerContainingNSView:self timeoutType:lockTimeoutNormal]) != nil)
	{
		if (![controller->myAppletView isDisplayable]) {
			if (debugUpdates) {
				NSString *info = [NSString stringWithFormat:
					@"%@ NSView setNeedsDisplayInRect: Stopping update to non-displayable applet %@",
					[NSDate date], controller->myAppletView->myAppletName];
				[AppletView printMessage:info javaEnv:nil];
			}
			skipUpdate = true;
		}
		checkUpdate = false;
	} else if ((controllersFound = [JEPController controllersInWindow:carbonWindow timeoutType:lockTimeoutNormal]) == nil)
		checkUpdate = false;

	if (checkUpdate) {
		// Convert invalidRect to something (in updateRect) that's comparable
		// to JEPController's currentFrame -- which has our window's (i.e.
		// it's content view's) coordinate system, with a bottom-left origin
		// (since we're in a Cocoa app).
		updateRect = [contentView convertRect:invalidRect fromView:self];
		count = [controllersFound count];
		for (i = 1; i <= count; ++i) {
			NSRect outsideRects[4] = {NSZeroRect};
			unsigned outsideRectsNum = 0;
			controller = [controllersFound objectAtIndex:i-1];
			if (controller->myWindowClosing)
				break;
			if ([controller destroying] || !controller->myAppletView)
				continue;
			if (controller->myAppletView->isInvisible)
				continue;
			if ([controller->myAppletView isCurrentClipEmpty])
				continue;
			if (![controller->myAppletView isAppletVisible])
				continue;
			if (controller->updatesOff)
				break;
			if (!SectNSRectFuzzy(updateRect, controller->currentFrame, nil, 0))
				continue;
		    // Skip updates that match one of our applets.
			if (ContainsNSRectFuzzy(controller->currentFrame, updateRect, 2)) {
				if (debugUpdates) {
					NSString *info = [NSString stringWithFormat:
						@"%@ NSView setNeedsDisplayInRect: Skipping spurious update in %@",
						[NSDate date], controller->myAppletView->myAppletName];
					[AppletView printMessage:info javaEnv:nil];
				}
				skipUpdate = true;
				break;
			// Also skip updates that partially match one of our applets, but
			// repost updates for those parts of invalidRect that don't match
			// it.  Break on the first partial match.  Other applets on the
			// same page will be dealt with when our reposted updates come
			// through here.
			} else if (DiffNSRect(updateRect, controller->currentFrame,
				&outsideRectsNum, outsideRects, 0))
			{
				CocoaUpdateParams *params;
				if (outsideRectsNum) {
					params = [[CocoaUpdateParams alloc]
						initWithParams:contentView updateNSRects:outsideRects
						numUpdateNSRects:outsideRectsNum];
					if (params) {
						[nsAppShadow performSelectorOnMainThread:
							@selector(postCocoaUpdateOnMainThread:)
							withObject:params waitUntilDone:false];
					}
				}
				if (debugUpdates) {
					NSString *info = [NSString stringWithFormat:
						@"%@ NSView setNeedsDisplayInRect: Skipping spurious part of update in %@",
						[NSDate date], controller->myAppletView->myAppletName];
					[AppletView printMessage:info javaEnv:nil];
				}
				skipUpdate = true;
				break;
			}
		}
		[controllersFound removeAllObjects];
		[controllersFound release];
	}

	if (skipUpdate)
		return;

	target.receiver = self;
	target.class = [MyNSView superclass];
	objc_msgSendSuper(&target, _cmd, invalidRect);
}

- (void)_drawRect:(NSRect)aRect clip:(BOOL)clip
{
	unsigned i, count;
	NSRect updateRect;
	NSView *contentView;
	NSWindow *cocoaWindow;
	WindowRef carbonWindow;
	JEPController *controller;
	NSMutableArray *controllersFound;
	BOOL checkUpdate = true;
	BOOL skipUpdate = false;
	struct objc_super target;

	if (!nsAppShadow || isCarbonApp || (osxVersion >= 0x1030))
		checkUpdate = false;
	else if ((aRect.size.width <= 0) || (aRect.size.height <= 0))
		checkUpdate = false;
	else if ((cocoaWindow = [self window]) == nil)
		checkUpdate = false;
	else if ((carbonWindow = [MyNSWindow existingCarbonWindowFor:cocoaWindow]) == nil)
		checkUpdate = false;
	else if ((contentView = [cocoaWindow contentView]) == nil)
		checkUpdate = false;
	/* Only check updates to views that aren't "ours".  But also stop "our"
	 * updates if they're being sent to a zero-clipped applet.  This helps
	 * get around problems introduced by the "fix" for Bugzilla bug 298961.
	 * For more info see AppletView _propagateDirtyRectsToOpaqueAncestors.
	 */
	else if ((controller = [JEPController
		controllerContainingNSView:self timeoutType:lockTimeoutNormal]) != nil)
	{
		if (![controller->myAppletView isDisplayable]) {
			if (debugUpdates) {
				NSString *info = [NSString stringWithFormat:
					@"%@ NSView _drawRect: Stopping update to non-displayable applet %@",
					[NSDate date], controller->myAppletView->myAppletName];
				[AppletView printMessage:info javaEnv:nil];
			}
			skipUpdate = true;
		}
		checkUpdate = false;
	} else if ([JEPController controllerContainingNSView:self timeoutType:lockTimeoutNormal] != nil)
		checkUpdate = false;
	else if ((controllersFound = [JEPController controllersInWindow:carbonWindow timeoutType:lockTimeoutNormal]) == nil)
		checkUpdate = false;

	if (checkUpdate) {
		// Convert aRect to something (in updateRect) that's comparable
		// to JEPController's currentFrame -- which has our window's (i.e.
		// it's content view's) coordinate system, with a bottom-left origin
		// (since we're in a Cocoa app).
		updateRect = [contentView convertRect:aRect fromView:self];
		count = [controllersFound count];
		for (i = 1; i <= count; ++i) {
			NSRect outsideRects[4] = {NSZeroRect};
			unsigned outsideRectsNum = 0;
			controller = [controllersFound objectAtIndex:i-1];
			if (controller->myWindowClosing)
				break;
			if ([controller destroying] || !controller->myAppletView)
				continue;
			if (controller->myAppletView->isInvisible)
				continue;
			if ([controller->myAppletView isCurrentClipEmpty])
				continue;
			if (![controller->myAppletView isAppletVisible])
				continue;
			if (controller->updatesOff)
				break;
			if (!SectNSRectFuzzy(updateRect, controller->currentFrame, nil, 0))
				continue;
		    // Skip updates that match one of our applets.
			if (ContainsNSRectFuzzy(controller->currentFrame, updateRect, 2)) {
				if (debugUpdates) {
					NSString *info = [NSString stringWithFormat:
						@"%@ NSView _drawRect: Skipping spurious update in %@",
						[NSDate date], controller->myAppletView->myAppletName];
					[AppletView printMessage:info javaEnv:nil];
				}
				skipUpdate = true;
				break;
			// Also skip updates that partially match one of our applets, but
			// repost updates for those parts of aRect that don't match
			// it.  Break on the first partial match.  Other applets on the
			// same page will be dealt with when our reposted updates come
			// through here.
			} else if (DiffNSRect(updateRect, controller->currentFrame,
				&outsideRectsNum, outsideRects, 0))
			{
				CocoaUpdateParams *params;
				if (outsideRectsNum) {
					params = [[CocoaUpdateParams alloc]
						initWithParams:self updateNSRects:outsideRects
						numUpdateNSRects:outsideRectsNum];
					if (params) {
						/* Prior to JEP 0.9.6.1 this was a call to 
						 * doCocoaUpdateOnMainThread.  I changed it because
						 * I noticed hangs on this in OS X 10.2.8.
						 */
						[nsAppShadow performSelectorOnMainThread:
							@selector(postCocoaUpdateOnMainThread:)
							withObject:params waitUntilDone:false];
					}
				}
				if (debugUpdates) {
					NSString *info = [NSString stringWithFormat:
						@"%@ NSView _drawRect: Skipping spurious part of update in %@",
						[NSDate date], controller->myAppletView->myAppletName];
					[AppletView printMessage:info javaEnv:nil];
				}
				skipUpdate = true;
				break;
			}
		}
		[controllersFound removeAllObjects];
		[controllersFound release];
	}

	if (skipUpdate)
		return;

	target.receiver = self;
	target.class = [MyNSView superclass];
	objc_msgSendSuper(&target, _cmd, aRect, clip);
}

/* This helps get around problems introduced by the "fix" for Bugzilla bug
 * 298961 -- by stopping updates to applets that have been zero-clipped.
 * For more info see AppletView _propagateDirtyRectsToOpaqueAncestors.
 */
- (void)setNeedsDisplay:(BOOL)flag
{
	JEPController *controller;
	struct objc_super target;
	if (!isCarbonApp && flag &&
		((controller = [JEPController controllerContainingNSView:self
		timeoutType:lockTimeoutNormal]) != nil))
	{
		if (![controller->myAppletView isDisplayable]) {
			if (debugUpdates) {
				NSString *info = [NSString stringWithFormat:
					@"%@ NSView setNeedsDisplay: Stopping update to non-displayable applet %@",
					[NSDate date], controller->myAppletView->myAppletName];
				[AppletView printMessage:info javaEnv:nil];
			}
			flag = false;
		}
	}
	target.receiver = self;
	target.class = [MyNSView superclass];
	objc_msgSendSuper(&target, _cmd, flag);
}

/* JEPController hookExistingObjects doesn't reset the Class of existing
 * NSView objects if they haven't yet been added to an NSView hierarchy.  So
 * we need to hook the two different addSubview methods to make sure we catch
 * these objects (those that hookExistingObjects has missed).  This fixes the
 * bug reported at https://bugzilla.mozilla.org/show_bug.cgi?id=302606.
 */
- (void)addSubview:(NSView *)aView
{
	Class clazz;
	Class *classPtr;
	NSString *className;
	struct objc_super target;
	if (!isCarbonApp && aView &&
		((className = NSStringFromClass([aView class])) != nil))
	{
		if ([className isEqualToString:@"%NSView"])
			className = @"NSView";
		if ((clazz = NSClassFromString(className)) != nil) {
			classPtr = (Class *) ((void *) aView);
			*classPtr = clazz;
		}
	}
	target.receiver = self;
	target.class = [MyNSView superclass];
	objc_msgSendSuper(&target, _cmd, aView);
}

- (void)addSubview:(NSView *)aView positioned:(NSWindowOrderingMode)place relativeTo:(NSView *)otherView
{
	Class clazz;
	Class *classPtr;
	NSString *className;
	struct objc_super target;
	if (!isCarbonApp && aView &&
		((className = NSStringFromClass([aView class])) != nil))
	{
		if ([className isEqualToString:@"%NSView"])
			className = @"NSView";
		if ((clazz = NSClassFromString(className)) != nil) {
			classPtr = (Class *) ((void *) aView);
			*classPtr = clazz;
		}
	}
	target.receiver = self;
	target.class = [MyNSView superclass];
	objc_msgSendSuper(&target, _cmd, aView, place, otherView);
}

@end

@implementation JavaRunner

- (id)init
{
	if ((self = [super init]) == nil)
		return nil;
	jrunnable = nil;
	exception = nil;
	done = false;
	abort = false;
	return self;
}

/* See JEPRunJRunnableOnMainThread() below for info. */
- (void)runJRunnableOnMainThread
{
	JNIEnv *env = nil;
	jclass runnableClass, nullExceptionClass;
	jmethodID run, newNullException;
	jthrowable exceptionLocal = nil;

	if (abort) return;

	if (![AppletView getJavaEnv:&env]) {
		done = true;
		return;
	}

	if (abort) return;

	runnableClass = (*env)->FindClass(env, "java/lang/Runnable");
	nullExceptionClass = (*env)->FindClass(env, "java/lang/NullPointerException");
	run = (*env)->GetMethodID(env, runnableClass, "run", "()V");
	newNullException = (*env)->GetMethodID(env, nullExceptionClass, "<init>", "()V");
	if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionClear(env);
		done = true;
		return;
	}

	if (!jrunnable) {
		exceptionLocal = (*env)->NewObject(env, nullExceptionClass, newNullException);
		exception = (*env)->NewGlobalRef(env, exceptionLocal);
		(*env)->DeleteLocalRef(env, exceptionLocal);
		if ((*env)->ExceptionCheck(env)) {
			(*env)->ExceptionClear(env);
			exception = nil;
			done = true;
			return;
		}
		done = true;
		return;
	}

	if (abort) return;

	(*env)->CallVoidMethod(env, jrunnable, run);

	if ((exceptionLocal = (*env)->ExceptionOccurred(env)) != nil) {
		(*env)->ExceptionClear(env);
		exception = (*env)->NewGlobalRef(env, exceptionLocal);
		(*env)->DeleteLocalRef(env, exceptionLocal);
		if ((*env)->ExceptionCheck(env)) {
			(*env)->ExceptionClear(env);
			exception = nil;
			done = true;
			return;
		}
	}

	(*env)->DeleteLocalRef(env, runnableClass);
	(*env)->DeleteLocalRef(env, nullExceptionClass);
	exception = exceptionLocal;
	done = true;
	return;
}

@end

/* We create one instance of this class to "shadow" NSApp -- to stay alive
 * as long as NSApp is running.  It receives notification when NSApp is
 * about to die ... and does a whole bunch of other things besides.
 */
@implementation NSAppShadow

/* In a Carbon app, this won't get called unless we call it ourselves (from
 * HandleApp() or HandleAEQuit() in Handlers.m), even if we make ourselves a
 * delegate of NSApp or register for a notification -- Cocoa event handling
 * is completely broken in NSCarbonWindow objects.  (Somehow in Mozilla it
 * _still_ doesn't get called, usually:  I've only seen it happen when you
 * quit Mozilla while a non-browser child window (e.g. a Java window) is
 * active.  This could be a bug in Mozilla.)
 */
- (void)applicationWillTerminate:(NSNotification *)notification
{
	NSInvocation *message = nil;
	NSMethodSignature *sig = nil;
	NSNotification *arg = nil;
	if (!nsAppShadow || nsAppTerminating)
		return;
	nsAppShadow = nil;
	nsAppTerminating = true;
#ifdef DEBUG_EVENTS
	NSLog(@"NSAppShadow applicationWillTerminate: Calling");
#endif
	if (isCarbonApp) {
		RemoveTSMAEEventHandler();
		RemoveAppEventHandlers();
		/* We only need to explicitly call windowWillClose in a Carbon browser.
		 * Before JEP 0.9.6 we (unnecessarily) also did it in Cocoa browsers,
		 * which (in 0.9.5 on Mac OS X 10.4.X, at least) caused a hang if you
		 * quit while the current page contained an applet that had failed to
		 * load.
		 */
		sig = [JEPController instanceMethodSignatureForSelector:
			@selector(windowWillClose:)];
		if (sig)
			message = [NSInvocation invocationWithMethodSignature:sig];
		if (message) {
			[message setSelector:@selector(windowWillClose:)];
			[message setArgument:&arg atIndex:2];
			[JEPController sendToAllControllers:message targetIsMyWindow:false
				timeoutType:lockTimeoutNone];
		}
	} else {
		[[NSNotificationCenter defaultCenter] removeObserver:self];
	}
	if (allocCAWLock && deletedCAWs) {
		[allocCAWLock lock];
		[deletedCAWs removeAllObjects];
		[allocCAWLock unlock];
	}
	[self autorelease];
}

/* The only reason this is here is for convenience -- we can call it very
 * early, in JEPController : checkBundle.
 */
- (void)setMainThreadVariable
{
	mainThread = [NSThread currentThread];
}

/* Another convenience method. */
- (void)setMainJavaThread
{
	[AppletView setMainJavaThread:nil];
}

/* Called on the main thread from routeEvent below.  On OS X 10.3.X and 10.4,
 * the root menu won't get closed if it's currently open and the user clicks
 * anywhere in a browser window.  (I used to think that this only happened
 * when the user clicked in a browser window outside any applet ... but
 * that's not true.)  So we must do this ourselves.  We need the dynamic
 * symbol lookup because CancelMenuTracking() is only available starting with
 * OS X 10.3.  We get crashes in Tiger (OS X 10.4) unless we perform this
 * after routeEvent has returned (but still on the main thread).
 */
- (void)cancelMenuTracking
{
	NSSymbol symbol = nil;
	if (!CancelMenuTrackingPtr) {
		if (symbol = NSLookupAndBindSymbol("_CancelMenuTracking")) {
			CancelMenuTrackingPtr = (OSStatus (*) (MenuRef, Boolean, UInt32))
				NSAddressOfSymbol(symbol);
		}
	}
	if (CancelMenuTrackingPtr)
		CancelMenuTrackingPtr(AcquireRootMenu(), true, 0);
}

- (void)dismissJavaPopupMenu
{
	PopupInfo *info;
	NSSymbol symbol = nil;
	if (!CancelMenuTrackingPtr) {
		if (symbol = NSLookupAndBindSymbol("_CancelMenuTracking")) {
			CancelMenuTrackingPtr = (OSStatus (*) (MenuRef, Boolean, UInt32))
				NSAddressOfSymbol(symbol);
		}
	}
	if (!CancelMenuTrackingPtr)
		return;
	if ((info = [PopupInfo getVisible]) == nil)
		return;
	CancelMenuTrackingPtr(_NSGetCarbonMenu([info getMenu]), true, 0);
	[PopupInfo setVisible:nil];
}

/* Our "hybrid view" (usually (always?) our immediate subview) doesn't always
 * get resized when we do.  This code fixes the problem.  (Before JEP 0.9.6 we
 * used to give OS X Tiger's Java versions special treatment, by making a
 * delayed call from here to updateCarbonWindow (which posted a
 * kEventWindowUpdate Carbon event).  But this didn't always work, and now it
 * seems to be unnecesary.)  (We also used to reset currentBounds and
 * currentFrame here.  But that turns out to be unnecessary, and sometimes
 * even harmful -- because doing this resets the current clip, it can cause an
 * applet to be displayed even after it's been made "invisible" by zero-
 * clipping.  For an example go to http://194.65.135.122/, switch from the
 * "Hierarchia" tab to either the "Favoritas" or "Pesquisas" tab, and resize
 * the page.)
 */
- (void)resetHybridView:(JEPController *)controller
{
	NSView *hybridView;
	NSRect appletViewFrame;
	NSArray *hvSubviews = nil;
	NSView *oneView = nil;
	if (!controller || !controller->appletWindow ||
		!controller->myAppletView || [controller destroying])
		return;

	appletViewFrame = [controller->myAppletView frame];
	[controller->myAppletView setBoundsSize:appletViewFrame.size];
	hybridView = [controller->appletWindow hybridView];
	if (hybridView) {
		[hybridView setFrameSize:appletViewFrame.size];
		/* hybridView should have exactly one subview -- an NSViewAWT object
		 * corresponding to our applet.  If it does, also resize that object.
		 */
		if ((hvSubviews = [hybridView subviews]) != nil) {
			if ([hvSubviews count] == 1) {
				oneView = (NSView *) [hvSubviews objectAtIndex:0];
				[oneView setFrameSize:appletViewFrame.size];
			}
		}
		/* hybridView's one subview may also have exactly one subview.  If
		 * so resize it, too.
		 */
		if (oneView && ((hvSubviews = [oneView subviews]) != nil)) {
			if ([hvSubviews count] == 1) {
				oneView = (NSView *) [hvSubviews objectAtIndex:0];
				[oneView setFrameSize:appletViewFrame.size];
			}
		}
	}
	[controller resumeAppletUpdates];
}

- (void)showURLOnMainThread:(ShowURLParams *)params
{
	JEPController *controller = nil;
	if (!params)
		return;
	if ((controller = [params getController]) == nil)
		return;
	if (!controller->showURLCallback || !controller->myAppletView ||
		[controller destroying])
		return;
	(*controller->showURLCallback)(controller->myAppletView->myApplet,
		(CFURLRef) [params getURL], (CFStringRef) [params getWindowName],
		controller->showURLUserData);
	[params autorelease];
}

- (void)showStatusOnMainThread:(ShowStatusParams *)params
{
	JEPController *controller = nil;
	NSString *statusMessage = nil;
	if (!params)
		return;
	if ((controller = [params getController]) == nil)
		return;
	if ((statusMessage = [params getMessage]) == nil)
		return;
	if (!controller->showStatusCallback || !statusMessage ||
		!controller->myAppletView || [controller destroying])
		return;
	(*controller->showStatusCallback)(controller->myAppletView->myApplet,
		(CFStringRef) statusMessage, controller->showStatusUserData);
	[params autorelease];
}

- (void)drawAppletOnMainThread:(JEPController *)controller
{
	NSView *contentView = nil;
	if (controller && !controller->drawAppletPending)
		return;
	if (!controller || !controller->myWindow ||
		!controller->myAppletView || controller->updatesOff ||
		[controller destroying] || ![controller->myAppletView isDisplayable])
	{
		if (controller)
			controller->drawAppletPending = false;
		return;
	}
	if (debugUpdates) {
		NSString *info = [NSString stringWithFormat:
			@"%@ JEPController drawAppletOnMainThread: About to redraw %@",
			[NSDate date], controller->myAppletView->myAppletName];
		[AppletView printMessage:info javaEnv:nil];
	}
	[controller->myAppletView setNeedsDisplayInRect:
		controller->myAppletView->currentBounds];
	if (controller->erasePreviousFrame &&
		!EqualNSRectFuzzy(controller->previousFrame, controller->currentFrame, 0) &&
		((contentView = [controller->myWindow contentView]) != nil))
		[contentView setNeedsDisplayInRect:controller->previousFrame];
}

- (void)suspendDrawAppletAfterDelay:(JEPController *)controller
{
	if (!controller || [controller destroying])
		return;
	++controller->suspendDrawAppletCounter;
}

/* suspendDrawAppletAfterDelay doesn't need to run on the main thread, but
 * performSelector:withObject:afterDelay doesn't work on Java threads.  So we
 * make sure that scheduleSuspendDrawApplet _does_ run on the main thread.
 */
- (void)scheduleSuspendDrawApplet:(JEPController *)controller
{
	if (!controller || [controller destroying])
		return;
	[self performSelector:@selector(suspendDrawAppletAfterDelay:)
		withObject:controller afterDelay:.5];
}

- (void)resumeDrawAppletAfterDelay:(JEPController *)controller
{
	if (!controller || [controller destroying])
		return;
	if (--controller->suspendDrawAppletCounter < 0)
		controller->suspendDrawAppletCounter = 0;
}

/* resumeDrawAppletAfterDelay doesn't need to run on the main thread, but
 * performSelector:withObject:afterDelay doesn't work on Java threads.  So we
 * make sure that scheduleResumeDrawApplet _does_ run on the main thread.
 */
- (void)scheduleResumeDrawApplet:(JEPController *)controller
{
	if (!controller || [controller destroying])
		return;
	[self performSelector:@selector(resumeDrawAppletAfterDelay:)
		withObject:controller afterDelay:.1];
}

- (void)createJavaVMOnMainThread:(CreateJavaVMParams *)params
{
	if (!params)
		return;
	[AppletView maybeCreateJavaVM:[params getExtraOpts]
		extraOptsNum:[params getExtraOptsNum]
		vmPtr:[params getVMPtr] envPtr:(void **)[params getEnvPtr]
		initAWT:[params getInitAWT]];
	[params release];
}

- (void)createCocoaWindowOnMainThread:(CreateCocoaWindowParams *)params
{
	JEPController *controller;
	if (!params)
		return;
	if ((controller = [params getController]) == nil) {
		[params release];
		return;
	}
	CreateCocoaWindow(controller, [params getAttributes],
		[params getParentWindow], [params getAppletClip],
		[params getAppletPosX], [params getAppletPosY],
		[params getAppletWidth], [params getAppletHeight]);
	[params release];
}

- (void)performWithNumberOnMainThread:(PerformWithNumberParams *)params
{
	SEL selector;
	JEPController *controller;
	if (!params)
		return;
	controller = [params getController];
	selector = [params getSelector];
	if (!controller || !selector) {
		[params release];
		return;
	}
	objc_msgSend(controller, selector, [params getNumber]);
	[params release];
}

- (void)sizeOnMainThread:(SizeParams *)params
{
	JEPController *controller;
	if (!params)
		return;
	if ((controller = [params getController]) == nil) {
		[params release];
		return;
	}
	objc_msgSend(controller, @selector(sizeAppletImpl:), [params getSize]);
	[params release];
}

- (void)positionOnMainThread:(PositionParams *)params
{
	JEPController *controller;
	if (!params)
		return;
	if ((controller = [params getController]) == nil) {
		[params release];
		return;
	}
	objc_msgSend(controller, @selector(positionAppletImpl:),
		[params getPosition]);
	[params release];
}

- (void)setClipOnMainThread:(ClipParams *)params
{
	JEPController *controller;
	if (!params)
		return;
	if ((controller = [params getController]) == nil) {
		[params release];
		return;
	}
	objc_msgSend(controller, @selector(setClipImpl:), [params getClip]);
	[params release];
}

- (void)moveAndClipOnMainThread:(MoveAndClipParams *)params
{
	JEPController *controller;
	NSRect appletFrame;
	NSRect clip;
	BOOL willZeroClip;
	if (!params)
		return;
	if ((controller = [params getController]) == nil) {
		[params release];
		return;
	}
	appletFrame = [controller->myAppletView frame];
	clip = [params getClip];
	if ((appletFrame.origin.x == clip.origin.x) &&
		(appletFrame.origin.y == clip.origin.y) &&
		(clip.origin.x == controller->currentFrame.origin.x) &&
		(clip.origin.y == controller->currentFrame.origin.y) &&
		(clip.size.width == controller->currentFrame.size.width) &&
		(clip.size.height == controller->currentFrame.size.height))
	{
		[params release];
		return;
	}
	willZeroClip = ((clip.size.width <= 0) || (clip.size.height <= 0) ||
		(clip.origin.x > appletFrame.origin.x+appletFrame.size.width) ||
		(clip.origin.x+clip.size.width < appletFrame.origin.x) ||
		(clip.origin.y > appletFrame.origin.y+appletFrame.size.height) ||
		(clip.origin.y+clip.size.height < appletFrame.origin.y));

	[controller suspendAppletUpdates];
	/* If we call positionApplet before setClip when we are about to zero-clip
	 * an applet that currently isn't zero-clipped, we may cause a subtle
	 * problem with the workaround for bmo bug 277067 in [AppletView setClip:]:
	 * If our call to positionApplet here changes the applet's position, then
	 * the update performed by NSAppShadow's postCarbonUpdateOnMainThread: or
	 * postCocoaUpdateOnMainThread: will effect the wrong "region" (the
	 * workaround's code will think that the new location (the one set here)
	 * is the "old location" (the one where the applet that's about to be
	 * zero-clipped is/was located)).
	 */
	if (willZeroClip && ![controller->myAppletView isCurrentClipEmpty]) {
		objc_msgSend(controller, @selector(setClipImpl:), clip);
		objc_msgSend(controller, @selector(positionAppletImpl:),
			[params getLocation]);
	} else {
		objc_msgSend(controller, @selector(positionAppletImpl:),
			[params getLocation]);
		objc_msgSend(controller, @selector(setClipImpl:), clip);
	}
	[controller resumeAppletUpdates];
	[params release];
}

static EventComparatorUPP IsUpdateForWindowUPP = nil;

static Boolean IsUpdateForWindow(EventRef inEvent, void *windowToMatch)
{
	UInt32 kind, class;
	WindowRef carbonWindow = nil;
	kind = GetEventKind(inEvent);
	class = GetEventClass(inEvent);
	if ((class != kEventClassWindow) || (kind != kEventWindowUpdate))
		return false;
	if (GetEventParameter(inEvent, kEventParamDirectObject,
		typeWindowRef, NULL, sizeof(WindowRef), NULL, &carbonWindow) != noErr)
		return false;
	if (windowToMatch != (void *) carbonWindow)
		return false;
	return true;
}

- (void)postCarbonUpdateOnMainThread:(CarbonUpdateParams *)params
{
	Rect boundsRect;
	GrafPtr savedPort;
	Point windowCorner = {0}, portOrigin = {0};
	Rect *updateRects;
	unsigned i, numUpdateRects;
	WindowRef carbonWindow;
	EventRef updateEvent;
	OSStatus retval;
	if (!IsUpdateForWindowUPP)
		IsUpdateForWindowUPP = NewEventComparatorUPP(IsUpdateForWindow);
	if (!params)
		return;
	if ((carbonWindow = [params getWindow]) == nil) {
		[params release];
		return;
	}
	if ((numUpdateRects = [params getNumUpdateRects]) == 0) {
		[params release];
		return;
	}
	if ((updateRects = [params getUpdateRects]) == nil) {
		[params release];
		return;
	}
	GetPort(&savedPort);
	SetPortWindowPort(carbonWindow);
	LocalToGlobal(&windowCorner);
	GetWindowBounds(carbonWindow, kWindowContentRgn, &boundsRect);
	portOrigin.h = windowCorner.h - boundsRect.left;
	portOrigin.v = windowCorner.v - boundsRect.top;
	for (i = 1; i <= numUpdateRects; ++i) {
		Rect updateRect = updateRects[i-1];
		OffsetRect(&updateRect, -portOrigin.h, -portOrigin.v);
		InvalWindowRect(carbonWindow, &updateRect);
	}
	retval = CreateEvent(nil, kEventClassWindow, kEventWindowUpdate, 0,
		kEventAttributeNone, &updateEvent);
	if (retval == noErr) {
		retval = SetEventParameter(updateEvent, kEventParamDirectObject,
			typeWindowRef, sizeof(carbonWindow), &carbonWindow);
	}
	if (retval == noErr) {
		FlushSpecificEventsFromQueue(GetMainEventQueue(),
			IsUpdateForWindowUPP, carbonWindow);
		PostEventToQueue(GetMainEventQueue(), updateEvent,
			kEventPriorityStandard);
	}
	SetPort(savedPort);
	[params release];
}

- (void)doCarbonUpdateOnMainThread:(CarbonUpdateParams *)params
{
	Rect boundsRect;
	GrafPtr savedPort;
	Point windowCorner = {0}, portOrigin = {0};
	Rect *updateRects;
	unsigned i, numUpdateRects;
	WindowRef carbonWindow;
	EventRef updateEvent;
	if (!IsUpdateForWindowUPP)
		IsUpdateForWindowUPP = NewEventComparatorUPP(IsUpdateForWindow);
	if (!params)
		return;
	if ((carbonWindow = [params getWindow]) == nil) {
		[params release];
		return;
	}
	if ((numUpdateRects = [params getNumUpdateRects]) == 0) {
		[params release];
		return;
	}
	if ((updateRects = [params getUpdateRects]) == nil) {
		[params release];
		return;
	}
	GetPort(&savedPort);
	SetPortWindowPort(carbonWindow);
	LocalToGlobal(&windowCorner);
	GetWindowBounds(carbonWindow, kWindowContentRgn, &boundsRect);
	portOrigin.h = windowCorner.h - boundsRect.left;
	portOrigin.v = windowCorner.v - boundsRect.top;
	if (CreateEvent(nil, kEventClassWindow, kEventWindowUpdate, 0,
		kEventAttributeNone, &updateEvent) != noErr)
		updateEvent = nil;
	if (updateEvent) {
		if (SetEventParameter(updateEvent, kEventParamDirectObject,
			typeWindowRef, sizeof(carbonWindow), &carbonWindow) != noErr)
			updateEvent = nil;
	}
	if (updateEvent) {
		BOOL alsoPost = false;
		EventRef updateEventCopy = CopyEvent(updateEvent);
		for (i = 1; i <= numUpdateRects; ++i) {
			Rect updateRect = updateRects[i-1];
			OffsetRect(&updateRect, -portOrigin.h, -portOrigin.v);
			InvalWindowRect(carbonWindow, &updateRect);
		}
		if (SendEventToWindow(updateEvent, carbonWindow) == eventNotHandledErr)
			alsoPost = true;
		FlushSpecificEventsFromQueue(GetMainEventQueue(),
			IsUpdateForWindowUPP, carbonWindow);
		if (updateEventCopy) {
			if (alsoPost) {
				PostEventToQueue(GetMainEventQueue(), updateEventCopy,
					kEventPriorityStandard);
			} else {
				ReleaseEvent(updateEventCopy);
			}
		}
	}
	SetPort(savedPort);
	[params release];
}

- (void)postCocoaUpdateOnMainThread:(CocoaUpdateParams *)params
{
	NSView *theView;
	NSRect *updateRects;
	unsigned i, numUpdateRects;
	if (!params)
		return;
	if ((theView = [params getView]) == nil) {
		[params release];
		return;
	}
	if (![theView window]) {
		[params release];
		return;
	}
	if ((numUpdateRects = [params getNumUpdateNSRects]) == 0) {
		[params release];
		return;
	}
	if ((updateRects = [params getUpdateNSRects]) == nil) {
		[params release];
		return;
	}
	for (i = 1; i <= numUpdateRects; ++i)
		[theView setNeedsDisplayInRect:updateRects[i-1]];
	[params release];
}

- (void)doCocoaUpdateOnMainThread:(CocoaUpdateParams *)params
{
	NSView *theView;
	NSView *contentView;
	NSWindow *ourWindow;
	NSRect *updateRects;
	unsigned i, numUpdateRects;
	if (!params)
		return;
	if ((theView = [params getView]) == nil) {
		[params release];
		return;
	}
	if ((ourWindow = [theView window]) == nil) {
		[params release];
		return;
	}
	if ((contentView = [ourWindow contentView]) == nil) {
		[params release];
		return;
	}
	if ((numUpdateRects = [params getNumUpdateNSRects]) == 0) {
		[params release];
		return;
	}
	if ((updateRects = [params getUpdateNSRects]) == nil) {
		[params release];
		return;
	}
	if (![theView lockFocusIfCanDraw]) {
		[params release];
		return;
	}
	for (i = 1; i <= numUpdateRects; ++i) {
		NSRect updateRect = [theView convertRect:updateRects[i-1] fromView:contentView];
		[theView _recursiveDisplayAllDirtyWithLockFocus:false visRect:updateRect];
	}
	[theView unlockFocus];
	[params release];
}

@end

@implementation ShowURLParams

/* We must make local copies of url and (if available) windowName -- otherwise
 * after a few seconds we'll start to get protection failures.
 */
- (id)initWithParams:(NSURL *)url windowName:(NSString *)windowName
	controller:(JEPController *)controller
{
	if ((self = [super init]) == nil)
		return nil;
	if (url) {
		myURL = [url copy];
	} else {
		myURL = nil;
	}
	if (windowName) {
		myWindowName = [windowName copy];
	} else {
		myWindowName = [[NSString alloc] initWithCString:"_self"];
	}
	if (controller) {
		myController = [controller retain];
	} else {
		myController = nil;
	}
	return self;
}

- (NSURL *)getURL
{
	return myURL;
}

- (NSString *)getWindowName
{
	return myWindowName;
}

- (JEPController *)getController
{
	return myController;
}

- (void)dealloc
{
	if (myURL)
		[myURL release];
	if (myWindowName)
		[myWindowName release];
	if (myController)
		[myController autorelease];
	[super dealloc];
}

@end

@implementation ShowStatusParams

- (id)initWithParams:(NSString *)message controller:(JEPController *)controller
{
	if ((self = [super init]) == nil)
		return nil;
	if (message) {
		myMessage = [message copy];
	} else {
		myMessage = nil;
	}
	if (controller) {
		myController = [controller retain];
	} else {
		myController = nil;
	}
	return self;
}

- (NSString *)getMessage
{
	return myMessage;
}

- (JEPController *)getController
{
	return myController;
}

- (void)dealloc
{
	if (myMessage)
		[myMessage release];
	if (myController)
		[myController autorelease];
	[super dealloc];
}

@end

@implementation CreateJavaVMParams

- (id)initWithParams:(JavaVMOption *)jvmExtraOpts
	extraOptsNum:(int)jvmExtraOptsNum vmPtr:(JavaVM **)vmPtr
	envPtr:(void **)envPtr initAWT:(BOOL)initAWT
{
	if ((self = [super init]) == nil)
		return nil;
	myExtraOpts = jvmExtraOpts;
	myExtraOptsNum = jvmExtraOptsNum;
	myVMPtr = vmPtr;
	myEnvPtr = (JNIEnv *) envPtr;
	myInitAWT = initAWT;
	return self;
}

- (JavaVMOption *)getExtraOpts
{
	return myExtraOpts;
}

- (int)getExtraOptsNum
{
	return myExtraOptsNum;
}

- (JavaVM **)getVMPtr
{
	return myVMPtr;
}

- (JNIEnv *)getEnvPtr
{
	return myEnvPtr;
}

- (BOOL)getInitAWT
{
	return myInitAWT;
}

@end

@implementation CreateCocoaWindowParams

- (id)initWithParams:(JEPController *)controller
	attributes:(NSDictionary *)attributes parentWindow:(WindowRef)parentWindow
	appletClip:(Rect)appletClip appletPosX:(int)appletPosX
	appletPosY:(int)appletPosY appletWidth:(unsigned)appletWidth
	appletHeight:(unsigned) appletHeight
{
	if ((self = [super init]) == nil)
		return nil;
	if (controller) {
		myController = [controller retain];
	} else {
		myController = nil;
	}
	if (attributes) {
		myAttributes = [attributes retain];
	} else {
		myAttributes = nil;
	}
	myParentWindow = parentWindow;
	myAppletClip = appletClip;
	myAppletPosX = appletPosX;
	myAppletPosY = appletPosY;
	myAppletWidth = appletWidth;
	myAppletHeight = appletHeight;
	return self;
}

- (JEPController *)getController
{
	return myController;
}

- (NSDictionary *)getAttributes
{
	return myAttributes;
}

- (WindowRef)getParentWindow
{
	return myParentWindow;
}

- (Rect)getAppletClip
{
	return myAppletClip;
}

- (int)getAppletPosX
{
	return myAppletPosX;
}

- (int)getAppletPosY
{
	return myAppletPosY;
}

- (unsigned)getAppletWidth
{
	return myAppletWidth;
}

- (unsigned)getAppletHeight
{
	return myAppletHeight;
}

- (void)dealloc
{
	if (myAttributes)
		[myAttributes release];
	if (myController)
		[myController autorelease];
	[super dealloc];
}

@end

@implementation PerformWithNumberParams

- (id)initWithParams:(JEPController *)controller selector:(SEL)selector
	number:(int)number
{
	if ((self = [super init]) == nil)
		return nil;
	if (controller) {
		myController = [controller retain];
	} else {
		myController = nil;
	}
	mySelector = selector;
	myNumber = number;
	return self;
}

- (JEPController *)getController
{
	return myController;
}

- (SEL)getSelector
{
	return mySelector;
}

- (int)getNumber
{
	return myNumber;
}

- (void)dealloc
{
	if (myController)
		[myController autorelease];
	[super dealloc];
}

@end

@implementation SizeParams

- (id)initWithParams:(JEPController *)controller size:(NSSize)size
{
	if ((self = [super init]) == nil)
		return nil;
	if (controller) {
		myController = [controller retain];
	} else {
		myController = nil;
	}
	mySize = size;
	return self;
}

- (JEPController *)getController
{
	return myController;
}

- (NSSize)getSize
{
	return mySize;
}

- (void)dealloc
{
	if (myController)
		[myController autorelease];
	[super dealloc];
}

@end

@implementation PositionParams

- (id)initWithParams:(JEPController *)controller position:(NSPoint)position
{
	if ((self = [super init]) == nil)
		return nil;
	if (controller) {
		myController = [controller retain];
	} else {
		myController = nil;
	}
	myPosition = position;
	return self;
}

- (JEPController *)getController
{
	return myController;
}

- (NSPoint)getPosition
{
	return myPosition;
}

- (void)dealloc
{
	if (myController)
		[myController autorelease];
	[super dealloc];
}

@end

@implementation ClipParams

- (id)initWithParams:(JEPController *)controller clip:(NSRect)clip
{
	if ((self = [super init]) == nil)
		return nil;
	if (controller) {
		myController = [controller retain];
	} else {
		myController = nil;
	}
	myClip = clip;
	return self;
}

- (JEPController *)getController
{
	return myController;
}

- (NSRect)getClip
{
	return myClip;
}

- (void)dealloc
{
	if (myController)
		[myController autorelease];
	[super dealloc];
}

@end

@implementation MoveAndClipParams

- (id)initWithParams:(JEPController *)controller location:(NSPoint)location
	clip:(NSRect)clip
{
	if ((self = [super init]) == nil)
		return nil;
	if (controller) {
		myController = [controller retain];
	} else {
		myController = nil;
	}
	myLocation = location;
	myClip = clip;
	return self;
}

- (JEPController *)getController
{
	return myController;
}

- (NSPoint)getLocation
{
	return myLocation;
}

- (NSRect)getClip
{
	return myClip;
}

- (void)dealloc
{
	if (myController)
		[myController autorelease];
	[super dealloc];
}

@end

@implementation CarbonUpdateParams

- (id)initWithParams:(WindowRef)window updateRects:(Rect *)updateRects
	numUpdateRects:(unsigned)numUpdateRects
{
	if ((self = [super init]) == nil)
		return nil;
	myWindow = nil;
	if (window && (RetainWindow(window) == noErr))
		myWindow = window;
	myNumUpdateRects = numUpdateRects;
	if ((myUpdateRects = updateRects) == nil)
		myNumUpdateRects = 0;
	return self;
}

- (WindowRef)getWindow
{
	return myWindow;
}

- (Rect *)getUpdateRects
{
	return myUpdateRects;
}

- (unsigned)getNumUpdateRects
{
	return myNumUpdateRects;
}

- (void)dealloc
{
	if (myUpdateRects)
		free(myUpdateRects);
	if (myWindow)
		ReleaseWindow(myWindow);
	[super dealloc];
}

@end

@implementation CocoaUpdateParams

- (id)initWithParams:(NSView *)aView updateNSRects:(NSRect *)updateNSRects
	numUpdateNSRects:(unsigned)numUpdateNSRects
{
	unsigned i;
	if ((self = [super init]) == nil)
		return nil;
	if (aView) {
		myView = [aView retain];
	} else {
		myView = nil;
	}
	if ((myNumUpdateNSRects = numUpdateNSRects) > 4)
		myNumUpdateNSRects = 4;
	memset(myUpdateNSRects, 0, sizeof(NSRect));
	if (!updateNSRects)
		myNumUpdateNSRects = 0;
	for (i = 1; i <= myNumUpdateNSRects; ++i)
		myUpdateNSRects[i-1] = updateNSRects[i-1];
	return self;
}

- (NSView *)getView
{
	return myView;
}

- (NSRect *)getUpdateNSRects
{
	return myUpdateNSRects;
}

- (unsigned)getNumUpdateNSRects
{
	return myNumUpdateNSRects;
}

- (void)dealloc
{
	if (myView)
		[myView autorelease];
	[super dealloc];
}

@end

/* This class (plus its associated variables and methods) is needed to deal
 * with a bug in how CPopupMenu objects are handled on OS X 10.3.X and 10.4.X
 * in Carbon-based browsers -- a CPopupMenu menu don't close when you click
 * outside it (anywhere in the current browser window).  (A CPopupMenu object
 * is a "model" corresponding to the "peer" of a Java PopupMenu object.)  This
 * bug seems to belong to the (very large) class of Cocoa-Carbon interface
 * bugs.  This class fixes bugzilla.mozilla.org bug 318424 (see
 * https://bugzilla.mozilla.org/show_bug.cgi?id=318424).  This problem seems
 * to be fixed in Java 1.6.0.
 */

@implementation PopupInfo

+ (void)addInfo:(PopupInfo *)info
{
	PopupInfo *oldInfo;
	if (!popupsArray || !popupsLock || !info)
		return;
	[popupsLock lock];
	if ((oldInfo = [PopupInfo infoForModel:[info getModel]]) != nil)
		[popupsArray removeObject:oldInfo];
	[popupsArray addObject:info];
	[popupsLock unlock];
}

+ (void)removeInfoForController:(JEPController *)controller
{
	unsigned i, count;
	PopupInfo *info;
	NSMutableArray *dead;
	if (!popupsArray || !popupsLock || !controller)
		return;
	[popupsLock lock];
	if ((dead = [[NSMutableArray alloc] init]) == nil) {
		[popupsLock unlock];
		return;
	}
	count = [popupsArray count];
	for (i = 1; i <= count; ++i) {
		info = [popupsArray objectAtIndex:i-1];
		if (![controller isEqual:[info getController]])
			continue;
		[dead addObject:info];
	}
	[popupsArray removeObjectsInArray:dead];
	[dead removeAllObjects];
	[dead release];
	[popupsLock unlock];
}

+ (PopupInfo *)infoForModel:(id)model
{
	PopupInfo *retval = nil;
	unsigned i, count;
	PopupInfo *info;
	if (!popupsArray || !popupsLock || !model)
		return nil;
	[popupsLock lock];
	count = [popupsArray count];
	for (i = 1; i <= count; ++i) {
		info = [popupsArray objectAtIndex:i-1];
		if (model != [info getModel])
			continue;
		retval = info;
		break;
	}
	[popupsLock unlock];
	return retval;
}

+ (PopupInfo *)getVisible
{
	PopupInfo *retval = nil;
	unsigned i, count;
	PopupInfo *info;
	if (!popupsArray || !popupsLock)
		return nil;
	[popupsLock lock];
	count = [popupsArray count];
	for (i = 1; i <= count; ++i) {
		info = [popupsArray objectAtIndex:i-1];
		if ([info isVisible]) {
			retval = info;
			break;
		}
	}
	[popupsLock unlock];
	return retval;
}

+ (void)setVisible:(id)model
{
	unsigned i, count;
	PopupInfo *info;
	if (!popupsArray || !popupsLock)
		return;
	[popupsLock lock];
	count = [popupsArray count];
	for (i = 1; i <= count; ++i) {
		info = [popupsArray objectAtIndex:i-1];
		if (model && (model == [info getModel])) {
			[info setVisible:true];
		} else {
			[info setVisible:false];
		}
	}
	[popupsLock unlock];
}

- (id)initWithParams:(id)model menu:(NSMenu *)menu controller:(JEPController *)controller;
{
	if ((self = [super init]) == nil)
		return nil;
	if (!model || !menu || !controller) {
		[self autorelease];
		return nil;
	}
	myModel = [model retain];
	myMenu = [menu retain];
	myController = [controller retain];
	mVisible = false;
	return self;
}

- (id)getModel
{
	return myModel;
}

- (NSMenu *)getMenu
{
	return myMenu;
}

- (JEPController *)getController
{
	return myController;
}

- (BOOL)isVisible
{
	return mVisible;
}

- (void)setVisible:(BOOL)visible
{
	mVisible = visible;
}

- (void)dealloc
{
	if (myModel)
		[myModel release];
	if (myMenu)
		[myMenu release];
	if (myController)
		[myController release];
	[super dealloc];
}

@end

/* When Mozilla (and presumably also Firefox) handles an activate event or a
 * resume event (old fashioned Event Manager events), it makes its own
 * TSMDocument active (see nsMacEventHandler::HandleActivateEvent() in
 * widget/src/mac/nsMacEventHandler.cpp).  But this will cause trouble if the
 * user focus is currently in an applet (because each input context in a Java
 * 1.4.X applet has its own TSMDocument).  So when (in HandleWindow() or
 * HandleApp() in Handlers.m) we observe an activate or resume event that we
 * know Mozilla will mis-handle, we create an ActivateInputContext object to
 * wait for that event to disappear from the event queue -- after which we can
 * reverse the damage Mozilla has caused.  Even if we're working with some
 * other (Carbon) browser that doesn't do this, a superfluous call to activate
 * one of our own TSMDocuments shouldn't cause any trouble.  (This problem
 * never happens with a Cocoa browser like Camino.)
 */
@implementation ActivateInputContext

- (id)initWithParams:(EventQueueRef)eventQueue event:(EventRef)event
	controller:(JEPController *)controller
{
	if ((self = [super init]) == nil)
		return nil;
	myEventQueue = eventQueue;
	myController = controller;
	if (event) {
		myEvent = RetainEvent(event);
	} else {
		myEvent = nil;
	}
	return self;
}

/* Run this on a secondary thread */
- (void)waitForEventComplete
{
	NSAutoreleasePool *localPool;
	[NSRunLoop currentRunLoop];
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	if (!myController || !myController->hasKeyWindow) {
		[localPool release];
		return;
	}
	if (myEventQueue && myEvent) {
		/* */
		while (IsEventInQueue(myEventQueue, myEvent))
			[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:.05]];
		/* */
		//NSTimeInterval interval = 0;
		//NSDate *started = [NSDate date];
		//for (;;) {
		//	if (!IsEventInQueue(myEventQueue, myEvent))
		//		break;
		//	[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:.05]];
		//	if ((interval = [started timeIntervalSinceNow]) < 0)
		//		interval = -interval;
		//	if (interval > 5)
		//		break;
		//}
		/* */
	}
	if (!myController->hasKeyWindow) {
		[localPool release];
		return;
	}
	[self performSelectorOnMainThread:@selector(performOnMainThread)
		withObject:nil waitUntilDone:FALSE];
	[localPool release];
	return;
}

- (void)performOnMainThread
{
	NSInputContext *currentContext;
	/* It's difficult to know _which_ of our applet's TSMDocuments to activate.
	 * But sending an _activate message to the current input context seems to
	 * do the right thing.
	 */
	currentContext = [NSInputContext currentInputContext];
	if (currentContext && myController && myController->hasKeyWindow)
		[currentContext _activate];
	[self autorelease];
}

- (void)dealloc
{
	if (myEvent)
		ReleaseEvent(myEvent);
	[super dealloc];
}

@end

void ShowEventInfo(CFNotificationCenterRef center, void *observer,
	CFStringRef name, const void *object, CFDictionaryRef userInfo)
{
	NSLog(@"CFCallback(): notification %@, poster %@, info %@",
		(NSString *) name, (NSString *) object,
		(NSDictionary *) userInfo);
}

void ShowClassInfo(Class ourClass)
{
	int count;
	int i, j, k, l;
	char *indent;
	NSString *ourClassName;
	Class aClass;
	struct objc_method *aMethod;
	BOOL isMetaclass = false;
	void *iterator;
	struct objc_method_list *methods;
	struct objc_protocol_list *protocols;

	if (ourClass == nil)
		return;

	if (ourClass->info & CLS_META)
		isMetaclass = true;

	if (isMetaclass) {
		ourClassName = [NSString stringWithFormat:@"%s (meta)", ourClass->name];
		indent = "  ";
	} else {
		ourClassName = [NSString stringWithFormat:@"%s", ourClass->name];
		indent = "";
		NSLog(@"---------------------------------------");
		NSLog(@"**Information For Class %@**", ourClassName);
	}

	/* Superclass */
	NSLog(@"%sSuperclass of class %@:", indent, ourClassName);
	NSLog(@"%s  %s", indent, ourClass->super_class->name);

	/* Version */
	NSLog(@"%sVersion of class %@:", indent, ourClassName);
	NSLog(@"%s  0x%.8X", indent, ourClass->version);

	/* Info */
	NSLog(@"%sInfo of class %@:", indent, ourClassName);
	NSLog(@"%s  0x%.8X", indent, ourClass->info);

	/* Instance size */
	NSLog(@"%sInstance size of class %@:", indent, ourClassName);
	NSLog(@"%s  %i", indent, ourClass->instance_size);

	/* Methods */
	if (!isMetaclass) {
		iterator = nil;
		if ((methods = class_nextMethodList(ourClass, &iterator)) != nil) {
			i = 1;
			NSLog(@"Instance methods of class %@:", ourClassName);
		}
		while (methods != nil) {
			NSLog(@"  Method list %i", i);
			for (j = 1; j <= methods->method_count; ++j) {
				NSLog(@"    %@", NSStringFromSelector(methods->method_list[j-1].method_name));
			}
			methods = class_nextMethodList(ourClass, &iterator);
			++i;
		}
		iterator = nil;
		if ((methods = class_nextMethodList(ourClass->isa, &iterator)) != nil) {
			i = 1;
			NSLog(@"Class methods of class %@:", ourClassName);
		}
		while (methods != nil) {
			NSLog(@"  Method list %i", i);
			for (j = 1; j <= methods->method_count; ++j) {
				NSLog(@"    %@", NSStringFromSelector(methods->method_list[j-1].method_name));
			}
			methods = class_nextMethodList(ourClass->isa, &iterator);
			++i;
		}
	}

	/* Protocols */
	if ((ourClass->protocols != nil) && (ourClass->protocols->count > 0)) {
		NSLog(@"%sProtocols of class %@:", indent, ourClassName);
		protocols = ourClass->protocols;
		i = 1;
		while (protocols != nil) {
			NSLog(@"%s  Protocol list %i", indent, i);
			count = protocols->count;
			for (j = 1; j <= count; ++j) {
				aClass = (struct objc_class *) protocols->list[j-1];
				NSLog(@"%s    Protocol %s:", indent, aClass->name);
				iterator = nil;
				if ((methods = class_nextMethodList(aClass, &iterator)) != nil) {
					k = 1;
					NSLog(@"%s      Instance methods of protocol %s:", indent, aClass->name);
				}
				while (methods != nil) {
					NSLog(@"%s        Method list %i", indent, k);
					for (l = 1; l <= methods->method_count; ++l) {
						NSLog(@"%s          %@", indent,
							NSStringFromSelector(methods->method_list[l-1].method_name));
					}
					methods = class_nextMethodList(aClass, &iterator);
					++k;
				}
				iterator = nil;
				if ((methods = class_nextMethodList(aClass->isa, &iterator)) != nil) {
					k = 1;
					NSLog(@"%s      Class methods of protocol %s:", indent, aClass->name);
				}
				while (methods != nil) {
					NSLog(@"%s        Method list %i", indent, k);
					for (l = 1; l <= methods->method_count; ++l) {
						NSLog(@"%s          %@", indent,
							NSStringFromSelector(methods->method_list[l-1].method_name));
					}
					methods = class_nextMethodList(aClass->isa, &iterator);
					++k;
				}
			}
			protocols = protocols->next;
			++i;
		}
	}

	/* Instance variables */
	if ((ourClass->ivars != nil) && (ourClass->ivars->ivar_count > 0)) {
		NSLog(@"%sInstance variables of class %@:", indent, ourClassName);
		for (i = 1; i <= ourClass->ivars->ivar_count; ++i) {
			NSLog(@"%s  %s (%s)", indent, ourClass->ivars->ivar_list[i-1].ivar_name,
				ourClass->ivars->ivar_list[i-1].ivar_type);
		}
	}

	/* Cached methods */
	if (!isMetaclass) {
		if (ourClass->cache && (ourClass->cache->occupied > 0)) {
			NSLog(@"Cached instance methods of class %@:", ourClassName);
			for (i = 1; i <= ourClass->cache->mask + 1; ++i) {
				if (ourClass->cache->buckets[i-1] == nil)
					continue;
				aMethod = (struct objc_method *) ourClass->cache->buckets[i-1];
				NSLog(@"  %@", NSStringFromSelector(aMethod->method_name));
			}
		}
		if (ourClass->isa->cache && (ourClass->isa->cache->occupied > 0)) {
			NSLog(@"Cached class methods of class %@:", ourClassName);
			for (i = 1; i <= ourClass->isa->cache->mask + 1; ++i) {
				if (ourClass->isa->cache->buckets[i-1] == nil)
					continue;
				aMethod = (struct objc_method *) ourClass->isa->cache->buckets[i-1];
				NSLog(@"  %@", NSStringFromSelector(aMethod->method_name));
			}
		}
	}

	if (isMetaclass) {
		NSLog(@"---------------------------------------");
	} else {
		NSLog(@"Information For Metaclass %@", ourClassName);
		ShowClassInfo(ourClass->isa);
	}
}

@implementation JEPController

/* Call this after loading our bundle to make sure it checks out OK.
 */
+ (BOOL)checkBundle
{
	NSBundle *javaVMFramework = nil;
	NSBundle *javaPluginCocoa = nil;
	NSArray *searchPaths;
	NSString *pathBase, *path;
	NSString *subpath = @"Internet Plug-Ins/JavaPluginCocoa.bundle";
	NSDictionary *infoPlist = nil;
	NSString *versionString = nil;
	NSString *webPluginDescriptionString = nil;
	NSComparisonResult result = 0;
	const struct mach_header *libawtImage = nil;
	id object = nil;
	unsigned i, count;

	if (nsAppInited) {
		if (installedPath)
			return true;
		return false;
	}

	nsAppInited = true;

	/* Initialize NSApplication, using an entry point that is specific to
	 * bundles.  This will be a no-op for Cocoa apps loading this bundle,
	 * but is required by Carbon apps.  This is not the final API.
	 */
	if (!NSApplicationLoad()) {
		NSLog(@"JEPController:checkBundle NSApplicationLoad() failed");
		return false;
	}

	Gestalt(gestaltSystemVersion, (long *) &osxVersion);
	osxVersion &= 0xFFFF;

	/* We used to use NSApp being non-null (before the call to
	 * NSApplicationLoad()) to determine whether we're running in a Cocoa
	 * app (as opposed to a Carbon one).  This seems to work fine in Jaguar
	 * and Panther, but it doesn't always work in Tiger (where certain user
	 * actions (such as choosing "Print" from the browser's main menu) can turn
	 * a Carbon browser into a half-Cocoa one (i.e. can run NSApplicationLoad()
	 * before we've had a chance to do it ourselves).  Fortunately there's an
	 * undocumented _NSIsCarbonApp() call in the AppKit Framework that's been
	 * present since Jaguar, at least.  It needs to be called after
	 * NSApplicationLoad() (which initializes its return value).
	 */
	isCarbonApp = _NSIsCarbonApp();

	InstallRunLoopSources();

	if (!isCarbonApp)
		[MyNSView poseAsClass:[NSView class]];
	[MyNSWindow poseAsClass:[NSWindow class]];
	[MyNSApplication poseAsClass:[NSApplication class]];

	/* At this point NSApp has already been created, with an "old" class
	 * pointer (one that points to the original NSApplication class, not
	 * our posing class).  As we noted above, NSApp sometimes (in Tiger)
	 * gets created even before JavaEmbeddingPlugin.bundle is loaded.  So
	 * the only option that covers all cases is to replace NSApp's
	 * original pointer with a new pointer to our posing class.  This
	 * doesn't seem to cause any trouble.
	 */
	[JEPController hookExistingObjects];

	/* Look for JavaPluginCocoa.bundle first in the user's home directory,
	 * then in the "local domain".  Its presence is required on Java 1.4.2
	 * Update 1 and above.
	 */
	searchPaths = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory,
		NSUserDomainMask + NSLocalDomainMask, YES);
	count = [searchPaths count];
	for (i = 1; i <= count; ++i) {
		pathBase = [searchPaths objectAtIndex:i-1];
		path = [pathBase stringByAppendingPathComponent:subpath];
		if ((javaPluginCocoa = [NSBundle bundleWithPath:path]) == nil)
			continue;
		javaPluginCocoaPath = [path retain];
		break;
	}

	/* Do initial versions checking here, before the Java VM is loaded -- some
	 * version-based decisions need to be made before than happens.  We'll
	 * ask the VM to reconfirm the Java version after it's been loaded (if
	 * that happens).
	 */
	javaVMFramework =
		[NSBundle bundleWithPath:@"/System/Library/Frameworks/JavaVM.framework"];
	if (javaVMFramework) {
		if ((infoPlist = [javaVMFramework infoDictionary]) != nil) {
			object = [infoPlist objectForKey:@"CFBundleShortVersionString"];
			if (object && [object isKindOfClass:[NSString class]])
				versionString = (NSString *) object;
			if (versionString) {
				result = [versionString compare:@"1.4.2" options:NSNumericSearch];
				if ((result == NSOrderedSame) || (result == NSOrderedDescending))
					isJava142 = true;
				result = [versionString compare:@"2.2.0" options:NSNumericSearch];
				if ((result == NSOrderedSame) || (result == NSOrderedDescending)) {
					isJava142 = true;
					isJava142Update1 = true;
					isJava142Update2 = true;
				}
				if (osxVersion >= 0x1040) {
					isJava142 = true;
					isJava142Update1 = true;
					isJava142Update2 = true;
					isJava142Tiger = true;
				}
			}
			if (!isJava142Update1) {
				versionString = nil;
				object = [infoPlist objectForKey:@"CFBundleVersion"];
				if (object && [object isKindOfClass:[NSString class]])
					versionString = (NSString *) object;
				if (versionString) {
					result = [versionString compare:@"1.4.2 Release 1" options:NSNumericSearch];
					if ((result == NSOrderedSame) || (result == NSOrderedDescending))
						isJava142Update1 = true;
					result = [versionString compare:@"1.4.2 Release 2" options:NSNumericSearch];
					if ((result == NSOrderedSame) || (result == NSOrderedDescending))
						isJava142Update2 = true;
				}
			}
			if (isJava142Update1 && !javaPluginCocoa) {
				NSLog(@"JEPController:checkBundle couldn't find JavaPluginCocoa.bundle");
				return false;
			}
			if (isJava142Tiger) {
				infoPlist = nil, versionString = nil;
				if ((infoPlist = [javaPluginCocoa infoDictionary]) != nil) {
					object = [infoPlist objectForKey:@"CFBundleShortVersionString"];
					if (object && [object isKindOfClass:[NSString class]])
						versionString = (NSString *) object;
				}
				if (versionString) {
					/* Starting with J2SE 5.0 Release 4, the JavaPluginCocoa.bundle
					 * in /Library/Internet Plug-Ins/ is a stub that loads the
					 * appropriate (version-specific) full copy from inside the
					 * JavaVM framework.  That stub's version appears to always
					 * be 11.4.0 (or higher).
					 */
					result = [versionString compare:@"11.4.0" options:NSNumericSearch];
					if ((result == NSOrderedSame) || (result == NSOrderedDescending)) {
						/* Also starting with J2SE 5.0 Release 4, the Java
						 * Preferences' "Plugin Version" setting is available
						 * from its preferences file.
						 */
						/* As of JEP 0.9.5+f, in J2SE 5.0 Release 4 (and above)
						 * we default to Java 1.5/5.0 when the "Plugin Version"
						 * setting is absent -- since that's how this version
						 * of Sun/Apple's JVM behaves when run from Safari.
						 * (In JEP 0.9.5+e we (inadvertently) defaulted to
						 * Java 1.4.2.)
						 */
						isJava150 = true;
						NSString *pluginVersion = (NSString *)
							CFPreferencesCopyAppValue(CFSTR("Plugin Version"),
							CFSTR("com.apple.java.JavaPreferences"));
						if (pluginVersion && [pluginVersion isKindOfClass:[NSString class]]) {
							result = [pluginVersion compare:@"1.4" options:NSNumericSearch];
							if (result == NSOrderedSame) {
								isJava150 = false;
							} else {
								result = [pluginVersion compare:@"1.6" options:NSNumericSearch];
								if (result == NSOrderedSame)
									isJava160 = true;
							}
						}
						if (pluginVersion)
							CFRelease((CFStringRef) pluginVersion);
						[javaPluginCocoaPath release];
						if (isJava160) {
							javaPluginCocoaPath = @"/System/Library/Frameworks/JavaVM.framework/Versions/1.6/Resources/JavaPluginCocoa.bundle";
						} else if (isJava150) {
							javaPluginCocoaPath = @"/System/Library/Frameworks/JavaVM.framework/Versions/1.5/Resources/JavaPluginCocoa.bundle";
						} else {
							javaPluginCocoaPath = @"/System/Library/Frameworks/JavaVM.framework/Versions/1.4/Resources/JavaPluginCocoa.bundle";
						}
					/* Previous OS X Tiger Java versions either only supported
					 * Java 1.4.2 or switched between Java 1.4.2 and Java 1.5.0
					 * by copying the appropriate JavaPluginCocoa.bundle to
					 * /Library/Internet Plug-Ins/.  (Amazing!)  Those that
					 * supported Java 1.4.2 had version numbers in the 10.0.X
					 * range, while those that supported Java 1.5.0 had version
					 * numbers in the 11.0.X range.  Except (groan) for the
					 * JVM(s) that appeared on the first Intel Macs (bundled
					 * with OS X 10.4.5, and (I think) equivalent to "Java
					 * 1.3.1 and Java 1.4.2 Release 2" plus "J2SE 5.0 Release
					 * 3" on PPC Macs) -- these JVM's JavaPluginCocoa.bundles
					 * had the same version number (11.3.0) for both Java 1.4.2
					 * and Java 1.5.0!  Fortunately their WebPluginDescription
					 * does indicate which Java version they support.  Before
					 * were able to decipher these JVMs' version information
					 * correctly, we always ran Java 1.5.0, even when the user
					 * had chosen Java 1.4.2 (which was the default).  In JEP
					 * 0.9.5+f and 0.9.5+g, which call Apple's registerNatives()
					 * method directly (see Java_callRegisterNatives() in
					 * AppletView.m), choosing Java 1.4.2 caused crashes in
					 * registerNatives() (since we were mixing Java versions).
					 * Our workaround for Apple's confusion (the
					 * WebPluginDescription trick) resolves bmo bug 355637
					 * (see particularly comment 3).
					 */
					} else {
						object = [infoPlist objectForKey:@"WebPluginDescription"];
						if (object && [object isKindOfClass:[NSString class]])
							webPluginDescriptionString = (NSString *) object;
						result = [versionString compare:@"11.3.0" options:NSNumericSearch];
						if (webPluginDescriptionString && (result == NSOrderedSame)) {
							NSArray *list = [webPluginDescriptionString
								componentsSeparatedByString:@" "];
							if (list && ([list count] >= 2)) {
								versionString = (NSString *) [list objectAtIndex:1];
								result = [versionString compare:@"1.5.0" options:NSNumericSearch];
								if ((result == NSOrderedSame) || (result == NSOrderedDescending))
									isJava150 = true;
							}
						} else {
							result = [versionString compare:@"11.0.0" options:NSNumericSearch];
							if ((result == NSOrderedSame) || (result == NSOrderedDescending))
								isJava150 = true;
						}
					}
				}
			}
		}
	}

	/* We need to load libawt.jnilib into our app's image so that our calls to
	 * PoseAsWhatever() below will work.  But we won't yet initialize it --
	 * that needs to wait until we call AppletView loadAWTLibraries (from
	 * AppletView maybeCreateJavaVM).
	 */
	if (isJava160) {
		libawtImage = NSAddImage(
			"/System/Library/Frameworks/JavaVM.framework/Versions/1.6/Libraries/libawt.jnilib",
			NSADDIMAGE_OPTION_RETURN_ON_ERROR);
	} else if (isJava150) {
		libawtImage = NSAddImage(
			"/System/Library/Frameworks/JavaVM.framework/Versions/1.5/Libraries/libawt.jnilib",
			NSADDIMAGE_OPTION_RETURN_ON_ERROR);
	} else if (isJava142) {
		libawtImage = NSAddImage(
			"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.2/Libraries/libawt.jnilib",
			NSADDIMAGE_OPTION_RETURN_ON_ERROR);
	} else {
		libawtImage = NSAddImage(
			"/System/Library/Frameworks/JavaVM.framework/Versions/1.4.1/Libraries/libawt.jnilib",
			NSADDIMAGE_OPTION_RETURN_ON_ERROR);
	}
	if (!libawtImage) {
		NSLog(@"JEPController:checkBundle couldn't find libawt.jnilib");
		return false;
	}
	/* The following trick can be used in place of the "..._Ptr" hack that
	 * I used in JEP 0.9.1 and 0.9.2.  It's better because it can handle
	 * calls to JNI_GetDefaultJavaVMInitArgs(), JNI_GetCreatedJavaVMs(),
	 * or JNI_CreateJavaVM() from inside jnilibs (there's a call to
	 * JNI_GetCreatedJavaVMs() in libjsound.jnilib).  See the following:
	 * http://lists.apple.com/archives/java-dev/2005/May/msg00421.html
	 * http://lists.apple.com/archives/java-dev/2005/May/msg00238.html
	 * As of J2SE 5.0 Release 4 (which defaults to Java 1.5), we need to
	 * specify both Java 1.4 and Java 1.5 explicitly.
	 */
	if (isJava160) {
		setenv("JAVA_JVM_VERSION", "1.6", true);
	} else if (isJava150) {
		setenv("JAVA_JVM_VERSION", "1.5", true);
	} else if (osxVersion >= 0x1040) {
		setenv("JAVA_JVM_VERSION", "1.4", true);
	}

	/* The calls to PoseAsWhatever() should be arranged so that superclasses
	 * are "posed as" before subclasses.
	 */
	if (!PoseAsControlModel())
		return false;
	if (!PoseAsNSViewAWT())
		return false;
	if (!PoseAsNSButtonAWT())
		return false;
	if (!PoseAsNSScroller())
		return false;
	if (!PoseAsNSScrollViewAWT())
		return false;
	if (!PoseAsNSChoiceAWT())
		return false;
	if (!PoseAsNSTableViewAWT())
		return false;
	if (!PoseAsAWTCombinedTextField())
		return false;
	if (!PoseAsNSTextFieldAWT())
		return false;
	if (!PoseAsNSTextViewAWT())
		return false;
	if (!PoseAsNSTextFieldFocusView())
		return false;
	if (!PoseAsNSWindowViewAWT())
		return false;
	if (!PoseAsAWTCombinedTextFieldCell())
		return false;
	if (!PoseAsAWTCombinedTextView())
		return false;
	if (!PoseAsCWindow())
		return false;
	if (!PoseAsEventFactory())
		return false;
	PoseAsFocusManager();
	PoseAsCTextField();
	PoseAsCAppletWindow();
	PoseAsCEmbedded();
	PoseAsNSViewAWTContext();
	PoseAsAWTDrawableContextManager();
	PoseAsThreadUtilities();
	PoseAsAWTModelUpdater();
	PoseAsAWTStarter();
	PoseAsStartupUtility();
	PoseAsAwtException();
	PoseAsCPopupMenu();
	PoseAsSyntheticEvent();
	if (isCarbonApp) {
		if (!PoseAsNSCarbonWindow())
			return false;
		if (!PoseAsNSCarbonWindowContentView())
			return false;
		if (!PoseAsNSCarbonWindowFrame())
			return false;
		/* For reasons that remain obscure, we can't use
		 * PoseAsUndocumentedClass() for this class in Tiger.  Apple's
		 * poseAsClass method (of NSObject) does different things (and more
		 * things) than Darwin's open-sourced method (which just uses
		 * class_poseAs())  So it works where Darwin's wouldn't.  The
		 * particular problem that we're avoiding here is:  If the user
		 * has chosen print from the browser's main menu before opening an
		 * applet, we'd otherwise get a (spurious) unhandled exception error
		 * (which causes a crash) -- "Failed to install Carbon Event Handler
		 * for text input".  My hunch is that Apple squelches (and ignores)
		 * this error using a Cocoa exception handler.  But, since we need
		 * to be able to run on Jaguar, that course isn't open to us.
		 */
		/* Apple seems to have recently done something to make this problem
		 * happen on OS X 10.3.X, too -- at least with laptops (possibly only
		 * certain models).  See
		 * http://sourceforge.net/tracker/index.php?func=detail&aid=1295817&group_id=107955&atid=649116.
		 */
		if (osxVersion >= 0x1030) {
			[MyNSTSMInputContext poseAsClass:[MyNSTSMInputContext superclass]];
		} else {
			PoseAsNSTSMInputContext();
		}
	}

	/* Useful in debugging */
	//CFNotificationCenterAddObserver(
	//	CFNotificationCenterGetDistributedCenter(),
	//	CFSTR("ShowEventInfo"), &ShowEventInfo, nil, nil,
	//	CFNotificationSuspensionBehaviorDeliverImmediately);

	if ((controllersLock = [[NSRecursiveLock alloc] init]) == nil) {
		NSLog(@"JEPController:checkBundle couldn't create lock for controller array");
		return false;
	}
	if ((controllerArray = [[NSMutableArray alloc] init]) == nil) {
		NSLog(@"JEPController:checkBundle couldn't create controller array");
		[controllersLock release];
		controllersLock = nil;
		return false;
	}
	if (osxVersion >= 0x1030) {
		if ((popupsLock = [[NSRecursiveLock alloc] init]) == nil) {
			NSLog(@"JEPController:checkBundle couldn't create lock for Java popups array");
			return false;
		}
		if ((popupsArray = [[NSMutableArray alloc] init]) == nil) {
			NSLog(@"JEPController:checkBundle couldn't create Java popups array");
			return false;
		}
	}

	installedPath = [[[NSBundle bundleForClass:[JEPController class]] bundlePath] retain];
	if (!installedPath || !strlen([installedPath UTF8String])) {
		NSLog(@"JEPController:checkBundle can't find our own location!");
		[controllerArray release];
		controllerArray = nil;
		[controllersLock release];
		controllersLock = nil;
		installedPath = nil;
		return false;
	}

	homeDirectory = [NSHomeDirectory() retain];

	asyncInitializeApplet =
		[AppletView getRuntimeParamBoolean:(NSString *)CFSTR(JEP_ASYNCINIT_OPTION)];
	debugVisibility =
		[AppletView getRuntimeParamBoolean:(NSString *)CFSTR(JEP_DEBUG_VISIBILITY_OPTION)];
	debugRelease =
		[AppletView getRuntimeParamBoolean:(NSString *)CFSTR(JEP_DEBUG_RELEASE_OPTION)];
	debugUpdates =
		[AppletView getRuntimeParamBoolean:(NSString *)CFSTR(JEP_DEBUG_UPDATES_OPTION)];

	nsAppShadow = [[NSAppShadow alloc] init];
	if (nsAppShadow && !isCarbonApp) {
		/* We can't make nsAppShadow a delegate of NSApp -- our parent
		 * app may (like Camino) already have set its delegate to something
		 * else.  Adding nsAppShadow to the default notification center
		 * works fine, though.
		 */
		[[NSNotificationCenter defaultCenter] addObserver:nsAppShadow
			selector:@selector(applicationWillTerminate:)
			name:@"NSApplicationWillTerminateNotification" object:NSApp];
	}
	if (nsAppShadow) {
		[nsAppShadow performSelectorOnMainThread:@selector(setMainThreadVariable)
			withObject:nil waitUntilDone:true];
	}

	screenUpdatesLock = [[NSRecursiveLock alloc] init];
	modelUpdaterLock = [[NSRecursiveLock alloc] init];
	allocCAWLock = [[NSRecursiveLock alloc] init];
	deletedCAWs = [[NSMutableArray alloc] init];
	customEventSpoolLock = [[NSRecursiveLock alloc] init];
	customEventSpool = [[NSMutableArray alloc] init];
	cEventSpoolLock = [[NSRecursiveLock alloc] init];
	cEventSpool = [[NSMutableArray alloc] init];

	Java_runningJobs = [[NSMutableArray alloc] init];
	Java_runningJobs_lock = [[NSLock alloc] init];

	return true;
}

/* External apps should use this instead of checkBundle, to avoid a version
 * mismatch between themselves and the Java Embedding Plugin.
 */
+ (BOOL)checkBundleWithVersion:(NSString *)versionString
{
	if (!versionString || strcasecmp(JEP_VERSION, [versionString UTF8String]))
		return false;
	return [JEPController checkBundle];
}

/* [NSObject poseAsClass] doesn't (by itself) effect objects that have already
 * been created.  But many "interesting" objects will already have been
 * created if we're running in a Cocoa app, and (on Tiger) NSApp may already
 * exist even in a Carbon app.  We can "hook" existing objects by resetting
 * their isa variables to the same class, by class name -- this allows the
 * current class structure (changed by our calls to poseAsClass) to be "read
 * back" into existing objects.
 */
+ (void)hookExistingObjects
{
	Class *classPtr;
	NSString *className;
	NSWindow *aWindow;
	NSView *aView, *contentView;
	NSArray *windowArray;
	NSMutableArray *viewArray;
	unsigned i, j, windowCount, viewCount;

	/* What we need to do is set the object's instance variable's isa pointer
	 * to its re-posed Class object (in this case, we just want to do
	 * "NSApp->isa = [NSApplication class]").  But doing this on NSApp causes
	 * a compiler warning, and doing it on other NSObjects actually causes a
	 * compiler error (the isa variable is protected!).  However, since the
	 * isa variable is always the first member of an objc_object structure,
	 * we can assume that the address of the isa variable is the same as the
	 * address of the objc_object structure as a whole.
	 */
	classPtr = (Class *) ((void *) NSApp);
	*classPtr = [NSApplication class];
	if (isCarbonApp)
		return;

	[JEPController disableScreenUpdates];
	// Do NOT release "windowArray"
	if ((windowArray = [NSApp windows]) == nil) {
		[JEPController enableScreenUpdates];
		return;
	}
	if ((viewArray = [[NSMutableArray alloc] init]) == nil) {
		[JEPController enableScreenUpdates];
		return;
	}
	windowCount = [windowArray count];
	for (i = 1; i <= windowCount; ++i) {
		aWindow = [windowArray objectAtIndex:i-1];
		className = NSStringFromClass([aWindow class]);
		if ([className isEqualToString:@"%NSWindow"])
			className = @"NSWindow";
		classPtr = (Class *) ((void *) aWindow);
		*classPtr = NSClassFromString(className);
		/* Don't iterate through the NSView objects of invisible windows --
		 * sometimes (though probably very rarely) an invisible window's
		 * contentView can be invalid.  In the one case I'm aware of, the
		 * window had just been dismissed, and was in the process of being
		 * destroyed (though _isTerminating and _windowDying both returned
		 * false).  If/when one of these windows is made visible, we'll check
		 * its NSView objects at that time (in the NSWindow _setVisible
		 * method).  Only skip windows that are _really_ invisible -- not just
		 * miniaturized.  This change resolves part of bmo bug 374260.
		 */
		if (![aWindow isVisible] && ![aWindow isMiniaturized])
			continue;
		if ((contentView = [aWindow contentView]) == nil)
			continue;
		[viewArray addObject:contentView];
		[JEPController allSubviews:contentView container:viewArray];
		viewCount = [viewArray count];
		for (j = 1; j <= viewCount; ++j) {
			aView = [viewArray objectAtIndex:j-1];
			className = NSStringFromClass([aView class]);
			if ([className isEqualToString:@"%NSView"])
				className = @"NSView";
			classPtr = (Class *) ((void *) aView);
			*classPtr = NSClassFromString(className);
		}
		[viewArray removeAllObjects];
	}
	[JEPController enableScreenUpdates];
	[viewArray release];
}

/* If called with handle == nil creates a new JEPController object, puts it
 * into controllerArray, and returns its "handle".  Otherwise checks that
 * handle exists in controllerArray (i.e. that it's currently valid) --
 * returns handle if it is, and otherwise nil (to indicate that handle is
 * invalid).
 */
+ (JEPController *)myController:(JEPController *)handle
{
	JEPController *retval = nil;
	if (!nsAppInited)
		return nil;
	if (!controllerArray || !controllersLock)
		return nil;
	if (!handle) {
		[JEPController purgeControllerArray];
		if ((retval = [[JEPController alloc] init]) != nil) {
			[controllersLock lock];
			[controllerArray addObject:retval];
			[controllersLock unlock];
		}
	} else {
		unsigned i, count;
		JEPController *aController;
		NSMutableArray *dead;
		if ((dead = [[NSMutableArray alloc] init]) == nil)
			return nil;
		[controllersLock lock];
		count = [controllerArray count];
		for (i = 1; i <= count; ++i) {
			aController = [controllerArray objectAtIndex:i-1];
			if (aController->fNeedsCleanup) {
				[dead addObject:aController];
				continue;
			}
			if ([aController isEqual:handle])
				retval = aController;
		}
		[controllerArray removeObjectsInArray:dead];
		[controllersLock unlock];
		count = [dead count];
		for (i = 1; i <= count; ++i) {
			aController = [dead objectAtIndex:i-1];
			//NSLog(@"myController: retain count %i", [aController retainCount]);
			[aController release];
		}
		[dead removeAllObjects];
		[dead release];
	}
	if (retval)
		retval = [[retval retain] autorelease];
	return retval;
}

+ (BOOL)isControllerValid:(JEPController *)controller
{
	unsigned i, count;
	JEPController *aController;
	BOOL retval = false;
	if (!nsAppInited || !controller)
		return false;
	if (!controllerArray || !controllersLock)
		return false;
	[controllersLock lock];
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		aController = [controllerArray objectAtIndex:i-1];
		if ([aController isEqual:controller]) {
			retval = true;
			break;
		}
	}
	[controllersLock unlock];
	return retval;
}

/* Destroys (and removes from controllerArray) JEPController objects that
 * have been marked for cleanup.
 */
+ (void)purgeControllerArray
{
	unsigned i, count;
	JEPController *controller;
	NSMutableArray *dead;
	if (!controllerArray || !controllersLock)
		return;
	[controllersLock lock];
	if ((dead = [[NSMutableArray alloc] init]) == nil) {
		[controllersLock unlock];
		return;
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (controller->fNeedsCleanup)
			[dead addObject:controller];
	}
	[controllerArray removeObjectsInArray:dead];
	[controllersLock unlock];
	count = [dead count];
	for (i = 1; i <= count; ++i) {
		controller = [dead objectAtIndex:i-1];
		//NSLog(@"purgeControllerArray: retain count %i", [controller retainCount]);
		[controller release];
	}
	[dead removeAllObjects];
	[dead release];
}

/* Destroys JEPController objects whose applets have been destroyed (i.e. marks
 * them for cleanup, destroys them, and removes them from controllerArray) that
 * are contained in a particular Carbon window.  (We cause subtle errors if we
 * purge them from _all_ of our browser's windows.  For example, if (in a
 * Carbon browser) we purge them from another window that once contained
 * applet(s) but no longer does so, we will no longer be able to send any mouse
 * events or keystrokes to that window -- because we'll have removed the
 * necessary Carbon event handlers.)
 */
+ (void)purgeIfAppletDestroyedInWindow:(WindowRef)carbonWindow
{
	unsigned i, count;
	JEPController *controller;
	if (!controllerArray || !controllersLock || !carbonWindow)
		return;
	[controllersLock lock];
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myWindowCarbon)
			continue;
		if (carbonWindow != controller->myWindowCarbon)
			continue;
		if (controller->fAppletDestroyed)
			controller->fNeedsCleanup = TRUE;
	}
	[controllersLock unlock];
	[JEPController purgeControllerArray];
}

+ (void)destroyControllerArray
{
	unsigned i, count;
	JEPController *controller;
	if (!controllerArray || !controllersLock)
		return;
	[controllersLock lock];
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		//NSLog(@"destroyControllerArray: retain count %i", [controller retainCount]);
		[controller release];
	}
	[controllerArray removeAllObjects];
	[controllerArray release];
	controllerArray = nil;
	[controllersLock unlock];
	[controllersLock release];
	controllersLock = nil;
}

/* The Carbon API has DisableScreenUpdates() and EnableScreenUpdates() to
 * globally disable (for up to one second) or enable screen updates.  But
 * the calls can't be nested, so it's not really safe to use them in a
 * multithreaded app.  The following two methods fix the problem.
 */

static int screenUpdatesDisabledCount = 0;
NSLock *screenUpdatesLock = nil;

+ (void)enableScreenUpdates
{
	/* EnableScreenUpdates() (and presumably also DisableScreenUpdates()) can
	 * crash if we use it while our app is being terminated.  The check for
	 * !nsAppShadow makes sure we don't do that.
	 */
	if (!screenUpdatesLock || !nsAppShadow)
		return;
	[screenUpdatesLock lock];
	if (--screenUpdatesDisabledCount > 0) {
		[screenUpdatesLock unlock];
		return;
	} else if (screenUpdatesDisabledCount < 0) {
		screenUpdatesDisabledCount = 0;
		[screenUpdatesLock unlock];
		return;
	}
	EnableScreenUpdates();
	[screenUpdatesLock unlock];
}

+ (void)disableScreenUpdates
{
	/* EnableScreenUpdates() (and presumably also DisableScreenUpdates()) can
	 * crash if we use it while our app is being terminated.  The check for
	 * !nsAppShadow makes sure we don't do that.
	 */
	if (!screenUpdatesLock || !nsAppShadow)
		return;
	[screenUpdatesLock lock];
	if (++screenUpdatesDisabledCount != 1) {
		[screenUpdatesLock unlock];
		return;
	}
	DisableScreenUpdates();
	[screenUpdatesLock unlock];
}

+ (BOOL)screenUpdatesDisabled
{
	return (screenUpdatesDisabledCount > 0);
}

/* Carbon's and Java's coordinate systems have the upper left corner as their
 * origin, but Cocoa uses the lower left corner.  This causes no end of grief
 * until you learn how to deal with it.  Among other things, you can only
 * translate a point from one coordinate system to the other in the context of
 * a parent view or window.  The flipScreenPoint, flipWindowPoint and
 * flipViewPoint methods help cope with this.
 */
+ (NSPoint)flipScreenPoint:(NSPoint)pointToFlip
{
	NSRect screenFrame = NSZeroRect;
	NSPoint flippedPoint = NSZeroPoint;
	screenFrame = [[NSScreen mainScreen] frame];
	flippedPoint.x = pointToFlip.x;
	flippedPoint.y = screenFrame.size.height - pointToFlip.y;
	return flippedPoint;
}

+ (NSPoint)flipWindowPoint:(NSPoint)pointToFlip window:(NSWindow *)window
{
	int windowHeight;
	WindowRef windowRef;
	Rect frameCarbon = {0};
	NSRect frameCocoa = NSZeroRect;
	NSPoint flippedPoint = NSZeroPoint;
	if (!window || !(windowRef = [MyNSWindow existingCarbonWindowFor:window]))
		return flippedPoint;
	if (isCarbonApp) {
		GetWindowBounds(windowRef, kWindowContentRgn, &frameCarbon);
		windowHeight = frameCarbon.bottom - frameCarbon.top;
	} else {
		frameCocoa = [window frame];
		if (osxVersion >= 0x1030) {
			frameCocoa = [window contentRectForFrameRect:frameCocoa];
		} else {
			frameCocoa = [window contentRectForFrameRect:frameCocoa
				styleMask:0];
		}
		windowHeight = frameCocoa.size.height;
	}
	flippedPoint.x = pointToFlip.x;
	flippedPoint.y = windowHeight - pointToFlip.y;
	return flippedPoint;
}

+ (NSPoint)flipViewPoint:(NSPoint)pointToFlip view:(NSView *)view
{
	NSRect viewFrame;
	NSPoint flippedPoint = NSZeroPoint;
	if (!view)
		return flippedPoint;
	viewFrame = [view frame];
	flippedPoint.x = pointToFlip.x;
	flippedPoint.y = viewFrame.size.height - pointToFlip.y;
	return flippedPoint;
}

/* Though a URL with the format "file://localhost/..." is legal, Sun's
 * internal methods for creating a Java applet will choke on it.  So we fix
 * the problem here.
 */
+ (NSURL *)fixDocbase:(NSURL *)docbaseToFix
{
	NSString *path;
	NSString *url;
	NSURL *fixedDocbase;
	if (!docbaseToFix)
		return nil;
	if (![docbaseToFix isFileURL])
		return docbaseToFix;
	if ((path = [docbaseToFix path]) == nil)
		return docbaseToFix;
	/* Needed if docbaseToFix contains spaces */
	path = (NSString *) CFURLCreateStringByAddingPercentEscapes(nil,
		(CFStringRef) path, nil, nil, kCFStringEncodingMacRoman);
	url = [NSString stringWithString:@"file://"];
	url = [url stringByAppendingString:path];
	fixedDocbase = [[[NSURL alloc] initWithString:url] autorelease];
	return fixedDocbase;
}

/* This is usually only safe in a Cocoa app:  In a Carbon app each Carbon
 * window may correspond to more than one Cocoa window.  (Though if any of
 * the Cocoa windows will do, this method can safely be used in a Carbon app,
 * too.)  Another reason this method might be unsuitable in a Carbon app is
 * that it returns the NSWindow object (corresponding to carbonWindow) that's
 * currently in the window cache.
 */
+ (NSWindow *)nsWindowFromCarbonWindow:(WindowRef)carbonWindow;
{
	unsigned i, count;
	NSArray *windowArray;
	NSWindow *window = nil;
	if (!carbonWindow)
		return nil;
	/* We globally disable screen updates for the duration of this method in
	 * order to stop a change in the window hierarchy from messing up our
	 * search.
	 */
	[JEPController disableScreenUpdates];
	// Do NOT release "windowArray"
	if ((windowArray = [NSApp windows]) == nil) {
		[JEPController enableScreenUpdates];
		return nil;
	}
	count = [windowArray count];
	for (i = 1; i <= count; ++i) {
		window = [windowArray objectAtIndex:i-1];
		if (![window isVisible])
			continue;
		if (carbonWindow == [MyNSWindow existingCarbonWindowFor:window])
			break;
	}
	if (i > count)
		window = nil;
	[JEPController enableScreenUpdates];
	return window;
}

/* This is only safe in a Carbon app:  In a Carbon app each NSWindow
 * corresponds to exactly one JEPController and vice versa.  But in
 * a Cocoa app a single NSWindow might correspond to more than one
 * JEPController.
 */
+ (JEPController *)controllerFromNSWindow:(NSWindow *)window
	timeoutType:(LockTimeout)timeoutType
{
	unsigned i, count;
	NSTimeInterval timeout;
	JEPController *controller = nil;
	if (!controllerArray || !controllersLock || !window)
		return nil;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return nil;
	} else {
		[controllersLock lock];
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myWindow)
			continue;
		if (window != controller->myWindow)
			continue;
		break;
	}
	if (i > count)
		controller = nil;
	[controllersLock unlock];
	return controller;
}

+ (JEPController *)controllerFromAppletView:(AppletView *)appletView
	timeoutType:(LockTimeout)timeoutType
{
	unsigned i, count;
	NSTimeInterval timeout;
	JEPController *controller = nil;
	if (!controllerArray || !controllersLock || !appletView)
		return nil;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return nil;
	} else {
		[controllersLock lock];
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myAppletView)
			continue;
		if (controller->myAppletView != appletView)
			continue;
		break;
	}
	if (i > count)
		controller = nil;
	[controllersLock unlock];
	return controller;
}

+ (JEPController *)controllerFromAWTContext:(id)awtContext
	timeoutType:(LockTimeout)timeoutType
{
	unsigned i, count;
	NSTimeInterval timeout;
	JEPController *controller = nil;
	if (!controllerArray || !controllersLock || !awtContext)
		return nil;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return nil;
	} else {
		[controllersLock lock];
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->appletAWTContext)
			continue;
		if (controller->appletAWTContext != awtContext)
			continue;
		break;
	}
	if (i > count)
		controller = nil;
	[controllersLock unlock];
	return controller;
}

+ (JEPController *)controllerFromCWindow:(id)cWindow
	timeoutType:(LockTimeout)timeoutType
{
	unsigned i, count;
	NSTimeInterval timeout;
	JEPController *controller = nil;
	if (!controllerArray || !controllersLock || !cWindow)
		return nil;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return nil;
	} else {
		[controllersLock lock];
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->appletWindow)
			continue;
		if (controller->appletWindow != cWindow)
			continue;
		break;
	}
	if (i > count)
		controller = nil;
	[controllersLock unlock];
	return controller;
}

+ (JEPController *)controllerFromPluginInstance:(void *)instance
	timeoutType:(LockTimeout)timeoutType
{
	unsigned i, count;
	NSTimeInterval timeout;
	JEPController *controller = nil;
	if (!controllerArray || !controllersLock || !instance)
		return nil;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return nil;
	} else {
		[controllersLock lock];
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->pluginInstance)
			continue;
		if (controller->pluginInstance != instance)
			continue;
		break;
	}
	if (i > count)
		controller = nil;
	[controllersLock unlock];
	return controller;
}

/* Returns all JEPController objects that share the same carbon window.
 */
+ (NSMutableArray *)controllersInWindow:(WindowRef)carbonWindow
	timeoutType:(LockTimeout)timeoutType
{
	unsigned i, count;
	NSTimeInterval timeout;
	JEPController *controller;
	NSMutableArray *controllersFound;
	if (!controllerArray || !controllersLock || !carbonWindow)
		return nil;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return nil;
	} else {
		[controllersLock lock];
	}
	if ((controllersFound = [[NSMutableArray alloc] init]) == nil) {
		[controllersLock unlock];
		return nil;
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myWindowCarbon)
			continue;
		if (carbonWindow != controller->myWindowCarbon)
			continue;
		[controllersFound addObject:controller];
	}
	[controllersLock unlock];
	if (![controllersFound count]) {
		[controllersFound release];
		controllersFound = nil;
	}
	return controllersFound;
}

+ (void)allSubviews:(NSView *)view container:(NSMutableArray *)container
{
	unsigned i, count;
	NSView *subview;
	NSArray *immediateSubviews;
	if (!view || !container)
		return;
	if ((immediateSubviews = [view subviews]) == nil)
		return;
	count = [immediateSubviews count];
	for (i = 1; i <= count; ++i) {
		subview = [immediateSubviews objectAtIndex:i-1];
		[container addObject:subview];
		[JEPController allSubviews:subview container:container];
	}
}

/* Returns the single JEPController object (if there is just one) associated
 * with "view" -- either whose applet view contains "view" or whose applet
 * view is (uniquely) contained in "view" (we aren't interested in "views"
 * that contain more than one JEPController's applet view, only in those of
 * their subviews that contain just one).
 */
+ (JEPController *)controllerFromNSView:(NSView *)view
	timeoutType:(LockTimeout)timeoutType
{
	JEPController *controller;
	NSTimeInterval timeout;
	NSMutableArray *controllersFound;
	if (!controllersLock)
		return nil;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return nil;
	} else {
		[controllersLock lock];
	}
	if ((controller = [JEPController controllerContainingNSView:view
		timeoutType:lockTimeoutNone]) == nil)
	{
		if ((controllersFound = [JEPController controllersInNSView:view
			timeoutType:lockTimeoutNone]) != nil)
		{
			if ([controllersFound count] == 1)
				controller = [controllersFound objectAtIndex:0];
			[controllersFound removeAllObjects];
			[controllersFound release];
		}
	}
	[controllersLock unlock];
	return controller;
}

/* Returns all the JEPController objects whose applet views are contained in
 * "view".
 */
+ (NSMutableArray *)controllersInNSView:(NSView *)view
	timeoutType:(LockTimeout)timeoutType
{
	NSView *oneView;
	NSTimeInterval timeout;
	JEPController *controller;
	NSMutableArray *viewsFound;
	NSMutableArray *controllersFound;
	unsigned i, j, count1, count2;
	if (!controllerArray || !controllersLock || !view)
		return nil;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return nil;
	} else {
		[controllersLock lock];
	}
	if ((viewsFound = [[NSMutableArray alloc] init]) == nil) {
		[controllersLock unlock];
		return nil;
	}
	if ((controllersFound = [[NSMutableArray alloc] init]) == nil) {
		[viewsFound release];
		[controllersLock unlock];
		return nil;
	}
	/* We globally disable screen updates for the duration of this method in
	 * order to stop a change in the window hierarchy from messing up our
	 * search.
	 */
	[JEPController disableScreenUpdates];
	[viewsFound addObject:view];
	[JEPController allSubviews:view container:viewsFound];
	count1 = [controllerArray count];
	for (i = 1; i <= count1; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myAppletView)
			continue;
		count2 = [viewsFound count];
		for (j = 1; j <= count2; ++j) {
			oneView = [viewsFound objectAtIndex:j-1];
			if (oneView != controller->myAppletView)
				continue;
			[controllersFound addObject:controller];
			break;
		}
	}
	[controllersLock unlock];
	[viewsFound removeAllObjects];
	[viewsFound release];
	[JEPController enableScreenUpdates];
	if (![controllersFound count]) {
		[controllersFound release];
		controllersFound = nil;
	}
	return controllersFound;
}

/* Returns the JEPController object (if there is just one) whose applet
 * view is (uniquely) contained in "view" (we aren't interested in "views"
 * that contain more than one JEPController's applet view, only in those of
 * their subviews that contain just one).
 */
+ (JEPController *)controllerContainingNSView:(NSView *)view
	timeoutType:(LockTimeout)timeoutType
{
	unsigned i, count;
	NSView *oneView;
	NSTimeInterval timeout;
	JEPController *controller = nil;
	if (!controllerArray || !controllersLock || !view)
		return nil;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return nil;
	} else {
		[controllersLock lock];
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myAppletView)
			continue;
		oneView = view;
		do {
			if (oneView == controller->myAppletView)
				break;
		} while ((oneView = [oneView superview]) != nil);
		if (oneView)
			break;
	}
	if (i > count)
		controller = nil;
	[controllersLock unlock];
	return controller;
}

/* Calls a method in all JEPController objects.
 */
+ (void)sendToAllControllers:(NSInvocation *)message
	targetIsMyWindow:(BOOL)targetIsMyWindow
	timeoutType:(LockTimeout)timeoutType
{
	id target;
	NSTimeInterval timeout;
	JEPController *controller;
	NSMutableArray *holder;
	unsigned i, count;
	if (!controllerArray || !controllersLock)
		return;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return;
	} else {
		[controllersLock lock];
	}
	if ((holder = [[NSMutableArray alloc] init]) == nil) {
		[controllersLock unlock];
		return;
	}
	[holder addObjectsFromArray:controllerArray];
	[controllersLock unlock];
	count = [holder count];
	for (i = 1; i <= count; ++i) {
		controller = [holder objectAtIndex:i-1];
		if (targetIsMyWindow) {
			if ((target = controller->myWindow) == nil)
				continue;
		} else {
			target = controller;
		}
		[message invokeWithTarget:target];
	}
	[holder removeAllObjects];
	[holder release];
}

/* Calls a method in all the JEPController objects which share the same
 * Carbon window.
 */
+ (void)sendToControllersInWindow:(NSInvocation *)message
	carbonWindow:(WindowRef)carbonWindow
	targetIsMyWindow:(BOOL)targetIsMyWindow
	timeoutType:(LockTimeout)timeoutType
{
	id target;
	NSTimeInterval timeout;
	JEPController *controller;
	NSMutableArray *inWindow;
	unsigned i, count;
	if (!controllerArray || !controllersLock || !carbonWindow)
		return;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return;
	} else {
		[controllersLock lock];
	}
	if ((inWindow = [[NSMutableArray alloc] init]) == nil) {
		[controllersLock unlock];
		return;
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myWindowCarbon)
			continue;
		if (carbonWindow != controller->myWindowCarbon)
			continue;
		[inWindow addObject:controller];
	}
	[controllersLock unlock];
	count = [inWindow count];
	for (i = 1; i <= count; ++i) {
		controller = [inWindow objectAtIndex:i-1];
		if (targetIsMyWindow)
			target = controller->myWindow;
		else target = controller;
		[message invokeWithTarget:target];
	}
	[inWindow removeAllObjects];
	[inWindow release];
}

/* Calls a method in all the "other" JEPController objects that share a
 * Carbon window.
 */
+ (void)sendToOtherControllersInWindow:(NSInvocation *)message
	sender:(JEPController *)sender carbonWindow:(WindowRef)carbonWindow
	targetIsMyWindow:(BOOL)targetIsMyWindow timeoutType:(LockTimeout)timeoutType
{
	id target;
	NSTimeInterval timeout;
	JEPController *controller;
	NSMutableArray *inOtherWindow;
	unsigned i, count;
	if (!controllerArray || !controllersLock || !sender || !carbonWindow)
		return;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return;
	} else {
		[controllersLock lock];
	}
	if ((inOtherWindow = [[NSMutableArray alloc] init]) == nil) {
		[controllersLock unlock];
		return;
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if ([sender isEqual:controller])
			continue;
		if (!controller->myWindowCarbon)
			continue;
		if (carbonWindow != controller->myWindowCarbon)
			continue;
		[inOtherWindow addObject:controller];
	}
	[controllersLock unlock];
	count = [inOtherWindow count];
	for (i = 1; i <= count; ++i) {
		controller = [inOtherWindow objectAtIndex:i-1];
		if (targetIsMyWindow)
			target = controller->myWindow;
		else target = controller;
		[message invokeWithTarget:target];
	}
	[inOtherWindow removeAllObjects];
	[inOtherWindow release];
}

+ (BOOL)windowHasKeyWindow:(WindowRef)windowRef
	timeoutType:(LockTimeout)timeoutType
{
	BOOL retval = false;
	int keyWindows = 0;
	NSTimeInterval timeout;
	JEPController *controller;
	unsigned i, count;
	if (!controllerArray || !controllersLock || !windowRef)
		return false;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return false;
	} else {
		[controllersLock lock];
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myWindowCarbon || [controller destroying])
			continue;
		if (windowRef != controller->myWindowCarbon)
			continue;
		if (!controller->hasKeyWindow)
			continue;
		++keyWindows;
	}
	[controllersLock unlock];
	if (keyWindows)
		retval = true;
	return retval;
}

/* Assumes that "point" uses the coordinate system of "windowRef"'s content
 * region (i.e. not including the title bar and "resize region"), and that
 * this is a Carbon-style coordinate system (i.e. its origin is the content
 * region's top-left corner).
 */
+ (BOOL)appletViewInWindowHasPoint:(WindowRef)windowRef point:(Point)point
	timeoutType:(LockTimeout)timeoutType
{
	BOOL retval = false;
	NSTimeInterval timeout;
	JEPController *controller;
	NSMutableArray *controllersFound;
	NSWindow *cocoaWindow = nil;
	NSView *contentView = nil;
	NSPoint orgPoint = {point.h, point.v};
	NSPoint convertedPoint = NSZeroPoint;
	unsigned i, count;

	if (!controllerArray || !controllersLock || !windowRef)
		return false;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return false;
	} else {
		[controllersLock lock];
	}

	/* We previously (before 0.8.6) used nsWindowFromCarbonWindow here (since
	 * any NSWindow object in windowRef will do).  But nsWindowFromCarbonWindow
	 * only returns the NSWindow that's currently in the window cache -- and
	 * even if one or more suitable NSWindow objects do exist, it might be that
	 * none of them is currently in the window cache.
	 */
	if ((controllersFound = [JEPController controllersInWindow:windowRef
		timeoutType:lockTimeoutNone]) == nil)
	{
		[controllersLock unlock];
		return false;
	} else {
		controller = [controllersFound objectAtIndex:0];
		cocoaWindow = controller->myWindow;
		[controllersFound removeAllObjects];
		[controllersFound release];
	}
	if ((contentView = [cocoaWindow contentView]) == nil) {
		[controllersLock unlock];
		return false;
	}
	/* "point" is assumed to use Carbon coordinates (upper-left origin).
	 * If we're in a Cocoa app we need to convert it to Cocoa coordinates
	 * (bottom-left origin).
	 */
	if (isCarbonApp) {
		convertedPoint = orgPoint;
	} else {
		convertedPoint =
			[JEPController flipWindowPoint:orgPoint window:cocoaWindow];
	}

	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myWindowCarbon || !controller->myAppletView)
			continue;
		if (windowRef != controller->myWindowCarbon)
			continue;
		if ([controller destroying])
			continue;
		if (controller->myAppletView->isInvisible)
			continue;
		if ([controller->myAppletView isCurrentClipEmpty])
			continue;
		if (![controller->myAppletView isAppletVisible])
			continue;
		if ([contentView mouse:convertedPoint inRect:controller->currentFrame]) {
			retval = true;
			break;
		}
	}

	[controllersLock unlock];
	return retval;
}

+ (JEPController *)focusedControllerInWindow:(WindowRef)windowRef
	timeoutType:(LockTimeout)timeoutType
{
	unsigned i, count;
	NSTimeInterval timeout;
	JEPController *controller = nil;
	if (!controllerArray || !controllersLock || !windowRef)
		return false;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return false;
	} else {
		[controllersLock lock];
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myWindowCarbon || [controller destroying])
			continue;
		if (windowRef != controller->myWindowCarbon)
			continue;
		if (!controller->hasKeyWindow)
			continue;
		break;
	}
	if (i > count)
		controller = nil;
	[controllersLock unlock];
	return controller;
}

/* When you click in the parent window outside of any applet, the parent
 * window's TSMDocument (if any) should become active.  But sometimes the
 * _activate method of NSTSMInputContext wants to make one of the applets'
 * TSMDocuments active.  We override this method in Handlers.m to change
 * this behavior.  This method tells our code (in _activate_TSMIC())
 * whether or not we need to clean up after the _activate method, and if so
 * which TSMDocument to make active (i.e. what is the parent window's
 * TSMDocument, if any).
 */
+ (TSMDocumentID)getWindowTSMDocIfShouldBeActive:(WindowRef)windowRef
	timeoutType:(LockTimeout)timeoutType
{
	TSMDocumentID tsmDoc = nil;
	NSTimeInterval timeout;
	JEPController *controller;
	unsigned i, count;
	if (!controllerArray || !controllersLock || !windowRef)
		return nil;
	switch (timeoutType) {
		case lockTimeoutShort:
			timeout = SHORT_LOCK_TIMEOUT;
			break;
		case lockTimeoutNormal:
			timeout = LOCK_TIMEOUT;
			break;
		case lockTimeoutNone:
		default:
			timeout = 0;
			break;
	}
	if (timeout) {
		if (![controllersLock
			lockBeforeDate:[NSDate dateWithTimeIntervalSinceNow:timeout]])
			return nil;
	} else {
		[controllersLock lock];
	}
	count = [controllerArray count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArray objectAtIndex:i-1];
		if (!controller->myWindowCarbon || [controller destroying])
			continue;
		if (windowRef != controller->myWindowCarbon)
			continue;
		tsmDoc = controller->tsmInfo.parentActiveDoc;
		if (controller->hasKeyWindow) {
			tsmDoc = nil;
			break;
		}
	}
	[controllersLock unlock];
	return tsmDoc;
}

/* Decides to which NSCarbonWindow object (among those that share the same
 * Carbon window) an event (originally sent to one of them) will be routed.
 * See sendEvent_CW() in Handlers.m for more info.
 */
+ (void)routeEvent:(NSEvent *)theEvent origTarget:(NSWindow *)origTarget
{
	NSThread *currentThread;
	WindowRef windowRef = nil;
	EventRef carbonEvent = nil;
	Boolean notForCocoa = false;
	NSWindow *nsWindowProperty = nil;
	NSWindow *defaultTarget;
	NSWindow *targetWindow;
	NSWindow *cacheWindow;
	NSMutableArray *controllerArrayHolder;
	JEPController *controller;
	JEPController *targetController;
	JEPController *prevKeyWindowController;
	JEPController *cacheController;
	unsigned i, count;
	NSEventType type;
	struct objc_super target;
	BOOL isKeyEvent;
	BOOL isMouseEvent;
	NSView *contentView;
	NSPoint locationInView;
	Point location;
	Rect boundsRect;
	WindowPartCode part;
	WindowRef foundWindow;
	int mouseDownMatchesCount;
	NSMutableArray *mouseDownMatches;
	BOOL mouseDownMatchesPrevKeyWindow;
	BOOL noKeyWindow;
	int windowNumber;
	NSView *topView = nil, *superView = nil;

	if (!mainThread)
		return;
	currentThread = [NSThread currentThread];

	/* As of Tiger, Mac OS X uses some NSCarbonWindow objects in Carbon apps
	 * on its own account.  An example is the OS's "Print" dialog for Carbon
	 * apps.  So now we need to make sure we only fiddle with "our" windows.
	 */
	if (origTarget) {
		/* We can't use controllerFromNSWindow or similar methods to check
		 * whether or not origTarget is "our" window -- they wouldn't work if
		 * the applet once associated with "self" had been destroyed.  In
		 * Controller.m's CreateCocoaWindow() we set our parent window's
		 * kPropertyNSWindow property to the latest NSCarbonWindow we've
		 * created in it.  myWindow's windowRef is the same as our parent
		 * window.  So the presence of this property indicates that "self" is
		 * "our" window, and its absence indicates that it isn't "our" window.
		 * (We don't care about the actual value of nsWindowProperty.)
		 */
		if ((windowRef = [MyNSWindow existingCarbonWindowFor:origTarget]) != nil) {
			if (GetWindowProperty(windowRef, kJEPCreatorCode,
				kPropertyNSWindow, sizeof(nsWindowProperty), nil,
				&nsWindowProperty) != noErr)
			{
				nsWindowProperty = nil;
			}
		}
		if (!nsWindowProperty) {
			target.class = objc_lookUpClass("NSWindow");
			target.receiver = origTarget;
			objc_msgSendSuper(&target, @selector(sendEvent:), theEvent);
			return;
		}
	}

	if (!controllerArray || !controllersLock || !origTarget || !windowRef)
		return;

	/* We don't want to process events that have been marked as "notForCocoa"
	 * here.
	 */
	type = [theEvent type];
	carbonEvent = [theEvent _eventRef];
	if (carbonEvent) {
		if (GetEventParameter(carbonEvent, kEventParamJEPNotForCocoa,
			typeBoolean, NULL, sizeof(Boolean), NULL, &notForCocoa) != noErr)
			notForCocoa = false;
	}
	if (notForCocoa) {
#ifdef DEBUG_EVENTS
		if (type != NSMouseMoved)
			NSLog(@"routeEvent: Discarding notForCocoa event %@", theEvent);
#endif
		return;
	}

	/* In a Carbon app each applet view has its own NSCarbonWindow object, but
	 * the OS expects there to be at most one such object for each parent
	 * Carbon window.  So if the parent window has more than one applet view,
	 * we need to fiddle with NSApplication's "window cache" each time the
	 * focus changes from one applet view to another to maintain the one-to-one
	 * correspondence.  (See below in this method and in cleanupApplet for more
	 * information.)  Here we assume that all the necessary fiddling has been
	 * done correctly, and use origTarget's window number (which is the same
	 * for all the applet views' NSCarbonWindow objects) to find the
	 * NSCarbonWindow object that currently has (or most recently had) the
	 * user focus.
	 */
	cacheWindow = [NSApp windowWithWindowNumber:[origTarget _realWindowNumber]];
	if (cacheWindow) {
		defaultTarget = cacheWindow;
	} else {
		defaultTarget = origTarget;
	}

	part = 0;
	/* Though the target will always be an NSCarbonWindow object, to avoid
	 * infinite loops we send our message to its superclass (NSWindow).
	 */
	target.class = objc_lookUpClass("NSWindow");
	switch (type) {
		case NSLeftMouseDown:
		case NSLeftMouseUp:
		case NSRightMouseDown:
		case NSRightMouseUp:
		case NSOtherMouseDown:
		case NSOtherMouseUp:
		case NSLeftMouseDragged:
		case NSRightMouseDragged:
		case NSOtherMouseDragged:
		case NSMouseMoved:
		case NSScrollWheel:
		case NSMouseEntered:
		case NSMouseExited:
			isMouseEvent = true;
			isKeyEvent = false;
			targetWindow = nil;
			/* A left mouse down can change the user focus (see below in this
			 * method), but we want this to happen only when the mouse-down
			 * occurs in the parent window's content region.
			 */
			if ((contentView = [defaultTarget contentView]) != nil) {
				/* Convert from the Cocoa coordinate system (bottom-left
				 * origin) to the Carbon one (top-left origin).
				 */
				locationInView = [contentView
					convertPoint:[theEvent locationInWindow] fromView:nil];
				GetWindowBounds(windowRef, kWindowContentRgn, &boundsRect);
				location.v = boundsRect.top + locationInView.y;
				location.h = boundsRect.left + locationInView.x;
				part = FindWindow(location, &foundWindow);
				if (foundWindow != windowRef)
					part = 0;
			}
			break;
		case NSKeyDown:
		case NSKeyUp:
		case NSFlagsChanged:
			isMouseEvent = false;
			isKeyEvent = true;
			targetWindow = nil;
			break;
		default:
			isMouseEvent = false;
			isKeyEvent = false;
			targetWindow = defaultTarget;
			break;
	}

	/* If the event is an unknown kind, send it to its default target without
	 * further processing.
	 */
	if (targetWindow) {
		target.receiver = targetWindow;
#ifdef DEBUG_EVENTS
		NSLog(@"routeEvent: Routing unknown event %@ to %@", theEvent, targetWindow);
#endif
		objc_msgSendSuper(&target, @selector(sendEvent:), theEvent);
		return;
	}

	if (part == inContent) {
		WindowModality carbonModal = 0;
		WindowRef unavailableWindow = nil;
		GetWindowModality(GetUserFocusWindow(), &carbonModal, &unavailableWindow);
		if (carbonModal) {
#ifdef DEBUG_EVENTS
			if (type != NSMouseMoved)
				NSLog(@"routeEvent: Discarding event %@ while running Carbon-modal", theEvent);
#endif
			return;
		}
	}

	/* If the event is a left mouse down it matches the NSCarbonWindow of
	 * whatever applet view it occurs in, or the default target if it occurs
	 * outside any of them.  A left mouse down will usually change the
	 * current focus (and key window).  Any other kind of event matches either
	 * the key window already set (if any) or the default target (if there is
	 * no key window -- which means that the focus is outside any applet view).
	 */
	targetController = nil;
	prevKeyWindowController = nil;
	mouseDownMatchesPrevKeyWindow = false;
	noKeyWindow = false;
	if ((controllerArrayHolder = [[NSMutableArray alloc] init]) == nil)
		return;
	if ((mouseDownMatches = [[NSMutableArray alloc] init]) == nil)
		return;
	[controllersLock lock];
	[controllerArrayHolder addObjectsFromArray:controllerArray];
	[controllersLock unlock];
	count = [controllerArrayHolder count];
	for (i = 1; i <= count; ++i) {
		controller = [controllerArrayHolder objectAtIndex:i-1];
		if (!controller->myWindow || !controller->myWindowCarbon ||
			[controller destroying])
			continue;
		if (windowRef != controller->myWindowCarbon)
			continue;
		if (!controller->myAppletView ||
			controller->myAppletView->isInvisible)
			continue;
		/* Unset the focus on an applet view that's been zero-clipped
		 * (because, for example, it's been scrolled off the page).
		 */
		if ([controller->myAppletView isCurrentClipEmpty] ||
			![controller->myAppletView isAppletVisible])
		{
			if (controller->hasKeyWindow) {
				controller->hasKeyWindow = false;
				[controller setCarbonTSMState];
				[controller activateCarbonInputContext];
			}
			continue;
		}
		if ((type != NSLeftMouseDown) || (part != inContent)) {
			if (controller->hasKeyWindow) {
				targetWindow = controller->myWindow;
				targetController = controller;
				break;
			}
			continue;
		}
		if (controller->hasKeyWindow) {
			prevKeyWindowController = controller;
			mouseDownMatchesPrevKeyWindow = false;
		}
		if ((contentView = [controller->myWindow contentView]) == nil) {
			controller->hasKeyWindow = false;
			continue;
		}
		/* Convert from the Cocoa coordinate system (bottom-left origin) to
		 * the Carbon one (top-left origin).
		 */
		locationInView = [contentView
			convertPoint:[theEvent locationInWindow] fromView:nil];
		if (![contentView mouse:locationInView inRect:controller->currentFrame]) {
			controller->hasKeyWindow = false;
			continue;
		}
		[mouseDownMatches addObject:controller];
		if (!controller->hasKeyWindow)
			continue;
		mouseDownMatchesPrevKeyWindow = true;
		controller->hasKeyWindow = false;	// Clear value, possibly temporarily
	}

	/* On OS X 10.3.X and 10.4, the root menu won't get closed if it's
	 * currently open and the user clicks anywhere in a browser window.
	 * See NSAppShadow:cancelMenuTracking for more info.  The same is true
	 * for Java PopupMenu objects.  See the PopupInfo class for more info.
	 */
	if ((osxVersion >= 0x1030) && nsAppShadow &&
		((type == NSLeftMouseDown) || (type == NSRightMouseDown) ||
		(type == NSOtherMouseDown)))
	{
		[nsAppShadow performSelectorOnMainThread:@selector(cancelMenuTracking)
			withObject:nil waitUntilDone:false];
		/* We need to call NSAppShadow dismissJavaPopupMenu (on the main
		 * thread) after a delay.  Otherwise clicking on a (Java) PopupMenu's
		 * "button" won't toggle it.  Instead it will always open or reopen
		 * it, since it it's open now it will get closed in dismissJavaPopupMenu
		 * and then reopened in Handlers.m's safeShowAtSpot_CPM().  Using the
		 * delay makes dismissJavaPopupMenu always get called after
		 * safeShowAtSpot_CPM().
		 */
		if ([PopupInfo getVisible] && [mainThread isEqual:currentThread]) {
			[nsAppShadow performSelector:@selector(dismissJavaPopupMenu)
				withObject:nil afterDelay:.1];
		}
	}

	mouseDownMatchesCount = [mouseDownMatches count];
#ifdef DEBUG_EVENTS
	if ((type == NSLeftMouseDown) && (part == inContent))
		NSLog(@"routeEvent: mouseDownMatchesCount is %i", mouseDownMatchesCount);
#endif

	/* Only reset the user focus to one of our applet views if exactly one of
	 * them is matched.
	 */
	if (mouseDownMatchesCount == 1) {
		targetController = [mouseDownMatches objectAtIndex:0];
		targetController->hasKeyWindow = true;
		targetController->everMouseFocused = true;
		targetWindow = targetController->myWindow;
#ifdef DEBUG_EVENTS
		NSLog(@"routeEvent: Setting focus to window %@ (applet %@) on left mouse click",
			targetWindow, targetController->myAppletView->myAppletName);
#endif
		if (targetController->myLock) {
			[targetController->myLock lock];
			/* If need be, change the NSCarbonWindow object in NSApplication's
			 * window cache to the one that's just been given the user focus.
			 */
			targetController->displayNeedsUpdate = false;
			if (!cacheWindow || (cacheWindow != targetWindow)) {
#ifdef DEBUG_EVENTS
				NSLog(@"routeEvent: Resetting window in cache from %@ to %@",
					defaultTarget, targetWindow);
#endif
				windowNumber = [targetWindow _realWindowNumber];
				if (cacheWindow) {
					cacheController = [JEPController
						controllerFromNSWindow:cacheWindow
						timeoutType:lockTimeoutNone];
					if (cacheController && cacheController->myLock) {
						[cacheController->myLock lock];
						cacheController->hasKeyWindow = false;
						[cacheController setCarbonTSMState];
						[cacheController activateCarbonInputContext];
						if ([cacheWindow isVisible]) {
							[cacheWindow _setVisible:false];
#ifdef DEBUG_EVENTS
							NSLog(@"routeEvent: Made invisible window %@", cacheWindow);
#endif
						}
						[NSApp _removeWindowFromCache:cacheWindow];
						[cacheController->myLock unlock];
					}
				}
				[targetWindow _setWindowNumber:0xFFFFFFFF];
				[targetWindow _setWindowNumber:windowNumber];
				[NSApp _getWindowCache:targetWindow add:true];
			}
			if ([NSApp keyWindow] != targetWindow)
				[targetWindow makeKeyWindow];
			/* NSWindow's documented isVisible method tells you whether or not
			 * the window is onscreen (possibly obscured by other windows).
			 * It's undocumented _setVisible method seems to determine what's
			 * returned by isVisible.  A window must be "visible" (in this
			 * sense) for any of its NSViews to accept user input.  And for
			 * user input to work in applets in tabs, we must set "invisible"
			 * those NSCarbonWindow objects that disappear from the window
			 * cache and set "visible" those that get put into the window
			 * cache.
			 */
			if (![targetWindow isVisible]) {
				/* _invalidateFocus helps avoid the error message
				 * "CGContextReplaceTopGState: invalid context"
				 */
				if ((topView = [targetWindow contentView]) != nil) {
					if ((superView = [topView superview]) != nil)
						topView = superView;
				}
				if (topView)
					[topView _invalidateFocus];
				[targetWindow _setVisible:true];
#ifdef DEBUG_EVENTS
				NSLog(@"routeEvent: Made visible window %@", targetWindow);
#endif
			}
			[targetController->myLock unlock];
		}
		/* If necessary, change the TSM state to one appropriate to the applet
		 * view that's now focused.  Since setting the Cocoa input context
		 * correctly depends on knowing (among other things) what the first
		 * responder will be set to after NSWindow processes our key-down
		 * event, we now (as of JEP 0.8.4) set the Cocoa input context only
		 * after having forwarded theEvent to NSWindow.  (Before JEP 0.9.5 we
		 * only did this if the focus was previously outside any applet.  But
		 * this misses the case of the focus having been in _another_ applet
		 * on the same page.  As a result the Cocoa input context sometimes
		 * got set when the focus was in a browser input field -- which
		 * prevented the browser from receiving any keyboard input in that
		 * field until the user switched the focus somewhere else and back.)
		 */
		if (!targetController->tsmInfo.cocoaStateSet) {
			target.receiver = targetWindow;
			objc_msgSendSuper(&target, @selector(sendEvent:), theEvent);
			[targetController setCocoaTSMState];
			[targetController activateCocoaInputContext];
#ifdef DEBUG_EVENTS
			NSLog(@"routeEvent: TSM state changed from Carbon to Cocoa (window %@, applet %@), first responder %@",
				targetWindow, targetController->myAppletView->myAppletName, [targetWindow firstResponder]);
#endif
			[mouseDownMatches removeAllObjects];
			[mouseDownMatches release];
			[controllerArrayHolder removeAllObjects];
			[controllerArrayHolder release];
			return;
		}
	}
	[mouseDownMatches removeAllObjects];
	[mouseDownMatches release];

	if (!targetWindow) {
		targetWindow = defaultTarget;
		if (mouseDownMatchesCount) {
			/* If a left mouse-down matches more than one of our applet views,
			 * set the current target to whichever of our applet views, if
			 * any, previously had the focus (even ones that weren't matched).
			 */
			if (mouseDownMatchesPrevKeyWindow && prevKeyWindowController) {
				if ((targetWindow = prevKeyWindowController->myWindow) != nil)
					targetController = prevKeyWindowController;
				else targetWindow = defaultTarget;
			}
		}
	}

	if (!targetController) {
		noKeyWindow = true;
#ifdef DEBUG_EVENTS
		if (type != NSMouseMoved)
			NSLog(@"routeEvent: No cocoa view has focus");
#endif
		targetController = [JEPController controllerFromNSWindow:targetWindow
			timeoutType:lockTimeoutNone];
		/* If the focus is now outside any applet view, make sure our TSM
		 * state is appropriate to the Carbon app that owns our parent
		 * window.
		 */
		if (targetController && !targetController->tsmInfo.carbonStateSet) {
			[targetController setCarbonTSMState];
			[targetController activateCarbonInputContext];
#ifdef DEBUG_EVENTS
			NSLog(@"routeEvent: TSM state changed from Cocoa to Carbon (window %@, applet %@), first responder %@",
				targetWindow, targetController->myAppletView ? targetController->myAppletView->myAppletName : nil,
				[targetWindow firstResponder]);
#endif
		}
	} else {
		targetController->hasKeyWindow = true;
		targetController->everMouseFocused = true;
#ifdef DEBUG_EVENTS
		if (type != NSMouseMoved) {
			NSLog(@"routeEvent: Cocoa view in window %@ (applet %@) has focus",
				targetWindow, targetController->myAppletView->myAppletName);
		}
#endif
	}

	/* NSWindow's sendEvent method doesn't handle mouse-moved events correctly
	 * if more than one applet is displayed on a single page (i.e. in a single
	 * Carbon window).  So we handle them ourselves, if we can.
	 */
	if (type == NSMouseMoved) {
		[JEPController routeMouseMovedEvent:theEvent
			targetController:targetController targetWindow:targetWindow];
		[controllerArrayHolder removeAllObjects];
		[controllerArrayHolder release];
		return;
	}

	[controllerArrayHolder removeAllObjects];
	[controllerArrayHolder release];

	/* Discard remaining mouse events that happen outside any applet --
	 * otherwise one of them might get fed to a hidden (i.e. clipped)
	 * applet control.
	 */
	if (noKeyWindow && isMouseEvent && (part == inContent)) {
#ifdef DEBUG_EVENTS
		if (type != NSMouseMoved) {
			NSLog(@"routeEvent: Discarding mouse event %@ outside any applet",
				theEvent);
		}
#endif
		return;
	}

	/* We need to swallow key events that happen outside any of our applet
	 * views -- otherwise two characters may appear in a browser text field
	 * for each one that we type!
	 */
	if (isKeyEvent && noKeyWindow && (type != NSFlagsChanged)) {
#ifdef DEBUG_EVENTS
		NSLog(@"routeEvent: Discarding key event %@ because focus not in Cocoa view", theEvent);
#endif
		return;
	}

#ifdef DEBUG_EVENTS
	if (type != NSMouseMoved) {
		NSLog(@"routeEvent: Routing event %@ to %@ (targetController %@, applet %@)",
			theEvent, targetWindow, targetController, targetController->myAppletView ?
			targetController->myAppletView->myAppletName : nil);
	}
#endif

	target.receiver = targetWindow;
	objc_msgSendSuper(&target, @selector(sendEvent:), theEvent);
}

+ (void)routeMouseMovedEvent:(NSEvent *)theEvent
	targetController:(JEPController *)targetController
	targetWindow:(NSWindow *)targetWindow
{
	WindowRef windowRef;
	unsigned i, count;
	JEPController *controller;
	NSView *contentView;
	NSView *frameView;
	NSView *panelView;
	NSArray *subviews;
	NSPoint locationInView;
	NSMutableArray *holder;

	if (!controllerArray || !controllersLock)
		return;
	if (!theEvent || !targetController || !targetWindow)
		return;

	if ((windowRef = [MyNSWindow existingCarbonWindowFor:targetWindow]) == nil)
		return;
	if ((contentView = [targetWindow contentView]) == nil)
		return;

	if ((holder = [[NSMutableArray alloc] init]) == nil)
		return;

	/* Convert from the Cocoa coordinate system (bottom-left origin) to the
	 * Carbon one (top-left origin).
	 */
	locationInView = [contentView
		convertPoint:[theEvent locationInWindow] fromView:nil];

	subviews = nil, frameView = nil;
	if (![targetController destroying] && targetController->myAppletView &&
		!targetController->myAppletView->isInvisible &&
		[targetController->myAppletView isAppletVisible])
	{
		subviews = [targetController->myAppletView subviews];
	}
	/* myAppletView should have one subview -- an NSViewAWT object
	 * corresponding to a Java Frame object.  We want to send our
	 * mouse-moved events to the NSViewAWT object.
	 */
	if (subviews && ([subviews count] == 1))
		frameView = [subviews objectAtIndex:0];
	/* Always send the mouseMoved event to the frameView corresponding to
	 * targetController, no matter what its location.
	 */
	if (frameView)
		[frameView mouseMoved:theEvent];

	[controllersLock lock];
	[holder addObjectsFromArray:controllerArray];
	[controllersLock unlock];
	count = [holder count];
	for (i = 1; i <= count; ++i) {
		controller = [holder objectAtIndex:i-1];
		if (!controller->myWindowCarbon || [controller destroying])
			continue;
		if ([controller isEqual:targetController])
			continue;
		if (windowRef != controller->myWindowCarbon)
			continue;
		if (!controller->myAppletView ||
			controller->myAppletView->isInvisible ||
			[controller->myAppletView isCurrentClipEmpty] ||
			![controller->myAppletView isAppletVisible])
			continue;
		frameView = panelView = nil;
		subviews = [controller->myAppletView subviews];
		/* myAppletView should have one subview -- an NSViewAWT object
		 * corresponding to a Java Frame object.  The Java Frame object should
		 * in turn have one subview corresponding to a Java Panel object.  We
		 * want to call mouseMovedAWT on the NSViewAWT object corresponding to
		 * the Java Panel object.
		 */
		if (subviews && ([subviews count] == 1))
			frameView = [subviews objectAtIndex:0];
		if (!frameView)
			continue;
		subviews = [frameView subviews];
		if (subviews && ([subviews count] == 1))
			panelView = [subviews objectAtIndex:0];
		if (!panelView)
			continue;
		if (![contentView mouse:locationInView inRect:controller->currentFrame])
			continue;
		/* Only call mouseMovedAWT on a panelView object not belonging to
		 * targetController if the mouseMoved event occurs over it.
		 */
		[panelView mouseMovedAWT:theEvent];
	}
	[holder removeAllObjects];
	[holder release];
}

- (id)init
{
	if ((self = [super init]) == nil)
		return nil;

	myWindow = nil;
	myWindowCarbon = nil;
	myWindowClosing = false;
	myAppletView = nil;
	appletWindow = nil;
	appletAWTContext = nil;
	javaVM = nil;
	pluginInstance = nil;
	createdInvisible = false;
	erasePreviousFrame = false;
	resizeAppletDirectly = false;
	updatesOffCounter = 0;
	updatesOff = false;
	asyncUpdatesOff = false;
	waitingForScreenUpdatesEnabled = false;
	skipDisplayOnce = false;
	skipUpdateOnce = false;
	suspendMouseDown = false;
	hasKeyWindow = false;
	everMouseFocused = false;
	drawAppletPending = false;
	suspendDrawAppletCounter = 0;
	skipDrawAppletOnce = false;
	allowDrawAppletOnce = false;
	displayNeedsUpdate = false;
	fNeedsCleanup = false;
	fAppletDestroyed = false;
	fPurgeOnAppletWindowCleanup = false;
	myLock = nil;
	initialPosition = NSZeroPoint;
	initialHeight = 0;
	initialWidth = 0;
	initialFrame = NSZeroRect;
	currentFrame = NSZeroRect;
	previousFrame = NSZeroRect;
	windowHandlerUPP = nil;
	tsmHandlerUPP = nil;
	dispatcherHandlerUPP = nil;
	windowHandlerRef = nil;
	tsmHandlerRef = nil;
	dispatcherHandlerRef = nil;
	memset(&tsmInfo, 0, sizeof(tsmInfo));
	lastKeyDownControlModifiers = 0;
	showStatusCallback = nil;
	showURLCallback = nil;
	appletCreatedCallback = nil;
	appletInitializedCallback = nil;
	appletStartedCallback = nil;
	showStatusUserData = nil;
	showURLUserData = nil;
	appletCreatedUserData = nil;
	appletInitializedUserData = nil;
	appletStartedUserData = nil;

	if ((myLock = [[NSRecursiveLock alloc] init]) == nil) {
		NSLog(@"JEPController:init couldn't create lock for current controller");
		[self autorelease];
		self = nil;
		return nil;
	}

	return self;
}

- (void)dealloc
{
	NSString *info;
	if (debugRelease) {
		info = [NSString stringWithFormat:@"%@ JEPController:dealloc for %@",
			[NSDate date], (myAppletView ? myAppletView->myAppletName : nil)];
		[AppletView printMessage:info javaEnv:nil];
	}
	[self performSelector:@selector(cleanupSelf)];
	[super dealloc];
}

- (BOOL)destroying
{
	if (fNeedsCleanup || fAppletDestroyed)
		return true;
	if (myAppletView && myAppletView->isDestroying)
		return true;
	if (![JEPController isControllerValid:self])
		return true;
	return false;
}

/* Mozilla's (and Firefox's) event handling on OS X is very old-fashioned --
 * in fact it uses the pre-Carbon EventRecord type when interpreting and
 * creating events (see nsMacMessagePump.cpp and nsMacEventHandler.cpp in
 * widget/src/mac).  But this means that two distinct Carbon events
 * (kEventWindowUpdate and kEventWindowBoundsChanged), which need different
 * handling, get conflated by Mozilla as an "updateEvt".  Furthermore,
 * Mozilla (and Firefox) can rapidly create many spurious updateEvts when
 * you left-click in an applet and then later move the mouse (I suspect this
 * happens in nsMacMessagePump::DoMouseDown() in widget/src/mac/nsMacMessagePump,
 * though updateEvts are only supposed to be created when the mouse is
 * dragged).  The upshot is that (unless we do something about it) each of
 * the spurious updateEvts gets translated into both a kEventWindowUpdate event
 * and a kEventWindowBoundsChanged event, and our entire applet gets rapidly
 * and completely redrawn many times in succession!  My solution is to swallow
 * all kEventWindowUpdate events that closely match any of our applets before
 * Mozilla can see them -- this somehow stops the offending code in
 * nsMacMessagePump::DoMouseDown() from being exercised.  (Also see
 * HandleWindow() in Handlers.m for a related fix.)  As of JEP 0.9.2 we only
 * check for spurious updates in the applet that currently has the focus (the
 * one whose window is in the window cache).  As of JEP 0.9.3 we no longer (as
 * in JEP 0.9.2) do Cocoa updates from here (by calling drawApplet) -- we're
 * letting in more excess updates from Mozilla-family browsers, which makes it
 * unnecessary.
 */
- (OSStatus)handleCarbonUpdate:(EventRef)theEvent
{
	OSStatus retval;
	WindowRef carbonWindow;
	GrafPtr savedPort;
	RgnHandle updateRgn;
	Point windowCorner = {0}, portOrigin = {0};
	Rect updateRect, currentRect, boundsRect;
	BOOL coordinatesMisplaced = false;
	Rect *outsideRects = nil;
	unsigned outsideRectsNum = 0;
	CarbonUpdateParams *params;

	/* Don't allow any updates if our window is closing.  This avoids crashes
	 * (particularly in Firefox 2.0 on OS X Tiger and above) in Handlers.m's
	 * HandleWindow(), on CallNextEventHandler(), when processing a
	 * kEventWindowUpdate event.
	 */
	if (myWindowClosing)
		return noErr;
	if ([self destroying])
		return eventNotHandledErr;
	if (!myWindow || !myAppletView)
		return eventNotHandledErr;
	if (myAppletView->isInvisible || [myAppletView isCurrentClipEmpty] ||
		![myAppletView isAppletVisible])
		return eventNotHandledErr;
	if ((carbonWindow = myWindowCarbon) == nil)
		return eventNotHandledErr;

	if ((updateRgn = NewRgn()) == nil)
		return eventNotHandledErr;
	GetWindowRegion(carbonWindow, kWindowUpdateRgn, updateRgn);
	if (EmptyRgn(updateRgn)) {
		DisposeRgn(updateRgn);
		return noErr;
	}

	GetPort(&savedPort);
	SetPortWindowPort(carbonWindow);

	// As of JEP 0.9.5+e, I've discovered that the "local" coordinate
	// system (the one we use for comparisons to currentFrame and to repost
	// an update) is often misplaced -- sometimes wildly so.  If correctly
	// placed, it will have the same coordinate system as currentFrame.  If
	// it's incorrectly placed, we need to compensate.  At least some of
	// this is caused by Mozilla.org code (in preparation for sending events
	// to plugins), but some may also be due to bugs in Apple's Cocoa-Carbon
	// interface.  My workaround fixes bugzilla.mozilla.org bug 333003.
	LocalToGlobal(&windowCorner);
	GetWindowBounds(carbonWindow, kWindowContentRgn, &boundsRect);
	portOrigin.h = windowCorner.h - boundsRect.left;
	portOrigin.v = windowCorner.v - boundsRect.top;

	// Translate updateRgn from screen coordinates to window "content
	// region" coordinates -- the same coordinates as currentFrame.
	OffsetRgn(updateRgn, -boundsRect.left, -boundsRect.top);
	GetRegionBounds(updateRgn, &updateRect);
	OffsetRect(&boundsRect, -boundsRect.left, -boundsRect.top);

	if (portOrigin.h || portOrigin.v) {
		if (debugUpdates) {
			NSString *info = [NSString stringWithFormat:
				@"%@ JEPController handleCarbonUpdate: Local coordinates misplaced in %@ (X (%i) Y (%i))",
				[NSDate date], myAppletView->myAppletName, portOrigin.h, portOrigin.v];
			[AppletView printMessage:info javaEnv:nil];
		}
		coordinatesMisplaced = true;
	}

	retval = eventNotHandledErr;
	currentRect.left = currentFrame.origin.x;
	currentRect.top = currentFrame.origin.y;
	currentRect.right = currentRect.left + currentFrame.size.width;
	currentRect.bottom = currentRect.top + currentFrame.size.height;
	if (ContainsRectFuzzy(currentRect, updateRect, 2)) {
		if (debugUpdates) {
			NSString *info = [NSString stringWithFormat:
				@"%@ JEPController handleCarbonUpdate: Skipping spurious update in %@",
				[NSDate date], myAppletView->myAppletName];
			[AppletView printMessage:info javaEnv:nil];
		}
		BeginUpdate(carbonWindow);
		EndUpdate(carbonWindow);
		retval = noErr;
	} else if (DiffRgnRect(updateRgn, currentRect, &outsideRectsNum,
		&outsideRects, 0))
	{
		if (debugUpdates) {
			NSString *info = [NSString stringWithFormat:
				@"%@ JEPController handleCarbonUpdate: Skipping spurious part of update in %@",
				[NSDate date], myAppletView->myAppletName];
			[AppletView printMessage:info javaEnv:nil];
		}
		BeginUpdate(carbonWindow);
		EndUpdate(carbonWindow);
		retval = noErr;
		if (outsideRectsNum) {
			// Now that I've discovered (and found out how to work around) the
			// "misplaced local coordinate system" bug (see above), I can
			// repost those parts of an update that don't overlap our plugin.
			// When I tried this in the past (before I understood these bugs),
			// I just generated a bunch of additional spurious updates.
			params = [[CarbonUpdateParams alloc]
				initWithParams:carbonWindow updateRects:outsideRects
				numUpdateRects:outsideRectsNum];
			if (params) {
				[nsAppShadow performSelectorOnMainThread:
					@selector(doCarbonUpdateOnMainThread:)
					withObject:params waitUntilDone:false];
			}
		}
	} else {
		if (outsideRects)
			free(outsideRects);
		/* This is a hack to allow us to expunge those few spurious updates
		 * that we still fail to catch (for example in BonEcho Alpha 2).
		 */
		//if (coordinatesMisplaced)
		//	allowDrawAppletOnce = true;
	}

	SetPort(savedPort);
	DisposeRgn(updateRgn);
	return retval;
}

- (BOOL)purgeOnAppletWindowCleanup
{
	return fPurgeOnAppletWindowCleanup;
}

- (void)saveParentTSMState:(WindowRef)parentWindow
{
	if (!myLock)
		return;
	[myLock lock];
	if (!parentWindow || tsmInfo.stateSaved) {
		[myLock unlock];
		return;
	}
	memset(&tsmInfo, 0, sizeof(tsmInfo));
	/* What TSMGetActiveDocument() returns won't be the browser's active TSM
	 * document if the currently focused window isn't the browser window.
	 * This change resolves bugzilla.mozilla.org's bug 331190, which was
	 * caused by tsmInfo.parentActiveDoc being (incorrectly) set here to the
	 * Java Console's TSM document, leading to crashes in Handlers.m's
	 * HandleAETSM() as a side effect of that routine's fix for
	 * bugzilla.mozilla.org bug 313807.  (So, though this is an old problem,
	 * it only started causing crashes in JEP 0.9.5+b.)  Because it (now)
	 * won't always be possible to set tsmInfo.parentActiveDoc here, I've
	 * added code elsewhere (in Handlers.m's HandleWindow()) that can also
	 * set this variable (in the appropriate circumstances).
	 */
	if (parentWindow == GetUserFocusWindow())
		tsmInfo.parentActiveDoc = TSMGetActiveDocument();
	if (tsmInfo.parentActiveDoc) {
		SetWindowProperty(parentWindow, kJEPCreatorCode,
			kPropertyBrowserTSMDoc, sizeof(TSMDocumentID),
			&(tsmInfo.parentActiveDoc));
	}
	tsmInfo.stateSaved = true;
	[myLock unlock];
}

- (void)copyParentTSMState:(JEPController *)controller
{
	if (!myLock)
		return;
	[myLock lock];
	if (!controller) {
		[myLock unlock];
		return;
	}
	memset(&tsmInfo, 0, sizeof(tsmInfo));
	tsmInfo.parentActiveDoc = controller->tsmInfo.parentActiveDoc;
	tsmInfo.stateSaved = true;
	[myLock unlock];
}

- (void)setCarbonTSMState
{
	if (!myLock)
		return;
	[myLock lock];
	if (tsmInfo.carbonStateSet) {
		[myLock unlock];
		return;
	}
	[self performSelector:@selector(cleanupTSMHandler)];
	tsmInfo.carbonStateSet = true;
	tsmInfo.cocoaStateSet = false;
	[myLock unlock];
}

- (void)setCocoaTSMState
{
	if (!myWindowCarbon || !myLock || [self destroying])
		return;
	[myLock lock];
	if (tsmInfo.cocoaStateSet) {
		[myLock unlock];
		return;
	}
	InstallTSMEventHandler(myWindowCarbon, self);
	tsmInfo.cocoaStateSet = true;
	tsmInfo.carbonStateSet = false;
	[myLock unlock];
}

- (void)activateCarbonInputContext
{
#ifdef USE_DUMMY_TEXTVIEW
	NSResponder *responder;
#endif
	/* We want activateCarbonInputContext to work even when our applet view is
	 * awaiting destruction -- otherwise keyboard input may not work when
	 * we've gone "back" or "forward" to a page without applets from one that
	 * contained at least one applet.
	 */
	if (!myWindow || !myWindowCarbon || hasKeyWindow || !myLock)
		return;
	[myLock lock];
	/* Don't do anything if the keyboard focus isn't in our window.  This may
	 * help with bmo bug 344006, comments 9 and 10.
	 */
	if (GetUserFocusWindow() != myWindowCarbon) {
		[myLock unlock];
		return;
	}
	/* Activate our parent window's TSM doc if we know it, but otherwise don't
	 * change the current TSM doc.
	 */
	if (tsmInfo.parentActiveDoc)
		ActivateTSMDocument(tsmInfo.parentActiveDoc);
	/* This ends up setting myWindow's firstResponder to myWindow itself.
	 */
	[myWindow makeFirstResponder:nil];
	[myLock unlock];
}

- (void)activateCocoaInputContext
{
	NSView *keyView = nil;
	NSResponder *firstResponder = nil;
	NSInputContext *currentContext = nil, *newContext = nil;

	if (!myWindow || !hasKeyWindow || !myAppletView ||
		!myLock || [self destroying])
		return;

	[myLock lock];

	firstResponder = [myWindow firstResponder];

	/* Make sure the current input context is set correctly. */
	firstResponder = [myWindow firstResponder];
	currentContext = [NSInputContext currentInputContext];
	if (firstResponder && [firstResponder isKindOfClass:[NSView class]]) {
		keyView = (NSView *) firstResponder;
		newContext = [keyView inputContext];
	}
	if (newContext && (currentContext != newContext)) {
		if (currentContext)
			[currentContext _deactivate];
		if (newContext)
			[newContext _activate];
	} else if (currentContext) {
		[currentContext _activate];
	}

	[myLock unlock];
}

/* Sun's internal methods for creating a Java applet expect attributes' names
 * to be all lowercase.  Lowercasing them is also more convenient for us (as
 * in CreateCocoaWindow() below).  If a "width" or "height" attribute is
 * expressed in a percentage of the space available (instead of an absolute
 * number of "pixels"), we need to convert it into an absolute value --
 * otherwise Sun's internal methods will choke on it.  Bizarrely, a "style"
 * attribute is legal here, and it may contain "width" or "height" attributes
 * which (apparently) need to override any that were specified at the top
 * level -- so we now (as of JEP 0.9.5+h) expand any "style" attribute we
 * encounter, and do so after we've dealt with the other parameters.  (The
 * "style" change resolves problems with the in-page applet ("Applet
 * Professional") at http://www.sferyx.com/htmleditorapplet/download.htm --
 * the applet didn't change size when you resized the page, and (as a result)
 * mouse events were usually sent to objects the mouse wasn't over.  This
 * problem is ultimately caused by an error in the HTML code that launches the
 * applet -- the non-style "width" tag is set incorrectly (to a fixed value of
 * 800) for non-IE browsers.  But Safari appears to read the "style" tag, and
 * it _is_ legal (though I don't think it really should be), so we might as
 * well support it ourselves, too.
 */
- (NSDictionary *)fixAttributes:(NSDictionary *)attributesToFix
	parentWindow:(WindowRef)parentWindow
{
	id key, object;
	NSArray *orgKeys, *expandedKeys;
	NSMutableDictionary *expandedAttributes, *styleAttributes, *fixedAttributes;
	NSString *lowercaseKey, *fixedObject;
	NSString *keyString, *objectString;
	NSString *styleKey, *styleObject;
	NSArray *styleSpec, *styleLine;
	NSWindow *windowCocoa;
	Rect boundsCarbon = {0};
	NSRect boundsCocoa = {{0},{0}};
	const char *valueChars;
	int value, length, numlen;
	double percent;
	unsigned i, j, count;
	unsigned styleCount;

	if (!attributesToFix || !parentWindow)
		return nil;
	if ((orgKeys = [attributesToFix allKeys]) == nil)
		return attributesToFix;
	if ((fixedAttributes = [[[NSMutableDictionary alloc] init] autorelease]) == nil)
		return attributesToFix;
	if ((expandedAttributes = [[NSMutableDictionary alloc] init]) == nil)
		return attributesToFix;
	if ((styleAttributes = [[NSMutableDictionary alloc] init]) == nil) {
		[expandedAttributes release];
		return attributesToFix;
	}
	if (isCarbonApp) {
		GetWindowBounds(parentWindow, kWindowContentRgn, &boundsCarbon);
		boundsCocoa.origin.x = boundsCocoa.origin.y = 0;
		boundsCocoa.size.width = boundsCarbon.right - boundsCarbon.left;
		boundsCocoa.size.height = boundsCarbon.bottom - boundsCarbon.top;
	} else {
		windowCocoa = [JEPController nsWindowFromCarbonWindow:parentWindow];
		if (!windowCocoa)
			return attributesToFix;
		boundsCocoa = [windowCocoa frame];
		if (osxVersion >= 0x1030) {
			boundsCocoa = [windowCocoa contentRectForFrameRect:boundsCocoa];
		} else {
			boundsCocoa = [windowCocoa contentRectForFrameRect:boundsCocoa
				styleMask:0];
		}
		boundsCocoa.origin.x = boundsCocoa.origin.y = 0;
	}

	count = [orgKeys count];
	for (i = 1; i <= count; ++i) {
		key = [orgKeys objectAtIndex:i-1];
		if (![key isKindOfClass:[NSString class]])
			continue;
		if ((lowercaseKey = [(NSString *)key lowercaseString]) == nil)
			continue;
		if ((object = [attributesToFix objectForKey:key]) == nil)
			continue;
		if (![object isKindOfClass:[NSString class]])
			continue;
		if (![lowercaseKey isEqualToString:@"style"]) {
			[expandedAttributes setObject:object forKey:lowercaseKey];
			continue;
		}
		if ((styleSpec =
			[(NSString *)object componentsSeparatedByString:@";"]) == nil)
			continue;
		styleCount = [styleSpec count];
		for (j = 1; j <= styleCount; ++j) {
			styleLine = [[styleSpec objectAtIndex:j-1]
				componentsSeparatedByString:@":"];
			if ([styleLine count] != 2)
				continue;
			styleKey = [[(NSString *)[styleLine objectAtIndex:0]
				stringByTrimmingCharactersInSet:
				[NSCharacterSet whitespaceCharacterSet]] lowercaseString];
			styleObject = [(NSString *)[styleLine objectAtIndex:1]
				stringByTrimmingCharactersInSet:
				[NSCharacterSet whitespaceCharacterSet]];
			[styleAttributes setObject:styleObject forKey:styleKey];
		}
	}
	[expandedAttributes addEntriesFromDictionary:styleAttributes];
	[styleAttributes removeAllObjects];
	[styleAttributes release];

	expandedKeys = [expandedAttributes allKeys];
	count = [expandedKeys count];
	for (i = 1; i <= count; ++i) {
		keyString = (NSString *) [expandedKeys objectAtIndex:i-1];
		if ((objectString = [expandedAttributes objectForKey:keyString]) == nil)
			continue;
		if (![keyString isEqualToString:@"height"] &&
			![keyString isEqualToString:@"width"])
		{
			[fixedAttributes setObject:objectString forKey:keyString];
			continue;
		}
		valueChars = [objectString UTF8String];
		length = strlen(valueChars);
		if ((numlen = strspn(valueChars, "0123456789.")) == length) {
			value = rint([objectString floatValue]);
		} else if ((numlen == length-1) && (valueChars[numlen] == '%')) {
			resizeAppletDirectly = true;
			percent = [objectString doubleValue] / 100;
			if ([keyString isEqualToString:@"height"]) {
				value = rint(percent * boundsCocoa.size.height);
			} else {
				value = rint(percent * boundsCocoa.size.width);
			}
		} else if ((numlen == length-2) && (strcmp(valueChars+length-2, "px") == 0)) {
			value = rint([objectString floatValue]);
		/* When converting "points" to "pixels", assume that a pixel is 1/96th
		 * inch and a point is 1/72nd inch (as per CSS 2.1, assuming the
		 * "viewing device" is a "computer display").
		 */
		} else if ((numlen == length-2) && (strcmp(valueChars+length-2, "pt") == 0)) {
			value = rint([objectString doubleValue] * (96.0 / 72.0));
		/* When converting "picas" to "pixels", assume that a pixel is 1/96th
		 * inch, a point is 1/72nd inch and a pica is 12 points (as per CSS
		 * 2.1, assuming the "viewing device" is a "computer display").
		 */
		} else if ((numlen == length-2) && (strcmp(valueChars+length-2, "pc") == 0)) {
			value = rint([objectString doubleValue] * (12.0 * 96.0 / 72.0));
		/* When converting inches to "pixels", assume that a pixel is 1/96th
		 * inch (as per CSS 2.1, assuming the "viewing device" is a "computer
		 * display").
		 */
		} else if ((numlen == length-2) && (strcmp(valueChars+length-2, "in") == 0)) {
			value = rint([objectString doubleValue] * 96.0);
		/* When converting centimeters to "pixels", assume that a pixel is .26
		 * mm (as per CSS 2.1, assuming the "viewing device" is a "computer
		 * display").
		 */
		} else if ((numlen == length-2) && (strcmp(valueChars+length-2, "cm") == 0)) {
			value = rint([objectString doubleValue] / .026);
		/* When converting millimeters to "pixels", assume that a pixel is .26
		 * mm (as per CSS 2.1, assuming the "viewing device" is a "computer
		 * display").
		 */
		} else if ((numlen == length-2) && (strcmp(valueChars+length-2, "mm") == 0)) {
			value = rint([objectString doubleValue] / .26);
		/* If we encounter some other unit, assume it's the same size as a
		 * pixel.
		 */
		} else {
			value = rint([objectString floatValue]);
		}
		fixedObject = [NSString stringWithFormat:@"%i", value];
		[fixedAttributes setObject:fixedObject forKey:keyString];
	}
	[expandedAttributes removeAllObjects];
	[expandedAttributes release];

	if (![fixedAttributes count])
		return attributesToFix;
	return (NSDictionary *) fixedAttributes;
}

/* Much of an EmbeddedAppletView's object's initialization takes place in
 * background threads, so this method only "starts" the initialization.
 * See doInitialization in AppletView.m for more information on why this
 * is necessary.
 */
- (AppletView *) startAppletInit:(NSURL *)docbase
	appletAttributes:(NSDictionary *)attributes
{
	AppletView *appletView;

	if (!myWindow)
		return nil;

	if ((appletView = [AppletView alloc]) == nil)
		return nil;
	appletView = [appletView initWithParams:self docbase:docbase
		attributes:attributes];
	if (!appletView)
		NSLog(@"JEPController.startAppletInit couldn't initialize applet");

	return appletView;
}

- (void)setClip:(NSRect)clip
{
	NSThread *currentThread;
	BOOL isMainThread;
	ClipParams *params;
	if (!mainThread || !nsAppShadow)
		return;
	currentThread = [NSThread currentThread];
	isMainThread = [mainThread isEqual:currentThread];
	if (isMainThread) {
		objc_msgSend(self, @selector(setClipImpl:), clip);
	} else {
		params = [[ClipParams alloc] initWithParams:self clip:clip];
		if (params) {
			[nsAppShadow performSelectorOnMainThread:
				@selector(setClipOnMainThread:)
				withObject:params waitUntilDone:true];
		}
	}
}

/* Expects "clip" to be in whatever coordinate system our parent window uses:
 * The Carbon one (origin top-left) if we're in a Carbon app, the Cocoa one
 * (origin bottom-left) if we're in a Cocoa one.
 */
- (void)setClipImpl:(NSRect)clip
{
	if (!myAppletView || !myWindow)
		return;
	if (isCarbonApp)
		[(NSCarbonWindow *) myWindow handleCarbonBoundsChange];
	[myAppletView setClip:clip];
}

/* The Cocoa equivalents of GetWindowBounds() sometimes show stale
 * information.  So we sometimes need to refresh Cocoa's notion of
 * what our current bounds are, especially just after they might
 * have been changed.
 */
- (void)refreshBounds
{
	Rect frameCarbon = {0};
	NSRect frameCocoa = {{0},{0}};
	NSView *contentView, *superView;
	if (!myWindow || !myWindowCarbon)
		return;
	GetWindowBounds(myWindowCarbon, kWindowContentRgn, &frameCarbon);
	frameCocoa.origin.x = frameCocoa.origin.y = 0;
	frameCocoa.size.width = frameCarbon.right - frameCarbon.left;
	frameCocoa.size.height = frameCarbon.bottom - frameCarbon.top;
	if ((contentView = [myWindow contentView]) != nil) {
		[contentView setFrame:frameCocoa];
		if ((superView = [contentView superview]) != nil) {
			[superView setFrame:frameCocoa];
		}
	}
}

- (OSStatus)setAppletState:(AppletStateJEP)newState
{
	OSStatus retval = noErr;
	NSThread *currentThread;
	BOOL isMainThread;
	PerformWithNumberParams *params;
	if (!mainThread || !nsAppShadow)
		return EINVAL;
	currentThread = [NSThread currentThread];
	isMainThread = [mainThread isEqual:currentThread];
	if (isMainThread) {
		retval = (OSStatus) objc_msgSend(self,
			@selector(setAppletStateImpl:), newState);
	} else {
		params = [[PerformWithNumberParams alloc]
			initWithParams:self selector:@selector(setAppletStateImpl:)
			number:newState];
		if (params) {
			[nsAppShadow performSelectorOnMainThread:
				@selector(performWithNumberOnMainThread:)
				withObject:params waitUntilDone:true];
		} else {
			retval = ENOMEM;
		}
	}
	return retval;
}

- (OSStatus)setAppletStateImpl:(AppletStateJEP)newState
{
	OSStatus retval = noErr;

	if (!myAppletView)
		return ENOENT;

	switch (newState) {
		case kAppletStartJEP:
			if (!myAppletView->myApplet) {
				retval = EINVAL;
				break;
			}
			if (myAppletView->appletState == kAppletInitJEP) {
				[myAppletView startApplet];
			} else if (myAppletView->appletState == kAppletStopJEP) {
				[myAppletView appletStartApplet];
			}
			break;
		case kAppletStopJEP:
			if (!myAppletView->myApplet) {
				retval = EINVAL;
				break;
			}
			if (myAppletView->appletState == kAppletStartJEP) {
				[myAppletView appletStopApplet];
			}
			break;
		case kAppletDestroyJEP:
			[self performSelector:@selector(destroyApplet)];
			break;
		default:
			retval = EINVAL;
			break;
	}

	return retval;
}

- (void)drawApplet
{
	NSThread *currentThread;
	if (!mainThread || !nsAppShadow || updatesOff ||
		[self destroying] || ![myAppletView isDisplayable])
		return;
	if (drawAppletPending)
		return;
	drawAppletPending = true;
	currentThread = [NSThread currentThread];
	if ([mainThread isEqual:currentThread]) {
		[nsAppShadow drawAppletOnMainThread:self];
	} else {
		[nsAppShadow performSelectorOnMainThread:@selector(drawAppletOnMainThread:)
			withObject:self waitUntilDone:false];
	}
}

- (void)maybeDrawApplet
{
	NSThread *currentThread;
	if (!mainThread || !nsAppShadow || updatesOff ||
		[self destroying] || ![myAppletView isDisplayable])
		return;
	if (skipDrawAppletOnce)
		allowDrawAppletOnce = false;
	if (allowDrawAppletOnce) {
		allowDrawAppletOnce = false;
	} else if (skipDrawAppletOnce || (suspendDrawAppletCounter > 0)) {
		if (skipDrawAppletOnce)
			skipDrawAppletOnce = false;
		if (debugUpdates) {
			NSString *info = [NSString stringWithFormat:
				@"%@ JEPController maybeDrawApplet: Skipping spurious update in %@",
				[NSDate date], myAppletView->myAppletName];
			[AppletView printMessage:info javaEnv:nil];
		}
		return;
	}
	if (drawAppletPending)
		return;
	drawAppletPending = true;
	currentThread = [NSThread currentThread];
	if ([mainThread isEqual:currentThread]) {
		[self performSelector:@selector(scheduleDrawApplet)];
	} else {
		[self performSelectorOnMainThread:@selector(scheduleDrawApplet)
			withObject:nil waitUntilDone:false];
	}
}

/* We want drawAppletOnMainThread to run on the main thread, so we must also
 * schedule it (using performSelector:withObject:afterDelay:) on the main
 * thread.
 */
- (void)scheduleDrawApplet
{
	if (!drawAppletPending || !nsAppShadow || [self destroying])
		return;
	[nsAppShadow performSelector:@selector(drawAppletOnMainThread:)
		withObject:self afterDelay:.2];
}

- (void)suspendAppletUpdates
{
	if (!myAppletView || [self destroying])
		return;
	updatesOff = true;
	asyncUpdatesOff = true;
	[myAppletView setUpdatesOff:true];
	if (++updatesOffCounter != 1)
		return;
	suspendDrawAppletCounter = 1;
	if ([myAppletView isDisplayable])
		previousFrame = currentFrame;
}

/* Mozilla (like Firefox) tends to suspend and resume updates very rapidly --
 * which can cause the applet to flash annoyingly.  In resumeAppletUpdates and
 * friends we cause fewer switches to happen, less rapidly.  The applet still
 * flashes, but not quite so often, which hopefully is less annoying.
 */
- (void)resumeAppletUpdates
{
	NSThread *currentThread;
	if (!mainThread || !myAppletView || [self destroying])
		return;
	/* The following returns before resumeAfterWait1 and so forth are called
	 * -- which is exactly what we want:  Otherwise we bog down the display
	 * in our parent browser (Mozilla, Firefox, Camino, or whatever).
	 */
	currentThread = [NSThread currentThread];
	if ([mainThread isEqual:currentThread]) {
		[self performSelector:@selector(resumeAfterWait1) withObject:nil];
	} else {
		[self performSelectorOnMainThread:@selector(resumeAfterWait1)
			withObject:nil waitUntilDone:false];
	}
}

/* We want resumeAfterWait2 to run on the main thread, so we must also
 * schedule it (using performSelector:withObject:afterDelay:) on the main
 * thread.
 */
- (void)resumeAfterWait1
{
	[self performSelector:@selector(resumeAfterWait2) withObject:nil
		afterDelay:.1];
}

/* As of JEP 0.9.1 we distinguish between "synchronous" updates (which take
 * place on the main thread) and "asynchronous" ones (which take place on
 * secondary threads).  Both kinds get turned off at the same time in
 * suspendAppletUpdates (above), but we turn asynchronous updates back on
 * slightly after synchronous ones.  This is because (at least on Java 1.4.2
 * Update 1 and later, where we can no longer clip "third thread" drawing)
 * asynchronous updates are much more prone to ghosts just after screen
 * updates have resumed.  Fortunately many of the ghosts disappear if we
 * insert a delay between resuming normal updates and resuming asynchronous
 * ones.
 */
- (void)resumeAfterWait2
{
	if (--updatesOffCounter != 0) {
		if (updatesOffCounter < 0)
			updatesOffCounter = 0;
		return;
	}
	if (!myAppletView || !myWindow || [self destroying])
		return;
	updatesOff = false;
	[myAppletView setUpdatesOff:false];
	suspendDrawAppletCounter = 0;
	[self performSelector:@selector(resumeAfterWait3) withObject:nil
		afterDelay:.1];
}

- (void)resumeAfterWait3
{
	if ((updatesOffCounter > 0) ||
		!nsAppShadow || !myAppletView || [self destroying])
		return;
	asyncUpdatesOff = false;
	[myAppletView setAsyncUpdatesOff:false];
	[nsAppShadow scheduleSuspendDrawApplet:self];
	[self drawApplet];
}

- (void)sizeApplet:(NSSize)size
{
	NSThread *currentThread;
	BOOL isMainThread;
	SizeParams *params;
	if (!mainThread || !nsAppShadow)
		return;
	currentThread = [NSThread currentThread];
	isMainThread = [mainThread isEqual:currentThread];
	if (isMainThread) {
		objc_msgSend(self, @selector(sizeAppletImpl:), size);
	} else {
		params = [[SizeParams alloc] initWithParams:self size:size];
		if (params) {
			[nsAppShadow performSelectorOnMainThread:
				@selector(sizeOnMainThread:)
				withObject:params waitUntilDone:true];
		}
	}
}

/* We resize our applet, not by calling its (Java-based) resize method, but
 * by changing its "frame" without changing its "bounds".  This makes Cocoa
 * do the resizing.  Cocoa does a fine job, and in any case some applets don't
 * natively support resizing.  NSView's "setFrame" methods normally change
 * the "bounds" at the same time they change the "frame", so we override them
 * in AppletView.m to correct this behavior.  (As of JEP 0.8.3 we now also
 * support resizing applets directly or "natively", when that's appropriate.
 * See resizeApplet() in AppletFrame.java for more info.)
 */
- (void)sizeAppletImpl:(NSSize)size
{
	if (!myAppletView || !myWindow || !myLock || [self destroying])
		return;

	[myLock lock];

	if ((size.width < 0) || (size.height < 0))
		size.width = size.height = 0;

	if ([myAppletView isDisplayable] && !updatesOff)
		previousFrame = currentFrame;

	[myAppletView setNeedsDisplay:false];
	[myAppletView setFrameSize:size];
	[self drawApplet];

	[myLock unlock];
}

- (void)positionApplet:(NSPoint)point
{
	NSThread *currentThread;
	BOOL isMainThread;
	PositionParams *params;
	if (!mainThread || !nsAppShadow)
		return;
	currentThread = [NSThread currentThread];
	isMainThread = [mainThread isEqual:currentThread];
	if (isMainThread) {
		objc_msgSend(self, @selector(positionAppletImpl:), point);
	} else {
		params = [[PositionParams alloc] initWithParams:self position:point];
		if (params) {
			[nsAppShadow performSelectorOnMainThread:
				@selector(positionOnMainThread:)
				withObject:params waitUntilDone:true];
		}
	}
}

/* Expects "point" to be in whatever coordinate system our parent window uses:
 * The Carbon one (origin top-left) if we're in a Carbon app, the Cocoa one
 * (origin bottom-left) if we're in a Cocoa one.
 */
- (void)positionAppletImpl:(NSPoint)point
{
	NSRect appletFrame;

	if (!myAppletView || !myWindow || !myLock || [self destroying])
		return;

	[myLock lock];

	appletFrame = [myAppletView frame];
	if ((appletFrame.origin.x == point.x) &&
		(appletFrame.origin.y == point.y))
	{
		[myLock unlock];
		return;
	}

	if ([myAppletView isDisplayable] && !updatesOff)
		previousFrame = currentFrame;

	[myAppletView setNeedsDisplay:false];
	[myAppletView setFrameOrigin:point];
	[self drawApplet];

	[myLock unlock];
}

- (void)showHideApplet:(BOOL)visible
{
	NSThread *currentThread;
	BOOL isMainThread;
	PerformWithNumberParams *params;
	if (!mainThread || !nsAppShadow)
		return;
	currentThread = [NSThread currentThread];
	isMainThread = [mainThread isEqual:currentThread];
	if (isMainThread) {
		objc_msgSend(self, @selector(showHideAppletImpl:), visible);
	} else {
		params = [[PerformWithNumberParams alloc]
			initWithParams:self selector:@selector(showHideAppletImpl:)
			number:visible];
		if (params) {
			[nsAppShadow performSelectorOnMainThread:
				@selector(performWithNumberOnMainThread:)
				withObject:params waitUntilDone:true];
		}
	}
}

- (void)showHideAppletImpl:(BOOL)visible
{
	int windowNumber;
	NSWindow *cacheWindow;
	NSView *topView, *superView;
	BOOL visibilityChanged;
	JEPController *cacheController = nil;
	NSString *info = nil;

	if (!myAppletView || !myWindow || [self destroying])
		return;

	visibilityChanged = (myAppletView->isInvisible == visible);

	if (debugVisibility) {
		info = [NSString stringWithFormat:
			@"%@ JEPController:showHideApplet making %@ %s", [NSDate date],
			myAppletView->myAppletName, (visible ? "visible" : "invisible")];
		[AppletView printMessage:info javaEnv:nil];
	}

	if ((myAppletView->appletState < kAppletInitJEP) ||
		(myAppletView->appletState == kAppletErrorJEP))
	{
		[myAppletView setAppletVisible:visible flagOnly:true wait:false];
		return;
	}

	/* The following code helps deal with applets in multiple tabs in the
	 * same window:  When switching between tabs, Mozilla-family browsers use
	 * clipping or the "ShowHide" function to make the applet views visible or
	 * invisible.  The NSCarbonWindow objects of applet views that become
	 * invisible should be removed from the NSApplication window cache, and
	 * the NSCarbonWindow objects of applet views that become visible should
	 * be added to the window cache.
	 */
	/* Mozilla and family use a wierd and seemingly arbitrary combination of
	 * clipping and the "ShowHide" function to show and hide applets.
	 * Specifically, they often use "ShowHide" to "show" an applet while it's
	 * still zero-clipped (for example when the current tab or window contains
	 * an applet and you open a new tab, or whenever you resize a window
	 * containing applets in more than one tab).  It'd be much better to use
	 * only the "ShowHide" function.  Browser support for applets in tabs
	 * might be made easier if I enhanced the JEP API to explicitly support
	 * tabs (e.g. JEPGetCurrentTab(), JEPCreateTab(), JEPDestroyTab(),
	 * JEPChangeTab() ...).
	 */
	[myAppletView setAppletVisible:visible flagOnly:false wait:false];
	if (isCarbonApp && myLock) {
		[myLock lock];
		if (!visible) {
			hasKeyWindow = false;
			[self setCarbonTSMState];
			[self activateCarbonInputContext];
			if ([myWindow isVisible]) {
				[myWindow _setVisible:false];
#ifdef DEBUG_EVENTS
				NSLog(@"showHideApplet: Made invisible (directly) window %@", myWindow);
#endif
			}
			/* As of JEP 0.9.6, we now always remove our NSWindow from the
			 * window cache here (after we've made sure that it's invisible).
			 * This avoids complicated wierdness -- such as sometimes having
			 * the wrong window in the window cache, which can lead to our
			 * code being mistaken about which applet is focused (if we have
			 * applets in several different tabs), which in turn can sometimes
			 * break keyboard input in browser fields.
			 */
			[NSApp _removeWindowFromCache:myWindow];
		}
		if (visible && ![myAppletView isCurrentClipEmpty]) {
			displayNeedsUpdate = false;
			windowNumber = [myWindow _realWindowNumber];
			cacheWindow = [NSApp windowWithWindowNumber:windowNumber];
			if (!cacheWindow || (cacheWindow != myWindow)) {
				if (cacheWindow) {
					cacheController = [JEPController
						controllerFromNSWindow:cacheWindow
						timeoutType:lockTimeoutShort];
					if (cacheController && cacheController->myLock) {
						[cacheController->myLock lock];
						cacheController->hasKeyWindow = false;
						[cacheController setCarbonTSMState];
						[cacheController activateCarbonInputContext];
						if ([cacheWindow isVisible]) {
							[cacheWindow _setVisible:false];
#ifdef DEBUG_EVENTS
							NSLog(@"showHideApplet: Made invisible (indirectly) window %@", cacheWindow);
#endif
						}
						[NSApp _removeWindowFromCache:cacheWindow];
						[cacheController->myLock unlock];
					}
				}
				[myWindow _setWindowNumber:0xFFFFFFFF];
				[myWindow _setWindowNumber:windowNumber];
				[NSApp _getWindowCache:myWindow add:true];
#ifdef DEBUG_EVENTS
				NSLog(@"showHideApplet: Reset window in cache from %@ to %@, first responder now %@",
					cacheWindow, myWindow, [myWindow firstResponder]);
#endif
			}
			if (![myWindow isVisible]) {
				/* _invalidateFocus helps avoid the error message
				 * "CGContextReplaceTopGState: invalid context"
				 */
				if ((topView = [myWindow contentView]) != nil) {
					if ((superView = [topView superview]) != nil)
						topView = superView;
				}
				if (topView)
					[topView _invalidateFocus];
				[myWindow _setVisible:true];
#ifdef DEBUG_EVENTS
				NSLog(@"showHideApplet: Made visible window %@", myWindow);
#endif
			}
		}
		[myLock unlock];
	}

	/* Our fix to registerModelForUpdate (in Handlers.m) doesn't work properly
	 * unless we sometimes "wake it up" (information on whether our applet is
	 * visible or invisible can sometimes be out of date).  This is
	 * unnecessary on Tiger (the problem seems to have been fixed) and also
	 * causes freezes (since we're triggering Java-related display activity on
	 * the main thread).
	 */
	if ((osxVersion < 0x1040) && isCarbonApp && appletWindow && myAppletView->myApplet &&
		visible && ![myAppletView isCurrentClipEmpty])
	{
		if (isJava142) {
			[NSClassFromString(@"AWTModelUpdater")
				registerModelForUpdate:appletWindow];
		} else {
			[NSClassFromString(@"ThreadUtilities")
				registerModelForUpdate:appletWindow];
		}
	}
}

- (void)setAppletCreatedCallback:(JEPAppletCreatedCallback)infoFunction
	userData:(void *)userData
{
	appletCreatedCallback = infoFunction;
	if (!infoFunction)
		appletCreatedUserData = nil;
	else appletCreatedUserData = userData;
}

- (void)setAppletInitializedCallback:(JEPAppletInitializedCallback)infoFunction
	userData:(void *)userData
{
	appletInitializedCallback = infoFunction;
	if (!infoFunction)
		appletInitializedUserData = nil;
	else appletInitializedUserData = userData;
}

- (void)setAppletStartedCallback:(JEPAppletStartedCallback)infoFunction
	userData:(void *)userData
{
	appletStartedCallback = infoFunction;
	if (!infoFunction)
		appletStartedUserData = nil;
	else appletStartedUserData = userData;
}

- (void)setShowStatusCallback:(JEPShowStatusCallback)showStatusFunction
	userData:(void *)userData
{
	showStatusCallback = showStatusFunction;
	if (!showStatusFunction)
		showStatusUserData = nil;
	else showStatusUserData = userData;
}

- (void)setShowURLCallback:(JEPShowDocumentCallback)showURLFunction
	userData:(void *)userData
{
	showURLCallback = showURLFunction;
	if (!showURLFunction)
		showURLUserData = nil;
	else showURLUserData = userData;
}

/* For safety's sake, we should always call this on the main thread (using
 * performSelectorOnMainThread).
 */
- (void)afterInitializeOnMainThread
{
	int windowNumber;
	NSWindow *cacheWindow;
	JEPController *cacheController = nil;
	NSView *topView = nil, *superView = nil;
	BOOL changeWindowCache = false;

	if (!isCarbonApp || !myAppletView || !myAppletView->myApplet || !myWindow)
		return;

	/* Some applets have no user interface of their own (e.g. the
	 * authentication applet used by Postbank.nl at
	 * https://www.p3.postbank.nl/sesam/SesamLoginServlet).  We should treat
	 * these like we do an empty applet view -- otherwise it will be
	 * impossible to input characters anywhere in the page.
	 */
	if (![myAppletView isDisplayable] || [myAppletView areAppletBoundsEmpty]) {
		return;
	}

	/* If our new applet has been created in a tab (and one or more other tabs
	 * also contain applets), we might need to fiddle with the window cache.
	 */
	displayNeedsUpdate = false;
	windowNumber = [myWindow _realWindowNumber];
	cacheWindow = [NSApp windowWithWindowNumber:windowNumber];
	if (!cacheWindow || (cacheWindow != myWindow)) {
		changeWindowCache = true;
		if (cacheWindow) {
			cacheController = [JEPController controllerFromNSWindow:cacheWindow
				timeoutType:lockTimeoutNone];
			if (cacheController && cacheController->myLock) {
				[cacheController->myLock lock];
				cacheController->hasKeyWindow = false;
				[cacheController setCarbonTSMState];
				[cacheController activateCarbonInputContext];
				if ([cacheWindow isVisible]) {
					[cacheWindow _setVisible:false];
#ifdef DEBUG_EVENTS
					NSLog(@"afterInitializeOnMainThread: Made invisible (indirectly) window %@", cacheWindow);
#endif
				}
				[NSApp _removeWindowFromCache:cacheWindow];
				[cacheController->myLock unlock];
			}
		}
		[myWindow _setWindowNumber:0xFFFFFFFF];
		[myWindow _setWindowNumber:windowNumber];
		[NSApp _getWindowCache:myWindow add:true];
#ifdef DEBUG_EVENTS
		NSLog(@"afterInitializeOnMainThread: Reset window in cache from %@ to %@, first responder now %@",
			cacheWindow, myWindow, [myWindow firstResponder]);
#endif
	}

	/* _invalidateFocus helps avoid the error message
	 * "CGContextReplaceTopGState: invalid context"
	 */
	if (changeWindowCache) {
		if ((topView = [myWindow contentView]) != nil) {
			if ((superView = [topView superview]) != nil)
				topView = superView;
		}
		if (topView)
			[topView _invalidateFocus];
	}
	/* As of JEP 0.8.4, _setVisible takes the place of orderFront and
	 * skipDisplayOnce.  This re-enables window ordering and allows
	 * keyboard input in applets, without forcing the current browser
	 * window to the front.
	 */
	/* The following seems more effective than makeKeyWindow, which we used
	 * before JEP 0.9.4.
	 */
	/* As of JEP 0.9.6/0.9.5+e, we're now turning on the plugin's "Cocoa
	 * state" (and if necessary making its window key and visible) when
	 * that's actually needed -- as the plugin is made first responder.  For
	 * more information see becomeFirstResponder_FM() in Handlers.m.
	 */
}

/* For safety's sake, we should always call this on the main thread (using
 * performSelectorOnMainThread).
 */
- (void)maybeStartAppletOnMainThread
{
	if (!myWindow || !myAppletView || !myAppletView->myApplet || [self destroying])
		return;

	[self performSelector:@selector(afterInitializeOnMainThread)];
	[self setAppletState:kAppletStartJEP];

	if (![self destroying]) {
		/* This largely eliminates the (very annoying) tendency for some
		 * applets to be partially or totally invisible after they've been
		 * loaded (particularly for the first time).  It seems that the JVM
		 * doesn't automatically make an applet visible when it calls its
		 * start() method.
		 */
		if (![myAppletView areAppletBoundsEmpty]) {
			[myAppletView setAppletVisible:true flagOnly:false wait:true];
			drawAppletPending = true;
			[self scheduleDrawApplet];
		}
		/* Purge older JEPController objects whose applets have been
		 * destroyed, but never purge ourselves -- that's unsafe.
		 */
		[JEPController purgeIfAppletDestroyedInWindow:myWindowCarbon];
	}

	return;
}

- (void)showURL:(NSURL *)url inFrame:(NSString *)windowName
{
	ShowURLParams *params;
	NSThread *currentThread;
	if (!mainThread || !nsAppShadow || !showURLCallback || !url ||
		!myAppletView || [self destroying])
		return;
	params = [[ShowURLParams alloc]
		initWithParams:url windowName:windowName controller:self];
	if (!params)
		return;
	currentThread = [NSThread currentThread];
	if ([mainThread isEqual:currentThread]) {
		[nsAppShadow showURLOnMainThread:params];
	} else {
		[nsAppShadow performSelectorOnMainThread:@selector(showURLOnMainThread:)
			withObject:params waitUntilDone:true];
	}
}

- (void)showStatus:(NSString *)statusMessage
{
	ShowStatusParams *params;
	NSThread *currentThread;
	if (!mainThread || !nsAppShadow || !showStatusCallback || !statusMessage ||
		!myAppletView || [self destroying])
		return;
	params = [[ShowStatusParams alloc]
		initWithParams:statusMessage controller:self];
	if (!params)
		return;
	currentThread = [NSThread currentThread];
	if ([mainThread isEqual:currentThread]) {
		[nsAppShadow showStatusOnMainThread:params];
	} else {
		[nsAppShadow performSelectorOnMainThread:@selector(showStatusOnMainThread:)
			withObject:params waitUntilDone:false];
	}
}

/* In a Carbon app, this won't get called unless we call it ourselves (from
 * close_CW() in Handlers.m and/or NSAppShadow:applicationWillTerminate here),
 * even if we register for a notification or make ourselves a delegate of
 * myWindow -- Cocoa event handling is completely broken in NSCarbonWindow
 * objects.
 */
- (void)windowWillClose:(NSNotification *)notification
{
#ifdef DEBUG_EVENTS
	NSLog(@"windowWillClose: Controller %@, applet 0x%.8X, window %@, carbonWindow 0x%.8X",
		self, (myAppletView == nil) ? nil : myAppletView->myApplet, myWindow, myWindowCarbon);
#endif
	myWindowClosing = true;
	[self performSelector:@selector(destroyApplet)];
	/* We sometimes get into trouble (i.e. we crash) if we don't call
	 * cleanupWindow as soon as possible.  Since we no longer explicitly
	 * "close" myWindow, we no longer need to wait to call cleanupWindow
	 * until after we've called cleanupAppletView.
	 */
	[self performSelector:@selector(cleanupWindow)];
	if (myAppletView && myAppletView->awaitingAppletWindowCleanup) {
		fPurgeOnAppletWindowCleanup = true;
	} else {
		fNeedsCleanup = true;
		[JEPController purgeControllerArray];
	}
}

/* When we destroy our applet we deliberately leave our per-applet-view
 * handlers installed.  We only uninstall them when our window gets closed
 * (along with the handlers for any other applet-view objects that window
 * contains).  Once an applet has been created in a Carbon browser's
 * window (even if it no longer exists), without our handlers that window
 * will no longer receive keyboard or mouse events.
 */
- (void)destroyApplet
{
#ifdef DEBUG_EVENTS
	NSLog(@"destroyApplet: Controller %@, applet 0x%.8X, window %@",
		self, (myAppletView == nil) ? nil : myAppletView->myApplet, myWindow);
#endif
	[self suspendAppletUpdates];
	[self performSelector:@selector(cleanupApplet)];
}

- (void)cleanupSelf
{
	[PopupInfo removeInfoForController:self];
	if (!myLock)
		return;
	[myLock lock];
#ifdef DEBUG_EVENTS
	NSLog(@"cleanupSelf: Controller %@, applet 0x%.8X, window %@",
		self, (myAppletView == nil) ? nil : myAppletView->myApplet, myWindow);
#endif
	if (!isCarbonApp)
		[[NSNotificationCenter defaultCenter] removeObserver:self];
	[self performSelector:@selector(cleanupApplet)];
	[self performSelector:@selector(cleanupWindow)];
	[self performSelector:@selector(cleanupAppletView)];
	[self performSelector:@selector(cleanupHandlers)];
	[myLock unlock];
	[myLock autorelease];
	myLock = nil;
}

- (void)cleanupWindow
{
	if (!myLock)
		return;
	[myLock lock];
	if (myWindow) {
		/* Autoreleasing myWindow sometimes causes crashes (it sometimes gets
		 * released before it can be used in close_CW() in Handlers.m), so
		 * we're no longer doing it (as of JEP 0.8.3).  NSWindow's (and
		 * NSCarbonWindow's) close message should take care of releasing it.
		 */
		myWindow = nil;
		myWindowCarbon = nil;
	}
	hasKeyWindow = false;
	[myLock unlock];
}

- (void)cleanupHandlers
{
	if (!myLock)
		return;
	[myLock lock];
#ifdef DEBUG_EVENTS
	NSLog(@"cleanupHandlers: Controller %@, window %@", self, myWindow);
#endif
	if (windowHandlerRef) {
		RemoveEventHandler(windowHandlerRef);
		windowHandlerRef = nil;
	}
	if (tsmHandlerRef) {
		RemoveEventHandler(tsmHandlerRef);
		tsmHandlerRef = nil;
	}
	if (dispatcherHandlerRef) {
		RemoveEventHandler(dispatcherHandlerRef);
		dispatcherHandlerRef = nil;
	}
	if (windowHandlerUPP) {
		DisposeEventHandlerUPP(windowHandlerUPP);
		windowHandlerUPP = nil;
	}
	if (tsmHandlerUPP) {
		DisposeEventHandlerUPP(tsmHandlerUPP);
		tsmHandlerUPP = nil;
	}
	if (dispatcherHandlerUPP) {
		DisposeEventHandlerUPP(dispatcherHandlerUPP);
		dispatcherHandlerUPP = nil;
	}
	hasKeyWindow = false;
	[myLock unlock];
}

- (void)cleanupTSMHandler
{
	if (!myLock)
		return;
	[myLock lock];
	if (tsmHandlerRef) {
		RemoveEventHandler(tsmHandlerRef);
		tsmHandlerRef = nil;
	}
	if (tsmHandlerUPP) {
		DisposeEventHandlerUPP(tsmHandlerUPP);
		tsmHandlerUPP = nil;
	}
	[myLock unlock];
}

- (void)cleanupApplet
{
	if (!myLock || !myAppletView)
		return;
#ifdef DEBUG_EVENTS
	NSLog(@"cleanupApplet: Controller %@, applet 0x%.8X, window %@",
		self, myAppletView->myApplet, myWindow);
#endif
	if (myAppletView->myApplet)
		[myAppletView stopApplet];
	[myAppletView destroyApplet];
	/* In a Carbon app we need to remove our NSCarbonWindow object from
	 * NSApplication's window cache (see above under routeEvent for more
	 * info).  When/if another applet view is created,  CreateCarbonWindow()
	 * (below) will add a new NSCarbonWindow object to this cache (it
	 * happens automatically when we send the initWithCarbonWindowRef
	 * message to our new NSCarbonWindow object).
	 */
	[myLock lock];
	hasKeyWindow = false;
	if (isCarbonApp && myWindow) {
		/* In case this NSCarbonWindow object remains in use for a while
		 * before a new one is created (along with a new applet).
		 */
		[self setCarbonTSMState];
		[self activateCarbonInputContext];
		[myWindow _setVisible:false];
		[NSApp _removeWindowFromCache:myWindow];
	}
	fAppletDestroyed = TRUE;
	[myLock unlock];
}

- (void)cleanupAppletView
{
	if (!myLock)
		return;
	[myLock lock];
	if (myAppletView) {
		[myAppletView removeFromSuperview];
		[myAppletView release];
		myAppletView = nil;
	}
	hasKeyWindow = false;
	[myLock unlock];
}

- (void)unsetHasKeyWindow
{
	hasKeyWindow = false;
}

- (void)setWindowClosing
{
	myWindowClosing = true;
}

- (void)unsetWindowClosing
{
	myWindowClosing = false;
}

- (void)setNeedsCleanup
{
	fNeedsCleanup = true;
}

- (void)setAppletDestroyed
{
	fAppletDestroyed = true;
}

- (unsigned)hash
{
	return (unsigned) self;
}

- (BOOL)isEqual:(id)anObject
{
	/* anObject may no longer be valid -- which means we may not be able to
	 * send it any messages.  This first test helps head off that possiblity.
	 */
	if (self != anObject)
		return FALSE;
	if ([self hash] != [anObject hash])
		return FALSE;
	if (![anObject isMemberOfClass:[self class]])
		return FALSE;
	return TRUE;
}

@end

/* Non-exported C utility functions */

/* Called with autorelease pool already in place. */
static OSStatus CreateCocoaWindow(JEPController *controller,
	NSDictionary *attributes, WindowRef parentWindow, Rect appletClip,
	int appletPosX, int appletPosY, unsigned appletWidth, unsigned appletHeight)
{
	id heightValue, widthValue;
	unsigned heightLocal, widthLocal;
	Rect frameCarbon = {0};
	NSRect frameCocoa = {{0},{0}};
	NSRect frameHolder = {{0},{0}};
	NSWindow *window = nil;
	NSInvocation *message = nil;
	NSMethodSignature *sig = nil;
	BOOL falseVal = false;
	NSMutableArray *controllersFound;
	JEPController *prevController;
	OSStatus retval = noErr;

	if (!controller || !parentWindow)
		return EINVAL;
	if (controller->myWindow)	// Don't create another window if one
		return noErr;			// already exists (shouldn't be possible).

	heightLocal = appletHeight;
	widthLocal = appletWidth;
	if (attributes) {
		if ((heightValue = [attributes objectForKey:@"height"]) != nil) {
			if ([heightValue isKindOfClass:[NSString class]])
				heightLocal = [(NSString *)heightValue intValue];
		}
		if ((widthValue = [attributes objectForKey:@"width"]) != nil) {
			if ([widthValue isKindOfClass:[NSString class]])
				widthLocal = [(NSString *)widthValue intValue];
		}
	}

	if (isCarbonApp) {
		/* Apple's documented method for "attaching" an NSWindow object to a
		 * Carbon window is to do the following:
		 *      myWindow = [[NSWindow alloc] initWithWindowRef:parentWindow];
		 * After the call myWindow gets turned into an NSCarbonWindow object.
		 * But Apple's implementation of this class is crawling with bugs
		 * (some of them big and nasty enough to be really design flaws), so
		 * we need to subclass NSCarbonWindow (and also do many other things)
		 * to fix the bugs.  In these circumstances it's more convenient to
		 * create our NSCarbonWindow object directly -- fortunately that's
		 * possible.
		 */
		if ((window = [carbonWindowClass alloc]) == nil)
			return ENOMEM;
		sig = [carbonWindowClass instanceMethodSignatureForSelector:
			@selector(initWithCarbonWindowRef:takingOwnership:disableOrdering:)];
		if (!sig)
			return ENOMEM;
		message = [NSInvocation invocationWithMethodSignature:sig];
		if (!message)
			return ENOMEM;
		[message setSelector:@selector(initWithCarbonWindowRef:takingOwnership:disableOrdering:)];
		[message setArgument:&parentWindow atIndex:2];
		[message setArgument:&falseVal atIndex:3];
		[message setArgument:&falseVal atIndex:4];
		[message invokeWithTarget:window];
		[message getReturnValue:&window];
		if (window)
			[window retain];
	} else {
		/* In a Cocoa app we can just use the parent window directly, but
		 * first we need to find the NSWindow object that corresponds to
		 * our parent window's WindowRef.
		 */
		window = [JEPController nsWindowFromCarbonWindow:parentWindow];
	}
	if (!window)
		return ENOMEM;

	/* We do this to label parentWindow as "ours".  We use this in Cocoa apps
	 * to help avoid crashes on OS X 10.2.8, especially in Camino (see
	 * drawRect_AWT() in Handlers.m).
	 */
	SetWindowProperty(parentWindow, kJEPCreatorCode,
		kPropertyNSWindow, sizeof(window), &window);

	if (isCarbonApp) {
		/* The event handler that gets installed when we create an
		 * NSCarbonWindow object is superfluous -- we replace it with our
		 * own below (HandleWindow() in Handlers.m, which also serves a
		 * lot of other purposes).  Luckily we can remove the original
		 * handler here.  Unless we do so, handleMouseDownEvent_CW() and
		 * handleCarbonBoundsChange_CW() in Handlers.m end up being called
		 * extra times on each applet view that we've created.
		 */
		RemoveEventHandler(((NSCarbonWindow *)window)->_eventHandler);
		DisposeEventHandlerUPP(((NSCarbonWindow *)window)->_handleEventUPP);
		((NSCarbonWindow *)window)->_eventHandler = nil;
		((NSCarbonWindow *)window)->_handleEventUPP = nil;
	}

	if ((appletClip.bottom > appletClip.top) &&
		(appletClip.right > appletClip.left))
	{
		frameHolder.origin.x = appletClip.left;
		frameHolder.origin.y = appletClip.top;
		frameHolder.size.width = appletClip.right - appletClip.left;
		frameHolder.size.height = appletClip.bottom - appletClip.top;
	}
	GetWindowBounds(parentWindow, kWindowContentRgn, &frameCarbon);
	frameCocoa.origin.x = frameCocoa.origin.y = 0;
	frameCocoa.size.width = frameCarbon.right - frameCarbon.left;
	frameCocoa.size.height = frameCarbon.bottom - frameCarbon.top;
	if ((frameHolder.origin.x >= frameCocoa.origin.x+frameCocoa.size.width) ||
		(frameHolder.origin.x+frameHolder.size.width <= frameCocoa.origin.x) ||
		(frameHolder.origin.y >= frameCocoa.origin.y+frameCocoa.size.height) ||
		(frameHolder.origin.y+frameHolder.size.height <= frameCocoa.origin.y))
	{
		frameHolder = NSZeroRect;
	}
	else if (frameHolder.origin.x + frameHolder.size.width >
		frameCocoa.origin.x + frameCocoa.size.width)
	{
		frameHolder.size.width = frameCocoa.origin.x +
			frameCocoa.size.width - frameHolder.origin.x;
	}
	else if (frameHolder.origin.y + frameHolder.size.height >
		frameCocoa.origin.y + frameCocoa.size.height)
	{
		frameHolder.size.height = frameCocoa.origin.y +
			frameCocoa.size.height - frameHolder.origin.y;
	}
	controller->initialPosition = NSMakePoint(appletPosX, appletPosY);
	controller->initialWidth = widthLocal;
	controller->initialHeight = heightLocal;
	controller->initialFrame = NSMakeRect(appletPosX, appletPosY,
		widthLocal, heightLocal);
	controller->currentFrame = frameHolder;

	if (isCarbonApp) {
		if ((controllersFound = [JEPController controllersInWindow:parentWindow
			timeoutType:lockTimeoutNone]) == nil)
		{
			/* Before JEP 0.9.5+h, we had to run NSWindow
			 * _installCarbonWindowEventHandlers here to install Apple's
			 * mouse and keyboard event handlers for its Cocoa-Carbon
			 * interface.  But now we don't, and life is much easier.  For
			 * more information see the comment above CreateNSEvent() in
			 * Handlers.m.
			 */
			[controller saveParentTSMState:parentWindow];
		} else {
			prevController = [controllersFound objectAtIndex:0];
			[controller copyParentTSMState:prevController];
			[controllersFound removeAllObjects];
			[controllersFound release];
		}

		/* Remove our Carbon event handlers for NSCarbonWindow objects that
		 * we've previously created in the same Carbon window -- each of them
		 * handles events for all the applet views that we've created in the
		 * same window.
		 */
		sig = [JEPController instanceMethodSignatureForSelector:
			@selector(cleanupHandlers)];
		if (sig)
			message = [NSInvocation invocationWithMethodSignature:sig];
		if (message) {
			[message setSelector:@selector(cleanupHandlers)];
			[JEPController sendToOtherControllersInWindow:message sender:controller
				carbonWindow:parentWindow targetIsMyWindow:false timeoutType:lockTimeoutNone];
		}

		/* Install the Carbon event handlers that help fix the bugs in Apple's
		 * implementation of NSCarbonWindow -- for the latest NSCarbonWindow
		 * object that we've created.
		 */
		if ((retval = InstallCarbonEventHandlers(parentWindow, controller)) != noErr)
			return retval;
	} else {
		/* We can't make our controller a delegate of "window" -- our parent
		 * app may (like Camino) already have set its delegate to something
		 * else.  Adding our controller to the default notification center
		 * works fine, though.
		 */
		[[NSNotificationCenter defaultCenter] addObserver:controller
			selector:@selector(windowWillClose:)
			name:@"NSWindowWillCloseNotification" object:window];
	}

	controller->myWindow = window;
	controller->myWindowCarbon = parentWindow;

	return noErr;
}

/* C-callable wrapper functions */

/* Call this after loading the bundle to make sure it can find what it
 * needs.  Returns true if things check out, otherwise false.  As of
 * JEP 0.8.7 I've added a version check to stop version mismatches
 * between JavaEmbeddingPlugin.bundle and the MRJ Plugin JEP.  This
 * function replaces JEPCheckBundle(), which is no longer supported.
 */
Boolean JEPCheckBundleWithVersion(CFStringRef version)
{
	BOOL retval = true;
	NSAutoreleasePool *localPool;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return false;
	retval = [JEPController checkBundleWithVersion:(NSString *)version];
	[localPool release];
	return retval;
}

CFStringRef JEPGetInstalledPath()
{
	CFStringRef path = nil;
	const char *installedPathUTF8 = nil;
	NSAutoreleasePool *localPool;
	if (!nsAppInited)
		return nil;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return nil;
	if (installedPath)
		installedPathUTF8 = [installedPath UTF8String];
	if (installedPathUTF8 && strlen(installedPathUTF8)) {
		path = CFStringCreateWithCString(NULL, installedPathUTF8,
			kCFStringEncodingMacRoman);
	}
	[localPool release];
	return path;
}

OSStatus JEPCreateJavaVM(JavaVMOption *jvmExtraOpts, int jvmExtraOptsNum,
	JavaVM **vm, void **env)
{
	OSStatus retval = noErr;
	NSAutoreleasePool *localPool;
	NSThread *currentThread;
	BOOL isMainThread;
	CreateJavaVMParams *params;
	if (jvmExtraOptsNum && !jvmExtraOpts)
		return EINVAL;
	if (!nsAppInited)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if (!mainThread || !nsAppShadow) {
		[localPool release];
		return EINVAL;
	}
	currentThread = [NSThread currentThread];
	isMainThread = [mainThread isEqual:currentThread];
	if (isMainThread) {
		retval = [AppletView maybeCreateJavaVM:jvmExtraOpts
			extraOptsNum:jvmExtraOptsNum vmPtr:vm envPtr:env initAWT:false];
	} else {
		params = [[CreateJavaVMParams alloc]
			initWithParams:jvmExtraOpts extraOptsNum:jvmExtraOptsNum
			vmPtr:vm envPtr:env initAWT:false];
		if (params) {
			[nsAppShadow performSelectorOnMainThread:
				@selector(createJavaVMOnMainThread:) withObject:params
				waitUntilDone:true];
		}
		if ((vm && !*vm) || (env && !*env))
			retval = ENOMEM;
	}
	[localPool release];
	return retval;
}

OSStatus JEPCreateJavaVMAndInitAWT(JavaVMOption *jvmExtraOpts,
	int jvmExtraOptsNum, JavaVM **vm, void **env,
	void (*runAfterInitAWT) (void *), void *runAfterInitAWTInfo)
{
	OSStatus retval = noErr;
	NSAutoreleasePool *localPool;
	NSThread *currentThread;
	BOOL isMainThread;
	CreateJavaVMParams *params;
	if (jvmExtraOptsNum && !jvmExtraOpts)
		return EINVAL;
	if (!nsAppInited)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if (!mainThread || !nsAppShadow) {
		[localPool release];
		return EINVAL;
	}
	afterInitAWT = runAfterInitAWT;
	afterInitAWTInfo = runAfterInitAWTInfo;
	currentThread = [NSThread currentThread];
	isMainThread = [mainThread isEqual:currentThread];
	if (isMainThread) {
		retval = [AppletView maybeCreateJavaVM:jvmExtraOpts
			extraOptsNum:jvmExtraOptsNum vmPtr:vm envPtr:env initAWT:true];
	} else {
		params = [[CreateJavaVMParams alloc]
			initWithParams:jvmExtraOpts extraOptsNum:jvmExtraOptsNum
			vmPtr:vm envPtr:env initAWT:true];
		if (params) {
			[nsAppShadow performSelectorOnMainThread:
				@selector(createJavaVMOnMainThread:) withObject:params
				waitUntilDone:true];
		}
		if ((vm && !*vm) || (env && !*env))
			retval = ENOMEM;
	}
	[localPool release];
	return retval;
}

/* Only returns an error if appletCreatedCallback will _not_ be called. */
OSStatus JEPCreateJavaApplet(CFURLRef docbase, CFDictionaryRef attributes,
	WindowRef parentWindow, JEPOptions *options, void **handle)
{
	OSStatus retval = noErr;
	JEPOptions opts = {{0},0};
	JEPController *controller;
	AppletView *appletView;
	NSURL *fixedDocbase;
	NSDictionary *fixedAttributes;
	JavaVM *vmHolder;
	NSThread *currentThread;
	BOOL isMainThread;
	CreateJavaVMParams *vmParams;
	CreateCocoaWindowParams *winParams;
	NSAutoreleasePool *localPool;

	if (!handle || !parentWindow || !mainThread || !nsAppShadow)
		return EINVAL;
	*handle = nil;
	if (options)
		memcpy(&opts, options, sizeof(JEPOptions));

	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:nil]) == nil) {
		[localPool release];
		return ENOMEM;
	}
	if (controller->myAppletView) {	// It shouldn't be possible that an applet
		[localPool release];		// already exists for "controller", but
		return noErr;				// just in case ...
	}

	currentThread = [NSThread currentThread];
	isMainThread = [mainThread isEqual:currentThread];

	if (isMainThread) {
		retval = [AppletView maybeCreateJavaVM:opts.jvmExtraOpts
			extraOptsNum:opts.jvmExtraOptsNum vmPtr:&vmHolder envPtr:nil
			initAWT:false];
	} else {
		vmParams = [[CreateJavaVMParams alloc]
			initWithParams:opts.jvmExtraOpts extraOptsNum:opts.jvmExtraOptsNum
			vmPtr:&vmHolder envPtr:nil initAWT:false];
		if (vmParams) {
			[nsAppShadow performSelectorOnMainThread:
				@selector(createJavaVMOnMainThread:) withObject:vmParams
				waitUntilDone:true];
		}
		if (!*vmHolder)
			retval = ENOMEM;
	}
	if ((retval != noErr) && (retval != EALREADY)) {
		[controller performSelector:@selector(setNeedsCleanup)];
		[localPool release];
		return retval;
	}
	controller->javaVM = vmHolder;
	if (isCarbonApp) {
		if ((retval = MaybeInstallAppEventHandlers()) != noErr) {
			[controller performSelector:@selector(setNeedsCleanup)];
			[localPool release];
			return retval;
		}
		if ((retval = MaybeInstallTSMAEEventHandler()) != noErr) {
			[controller performSelector:@selector(setNeedsCleanup)];
			[localPool release];
			return retval;
		}
	}

	fixedDocbase = [JEPController fixDocbase:(NSURL *)docbase];
	fixedAttributes = [controller fixAttributes:(NSDictionary *)attributes
		parentWindow:parentWindow];
	if (isMainThread) {
		retval = CreateCocoaWindow(controller, fixedAttributes,
			parentWindow, opts.appletClip, opts.appletPosX, opts.appletPosY,
			opts.appletWidth, opts.appletHeight);
	} else {
		winParams = [[CreateCocoaWindowParams alloc]
			initWithParams:controller attributes:fixedAttributes
			parentWindow:parentWindow appletClip:opts.appletClip
			appletPosX:opts.appletPosX appletPosY:opts.appletPosY
			appletWidth:opts.appletWidth appletHeight:opts.appletHeight];
		if (winParams) {
			[nsAppShadow performSelectorOnMainThread:
				@selector(createCocoaWindowOnMainThread:)
				withObject:winParams waitUntilDone:true];
		}
		if (!controller->myWindow)
			retval = ENOMEM;
	}
	if (retval != noErr) {
		[controller performSelector:@selector(setNeedsCleanup)];
		[localPool release];
		return retval;
	}

	controller->pluginInstance = opts.instance;
	controller->createdInvisible = opts.createdInvisible;
	controller->erasePreviousFrame = opts.erasePreviousFrame;
	if (opts.resizeAppletDirectly)
		controller->resizeAppletDirectly = true;
	[controller setAppletCreatedCallback:opts.appletCreatedCallback
		userData:opts.appletCreatedUserData];
	[controller setAppletInitializedCallback:opts.appletInitializedCallback
		userData:opts.appletInitializedUserData];
	[controller setAppletStartedCallback:opts.appletStartedCallback
		userData:opts.appletStartedUserData];
	[controller setShowURLCallback:opts.showDocumentCallback
		userData:opts.showDocumentUserData];
	[controller setShowStatusCallback:opts.showStatusCallback
		userData:opts.showStatusUserData];

	appletView = [controller startAppletInit:fixedDocbase
		appletAttributes:fixedAttributes];
	if (!appletView) {
		[controller performSelector:@selector(setNeedsCleanup)];
		retval = ENOMEM;
	} else {
		*handle = (void *) controller;
	}

	[localPool release];
	return retval;
}

OSStatus JEPGetJavaApplet(void *handle, jobject *applet)
{
	OSStatus retval = noErr;
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (applet)
		*applet = nil;
	if (!handle || !applet)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (controller->myAppletView && controller->myAppletView->myApplet)
		*applet = controller->myAppletView->myApplet;
	else retval = ENOENT;

	[localPool release];
	return retval;
}

OSStatus JEPGetJavaVM(void *handle, JavaVM **vm)
{
	OSStatus retval = noErr;
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (vm)
		*vm = nil;
	if (!handle || !vm)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (controller && controller->javaVM)
		*vm = controller->javaVM;
	else retval = ENOENT;

	[localPool release];
	return retval;
}

OSStatus JEPSetJavaAppletState(void *handle, AppletStateJEP newState)
{
	OSStatus retval = noErr;
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (!handle || !newState)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (!controller->myAppletView) {
		[localPool release];
		return ENOENT;
	}

	retval = [controller setAppletState:newState];

	[localPool release];
	return retval;
}

OSStatus JEPGetJavaAppletState(void *handle, AppletStateJEP *state)
{
	OSStatus retval = noErr;
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (state)
		*state = nil;
	if (!handle || !state)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (controller->myAppletView && controller->myAppletView->appletState)
		*state = controller->myAppletView->appletState;
	else retval = ENOENT;

	[localPool release];
	return retval;
}

OSStatus JEPDrawJavaApplet(void *handle)
{
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (!handle)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (!controller->myAppletView) {
		[localPool release];
		return ENOENT;
	}

	controller->previousFrame = controller->currentFrame;
	[controller drawApplet];

	[localPool release];
	return noErr;
}

OSStatus JEPMaybeDrawJavaApplet(void *handle)
{
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (!handle)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (!controller->myAppletView) {
		[localPool release];
		return ENOENT;
	}

	controller->previousFrame = controller->currentFrame;
	[controller maybeDrawApplet];

	[localPool release];
	return noErr;
}

OSStatus JEPSuspendJavaAppletUpdates(void *handle)
{
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (!handle)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (!controller->myAppletView) {
		[localPool release];
		return ENOENT;
	}

	[controller suspendAppletUpdates];

	[localPool release];
	return noErr;
}

OSStatus JEPResumeJavaAppletUpdates(void *handle)
{
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (!handle)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (!controller->myAppletView) {
		[localPool release];
		return ENOENT;
	}

	[controller resumeAppletUpdates];

	[localPool release];
	return noErr;
}

OSStatus JEPSizeJavaApplet(void *handle, int width, int height)
{
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (!handle)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (!controller->myAppletView) {
		[localPool release];
		return ENOENT;
	}

	[controller sizeApplet:NSMakeSize(width, height)];

	[localPool release];
	return noErr;
}

OSStatus JEPPositionJavaApplet(void *handle, int posX, int posY)
{
	JEPController *controller;
	NSAutoreleasePool *localPool;
	NSPoint fixedPoint;
	NSRect appletFrame;

	if (!handle)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (!controller->myAppletView) {
		[localPool release];
		return ENOENT;
	}

	fixedPoint = NSMakePoint(posX, posY);
	appletFrame = [controller->myAppletView frame];
	if (!isCarbonApp) {
		fixedPoint.y += appletFrame.size.height;
		fixedPoint = [JEPController flipWindowPoint:fixedPoint
			window:controller->myWindow];
	}

	if ((appletFrame.origin.x == fixedPoint.x) &&
		(appletFrame.origin.y == fixedPoint.y))
	{
		[localPool release];
		return noErr;
	}

	[controller positionApplet:fixedPoint];

	[localPool release];
	return noErr;
}

OSStatus JEPMoveAndClipJavaApplet(void *handle, int posX, int posY,
	int clipX, int clipY, int clipWidth, int clipHeight)
{
	JEPController *controller;
	NSThread *currentThread;
	BOOL isMainThread;
	MoveAndClipParams *params;
	NSAutoreleasePool *localPool;
	NSPoint fixedPoint;
	NSRect fixedClip;
	NSRect appletFrame = NSZeroRect;
	BOOL willZeroClip;

	if (!handle || !mainThread || !nsAppShadow)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (!controller->myAppletView || !controller->myLock) {
		[localPool release];
		return ENOENT;
	}

	[controller->myLock lock];

	/* Whether we're in a Carbon app or a Cocoa one, the coordinate system
	 * of the parameters with which we're called is the Carbon one (origin
	 * top-left).  But our positionApplet and setClip methods use whatever
	 * coordinate system our parent window uses.  So in a Cocoa app we need
	 * to translate our parameters from the Carbon coordinate system to the
	 * Cocoa one.
	 */
	fixedPoint = NSMakePoint(posX, posY);
	fixedClip = NSMakeRect(clipX, clipY, clipWidth, clipHeight);
	appletFrame = [controller->myAppletView frame];
	if (!isCarbonApp) {
		fixedPoint.y += appletFrame.size.height;
		fixedPoint = [JEPController flipWindowPoint:fixedPoint
			window:controller->myWindow];
		fixedClip.origin.y += clipHeight;
		fixedClip.origin = [JEPController flipWindowPoint:fixedClip.origin
			window:controller->myWindow];
	}
	if ((appletFrame.origin.x == fixedPoint.x) &&
		(appletFrame.origin.y == fixedPoint.y) &&
		(fixedClip.origin.x == controller->currentFrame.origin.x) &&
		(fixedClip.origin.y == controller->currentFrame.origin.y) &&
		(fixedClip.size.width == controller->currentFrame.size.width) &&
		(fixedClip.size.height == controller->currentFrame.size.height))
	{
		[controller->myLock unlock];
		[localPool release];
		return noErr;
	}
	willZeroClip =
		((fixedClip.size.width <= 0) || (fixedClip.size.height <= 0) ||
		(fixedClip.origin.x > appletFrame.origin.x+appletFrame.size.width) ||
		(fixedClip.origin.x+fixedClip.size.width < appletFrame.origin.x) ||
		(fixedClip.origin.y > appletFrame.origin.y+appletFrame.size.height) ||
		(fixedClip.origin.y+fixedClip.size.height < appletFrame.origin.y));

	currentThread = [NSThread currentThread];
	isMainThread = [mainThread isEqual:currentThread];
	if (isMainThread) {
		[controller suspendAppletUpdates];
		/* If we call positionApplet before setClip when we are about to zero-
		 * clip an applet that currently isn't zero-clipped, we may cause a
		 * subtle problem with the workaround for bmo bug 277067 in [AppletView
		 * setClip:]:  If our call to positionApplet here changes the applet's
		 * position, then the update performed by NSAppShadow's
		 * postCarbonUpdateOnMainThread: or postCocoaUpdateOnMainThread: will
		 * effect the wrong "region" (the workaround's code will think that
		 * the new location (the one set here) is the "old location" (the one
		 * where the applet that's about to be zero-clipped is/was located)).
		 * This change resolves bmo bug 379058.
		 */
		if (willZeroClip && ![controller->myAppletView isCurrentClipEmpty]) {
			objc_msgSend(controller, @selector(setClipImpl:), fixedClip);
			objc_msgSend(controller, @selector(positionAppletImpl:),
				fixedPoint);
		} else {
			objc_msgSend(controller, @selector(positionAppletImpl:),
				fixedPoint);
			objc_msgSend(controller, @selector(setClipImpl:), fixedClip);
		}
		[controller resumeAppletUpdates];
	} else {
		params = [[MoveAndClipParams alloc] initWithParams:controller
			location:fixedPoint clip:fixedClip];
		if (params) {
			[nsAppShadow performSelectorOnMainThread:
				@selector(moveAndClipOnMainThread:)
				withObject:params waitUntilDone:true];
		}
	}

	[controller->myLock unlock];

	[localPool release];
	return noErr;
}

OSStatus JEPShowHideJavaApplet(void *handle, Boolean visible)
{
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (!handle)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}
	if (!controller->myAppletView) {
		[localPool release];
		return ENOENT;
	}

	[controller showHideApplet:visible];

	[localPool release];
	return noErr;
}

OSStatus JEPRegisterStatusCallback(void *handle,
	JEPShowStatusCallback showStatusFunction, void *userData)
{
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (!handle)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}

	[controller setShowStatusCallback:showStatusFunction userData:userData];

	[localPool release];
	return noErr;
}

OSStatus JEPRegisterShowDocumentCallback(void *handle,
	JEPShowDocumentCallback showDocumentFunction, void * userData)
{
	JEPController *controller;
	NSAutoreleasePool *localPool;

	if (!handle)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	if ((controller = [JEPController myController:(JEPController *)handle]) == nil) {
		[localPool release];
		return ENOENT;
	}

	[controller setShowURLCallback:showDocumentFunction userData:userData];

	[localPool release];
	return noErr;
}

Boolean JEPGetJavaRuntimeParameterBoolean(CFStringRef key)
{
	Boolean retval = false;
	NSAutoreleasePool *localPool;
	if (!nsAppInited)
		return false;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return false;
	retval = [AppletView getRuntimeParamBoolean:(NSString *)key];
	[localPool release];
	return retval;
}

CFStringRef JEPGetJavaRuntimeParameterString(CFStringRef key)
{
	CFStringRef retval = nil;
	NSString *value = nil;
	const char *valueUTF8 = nil;
	NSAutoreleasePool *localPool;
	if (!nsAppInited)
		return nil;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return nil;
	if ((value = [AppletView getRuntimeParamString:(NSString *)key]) != nil)
		valueUTF8 = [value UTF8String];
	if (valueUTF8 && strlen(valueUTF8)) {
		retval = CFStringCreateWithCString(NULL, valueUTF8,
			kCFStringEncodingMacRoman);
	}
	[localPool release];
	return retval;
}

OSStatus JEPLogJavaMessage(CFStringRef message)
{
	OSStatus retval = noErr;
	NSAutoreleasePool *localPool;
	NSString *info;
	if (!nsAppInited)
		return EINVAL;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return ENOMEM;
	info = [NSString stringWithFormat:@"%@ %@", [NSDate date],
		(NSString *) message];
	retval = [AppletView printMessage:info javaEnv:nil];
	[localPool release];
	return retval;
}

/* Needed to deal with Tiger (OS X 10.4).  For more info see
 * JavaEmbeddingPlugin.h.
 */
void JEPDoLiveConnectProxy(jclass proxyClass, jobject runnable, jobject applet,
	jboolean isCallMethodCall)
{
	JNIEnv *env = nil;
	NSThread *currentThread = nil;
	BOOL isMainThread = false;
	jobject old;
	static jclass liveConnectClass = nil;
	static jmethodID doProxy = nil;
	jmethodID proxyRun = nil;
	NSAutoreleasePool *localPool;
	if (!mainThread)
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	if (![AppletView getJavaEnv:&env]) {
		[localPool release];
		return;
	}
	if ((*env)->ExceptionCheck(env))
		(*env)->ExceptionClear(env);
	if (!doProxy) {
		liveConnectClass =
			(*env)->FindClass(env, "jep/LiveConnect");
		if ((*env)->ExceptionCheck(env)) {
			(*env)->ExceptionClear(env);
			[localPool release];
			return;
		}
		old = liveConnectClass;
		liveConnectClass = (*env)->NewGlobalRef(env, liveConnectClass);
		(*env)->DeleteLocalRef(env, old);
		doProxy = (*env)->GetStaticMethodID(env, liveConnectClass,
			"doProxy", "(Ljava/lang/Class;Ljava/lang/Runnable;Ljava/applet/Applet;)V");
		if ((*env)->ExceptionCheck(env)) {
			doProxy = nil;
			(*env)->ExceptionClear(env);
			[localPool release];
			return;
		}
	}
	currentThread = [NSThread currentThread];
	if ([mainThread isEqual:currentThread])
		isMainThread = true;
	if ((osxVersion >= 0x1040) && isMainThread && isCallMethodCall) {
		(*env)->CallStaticVoidMethod(env, liveConnectClass,
			doProxy, proxyClass, runnable, applet);
	} else {
		proxyRun = (*env)->GetStaticMethodID(env, proxyClass, "run",
			"(Ljava/lang/Runnable;)V");
		if (proxyRun)
			(*env)->CallStaticVoidMethod(env, proxyClass, proxyRun, runnable);
	}
	[localPool release];
}

void JEPRunOnMainThread(void (*runnable) (void *), void *info, Boolean wait)
{
	NSAutoreleasePool *localPool;
	if (!runnable)
		return;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return;
	[AppletView performCEventOnMainThread:runnable userInfo:info
		waitUntilDone:wait];
	[localPool release];
}

/* As of JEP 0.8.3, I added this call to deal with occasional hangs in
 * JSObject methods (i.e. the Java-to-JavaScript side of LiveConnect).
 * (See Java_netscape_oji_JNIUtils_RunnableMessage_run() in
 * LiveConnectNativeMethods.cpp for more info.)  It doesn't fit terribly
 * well with the rest of the API, and I'd like to get rid of it at some
 * point.  But for now it solves a problem to which I have no other
 * solution.
 */
/* As of JEP 0.8.6 we no longer use this API -- I've found a better (more
 * general) solution to this problem that doesn't require it.  (See the
 * JavaMessage class in the MRJ Plugin JEP's JavaMessageQueue.h and
 * Java_netscape_javascript_JSObject_getMember() and so forth (the JSObject
 * methods that return values) in LiveConnectNativeMethods.cpp.)  But I'm
 * leaving this API in for the time being because browsers that (unlike the
 * Mozilla family) can't easily call something on the main thread might
 * find it useful -- especially now that the info callbacks (like
 * appletCreatedCallback) might not get called on the main thread.
 */
jthrowable JEPRunJRunnableOnMainThread(jobject jrunnable)
{
	NSDate *started;
	jthrowable retval = nil;
	JavaRunner *runner = nil;
	Boolean aborted = false;
	NSTimeInterval interval = 0;
	NSThread *currentThread = [NSThread currentThread];
	NSAutoreleasePool *localPool;

	if (!mainThread)
		return nil;
	if ((localPool = [[NSAutoreleasePool alloc] init]) == nil)
		return nil;

	if ((runner = [[JavaRunner alloc] init]) == nil) {
		[localPool release];
		return nil;
	}
	runner->jrunnable = jrunnable;
	if ([mainThread isEqual:currentThread]) {
		[runner runJRunnableOnMainThread];
		retval = runner->exception;
	} else {
		[runner performSelectorOnMainThread:@selector(runJRunnableOnMainThread)
			withObject:nil waitUntilDone:false];
		started = [NSDate date];
		for (;;) {
			if (runner->done) {
				[runner release];
				break;
			}
			[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:0.01]];
			if ((interval = [started timeIntervalSinceNow]) < 0)
				interval = -interval;
			if (interval > 0.5) {
				runner->abort = (aborted = true);
				[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:0.1]];
				break;
			}
		}
		if (!aborted)
			retval = runner->exception;
	}

	[localPool release];
	return retval;
}

